#ifndef lint
static const char Rcs_Id[] =
    "$Id: cmdstream.cc,v 1.7 1998/10/23 23:03:21 geoff Exp $";
#endif

/*
 * $Id: cmdstream.cc,v 1.7 1998/10/23 23:03:21 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: cmdstream.cc,v $
 * Revision 1.7  1998/10/23 23:03:21  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.6  1998/01/31 06:30:22  geoff
 * Panic on errors, rather than just aborting.
 *
 * Revision 1.5  1997/07/16  22:19:28  geoff
 * Only prototype getdtablesize and wait4 on Sun4 machines.
 *
 * Revision 1.4  1996/06/11  03:08:18  geoff
 * Add support for the openErrno.
 *
 * Revision 1.3  1996/05/03  01:31:33  geoff
 * Update the copyright.
 *
 * Revision 1.2  1995/11/10  03:00:52  geoff
 * Open streams with the given protection, not the irrelevant mode, which
 * was causing some files to be created with *very* strange permissions.
 *
 * Revision 1.1  1995/09/15  00:14:51  geoff
 * Initial revision
 *
 */

/*
 * A CommandStream is like an fstream, except that it is also possible
 * to open pipes to and from Unix commands.  This implements the
 * equivalent of "popen" in stdio.
 */
#include "cmdstream.hh"
#include "panic.hh"
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
int			CommandStream::open(const char* name,
			  int mode, int protection,
			  CommandStream::OpenOptions options);
					// Open file or pipe
void			CommandStream::close();
					// Close, waiting for subprocess
const char*		CompressedStream::mangleName(const char* name,
			  int mode, CompressionOptions compression);
					// Change name to force compression
#endif // 0

int			CommandStream::openErrno = 0;
					// Errno from last open, if NZ

/*
 * The following routines aren't prototyped under SunOS 4.1.1, so I
 * have to declare them myself.  Sigh.
 */
#ifdef __sparc__
extern "C" int		getdtablesize(void);
extern "C" int		wait4(int pid, void* statusp, int options,
			  struct rusage* rusagep);
#endif // __sparc__

/*
 * Open the named file and return an fd that can access it.  This is
 * necessary because that's what the underlying streambuf wants out of
 * life.
 */
int CommandStream::open(
    const char*		name,		// Name of file to open
    int			mode,		// Open mode to use
    int			protection,	// File protection, if creating
    CommandStream::OpenOptions options)	// Open options to use.
    {
    commandPID = 0;
    openErrno = 0;
    totalBytesRead = totalBytesWritten = 0;
    int nameLength = strlen(name);
    if (options == FileOrCommand)
	{
	/*
	 * Decide what sort of open to use.
	 */
	if ((mode & (ios::in | ios::out)) == ios::out  &&  name[0] == '|')
	    {
	    options = CommandOnly;
	    nameLength--;
	    name++;
	    }
	else if ((mode & (ios::in | ios::out)) == ios::in
	  &&  name[nameLength - 1] == '|')
	    {
	    options = CommandOnly;
	    nameLength--;
	    }
	else if (name[0] == '|'  ||  name[nameLength - 1] == '|')
	    {
	    errno = openErrno = EINVAL;
	    return -1;
	    }
	else
	    options = FileOnly;
	}
    if (options == FileOnly)
	{
	int openOptions = 0;
	switch (mode & (ios::in | ios::out))
	    {
	    case ios::in:
		openOptions = O_RDONLY;
		break;
	    case ios::out:
		openOptions = O_WRONLY | O_CREAT;
		break;
	    case ios::in | ios::out:
		openOptions = O_RDWR;
		break;
	    default:
		errno = openErrno = EINVAL;
		return -1;
		break;
	    }
	if (mode & ios::app)
	    openOptions |= O_APPEND;
	if (mode & ios::trunc)
	    openOptions |= O_TRUNC;
	if (mode & ios::nocreate)
	    openOptions &= ~O_CREAT;
	if (mode & ios::noreplace)
	    openOptions |= O_CREAT | O_EXCL;
	int fd = ::open(name, openOptions, protection);
	openErrno = fd < 0 ? errno : 0;
	if (mode & ios::ate)
	    (void)::lseek(fd, 0L, 2);
	return fd;
	}
    else
	{
	/*
	 * This is the fun case, where we get to build pipes and spawn
	 * a subprocess.  The subprocess will exec "sh -c 'command'"
	 * with appropriate redirection.
	 */
	if ((mode & (ios::in | ios::out)) == (ios::in | ios::out))
	    {
	    errno = openErrno = EINVAL;
	    return -1;
	    }
	int pipeDescriptors[2];
	if (pipe(pipeDescriptors) == -1)
	    {
	    /*
	     * Probably there's no room for more files.
	     */
	    openErrno = errno;
	    return -1;
	    }
	cout.flush();
	cerr.flush();
	int childPID;
	// NEEDSWORK: perhaps shouldn't try fork forever
	while ((childPID = fork()) == -1)
	    sleep (5);
	if (childPID != 0)
	    {
	    /*
	     * This is the parent process.  Clean up the unused part
	     * of the pipe, finish opening the stream, and record the
	     * child's PID for later use by the close routine.
	     */
	    commandPID = childPID;
	    if (mode & ios::in)
		{
		::close(pipeDescriptors[1]);
		return pipeDescriptors[0];
		}
	    else
		{
		::close(pipeDescriptors[0]);
		return pipeDescriptors[1];
		}
	    }
	else
	    {
	    /*
	     * This is the child process.  Life is more difficult
	     * here.  First we must dup(2) the appropriate descriptor
	     * back into standard input or output.  Then we have to
	     * clean up all open descriptors except standard input,
	     * output, and error.
	     */
	    if (mode & ios::in)
		{
		/*
		 * The parent wants to do input, so we must do output.
		 */
		::close(pipeDescriptors[0]);
		::close(1);
		dup(pipeDescriptors[1]);
		::close(pipeDescriptors[1]);
		}
	    else
		{
		/*
		 * The parent wants to do output, so we must read.
		 */
		::close(pipeDescriptors[1]);
		::close(0);
		dup(pipeDescriptors[0]);
		::close(pipeDescriptors[0]);
		}
	    for (int descriptor = getdtablesize();  --descriptor >= 3;  )
		::close(descriptor);
	    /*
	     * Finally, we must exec a shell to run the command.
	     */
	    char* command = new char[nameLength + 1];
	    (void)strncpy(command, name, nameLength);
	    command[nameLength] = '\0';
	    execl("/bin/sh", "sh", "-c", command, NULL);
	    /*
	     * We only get here if the exec of /bin/sh failed.
	     * There's no real need to delete the new character array,
	     * since we're exiting.
	     */
	    exit(1);
	    }
	}
    }

/*
 * Close a command stream.  If a subprocess existed, wait for it to
 * terminate before returning.  (Is this a good idea?)
 */
void CommandStream::close()
    {
    fstream::close();
    if (commandPID != 0)
	(void)wait4(commandPID, (int *)NULL, 0, (struct rusage*)NULL);
    }

/*
 * Look at the suffix of a file to see if it's ".gz".  If so, and
 * compression is selected, return a "name" that is a command that
 * will perform the desired compression or uncompression.  Otherwise
 * just return the name passed.
 */
const char* CompressedStream::mangleName(
    const char*		name,		// Name of file or command to open
    int			mode,		// Open mode (read/write)
    CompressionOptions	compression)	// Whether to auto-compress
    {
    static char*	mangledName = NULL; // Return value, see below
    /*
     * Start by freeing any mangled name from last time.  See the
     * comments below for more information.
     */
    delete[] mangledName;
    mangledName = NULL;
    switch (compression)
	{
	case AsIs:
	    return name;
	case AutoCompress:
	    break;
	default:
	    // NEEDSWORK: should throw exception.
	    panic("bad compression type in CompressedStream::mangleName");
	    break;
	}
    /*
     * We may need to auto-compress.  If this is a piped command, we
     * never auto-compress.  Otherwise, look for a ".gz" suffix.
     */
    int nameLength = strlen(name);
    if (name[0] == '|'  ||  name[nameLength - 1] == '|')
	return name;			// Pipe: always do as-is
    else if (strcmp(&name[nameLength - 3], ".gz") != 0)
	return name;			// Not .gz-suffixed, do as-is
    /*
     * We need to auto-compress, which means we have to return a
     * constructed command to our caller.  Since there's no easy way
     * for our caller (the constructor) to delete the allocated
     * storage, we'll use a static variable here and free it up the
     * next time we're called (see above).  This is ugly and slightly
     * wasteful, but I don't know any other way to get the proper
     * effect.
     */
    switch (mode & (ios::in | ios::out))
	{
	case ios::in:
	    mangledName = new char[nameLength + sizeof "gunzip < |"];
	    (void)strcpy(mangledName, "gunzip < ");
	    (void)strcat(mangledName, name);
	    (void)strcat(mangledName, "|");
	    break;
	case ios::out:
	    mangledName = new char[nameLength + sizeof "|gzip -9 >"];
	    (void)strcpy(mangledName, "|gzip -9 >");
	    (void)strcat(mangledName, name);
	    break;
	default:
	    // NEEDSWORK: should throw exception
	    panic("bad I/O mode in CompressedStream::mangleName");
	    break;
	}
    return mangledName;
    }
