#ifndef _CMDSTREAM_HH
#define _CMDSTREAM_HH
/*
 * $Id: cmdstream.hh,v 1.4 1996/06/11 03:08:18 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: cmdstream.hh,v $
 * Revision 1.4  1996/06/11 03:08:18  geoff
 * Add a bunch of missing inline declarations.  Add operator!() so the
 * underlying operator!() won't get called.
 *
 * Revision 1.3  1996/05/03  01:31:34  geoff
 * Update the copyright.
 *
 * Revision 1.2  1996/02/02  18:55:39  geoff
 * Add the reportBytesRead/Written routines.
 *
 * Revision 1.1  1995/09/15  00:14:51  geoff
 * Initial revision
 *
 */

#include <fstream.h>
#include <sys/types.h>

/*
 * A CommandStream is similar to fstream, except that it is also possible
 * to open pipes to and from Unix commands.  This implements the
 * equivalent of "popen" in stdio.
 *
 * Due to fstream/iostream/streambuf and C++ limitations
 * CommandStreams must be created using a constructor which also
 * opens the named file.  Depending on the options, the file may be
 * either a plain file or a pipe, as follows:
 *
 *	FileOnly	"name" refers to a plain file, just as in fstream.
 *	CommandOnly	"name" is a shell command to be run.  If
 *			"mode & ios::in" is nonzero, the standard
 *			output of the command is connected to the
 *			stream object.  If "mode & ios::out" is
 *			nonzero, the stream object is connected to the
 *			standard input of the command.  It is not
 *			permissible to specify a mode of
 *			"ios::in|ios::out".
 *	FileOrCommand	"name" is examined, in conjunction wih "mode".
 *			If "name" ends with a pipe symbol ('|') and
 *			"mode & ios::in" is nonzero, or if "name"
 *			begins with a pipe symbol and "mode &
 *			ios::out" is nonzero, the constructor behaves
 *			as if CommandOnly had been specified.
 *			Otherwise, if "name" begins or ends with a
 *			pipe symbol, an error occurs.  For all other
 *			cases, the constructor behaves as if
 *			"FileOnly" had been specified.
 *
 * A CommandStream may also be created from an fd, just like an
 * fstream, but in this case the caller is responsible for providing
 * the child PID so that the close routine can wait for the child.
 *
 * Regardless of the creation method, the close routine ignores the
 * child's completion status.
 */
class CommandStream : public fstream
    {
    public:
	enum OpenOptions
	    {
	    FileOnly,			// Only allow access to files
	    CommandOnly,		// Interpret name as command to run
	    FileOrCommand		// Look for '|' symbol to decide
	    };
	inline		CommandStream(const char* name, int mode = ios::out,
			  int protection = 0664,
			  OpenOptions options = FileOnly);
					// Open file or pipe
	inline		CommandStream(int fd, pid_t childPID = 0);
	inline		~CommandStream();
	void		close();	// Close, waiting for subprocess
	inline		operator bool() const; // Test for success
	inline bool	operator!() const; // Test for failure
	inline CommandStream&
			read(void* buffer, int numberOfBytes);
					// Read from stream, keeping count
	inline CommandStream&
			write(const void* buffer, int numberOfBytes);
					// Write to stream, keeping count
	/*
	 * Return the number of bytes read/written through the
	 * read/write routines.  These are necessary because tellg()
	 * et al. will not only fail to return correct values for
	 * pipes, but will actually set a flag that inhibits further
	 * use of the stream.  Note that these routines will only
	 * produce valid results if read/write are the *only* methods
	 * called for the file.
	 */
	inline unsigned long
			bytesRead() const; // Return number of bytes read
	inline unsigned long
			bytesWritten() const; // Return number of bytes written
	/*
	 * Provide the user of the class with a way to inform the
	 * class of bytes read or written through means other than
	 * read/write.
	 */
	inline void	reportBytesRead(unsigned long n);
					// Admit to reading n bytes
	inline void	reportBytesWritten(unsigned long n);
					// Admit to writing n bytes
    protected:
	static int	openErrno;	// Errno from last open, if NZ
    private:
	int		open(const char* name, int mode, int protection,
			  OpenOptions options);
					// Do the work of opening the file
	/*
	 * Private data
	 */
	pid_t		commandPID;	// PID of pipe command, if NZ
	unsigned long	totalBytesRead;	// Count of bytes read from stream
					// ..using read()
	unsigned long	totalBytesWritten; // Count of bytes written to stream
					// ..using write()
    };

/*
 * A CompressedStream is a CommandStream which will handle compressed
 * files.  If AutoCompression is selected, opening a ".gz"-suffixed
 * file (but not any command) will automatically interpose a gzip or
 * gunzip, as appropriate, into the stream.  All other files are
 * always opened in FileorCommand mode.  Note that the protection
 * argument is ignored when a gzipped output file is being created!
 */
class CompressedStream : public CommandStream
    {
    public:
	enum CompressionOptions
	    {
	    AutoCompress,		// Automatically handle .gz suffixes
	    AsIs			// Read/write file directly
	    };
	inline		CompressedStream(const char* name, int mode = ios::out,
			  int protection = 0664,
			  CompressionOptions compression = AutoCompress);
					// Open compressed file or pipe
	inline		CompressedStream(int fd, pid_t childPID = 0);
					// Open from FD; can't compress
	inline		~CompressedStream();
    private:
	const char*	mangleName(const char* name, int mode,
			  CompressionOptions compression);
					// Change name to force compression
    };

#include "cmdstream.icc"

#endif // _CMDSTREAM_HH
