#ifndef _FIELDS_HH
#define _FIELDS_HH
/*
 * $Id: fields.hh,v 1.2 1998/10/23 23:03:23 geoff Exp $
 */

/*
 * Classes to be used to conveniently access fields within an ASCII
 * line, such as those processed by sort, awk, cut, etc.
 */

/*
 * $Log: fields.hh,v $
 * Revision 1.2  1998/10/23 23:03:23  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.1  1996/02/14  19:16:38  geoff
 * Initial revision
 *
 */

class istream;
class ostream;

class ParsedFields
    {
	friend inline ostream&
			operator<<(ostream& stream,
			  const ParsedFields& fields);
    public:
	/*
	 * Flags for constructor options.  Should be const ints, but ANSI
	 * C++ prevents this, so I'll use an enum to kludge it.
	 */
	enum
	    {
	    RUNS =			0x0001,	// Treat runs of delimiters
						// as if they were one
	    SINGLE_QUOTES =		0x0002,	// Accept single-quoted fields
	    BACK_QUOTES =		0x0004,	// Accept back-quoted fields
	    DOUBLE_QUOTES =		0x0008,	// Accept double-quoted fields
	    SHELL_QUOTES =		0x0010,	// Use shell-style (embedded)
						// quoting rules
	    STRIP_QUOTES =		0x0020,	// Strip quotes from fields
	    BACKSLASH =			0x0040	// Process C-style backslashes
	    };
			ParsedFields(istream& stream,
			  const char* delimiters = " \t",
			  unsigned int flags = RUNS,
			  unsigned int maxFields = 0);
			ParsedFields(const char* string,
			  const char* delimiters = " \t",
			  unsigned int flags = RUNS,
			  unsigned int maxFields = 0);
	inline		~ParsedFields();
	inline		operator bool(); // Convert to boolean
					// ..for EOF testing on stream
	inline unsigned int
			count() const;	// Return number of active fields
	inline unsigned int
			inputLineLength() const;
					// Return length of input line
	inline const char* operator[](unsigned int index) const;
					// Return a chosen field
	ostream&	write(ostream& stream, char delimiter = ' ') const;
	static inline void
			setOptions(unsigned int fieldIncrementValue = 20,
			  unsigned int lineIncrementValue = 120);
					// Set global options
    private:
			ParsedFields(const ParsedFields& source);
					// Disable copy constructor
	ParsedFields&	operator=(const ParsedFields& source);
					// Disable assignment
	void		parse(const char* delimiters, unsigned int flags,
			  unsigned int maxFields);
					// Parse buffer into fields
	unsigned int	processBackslash(const char* string,
			  char** outputPointer, bool strip);
					// Process backslash sequences
	/*
	 * Static data
	 */
	static unsigned int
			fieldIncrement;	// Increment to use allocating fields
	static unsigned int
			lineIncrement;	// Increment to use allocating lines
	/*
	 * Private data
	 */
	unsigned int	lineLength;	// Length of unparsed input line
	unsigned int	numFields;	// Number of fields in the line
	char*		lineBuffer;	// Allocated buffer containing the line
	char**		fields;		// Malloc'ed array of pointers to
					// ..fields
	bool		hadNewline;	// True if line ended with a newline
	bool		valid;		// True if structure has valid data
    };

#include "fields.icc"

#endif // _FIELDS_HH
