#ifndef _HASHTABLE_HH
#define _HASHTABLE_HH
/*
 * $Id: hashtable.hh,v 1.13 1999/11/17 07:56:58 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: hashtable.hh,v $
 * Revision 1.13  1999/11/17 07:56:58  geoff
 * Add some missing friendship declarations that the compiler now catches.
 *
 * Revision 1.12  1998/10/23 23:03:25  geoff
 * Add HASH_CHECK_CHAIN and HASH_CHECK_BUCKETS.
 *
 * Revision 1.11  1998/03/05 04:41:10  geoff
 * Remove some leftover #if 0's
 *
 * Revision 1.10  1998/03/05 03:41:30  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.9  1998/01/31 06:30:23  geoff
 * Make using the type allocator be the default.
 *
 * Revision 1.8  1997/07/11 17:50:54  geoff
 * Add the BUZ (Bob Uzgalis) hashing algorithm.
 *
 * Revision 1.7  1996/09/11 23:07:56  geoff
 * Use a type-specific allocator.
 *
 * Revision 1.6  1996/06/11  03:08:19  geoff
 * Add a bunch of missing inline declarations.
 *
 * Revision 1.5  1996/05/03  01:31:38  geoff
 * Update the copyright.
 *
 * Revision 1.4  1996/02/02  18:55:47  geoff
 * Move the declaration of operator bool to reflect where it is in the
 * code file.
 *
 * Revision 1.3  1995/10/10  19:39:45  geoff
 * Get rid of the obsolete memory-leak-tracking stuff.  Turn off
 * HASH_DEBUG_LEAKS.
 *
 * Revision 1.2  1995/09/15  00:12:59  geoff
 * Add abortive memory-leak-catching support.
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

/*
 * Header file for the generalized hash package.
 */

#include <sys/types.h>
#include "panic.hh"
#include "typealloc.hh"

#ifndef NULL
#define NULL	    0		/* Since we don't include stdio.h */
#endif // NULL

/*
 * Define HASH_DEBUG_LEAKS to add code that's intended to help catch
 * memory leaks.
 */
#undef HASH_DEBUG_LEAKS

/*
 * Define HASH_CHECK_CHAIN to add code that will, on every lookup,
 * check the entire collision chain to ensure that there is exactly
 * one entry that matches the given key.
 */
#undef HASH_CHECK_CHAIN

/*
 * Define HASH_CHECK_BUCKETS to add code that will, on every lookup,
 * check each data item accessed to ensure that it has hashed to the
 * correct bucket.  If HASH_CHECK_CHAIN is also defined, every entry
 * on the collision chain will be checked.  This option is useful when
 * debugging programs that muck around with the keys belonging to
 * entries in the hash table.
 */
#undef HASH_CHECK_BUCKETS

#ifdef USE_TEMPLATES
#define TEMPLATIZED	<DATA_TYPE>
#else // USE_TEMPLATES
#define DATA_TYPE	void
#define TEMPLATIZED
#endif // USE_TEMPLATES

/*
 * Class used to describe an entry in a hash table
 */
class HashTable;

/*
 * Class used to describe a hash table.
 */
#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
class HashTable
    {
    public:
	class			Iterator;
    private:
	/*
	 * Helper class for holding table entries
	 */
#ifdef USE_TEMPLATES
	template <DATA_TYPE>
#endif // USE_TEMPLATES
	class Entry
	    {
		friend		HashTable;
		friend		HashTable::Iterator;
	    private:
#ifdef HASH_DEBUG_LEAKS
		inline		Entry(HashTable& containingTable,
				  Entry* link = NULL, void* value = NULL);
#else // HASH_DEBUG_LEAKS
		inline		Entry(Entry* link = NULL, void* value = NULL);
#endif // HASH_DEBUG_LEAKS
		inline		~Entry();
				Entry(const Entry&);
					// Disallow copying
		Entry&		operator=(const Entry&);
					// Disallow assignment
		inline void*	operator new(size_t size);
					// Allocate a hash element
		inline void	operator delete(void* element);
					// Free a single hash element
#ifdef HASH_DEBUG_LEAKS
		HashTable&	table;	// Table containing this entry
#endif // HASH_DEBUG_LEAKS
		Entry*		next;	// Next entry in collision chain
		DATA_TYPE*	datum;	// Datum provided by caller
	    };
    public:
	friend			Entry;
	/*
	 * Class for iterating through hash tables
	 */
#ifdef USE_TEMPLATES
	template <DATA_TYPE>
#endif // USE_TEMPLATES
	class Iterator
	    {
		friend		HashTable;
	    public:
		inline		Iterator(HashTable& target);
		inline		Iterator(const Iterator&);
		inline		~Iterator();
		inline Iterator&
				operator=(const Iterator& that);
		inline void	reset(); // Reset to start of table
		inline Iterator&
				operator++();
					// Step to next entry
		inline		operator bool();
					// For completion testing
		void		removeCurrent();
					// Zap current w/o affecting iter
	    private:
		HashTable*	table;	// Table we are iterating over
		unsigned int	hashValue;
					// Hash value we're currently doing
		Entry*		current; // Current entry
		bool		suppressIncrement;
					// Suppress increment in ++ because
					// ..removeCurrent already did it
	    };
	friend			Iterator;
	/*
	 * Actions to take in hash::find
	 */
	enum action
	    {
	    FAIL,			// Fail if not found
	    INSERT			// Insert if not found
	    };
	/*
	 * Public functions
	 */
			HashTable(unsigned int initialSize,
			  unsigned int (*hashFunction)(const DATA_TYPE* key),
			  int (*compareFunction)(const DATA_TYPE* ent1,
			    const DATA_TYPE* ent2));
			~HashTable();
#ifdef USE_TEMPLATES
	inline DATA_TYPE& operator[](const Iterator& iterator) const;
#else // USE_TEMPLATES
	inline DATA_TYPE* operator[](const Iterator& iterator) const;
#endif // USE_TEMPLATES
	DATA_TYPE*	find(DATA_TYPE* datum, action notFoundAction = FAIL);
	bool		remove(DATA_TYPE* datum);
					// Remove entry, return true if not
					// ..found
	inline unsigned long
			size() const;	// Return # entries in table
	void		showStatistics(class ostream& stream) const;
    private:
			HashTable(const HashTable&); // Disallow copying
	HashTable&	operator=(const HashTable&); // Disallow assignment
	bool		expand(unsigned int newSize);
	unsigned long	entryCount;	// Number of entries currently in table
	unsigned long	tableSize;	// Size of the hash table
	bool		cantExpand;	// NZ if can't expand further
	unsigned int	expansions;	// Number of expansions performed
	unsigned long	totalCollisions; // Total collisions ever experienced
	unsigned long	collisionsNow;	// Current no. of collisions in table
	unsigned long	recentCollisions; // Total colls since last expansion
	unsigned long	longestChain;	// Longest collision chain ever seen
	unsigned long	recentChain;	// Longest chain since last expansion
	Entry*		entries;	// Entries in the table
	unsigned int	(*hash)(const DATA_TYPE* key);
					// Routine to calculate a hash value
	int		(*compare)(const DATA_TYPE* ent1,
			  const DATA_TYPE* ent2);
					// Routine to compare two entries
#ifdef HASH_DEBUG_LEAKS
	unsigned long	elementCreations; // No. of elements created,
					// .. this table
	unsigned long	elementDeletions; // No. of elements destroyed,
					// .. this table
	static unsigned long
			globalElementCreations;
					// No. of elements created, all tables
	static unsigned long
			globalElementDeletions;
					// No. of elements destroyed, all tbls
#endif // HASH_DEBUG_LEAKS
	static TypeAllocator
			allocator;	// Type-specific allocator for elements
    };

/*
 * Handy hashing routines
 */
extern unsigned int	hashString(const char* key);
extern unsigned int	hashPJW(const char* key);
extern unsigned int	hashInteger(int key);
extern unsigned int	hashStringBUZ(const char* key);
extern unsigned int	hashIntegerBUZ(int key);

#include "hashtable.icc"

#endif // _HASHTABLE_HH
