#ifndef lint
static const char Rcs_Id[] =
    "$Id: list.cc,v 1.13 2000/12/30 09:05:45 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: list.cc,v $
 * Revision 1.13  2000/12/30 09:05:45  geoff
 * Add self-assignment protection
 *
 * Revision 1.12  1998/10/23 23:03:26  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.11  1998/10/13 05:57:30  geoff
 * Make hasMember's argument be const.
 *
 * Revision 1.10  1998/03/05 03:41:30  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.9  1998/02/19 01:03:32  geoff
 * Add support for copying and assignment of lists, including destructive
 * copying.
 *
 * Revision 1.8  1998/01/31 06:30:24  geoff
 * Get rid of LIST_OWN_NEW and LIST_USE_TYPE_ALLOCATOR; make them the
 * defaults instead.  Panic on errors, rather than just aborting.
 *
 * Revision 1.7  1996/09/11 23:07:58  geoff
 * Use a type-specific allocator.  Separate lists into dlists and slists.
 *
 * Revision 1.6  1996/05/03  01:31:40  geoff
 * Update the copyright.  Delete an obsolete dummy allocation.
 *
 * Revision 1.5  1996/02/02  18:55:52  geoff
 * Add FinalListDestructor and LIST_OWN_NEW.
 *
 * Revision 1.4  1995/10/10  19:39:46  geoff
 * Get rid of the obsolete memory-leak-tracking stuff.
 *
 * Revision 1.3  1995/09/15  00:13:04  geoff
 * Add abortive memory-leak-catching support.
 *
 * Revision 1.2  1995/04/20  20:19:27  geoff
 * Add hasMember
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

#include "list.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
		DList::DList(const DList& source); // Copy constructor
DList&		DList::operator=(const DList& source); // Assignment operator
void		DList::destructiveCopyFrom(DList& source);
					// Copy efficiently, destroying source
bool		DList::remove(DATA_TYPE* item);
					// Remove chosen item
void		DList::join(DList& that); // Append one list to another
void		DList::popAll();	// Pop elements until empty
unsigned long	DList::count() const;	// Count elements on the list
bool		DList::hasMember(const DATA_TYPE* item) const;
					// Test if item is on the list
void		DList::Iterator::removeCurrent();
					// Zap current w/o affecting iter
		SList::SList(const SList& source); // Copy constructor
SList&		SList::operator=(const SList& source); // Assignment operator
void		SList::destructiveCopyFrom(SList& source);
					// Copy efficiently, destroying source
bool		SList::remove(DATA_TYPE* item);
					// Remove chosen item
void		SList::join(SList& that); // Append one list to another
void		SList::popAll();	// Pop elements until empty
unsigned long	SList::count() const;	// Count elements on the list
bool		SList::hasMember(const DATA_TYPE* item) const;
					// Test if item is on the list
void		SList::Iterator::removeCurrent();
					// Zap current w/o affecting iter
#endif // 0

DList::Element*	DList::Element::freeElements = NULL; // List of freed elements
int		DList::Element::freeCount = 0; // Size of free-element list
TypeAllocator	DList::allocator(sizeof(DList::Element), 5000);
#ifdef LIST_DEBUG_LEAKS
unsigned long	DList::globalElementCreations = 0;
					// No. of elements created, all lists
unsigned long	DList::globalElementDeletions = 0;
					// No. of elements destroyed, all lists
#endif // LIST_DEBUG_LEAKS
SList::Element*	SList::Element::freeElements = NULL; // List of freed elements
int		SList::Element::freeCount = 0; // Size of free-element list
TypeAllocator	SList::allocator(sizeof(SList::Element), 5000);
#ifdef LIST_DEBUG_LEAKS
unsigned long	SList::globalElementCreations = 0;
					// No. of elements created, all lists
unsigned long	SList::globalElementDeletions = 0;
					// No. of elements destroyed, all lists
#endif // LIST_DEBUG_LEAKS

/*
 * Copy constructor for DLists.  See the discussion of the assignment
 * operator below.
 */
DList::DList(
    const DList&		source)	// Source list to copy
#ifdef LIST_DEBUG_LEAKS
    : listHead(this, &listHead, &listHead, NULL),
      elementCreations(0),
      elementDeletions(0)
#else // LIST_DEBUG_LEAKS
    : listHead(&listHead, &listHead, NULL)
#endif // LIST_DEBUG_LEAKS
    {
    *this = source;
    }

/*
 * Assignment operator for DLists.  The copy list contains its own
 * elements, but they point to the same data as the original.  Thus,
 * it is important that the user of this constructor is extremely
 * careful about allocation management.
 */
DList& DList::operator=(
    const DList&		source)	// Source list to copy
    {
    if (this == &source)
	return *this;
    if (mustBeEmptyToDestroy  &&  !isEmpty())
	panic("attempt to assign to non-empty DList");
    popAll();
    mustBeEmptyToDestroy = source.mustBeEmptyToDestroy;
    listHead.next = listHead.previous = &listHead;
    for (DList::Iterator sourceElement((DList&)source);
      sourceElement;
      ++sourceElement)
	pushTail(source[sourceElement]);
    return *this;
    }

/*
 * Copy a DList, destroying the source in the process.  This is more
 * efficient than the non-destructive copy or assignment, because it
 * allows the list elements to be re-used.
 */
void DList::destructiveCopyFrom(
    DList&			source)	// Source list to copy
    {
    if (mustBeEmptyToDestroy  &&  !isEmpty())
	panic("attempt to assign to non-empty DList");
    popAll();
    mustBeEmptyToDestroy = source.mustBeEmptyToDestroy;
    if (!source.isEmpty())
	{
	listHead.next = source.listHead.next;
	source.listHead.next->previous = &listHead;
	listHead.previous = source.listHead.previous;
	source.listHead.previous->next = &listHead;
	source.listHead.previous = source.listHead.next = &source.listHead;
	}
#ifdef LIST_DEBUG_LEAKS
    for (DList::Iterator element(*this);  element;  ++element)
	{
	element.current->list = this;
	elementCreations++;
	source.elementDeletions++;
	}
#endif // LIST_DEBUG_LEAKS
    }

/*
 * Remove a selected item from the list.  Returns true if the item
 * actually existed, false otherwise.
 */
bool DList::remove(			// Remove chosen item
    DATA_TYPE*			item)	// Item to remove
    {
    for (DList::Element* current = listHead.next;
      current != &listHead;
      current = current->next)
	{
#ifdef LIST_DEBUG_LEAKS
	if (this != current->list)
	    panic("leak in DList");
#endif // LIST_DEBUG_LEAKS
	if (current->data == item)
	    {
	    current->previous->next = current->next;
	    current->next->previous = current->previous;
	    current->data = NULL;
	    delete current;
	    return true;
	    }
	}
    return false;
    }

/*
 * Join appends the items of the DList "that" onto the end of the DList
 * "this".  The DList "that" is emptied as a side effect.  The join is
 * done by simple pointer maniplulation unless LIST_DEBUG_LEAKS is
 * defined, in which case "that" must be traversed so that we can
 * maintain the invariants that elementCreations + count() ==
 * elementDeletions and that all list elements point to their list
 * heads.
 */
void DList::join(			// Append one DList to another
    DList&			that)	// DList to be appended
    {
    if (!that.isEmpty())
	{
#ifdef LIST_DEBUG_LEAKS
	for (DList::Element* element = that.listHead.next;
	  element != &that.listHead;
	  element = element->next)
	    {
	    if (element->list != &that)
		panic("leak in DList");
	    that.elementDeletions++;
	    this->elementCreations++;
	    element->list = this;
	    }
#endif // LIST_DEBUG_LEAKS
	listHead.previous->next = that.listHead.next;
	that.listHead.next->previous = listHead.previous;
	listHead.previous = that.listHead.previous;
	listHead.previous->next = &listHead;
	that.listHead.previous = that.listHead.next = &that.listHead;
	}
    }

void DList::popAll()			// Pop elements until empty
    {
    while (!this->isEmpty())
	{
	DList::Element* nextEntry = listHead.next->next;
	listHead.next->data = NULL;
	delete listHead.next;
	listHead.next = nextEntry;
	}
    listHead.previous = &listHead;
    }

unsigned long DList::count()		// Count elements on the list
    const
    {
    unsigned long length = 0;
    for (DList::Element* nextEntry = listHead.next;
      nextEntry != &listHead;
      nextEntry = nextEntry->next)
	length++;
    return length;
    }

bool DList::hasMember(			// Test if item is on the list
    const DATA_TYPE*		item)	// Item to test for
    const
    {
    for (DList::Element* nextEntry = listHead.next;
      nextEntry != &listHead;
      nextEntry = nextEntry->next)
	{
	if (nextEntry->data == item)
	    return true;
	}
    return false;
    }

/*
 * Remove the current element without affecting the validity of the
 * iterator.  This is more efficient than searching the list with an
 * iterator and then calling list.remove().
 */
void DList::Iterator::removeCurrent()
    {
    DList::Element* removedElement = current;
    current = current->next;
    removedElement->previous->next = removedElement->next;
    removedElement->next->previous = removedElement->previous;
    removedElement->data = NULL;
    suppressIncrement = true;
    delete removedElement;
    }

/*
 * Copy constructor for SLists.  See the discussion of the assignment
 * operator below.
 */
SList::SList(
    const SList&		source)	// Source list to copy
#ifdef LIST_DEBUG_LEAKS
    : listHead(NULL),
      elementCreations(0),
      elementDeletions(0)
#else // LIST_DEBUG_LEAKS
    : listHead(NULL)
#endif // LIST_DEBUG_LEAKS
    {
    *this = source;
    }

/*
 * Assignment operator for SLists.  The copy list contains its own
 * elements, but they point to the same data as the original.  Thus,
 * it is important that the user of this constructor is extremely
 * careful about allocation management.
 */
SList& SList::operator=(
    const SList&		source)	// Source list to copy
    {
    if (this == &source)
	return *this;
    if (mustBeEmptyToDestroy  &&  !isEmpty())
	panic("attempt to assign to non-empty SList");
    popAll();
    mustBeEmptyToDestroy = source.mustBeEmptyToDestroy;
    SList::Element* lastElement = NULL;
    for (SList::Iterator sourceElement((SList&)source);
      sourceElement;
      ++sourceElement)
	{
#ifdef LIST_DEBUG_LEAKS
    SList::Element* newElement =
      new SList::Element(this, NULL, source[sourceElement]);
#else // LIST_DEBUG_LEAKS
    SList::Element* newElement =
      new SList::Element(NULL, source[sourceElement]);
#endif // LIST_DEBUG_LEAKS
	if (lastElement == NULL)
	    listHead = newElement;
	else
	    lastElement->next = newElement;
	lastElement = newElement;
	}
    return *this;
    }

/*
 * Copy a SList, destroying the source in the process.  This is more
 * efficient than the non-destructive copy or assignment, because it
 * allows the list elements to be re-used.
 */
void SList::destructiveCopyFrom(
    SList&			source)	// Source list to copy
    {
    if (mustBeEmptyToDestroy  &&  !isEmpty())
	panic("attempt to assign to non-empty SList");
    popAll();
    mustBeEmptyToDestroy = source.mustBeEmptyToDestroy;
    listHead = source.listHead;
    source.listHead = NULL;
#ifdef LIST_DEBUG_LEAKS
    for (SList::Iterator element(*this);  element;  ++element)
	{
	element.current->list = this;
	elementCreations++;
	source.elementDeletions++;
	}
#endif // LIST_DEBUG_LEAKS
    }

/*
 * Remove a selected item from the list.  Returns true if the item
 * actually existed, false otherwise.
 */
bool SList::remove(			// Remove chosen item
    DATA_TYPE*			item)	// Item to remove
    {
    SList::Element* previous = NULL;
    for (SList::Element* current = listHead;
      current != NULL;
      previous = current, current = current->next)
	{
#ifdef LIST_DEBUG_LEAKS
	if (this != current->list)
	    panic("leak in SList");
#endif // LIST_DEBUG_LEAKS
	if (current->data == item)
	    {
	    if (previous == NULL)
		listHead = current->next;
	    else
		previous->next = current->next;
	    current->data = NULL;
	    delete current;
	    return true;
	    }
	}
    return false;
    }

/*
 * Join appends the items of the SList "that" onto the end of the SList
 * "this".  The SList "that" is emptied as a side effect.
 */
void SList::join(			// Append one SList to another
    SList&			that)	// SList to be appended
    {
    if (!that.isEmpty())
	{
	if (listHead == NULL)
	    listHead = that.listHead;
	else
	    {
#ifdef LIST_DEBUG_LEAKS
	    for (SList::Element* element = that.listHead;
	      element != NULL;
	      element = element->next)
		{
		if (element->list != &that)
		    panic("leak in SList");
		that.elementDeletions++;
		this->elementCreations++;
		element->list = this;
		}
#endif // LIST_DEBUG_LEAKS
	    SList::Element* lastElement(listHead);
	    for (  ;
	      lastElement->next != NULL;
	      lastElement = lastElement->next)
		;
	    lastElement->next = that.listHead;
	    }
	that.listHead = NULL;
	}
    }

void SList::popAll()			// Pop elements until empty
    {
    while (!this->isEmpty())
	{
	SList::Element* nextEntry = listHead->next;
	listHead->data = NULL;
	delete listHead;
	listHead = nextEntry;
	}
    }

unsigned long SList::count()		// Count elements on the list
    const
    {
    unsigned long length = 0;
    for (SList::Element* nextEntry = listHead;
      nextEntry != NULL;
      nextEntry = nextEntry->next)
	length++;
    return length;
    }

bool SList::hasMember(			// Test if item is on the list
    const DATA_TYPE*		item)	// Item to test for
    const
    {
    for (SList::Element* nextEntry = listHead;
      nextEntry != NULL;
      nextEntry = nextEntry->next)
	{
	if (nextEntry->data == item)
	    return true;
	}
    return false;
    }

/*
 * Remove the current element without affecting the validity of the
 * iterator.  This is more efficient than searching the list with an
 * iterator and then calling list.remove().
 */
void SList::Iterator::removeCurrent()
    {
    SList::Element* removedElement = current;
    if (previous == NULL)
	target->listHead = current->next;
    else
	previous->next = current->next;
    current = current->next;
    suppressIncrement = true;
    delete removedElement;
    }
