#ifndef _LIST_HH
#define _LIST_HH
/*
 * $Id: list.hh,v 1.15 1999/11/17 07:56:58 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: list.hh,v $
 * Revision 1.15  1999/11/17 07:56:58  geoff
 * Add some missing friendship declarations that the compiler now catches.
 *
 * Revision 1.14  1998/10/13 05:57:30  geoff
 * Make hasMember's argument be const.
 *
 * Revision 1.13  1998/03/05 03:41:31  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.12  1998/02/19 01:03:33  geoff
 * Add support for copying and assignment of lists, including destructive
 * copying.
 *
 * Revision 1.11  1998/01/31 06:30:25  geoff
 * Get rid of LIST_OWN_NEW and LIST_USE_TYPE_ALLOCATOR; make them the
 * defaults instead.
 *
 * Revision 1.10  1996/09/11 23:07:59  geoff
 * Use a type-specific allocator.  Separate lists into dlists and slists.
 *
 * Revision 1.9  1996/06/11  03:08:19  geoff
 * Add a bunch of missing inline declarations.
 *
 * Revision 1.8  1996/05/03  01:31:41  geoff
 * Update the copyright.  Declare operator* as const (for the class).
 *
 * Revision 1.7  1996/02/02  18:55:55  geoff
 * Add FinalListDestructor and LIST_OWN_NEW.
 *
 * Revision 1.6  1995/11/10  03:00:53  geoff
 * Add ListIterator::retarget.  To support this, make the target be a
 * pointer rather than a reference.
 *
 * Revision 1.5  1995/10/10  19:39:48  geoff
 * Get rid of the obsolete memory-leak-tracking stuff.  Turn off
 * LIST_DEBUG_LEAKS.
 *
 * Revision 1.4  1995/09/15  00:13:06  geoff
 * Add abortive memory-leak-catching support.
 *
 * Revision 1.3  1995/05/05  21:45:20  geoff
 * Fix the comment leader.
 *
 * Revision 1.2  1995/04/20  20:18:57  geoff
 * Add hasMember
 *
 * Revision 1.1  1995/03/13  19:55:44  geoff
 * Initial revision
 *
 */

/*
 * Single- and doubly-linked lists, similar to those provided by templates, but
 * using a generic data type.
 */

#include <sys/types.h>
#include "panic.hh"
#include "typealloc.hh"

#ifndef NULL
#define NULL	    0		/* Since we don't include stdio.h */
#endif // NULL

/*
 * Define LIST_DEBUG_LEAKS to add code that's intended to help catch
 * memory leaks.
 */
#undef LIST_DEBUG_LEAKS

/*
 * LIST_MAX_FREE gives the maximum size of the free-element list.
 */
const int		LIST_MAX_FREE = 250;

#ifndef USE_TEMPLATES
#define DATA_TYPE	void
#endif // USE_TEMPLATES

class DList;
class SList;

#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
class DList
    {
    public:
	class			Iterator;
    private:
#ifdef USE_TEMPLATES
	template <DATA_TYPE>
#endif // USE_TEMPLATES
	/*
	 * Class for holding elements of a DList.  This must be
	 * declared here so that an element can be used as the list
	 * head.
	 */
	class Element
	    {
		friend		DList;
		friend		DList::Iterator;
	    private:
		DATA_TYPE*	operator*() const; // Return datum
#ifdef LIST_DEBUG_LEAKS
		inline		Element(DList* containingList,
				  Element* nextElement = NULL,
				  Element* previousElement = NULL,
				  DATA_TYPE* datum = NULL);
#else // LIST_DEBUG_LEAKS
		inline		Element(Element* nextElement = NULL,
				  Element* previousElement = NULL,
				  DATA_TYPE* datum = NULL);
#endif // LIST_DEBUG_LEAKS
		inline		~Element();
				Element(Element&); // Disallow copying
		Element&	operator=(Element&); // Disallow assignment
		inline void*	operator new(size_t size);
						// Allocate a list element
		inline void	operator delete(void* element);
						// Free a single list element
		/*
		 * Private data
		 */
#ifdef LIST_DEBUG_LEAKS
		DList*		list;		// List containing this element
#endif // LIST_DEBUG_LEAKS
		Element*	next;		// Next item on list
		Element*	previous;	// Previous item on list
		DATA_TYPE*	data;		// Data the list stores
		static Element* freeElements;	// List of freed elements
		static int	freeCount;	// Size of free-element list
	    };

	friend		Element;
	friend		Iterator;
    public:
	inline		DList(bool destroyOnlyIfEmpty = true);
			DList(const DList& source); // Copy constructor
	inline		~DList();
	inline bool	isEmpty() const; // Test if list is empty
	inline void	pushHead(DATA_TYPE* item);
					// Add an element to front of list
	inline void	pushTail(DATA_TYPE* item);
					// Add an element to back of list
	inline DATA_TYPE* popHead();	// Remove and return head element
	inline DATA_TYPE* popTail();	// Remove and return tail element
#ifdef USE_TEMPLATES
	inline DATA_TYPE& head() const;	// Return head element
	inline DATA_TYPE& tail() const;	// Return tail element
#else // USE_TEMPLATES
	inline DATA_TYPE* head() const;	// Return head element
	inline DATA_TYPE* tail() const;	// Return tail element
#endif // USE_TEMPLATES
	DList&		operator=(const DList& source); // Assignment operator
	void		destructiveCopyFrom(DList& source);
					// Copy efficiently, destroying source
	bool		remove(DATA_TYPE* item); // Remove chosen item
	void		join(DList& that); // Put that list on tail of this
	void		popAll();	// Pop until empty
	unsigned long	count() const;	// Count elements on the list
	bool		hasMember(const DATA_TYPE* item) const;
					// Test if item is on the list
#ifdef USE_TEMPLATES
	inline DATA_TYPE& operator[](const DList::Iterator& iterator) const;
#else // USE_TEMPLATES
	inline DATA_TYPE* operator[](const DList::Iterator& iterator) const;
#endif // USE_TEMPLATES
    private:
	Element		listHead;	// Dummy element for list head
	bool		mustBeEmptyToDestroy;
					// True if list must be empty to zap it
#ifdef LIST_DEBUG_LEAKS
	unsigned long	elementCreations; // No. of elements created, this list
	unsigned long	elementDeletions; // No. of elements destroyed,
					// ..this list
	static unsigned long 
			globalElementCreations;
					// No. of elements created, all lists
	static unsigned long
			globalElementDeletions;
					// No. of elements destroyed, all lists
#endif // LIST_DEBUG_LEAKS
	static TypeAllocator
			allocator;	// Type-specific allocator for elements
    };

class DList::Iterator
    {
	friend		DList;
    public:
	inline		Iterator(DList& targetList,
			  bool reverse = false);
	inline		~Iterator();
	inline void	retarget(DList& newTarget, bool reverse = false);
					// Point iterator at a new target list
	inline		operator bool() const; // For completion testing
	inline void	reset(bool reverse = false);
					// Reset to start of list
	inline Iterator& operator++();	// Step to next entry
	inline Iterator& operator--();	// Step to previous entry
	void		removeCurrent(); // Zap current w/o affecting iter
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			Iterator(const Iterator&);
	Iterator&	operator=(const Iterator&);
			Iterator();
	DList*		target;		// List we are iterating over
	DList::Element*	current;	// Current entry
	bool		suppressIncrement; // Suppress increment in ++ because
					// ..removeCurrent already did it
    };

#ifdef USE_TEMPLATES
template <DATA_TYPE>
#endif // USE_TEMPLATES
class SList
    {
    public:
	class			Iterator;
    private:
	/*
	 * Class for holding elements of an SList.  This must be
	 * declared here so that an element can be used as the list
	 * head.
	 */
#ifdef USE_TEMPLATES
	template <DATA_TYPE>
#endif // USE_TEMPLATES
	class Element
	    {
		friend		SList;
		friend		SList::Iterator;
	    private:
		DATA_TYPE*	operator*() const; // Return datum
#ifdef LIST_DEBUG_LEAKS
		inline		Element(SList* containingList,
				  Element* nextElement = NULL,
				  DATA_TYPE* datum = NULL);
#else // LIST_DEBUG_LEAKS
		inline		Element(Element* nextElement = NULL,
				  DATA_TYPE* datum = NULL);
#endif // LIST_DEBUG_LEAKS
		inline		~Element();
				Element(Element&); // Disallow copying
		Element&	operator=(Element&); // Disallow assignment
		inline void*	operator new(size_t size);
						// Allocate a list element
		inline void	operator delete(void* element);
						// Free a single list element
		/*
		 * Private data
		 */
#ifdef LIST_DEBUG_LEAKS
		SList*		list;		// List containing this element
#endif // LIST_DEBUG_LEAKS
		Element*	next;		// Next item on list
		DATA_TYPE*	data;		// Data the list stores
		static Element*	freeElements;	// List of freed elements
		static int	freeCount;	// Size of free-element list
	    };
	friend		Element;
	friend		Iterator;
    public:
	inline		SList(bool destroyOnlyIfEmpty = true);
			SList(const SList& source); // Copy constructor
	inline		~SList();
	inline bool	isEmpty() const; // Test if list is empty
	inline void	pushHead(DATA_TYPE* item);
					// Add an element to front of list
	inline DATA_TYPE* popHead();	// Remove and return head element
#ifdef USE_TEMPLATES
	inline DATA_TYPE& head() const;	// Return head element
#else // USE_TEMPLATES
	inline DATA_TYPE* head() const;	// Return head element
#endif // USE_TEMPLATES
	SList&		operator=(const SList& source); // Assignment operator
	void		destructiveCopyFrom(SList& source);
					// Copy efficiently, destroying source
	bool		remove(DATA_TYPE* item); // Remove chosen item
	void		join(SList& that); // Put that list on tail of this
	void		popAll();	// Pop until empty
	unsigned long	count() const;	// Count elements on the list
	bool		hasMember(const DATA_TYPE* item) const;
					// Test if item is on the list
#ifdef USE_TEMPLATES
	inline DATA_TYPE& operator[](const SList::Iterator& iterator) const;
#else // USE_TEMPLATES
	inline DATA_TYPE* operator[](const SList::Iterator& iterator) const;
#endif // USE_TEMPLATES
    private:
	Element*	listHead;	// Pointer to first element on list
	bool		mustBeEmptyToDestroy;
					// True if list must be empty to zap it
#ifdef LIST_DEBUG_LEAKS
	unsigned long	elementCreations; // No. of elements created, this list
	unsigned long	elementDeletions; // No. of elements destroyed,
					// ..this list
	static unsigned long 
			globalElementCreations;
					// No. of elements created, all lists
	static unsigned long
			globalElementDeletions;
					// No. of elements destroyed, all lists
#endif // LIST_DEBUG_LEAKS
	static TypeAllocator
			allocator;	// Type-specific allocator for elements
    };

class SList::Iterator
    {
    friend SList;
    public:
	inline		Iterator(SList& targetList);
	inline		~Iterator();
	inline void	retarget(SList& newTarget);
					// Point iterator at a new target list
	inline		operator bool() const; // For completion testing
	inline void	reset();	// Reset to start of list
	inline Iterator& operator++();	// Step to next entry
	void		removeCurrent(); // Zap current w/o affecting iter
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			Iterator(const Iterator&);
	Iterator&	operator=(const Iterator&);
			Iterator();
	SList*		target;		// List we are iterating over
	SList::Element*	current;	// Current entry
	SList::Element*	previous;	// Previous entry (for removeCurrent)
	bool		suppressIncrement; // Suppress increment in ++ because
					// ..removeCurrent already did it
    };

#include "list.icc"

#endif // _LIST_HH
