#ifndef lint
static const char Rcs_Id[] =
    "$Id: memleak.cc,v 1.1 1995/08/08 22:25:18 geoff Exp $";
#endif

/*
 * $Id: memleak.cc,v 1.1 1995/08/08 22:25:18 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: memleak.cc,v $
 * Revision 1.1  1995/08/08  22:25:18  geoff
 * Initial revision
 *
 */

/*
 * Routines to help detect memory leaks
 */

#include "memleak.hh"
#include <iostream.h>

void			MemoryMarker::clear();
					// Clear all current markings
void			MemoryMarker::mark(const void* area,
			  unsigned long size);
					// Mark an area as reachable
ostream&		MemoryMarker::showHoles(ostream& stream);
					// Show holes on an output stream
void			MemoryMarker::expand(const void* area,
			   unsigned long size);
					// Expand size of marked-bit array

/*
 * Clear all current allocated-memory markings, in preparation for a
 * new accessibility pass.
 */
void MemoryMarker::clear()
    {
    if (markedBits != NULL)
	{
	unsigned long clearSize = maskSize(markerSize);
	while (--clearSize > 0)
	    markedBits[clearSize] = 0;
	markedBits[0] = 0;
	}
    }

/*
 * Mark an area as allocated and reachable
 */
void MemoryMarker::mark(
    const void*		area,		// Address of reachable area
    unsigned long	size)		// Size of the area
    {
    /*
     * Adjust for the malloc overhead.
     */
    area -= MALLOC_EXTRA_BEFORE;
    size += MALLOC_EXTRA_BEFORE + MALLOC_EXTRA_AFTER;
    /*
     * Round the area pointer and size to match the resolution of our
     * bit map.
     */
    size += (unsigned long)area & (BYTES_PER_MARKER_BIT - 1);
    area = (const void*)((unsigned long)area & ~(BYTES_PER_MARKER_BIT - 1));
    size = (size + BYTES_PER_MARKER_BIT - 1) & ~(BYTES_PER_MARKER_BIT - 1);
    /*
     * Make sure the bit map has room to map this area.
     */
    if (area < markerBase  ||  area + size >= markerBase + markerSize)
	expand(area, size);
    /*
     * Turn on every bit covered by the area.  Note that this could be
     * optimized quite a bit, by batching bit sets of full longwords.
     * However, I expect most malloc'ed areas to be small, so it's not
     * worth the effort of writing the code unless profiling proves
     * this to be a problem.
     */
    size /= BYTES_PER_MARKER_BIT;
    for (unsigned long areaBit = (area - markerBase) / BYTES_PER_MARKER_BIT;
      size > 0;
      size--, areaBit++)
	markedBits[areaBit / BITS_PER_LONG] |=
	  1 << (areaBit & (BITS_PER_LONG - 1));
    }

/*
 * Show the unreachable holes in memory on an output stream
 */
ostream& MemoryMarker::showHoles(
    ostream&		stream)		// Stream to write to
    {
    unsigned long loopLimit = markerSize / BYTES_PER_MARKER_BIT;
    stream << "Holes between " << markerBase << " and "
      << markerBase + markerSize << ":\n";
    for (unsigned long holeBase = 0;  holeBase < loopLimit;  )
	{
	if (markedBits[holeBase / BITS_PER_LONG]
	  & (1 << (holeBase & (BITS_PER_LONG - 1))))
	    holeBase++;
	else
	    {
	    /*
	     * We've found the beginning of an unmarked hole.  Find
	     * the extent, and then print it.
	     */
	    for (unsigned long holeEnd = holeBase + 1;
	      holeEnd < loopLimit;
	      holeEnd++)
		{
		if (markedBits[holeEnd / BITS_PER_LONG]
		  & (1 << (holeEnd & (BITS_PER_LONG - 1))))
		    break;
		}
	    stream << "Hole from "
	      << (markerBase + holeBase * BYTES_PER_MARKER_BIT)
	      << " to "
	      << (markerBase + (holeEnd - holeBase) * BYTES_PER_MARKER_BIT)
	      << " (" << (holeEnd - holeBase) * BYTES_PER_MARKER_BIT
	      << " bytes).\n";
	    if (markerBase + holeBase <= (void*)markedBits
	      &&  markerBase + holeEnd >=
		(void*)markedBits + maskSize(markerSize))
		stream << "  (includes marker area at " << markedBits
		  << " to " << markedBits + maskSize(markerSize) << ".)\n";
	    holeBase = holeEnd;
	    }
	}
    return stream;
    }

/*
 * Expand the size of the markedBits array to be large enough to
 * encompass the given block.
 */
void MemoryMarker::expand(
    const void*		area,		// Area to be contained
    unsigned long	size)		// Size of the area
    {
    /*
     * Start by rounding the area pointer and size so that they cover
     * an integral number of blocks.
     */
    size += (unsigned long)area & (BLOCK_ROUNDING - 1);
    area = (const void*)((unsigned long)area & ~(BLOCK_ROUNDING - 1));
    size = (size + BLOCK_ROUNDING - 1) & ~(BLOCK_ROUNDING - 1);
    if (markedBits == NULL)
	{
	markerBase = area;
	markerSize = size;
	unsigned long unitsToClear = maskSize(size);
	markedBits = new unsigned long[unitsToClear];
	for (unsigned long i = 0;  i < unitsToClear;  i++)
	    markedBits[i] = 0;
	return;
	}
    else
	{
	/*
	 * We must expand the marked area downwards, upwards, or both.
	 * This involves allocating new space, copying the information
	 * from the old space, and clearing the new mask bits.
	 */
	const void* newStart = markerBase;
	if (area < markerBase)
	    newStart = area;
	const void* newEnd = markerBase + markerSize;
	if (area + size > newEnd)
	    newEnd = area + size;
	size = newEnd - newStart;
	unsigned long* newMarkedBits = new unsigned long[maskSize(size)];
	unsigned long i = 0;
	if (newStart < markerBase)
	    {
	    unsigned long unitsToClear = maskSize(markerBase - newStart);
	    for (i = 0;  i < unitsToClear;  i++)
		newMarkedBits[i] = 0;
	    }
	unsigned long unitsToCopy = maskSize(markerSize);
	for (unsigned long j = 0;  j < unitsToCopy;  i++, j++)
	    newMarkedBits[i] = markedBits[j];
	for (unsigned long clearLimit = maskSize(size);  i < clearLimit; i++)
	    newMarkedBits[i] = 0;
	delete[] markedBits;
	markedBits = newMarkedBits;
	markerSize = size;
	markerBase = newStart;
	return;
	}
    }
