#ifndef _SIMPLESTRING_HH
#define _SIMPLESTRING_HH
/*
 * $Id: simplestring.hh,v 1.4 2001/11/08 08:18:21 geoff Exp geoff $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: simplestring.hh,v $
 * Revision 1.4  2001/11/08 08:18:21  geoff
 * Add new operators: construction from a character, non-assignment
 * concatenation, and the rest of the comparisons.
 *
 * Revision 1.3  2000/11/30 06:41:38  geoff
 * Make the length-specified constructor be explicit.
 *
 * Revision 1.2  1998/10/23 23:03:28  geoff
 * Fix things to compile with -pedantic.  Add some new functions.  Make a
 * number of things non-inline.
 *
 * Revision 1.1  1997/12/04 23:37:40  geoff
 * Initial revision
 *
 */

/*
 * Simple string class.  This provides a clean fundamental
 * implementation of strings, without all the excess baggage dragged
 * in by classes that thing basic strings should include complex
 * facilities like regular-expression searching.
 *
 * A WriteableSimpleString is just like a SimpleString, except that
 * the subscripting operator can return pointer to a non-const
 * character if the string itself isn't const.
 */

#include <iostream.h>
extern "C"
    {
#include <string.h>
    }
#include "panic.hh"
#include "typealloc.hh"

#ifndef NULL
#define NULL	    0		/* Since we don't include stdio.h */
#endif // NULL

class SimpleString;
class WritableSimpleString;

class SimpleString
    {
	friend		WritableSimpleString;
    public:
	/*
	 * A note on constructors: the syntax SimpleString(NULL) will
	 * generate a compilation error, because it can't decide
	 * whether to use SimpleString(unsigned int size = 0) or
	 * SimpleString(const char* initialValue = 0, unsigned int size
	 * = 0).  The solution is to use one of
	 * SimpleString((char*)NULL), SimpleString(NULL, 0), or just
	 * SimpleString with no arguments.
	 */
	inline explicit	SimpleString(unsigned int size = 0);
	inline		SimpleString(const char* initialValue,
			  unsigned int size = 0);
	inline		SimpleString(char initialValue, unsigned int size = 0);
			SimpleString(const SimpleString& initialValue,
			  unsigned int size = 0);
	inline		~SimpleString();
	inline void*	operator new(size_t size);
					// Allocate a string
	inline void	operator delete(void* string);
					// Free a string
	/*
	 * Simple strings provide the following operators:
	 *
	 *  (const char *)	Return the null-terminated value of the string.
	 *  =			Assign a new value.
	 *  +=			Concatenate a value to the end (strcat).
	 *  +			Concatenate a value to the end.
	 *  ==			Compare strings.  Comes in 2 versions,
	 *			to allow comparison against
	 *			SimpleStrings or character strings.
	 *  !=			Compare strings.  Again, comes in 2
	 *			versions.
	 *  []			Provide individual-character access.
	 */
	inline		operator const char*() const;
	SimpleString&	operator=(const SimpleString& source);
	SimpleString&	operator=(const char* source);
	SimpleString&	operator+=(const SimpleString& source);
	SimpleString&	operator+=(const char* source);
	SimpleString&	operator+=(char source);
	SimpleString	operator+(const SimpleString& rhs);
	SimpleString	operator+(const char* rhs);
	SimpleString	operator+(char source);
	inline bool	operator==(const SimpleString& right) const;
	inline bool	operator==(const char* right) const;
	inline bool	operator!=(const SimpleString& right) const;
	inline bool	operator!=(const char* right) const;
	inline bool	operator<(const SimpleString& right) const;
	inline bool	operator<(const char* right) const;
	inline bool	operator<=(const SimpleString& right) const;
	inline bool	operator<=(const char* right) const;
	inline bool	operator>(const SimpleString& right) const;
	inline bool	operator>(const char* right) const;
	inline bool	operator>=(const SimpleString& right) const;
	inline bool	operator>=(const char* right) const;
	inline const char&
			operator[](unsigned int i) const;
	inline const char&
			operator[](int i) const;
	/*
	 * The following non-operator functions provide important libc
	 * functionality.
	 */
	inline unsigned int
			strlen() const;
	/*
	 * The following non-operator functions provide basic string
	 * manipulation.
	 */
	SimpleString	subString(unsigned int start, unsigned int subLength)
			  const;
					// Return substring
	/*
	 * Other non-operator functions.
	 */
	void		truncate(bool force = false);
					// Truncate and free up space
	/*
	 * I/O functions.  Some of these are similar to the functions in
	 * iostream.h.
	 */
	inline bool	get(istream& stream, char terminator = '\n');
					// Read a line, pushing newline back
	inline bool	gets(istream& stream, char terminator = '\n');
					// Read a line, discarding newline
    private:
	/*
	 * Internal functions
	 */
	enum SimpleStringTerminatorAction
	    {
	    GET_PUT_BACK,		// Put terminator back on stream
	    GET_DISCARD			// Discard terminator
	    };
					// What to do with terminator
	bool		underlyingGet(istream& stream, char terminator,
			  SimpleStringTerminatorAction action);
					// Do get/gets
	/*
	 * Parameters of the implementation.  Should be const ints,
	 * but ANSI C++ prevents this, so I'll use an enum to kludge
	 * it.
	 */
	enum
	    {
	    TRUNCATION_THRESHOLD = 16,	// Don't truncate unless you
					// can free up at least this
					// many bytes
	    OVERHEAD_SIZE = 1 + 1,	// Number of overhead bytes to
					// allocate: 1 for the null
					// terminator, and 1 for the
					// reference count
	    REFERENCE_OFFSET = 1,	// Value to add to maxSize to
					// get offset of reference
					// count
	    MAXIMUM_REFERENCES = (char)255
					// Maximum value reference
					// count can reach
	    };
	void		clone(unsigned int newSize) const;
					// Clone the value field
	inline void forgetValue(bool preserveLength);
					// Remove a reference to the value
	/*
	 * Private data.
	 *
	 * A SimpleString is represented by a null-terminated string,
	 * pointed to from this data structure.  The length is kept
	 * for efficiency; the maximum size records the amount of
	 * storage allocated (this can be used to make concatenation,
	 * etc., more efficient if the final length is known in
	 * advance).
	 *
	 * To use storage more efficiently, we use reference counting.
	 * The reference count is kept in the final byte of the value.
	 * Since this limits us to 255 references, the value will be
	 * cloned if the reference count overflows.  Note that this
	 * can cause inefficiency in some applications.  For example,
	 * the following loop will create 1000 extra copies of the string
	 * "source", each with a reference count of 1, as well as the
	 * original copy which will have 255 references:
	 *
	 *	SimpleString source("source");
	 *	SimpleString dest[1255];
	 *	for (int i = 0;  i < 1255;  i++)
	 *	    dest[i] = source;
	 */
	unsigned int	length;		// Length of the string
	mutable unsigned int
			maxSize;	// Maximum length string can hold
	mutable char*	value;		// Value of the string
	static TypeAllocator
			allocator;	// Type-specific allocator for strings
    };

class WritableSimpleString : public SimpleString
    {
    public:
	inline		WritableSimpleString(unsigned int size = 0);
	inline		WritableSimpleString(const char* initialValue,
			  unsigned int size = 0);
	inline		WritableSimpleString(const SimpleString& initialValue,
			  unsigned int size = 0);
	inline		~WritableSimpleString();
	inline char&	operator[](unsigned int i);
	inline char&	operator[](int i);
	inline const char&
			operator[](unsigned int i) const;
	inline const char&
			operator[](int i) const;
    };

#include "simplestring.icc"

#endif // _SIMPLESTRING_HH
