#ifndef _TYPEALLOC_HH
#define _TYPEALLOC_HH
/*
 * $Id: typealloc.hh,v 1.3 1998/10/23 23:03:30 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: typealloc.hh,v $
 * Revision 1.3  1998/10/23 23:03:30  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.2  1998/03/05 03:41:32  geoff
 * Convert to use private embedded classes for helpers.
 *
 * Revision 1.1  1997/07/16  22:16:45  geoff
 * Initial revision
 *
 */

/*
 * Generic support for type-specific allocation.  This is useful for
 * types that need to allocate lots of constant-size small instances,
 * because it eliminates most of the normal allocator overhead.  The
 * drawbacks are that freed memory is much less likely to be returned
 * to the heap, and there is no support for allocating arrays of
 * elements (although the latter could be added if necessary).
 */

#ifndef NULL
#define NULL	    0		/* Since we don't include stdio.h */
#endif // NULL

class TypeAllocator;

/*
 * The type-specific allocator class itself.  There should be
 * one instance of this class for each distinct type that uses the
 * allocator.  The class holds information specific to the client
 * type, plus critical pointers used to locate allocatable memory.
 *
 * An important point is that free areas are kept sorted by address.
 * This fact is used in the freeElement routine.
 *
 * Destroying an instance of this class also returns all memory
 * allocated by it to the heap.  Be *very* careful!  In most cases,
 * you'll want to allocate class instances statically.
 */
class TypeAllocator
    {
	/*
	 * Private embedded classes:
	 */
    private:
	/*
	 * Class used to link together free elements within a free
	 * area.  This is really more of a struct rather than a class.
	 * The constructor and destructor are deliberately omitted,
	 * for efficiency.
	 */
	class FreeElement
	    {
		friend		TypeAllocator;
	    private:
		FreeElement*	next;	// Next free element in list
	    };

	/*
	 * Class used to group large numbers of free elements into an
	 * allocation area.  When we need to get memory from the heap,
	 * we grab a chunk large enough to hold a given number of
	 * elements (see TypeAllocator) and then link them together
	 * into a list.  When an element is freed, it is returned to
	 * the list within its allocation area.  If all elements in
	 * the area ever become free, we drop the entire area.
	 */
	class FreeArea
	    {
 		friend		TypeAllocator;
	    private:
				FreeArea();
				~FreeArea();
		/*
		 * Private data
		 */
		FreeArea*	next;	// Next area in allocator list
		FreeArea*	previous;
					// Previous area in allocator list
		unsigned long	elementCount;
					// Count of all elements in list
		unsigned long	freeCount;
					// Count of free elements in list
		void*		areaEnd; // Last element in this area
		FreeElement*	elements;
					// List of available elements
		/*
		 * The space for the elements themselves is allocated here by
		 * over-allocating space for the free area.  "firstElement" is
		 * just a handle for referring to the location.
		 */
		unsigned long	firstElement;
	    };
	/*
	 * Public declarations for the type allocator class.
	 */
    public:
	/*
	 * Constants used by the allocator.  Should be const ints, but
	 * ANSI C++ prevents this, so I'll use an enum to kludge it.
	 */
	enum
	    {
			ALIGNMENT = sizeof(void*),
					// Alignment for allocated blocks
			MALLOC_OVERHEAD = 2 * sizeof(void*),
					// Overhead added by ::operator new()
					// ..(guess too big if unsure)
			PAGE_SIZE = 1 << 14
					// Size of a hardware page (guessing
					// ..too big is better than too small)
	    };
			TypeAllocator(unsigned long normalElementSize,
			  unsigned long chunkSize = 1000);
			~TypeAllocator();
	void*		allocElement();	// Allocate an element
	void		freeElement(void* element);
					// Return an element to the allocator
    private:
	unsigned long	elementSize;	// Size of an element of this type
	unsigned long	allocationIncrement;
					// Number of elements to alloc at once
	FreeArea*	firstArea;	// First area in allocator list
	FreeArea*	lastArea;	// Last area in allocator list
	FreeArea*	areaWithFirstFreeElement;
					// Where to find a free element
	FreeArea*	areaWithLastFreedElement;
					// Where last free was done
    };

#endif // _TYPEALLOC_HH
