#ifndef lint
static char Rcs_Id[] =
    "$Id: mtseeds.c,v 1.5 2003/09/16 05:19:40 geoff Exp $";
#endif

/*
 * Special program to find seeds for the Mersenne Twist algorithm.
 * The code doesn't actually generate random numbers, it just cycles
 * through the refresh code.  Occasionally, it saves the current state
 * of the PRNG to a file.
 *
 * Note that the save interval given by the user is multiplied by a
 * constant (currently 1 million), and refers to the number of state
 * refreshes, not the number of random values.  The number of random
 * values between saves is MT_STATE_SIZE times the requested save
 * interval.  Thus, to get an actual save interval of about 1 trillion
 * random numbers, use a specified save interval of 1602.
 *
 * $Log: mtseeds.c,v $
 * Revision 1.5  2003/09/16 05:19:40  geoff
 * Declare mt_default_state explicitly, since without inlining its not
 * declared in the header file.
 *
 * Revision 1.4  2001/08/18 02:11:30  geoff
 * Add control of the maximum number of output lines.  Change the code so
 * that the save interval is a refresh count, not a random-number count,
 * and change the interval base to a million to (roughly) compensate.
 * Add a bit more header commentary.
 *
 * Revision 1.3  2001/08/18 01:26:40  geoff
 * Don't abort if fdatasync fails (e.g., on writing to a pipe), and ifdef
 * the fdatasync code on having Posix I/O.  Don't generate an output
 * state that's identical to the input state.  Get rid of the obsolete
 * if'ed-out state-generation code.
 *
 * Revision 1.2  2001/06/21 08:13:46  geoff
 * Change the interval base to a billion, and make sure it's computed as a
 * long long.
 *
 * Revision 1.1  2001/06/21 08:11:27  geoff
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "mtwist.h"

/*
 * Table of contents:
 */
int			main(int argc, char* argv[]);

/*
 * Default 32-bit random seed used by mtwist.c.
 */
#define DEFAULT_SEED32	4357

/*
 * mtwist internal variable that we muck with.
 */
extern mt_state		mt_default_state;
					/* State of the default generator */
/*
 * Base unit for save count, so user can think in a convenient unit.
 */
#define INTERVAL_BASE	((unsigned long long)1000000)

int main(int argc, char* argv[])
    {
    unsigned long long	max_outputs = 0; /* Max output lines to create */
    unsigned long long	save_interval;	/* How often to save state */
    unsigned long long	statecount;	/* Number of states generated */
#ifdef _POSIX_SYNCHRONIZED_IO
    int			syncfailed = 0;	/* NZ if datasync has failed */
#endif /* _POSIX_SYNCHRONIZED_IO */

    if (argc != 2  &&  argc != 3)
	{
	(void) fprintf (stderr, "Usage: mtseeds save-interval [max_lines]\n");
	(void) fprintf (stderr, "Original seed state comes from stdin.\n");
	return 2;
	}

    if (!mt_loadstate(stdin))
	mt_seed32(DEFAULT_SEED32);	/* No seed given, use default */

    save_interval = (unsigned long long)atoll(argv[1]) * INTERVAL_BASE;

    if (argc > 2)
	max_outputs = (unsigned long long)atoll(argv[2]);

    while (1)
	{
	for (statecount = 0;  statecount < save_interval;  statecount++)
	    mts_refresh(&mt_default_state);

	if (!mt_savestate(stdout))
	    {
	    (void) fprintf (stderr, "mt_savestate failed.\n");
	    return 1;
	    }
	if (fflush(stdout))
	    {
	    (void) fprintf (stderr, "fflush failed.\n");
	    return 1;
	    }
#ifdef _POSIX_SYNCHRONIZED_IO
	if (fdatasync(fileno(stdout)) < 0)
	    {
	    if (!syncfailed)
		(void) fprintf (stderr, "fdatasync failed.\n");
	    syncfailed = 1;
	    }
#endif /* _POSIX_SYNCHRONIZED_IO */

	if (max_outputs != 0)
	    {
	    if (--max_outputs == 0)
		break;
	    }
	}
    return 0;
    }
