#ifndef lint
static char Rcs_Id[] =
    "$Id: rdcctest.cc,v 1.3 2004/03/30 07:29:43 geoff Exp $";
#endif

/*
 * Test the random-distribution library.  Usage:
 *
 * rdcctest seed how_many distribution [params...]
 *
 * where:
 *
 *	seed	is the random seed.  If seed is zero, the package's
 *		automatic seed generator is used.
 *	how_many
 *		is how many random numbers to generate.
 *	distribution
 *		is the distribution to draw from (see below).
 *	params	are the parameters to the distribution (see below).
 *
 * Distributions supported:
 *
 *   iuniform	A uniform distribution of integers on the interval (p1, p2].
 *   uniform	A uniform distribution on the interval (p1, p2].
 *   exponential	Exponential with mean p1, default 1.
 *   erlang	p1-Erlang with mean p2.
 *   weibull	Weibull with shape parameter p1 and scale parameter p2.
 *   normal	Normal with mean p1 and standard deviation p2.
 *   lognormal	Lognormal with scale parameter p1 and shape parameter p2.
 *   triangular	Triangular on the interval (p1, p2) with mode at p3.
 *   empirical	p1 with probability p2, p3 with probability p4, ...,
 *		p(2n+1) with probability 1-p2-p4-...-p(2n)
 *
 * $Log: rdcctest.cc,v $
 * Revision 1.3  2004/03/30 07:29:43  geoff
 * Document the program better, and allow the random seed to be
 * controlled for verification testing.
 *
 * Revision 1.2  2003/09/11 05:50:53  geoff
 * Use the standard namespace.
 *
 * Revision 1.1  2001/06/18 10:09:35  geoff
 * Initial revision
 *
 */

#include "randistrs.h"
#include <iostream>
#include <stdlib.h>

using namespace std;

int			main(int argc, char* argv[]);
static void		usage();

#define SEED_PARAM	1		/* Offset of seed param in argv */
#define COUNT_PARAM	2		/* Offset of count param in argv */
#define DISTR_PARAM	3		/* Offset of distr param in argv */
#define PARAM_OFFSET	4		/* Offset of params in argv */

int main(
    int			argc,		/* Argument count */
    char*		argv[])		/* Argument vector */
    {
    if (argc <= PARAM_OFFSET)
	usage();

    int seed = atoi (argv[SEED_PARAM]);
    int how_many = atoi(argv[COUNT_PARAM]);

    int n_params = argc - PARAM_OFFSET;
    double* params = new double[n_params];

    int			needed_params = 0;
					/* Number of params needed by distr */
    int			n_probs = 0;	/* Number of empirical probabilites */
    double*		probs = NULL;	/* Probabilities for empirical */
    double*		values = NULL;	/* Values for empirical */

    for (int i = 0;  i < n_params;  i++)
	params[i] = atof(argv[i + PARAM_OFFSET]);

    if (strcmp(argv[DISTR_PARAM], "iuniform") == 0)
	needed_params = 2;
    else if (strcmp(argv[DISTR_PARAM], "uniform") == 0)
	needed_params = 2;
    else if (strcmp(argv[DISTR_PARAM], "exponential") == 0)
	needed_params = 1;
    else if (strcmp(argv[DISTR_PARAM], "erlang") == 0)
	{
	if (n_params < 2  ||  params[0] < 1.0)
	    usage();
	needed_params = 2;
	}
    else if (strcmp(argv[DISTR_PARAM], "weibull") == 0)
	needed_params = 2;
    else if (strcmp(argv[DISTR_PARAM], "normal") == 0)
	needed_params = 2;
    else if (strcmp(argv[DISTR_PARAM], "lognormal") == 0)
	needed_params = 2;
    else if (strcmp(argv[DISTR_PARAM], "triangular") == 0)
	needed_params = 3;
    else if (strcmp(argv[DISTR_PARAM], "empirical") == 0)
	{
	if (n_params % 2 == 0  ||  n_params < 3)
	    usage();

	n_probs = n_params / 2;
	probs = new double[n_probs];
	values = new double[n_probs + 1];

	for (int i = 0;  i < n_probs;  i++)
	    {
	    values[i] = params[i * 2];
	    probs[i] = params[i * 2 + 1];
	    if (i != 0)
		probs[i] += probs[i - 1];
	    }

	values[n_probs] = params[n_probs * 2];
	needed_params = n_params;
	}
    else
	usage();

    if (n_params != needed_params)
	usage();

    /*
     * Create a generator and seed it.
     */
    mt_distribution distr(seed == 0);
    if (seed != 0)
	distr.seed32(seed);

    for (int i = 0;  i < how_many;  i++)
	{
	double ran_value = 0.0;
	if (strcmp(argv[DISTR_PARAM], "iuniform") == 0)
	    ran_value = distr.iuniform((long)params[0], (long)params[1]);
	else if (strcmp(argv[DISTR_PARAM], "uniform") == 0)
	    ran_value = distr.uniform(params[0], params[1]);
	else if (strcmp(argv[DISTR_PARAM], "exponential") == 0)
	    ran_value = distr.exponential(params[0]);
	else if (strcmp(argv[DISTR_PARAM], "erlang") == 0)
	    ran_value = distr.erlang((int) params[0], params[1]);
	else if (strcmp(argv[DISTR_PARAM], "weibull") == 0)
	    ran_value = distr.weibull(params[0], params[1]);
	else if (strcmp(argv[DISTR_PARAM], "normal") == 0)
	    ran_value = distr.normal(params[0], params[1]);
	else if (strcmp(argv[DISTR_PARAM], "lognormal") == 0)
	    ran_value = distr.lognormal(params[0], params[1]);
	else if (strcmp(argv[DISTR_PARAM], "triangular") == 0)
	    ran_value = distr.triangular(params[0], params[1], params[2]);
	else if (strcmp(argv[DISTR_PARAM], "empirical") == 0)
	    ran_value = distr.empirical(n_probs, values, probs);
	cout << ran_value << '\n';
	}
    return 0;
    }

static void usage()
    {
    cerr << "Usage: rdtest count distribution p0 ...\n";
    cerr << "Distributions:\n";
    cerr << "\tiuniform lower upper\n";
    cerr << "\tuniform lower upper\n";
    cerr << "\texponential mean\n";
    cerr << "\terlang p mean\n";
    cerr << "\tweibull shape scale\n";
    cerr << "\tnormal mean sigma\n";
    cerr << "\tlognormal shape scale\n";
    cerr << "\ttriangular lower upper mode\n";
    cerr << "\tempirical v0 p0 v1 p1 ... vn-1 pn-1 vn\n";
    exit(2);
    }
