#ifndef lint
static const char Rcs_Id[] =
    "$Id: byteswappedstream.cc,v 1.13 1998/10/23 23:14:54 geoff Exp $";
#endif

/*
 * $Id: byteswappedstream.cc,v 1.13 1998/10/23 23:14:54 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: byteswappedstream.cc,v $
 * Revision 1.13  1998/10/23 23:14:54  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.12  1996/10/11 22:08:51  geoff
 * Only set the endedness if the read succeeds, so that zero-length files
 * get set to the caller's chosen endedness.
 *
 * Revision 1.11  1996/05/03 01:11:28  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/02/02  20:09:38  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.9  1996/02/02  18:28:41  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.8  1995/11/10  02:58:16  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.7  1995/10/10  20:37:41  geoff
 * Disable the default constructor and the open routine.
 *
 * Revision 1.6  1995/09/15  00:04:37  geoff
 * Change the RCS ID to be a const.  Add compressed-stream support by
 * deriving from the CompressedStream class.
 *
 * Revision 1.5  1995/06/17  00:10:49  geoff
 * Flush cerr/cout before all aborts.
 *
 * Revision 1.4  1995/05/23  00:38:54  geoff
 * Depend on compilation options.  Change a bunch of exits into aborts so
 * I'll get a core dump to look at.
 *
 * Revision 1.3  1995/05/05  22:40:55  geoff
 * Swap the endedness and protection arguments to the constructors.  Add
 * readEndedness and writeEndedness.
 *
 * Revision 1.2  1995/03/29  21:03:13  geoff
 * Numerous debugging changes.  It now works.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cmdstream.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			ByteSwappedStream::ByteSwappedStream(
			  int fd, Endedness desiredEndedness, int childPID);
			ByteSwappedStream::ByteSwappedStream(
			  const char* name, int mode,
			  Endedness desiredEndedness, int protection);
ByteSwappedStream::Endedness ByteSwappedStream::getProcessorEndedness();
ByteSwappedStream&	ByteSwappedStream::writeEndedness();
					// Write chosen endedness to stream
ByteSwappedStream&	ByteSwappedStream::readEndedness();
					// Read and set chosen endedness
void			ByteSwappedStream::findEndedness();
					// Intuit processor endedness
ByteSwappedStream&	ByteSwappedStream::swappedRead(void* buffer,
			  unsigned int typeSize, int numberOfBytes);
ByteSwappedStream&	ByteSwappedStream::swappedWrite(const void* buffer,
			  unsigned int typeSize, int numberOfBytes);
#endif // 0

/*
 * Size of largest intermediate buffer that will be allocated by
 * swappedWrite.
 */
const int		bufferLimit = 8192;

/*
 * Processor endedness.  This is initialized to naturalEndedness, and
 * then modified by the findEndedness routine the first time a stream
 * is opened.
 */
ByteSwappedStream::Endedness ByteSwappedStream::processorEndedness =
			  naturalEndedness;

extern "C" void		exit(int status);

ByteSwappedStream::ByteSwappedStream(	// Constructor
    int			fd,		// Fd that file is already open on
    Endedness		desiredEndedness, // Endedness to use for the file
    int			childPID)	// PID of child process, if any
    : CompressedStream(fd, childPID)
    {
    findEndedness();
    if (desiredEndedness == naturalEndedness)
	fileEndedness = processorEndedness;
    else
	fileEndedness = desiredEndedness;
    }

ByteSwappedStream::ByteSwappedStream(	// Constructor
    const char*		name,		// Name of file to open for stream
    int			mode,		// Open mode (options) to use
    Endedness		desiredEndedness, // Endedness to use for the file
    int			protection)	// Protection to use if creating
    : CompressedStream(name, mode, protection)
    {
    findEndedness();
    if (desiredEndedness == naturalEndedness)
	fileEndedness = processorEndedness;
    else
	fileEndedness = desiredEndedness;
    }

ByteSwappedStream::Endedness ByteSwappedStream::getProcessorEndedness()
    {
    findEndedness();
    return processorEndedness;
    }

/*
 * Write chosen endedness to stream, for reading by readEndedNess()
 */
ByteSwappedStream& ByteSwappedStream::writeEndedness()
    {
    char endedness = (int) fileEndedness;
    return write(&endedness, sizeof endedness);
    }

/*
 * Read and set chosen endedness.  Must be called before any other I/O
 * is done.
 */
ByteSwappedStream& ByteSwappedStream::readEndedness()
    {
    char endedness;
    if (read(&endedness, sizeof endedness))
	fileEndedness = (Endedness) endedness;
    return *this;
    }

/*
 * Determine the endedness of the host processor.  We use tricks with
 * unions to figure out how bytes are stored.  This may not work on
 * all architectures.  This routine assumes that machines are strict
 * about being either big- or little-endian; historic "mixed"
 * architectures such as some PDP-11 data types are not supported.
 * Also, only architectures in which longs are either 32 or 64 bits
 * are supported.  If a violation of these rules is detected, a
 * message is issued to cerr and the program is terminated.
 */

void ByteSwappedStream::findEndedness()	// Intuit processor endedness
    {
    if (processorEndedness != naturalEndedness)
	return;
    union
	{
	char		charVersion[sizeof(long)];
					// Buffer for byte-wise access to long
	long		longVersion;	// Long version of buffer
	}
			deductionUnion; // Union used to deduce endedness
    /*
     * Ignore any warnings about lost precision in the following statement.
     */
    deductionUnion.longVersion = 0x0102030405060708L;
    if (deductionUnion.charVersion[0] == 0x01)
	{
	/*
	 * 32-bit or 64-bit big-endian architecture.  Verify validity.
	 */
	if (deductionUnion.charVersion[1] != 0x02
	  ||  deductionUnion.charVersion[2] != 0x03
	  ||  deductionUnion.charVersion[3] != 0x04
	  ||  (sizeof(long) > 4
	    && (deductionUnion.charVersion[4] != 0x05
	      ||  deductionUnion.charVersion[5] != 0x06
	      ||  deductionUnion.charVersion[6] != 0x07
	      ||  deductionUnion.charVersion[7] != 0x08)))
	    {
	    // NEEDSWORK:  should throw exception
	    panic("findEndedness:  mixed-architecture endedness violation");
	    }
	processorEndedness = bigEnded;
	}
    else if (deductionUnion.charVersion[0] == 0x05)
	{
	/*
	 * 32-bit big-endian architecture.  Verify validity.
	 */
	if (deductionUnion.charVersion[1] != 0x06
	  ||  deductionUnion.charVersion[2] != 0x07
	  ||  deductionUnion.charVersion[3] != 0x08
	  ||  sizeof(long) > 4)
	    {
	    // NEEDSWORK:  should throw exception
	    panic("findEndedness:  mixed-architecture endedness violation");
	    }
	processorEndedness = bigEnded;
	}
    else if (deductionUnion.charVersion[0] == 0x08)
	{
	/*
	 * 32-bit or 64-bit little-endian architecture.  Verify validity.
	 */
	if (deductionUnion.charVersion[1] != 0x07
	  ||  deductionUnion.charVersion[2] != 0x06
	  ||  deductionUnion.charVersion[3] != 0x05
	  ||  (sizeof(long) > 4
	    && (deductionUnion.charVersion[4] != 0x04
	      ||  deductionUnion.charVersion[5] != 0x03
	      ||  deductionUnion.charVersion[6] != 0x02
	      ||  deductionUnion.charVersion[7] != 0x01)))
	    {
	    // NEEDSWORK:  should throw exception
	    panic("findEndedness:  mixed-architecture endedness violation");
	    }
	processorEndedness = littleEnded;
	}
    else if (deductionUnion.charVersion[0] == 0x04)
	{
	/*
	 * 32-bit little-endian architecture.  Verify validity.
	 */
	if (deductionUnion.charVersion[1] != 0x03
	  ||  deductionUnion.charVersion[2] != 0x02
	  ||  deductionUnion.charVersion[3] != 0x01
	  ||  sizeof(long) > 4)
	    {
	    // NEEDSWORK:  should throw exception
	    panic("findEndedness:  mixed-architecture endedness violation");
	    }
	processorEndedness = littleEnded;
	}
    else
	{
	/*
	 * Unknown architecture.  Reject.
	 */
	// NEEDSWORK:  should throw exception
	panic("findEndedness:  mixed-architecture endedness violation");
	}
    }

/*
 * Perform a byte-swapped read.  First the data is read into the
 * buffer; then a second pass is made (if necessary) to actually do
 * the byte swapping.
 *
 * LIMITATIONS:  typeSize must be even.
 */
ByteSwappedStream& ByteSwappedStream::swappedRead(
    void*		buffer,		// Buffer to read into
    unsigned int	typeSize,	// Size of buffer's underlying type
    int			numberOfBytes)	// Number of bytes to read
    {
    CommandStream::read(buffer, numberOfBytes);
    if (typeSize != 1  &&  processorEndedness != fileEndedness)
	{
	/*
	 * Swap the buffer we just read in
	 */
	for (char* nextSwap = (char*)buffer;
	  numberOfBytes > 0;
	  numberOfBytes -= typeSize, nextSwap += typeSize)
	    {
	    /*
	     * Swap one element.  Note that it doesn't really matter
	     * which of the file and processor is big-endian, because
	     * the swapping code is the same either way.
	     */
	    for (unsigned int byteNo = 0;  byteNo < typeSize / 2;  byteNo++)
		{
		char temp = nextSwap[byteNo];
		nextSwap[byteNo] = nextSwap[typeSize - byteNo - 1];
		nextSwap[typeSize - byteNo - 1] = temp;
		}
	    }
	}
    return *this;
    }

/*
 * Perform a byte-swapped write.  If the processor endedness matches
 * the file endedness, we just pass the operation to the CommandStream
 * write routine.  Otherwise we allocate a buffer and copy the data
 * into it, swapping as we go, and then write from the buffer.
 *
 * LIMITATIONS:  typeSize must be even.
 */
ByteSwappedStream& ByteSwappedStream::swappedWrite(
    const void*		buffer,		// Buffer to write from
    unsigned int	typeSize,	// Size of buffer's underlying type
    int			numberOfBytes)	// Number of bytes to write
    {
    if (typeSize == 1  ||  processorEndedness == fileEndedness)
	{
	CommandStream::write(buffer, numberOfBytes);
	return *this;
	}
    int bufferSize = numberOfBytes;
    if (bufferSize > bufferLimit)
	bufferSize = bufferLimit;
    char* swappedBuffer = new char[bufferSize];
    // NEEDSWORK:  should throw exception on failure
    for (char* nextItem = (char*)buffer;  numberOfBytes > 0;  )
	{
	/*
	 * Fill swappedBuffer with the next group of swapped data
	 */
	int swapLimit = bufferSize;
	char* nextSwap = swappedBuffer;
	for (  ;
	  swapLimit > 0  &&  numberOfBytes > 0;
	  swapLimit -= typeSize, numberOfBytes -= typeSize,
	    nextItem += typeSize, nextSwap += typeSize)
	    {
	    /*
	     * Copy and swap one element.
	     */
	    for (unsigned int byteNo = 0;  byteNo < typeSize;  byteNo++)
		nextSwap[byteNo] = nextItem[typeSize - byteNo - 1];
	    }
	/*
	 * Write the bufferfull of swapped data.
	 */
	CommandStream::write(swappedBuffer, nextSwap - swappedBuffer);
	}
    delete[] swappedBuffer;
    swappedBuffer = NULL;
    return *this;
    }
