#ifndef _BYTESWAPPEDSTREAM_HH
#define _BYTESWAPPEDSTREAM_HH
/*
 * $Id: byteswappedstream.hh,v 1.13 1998/10/23 23:14:55 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: byteswappedstream.hh,v $
 * Revision 1.13  1998/10/23 23:14:55  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.12  1996/09/16 22:16:39  geoff
 * Add default sizes for the character read/write routines.
 *
 * Revision 1.11  1996/05/03  01:11:29  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/02/02  18:28:42  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.9  1995/11/10  02:58:17  geoff
 * Don't include the icc file, since zzinlinefunctions now does that.
 *
 * Revision 1.8  1995/10/10  20:37:42  geoff
 * Disable the default constructor and the open routine.
 *
 * Revision 1.7  1995/09/15  00:04:46  geoff
 * Add compressed-stream support by deriving from the CompressedStream
 * class.
 *
 * Revision 1.6  1995/08/07  20:02:16  geoff
 * Add support for IEEE floats and doubles.
 *
 * Revision 1.5  1995/05/23  00:38:56  geoff
 * Add the very ugly TestableStreambuf abstraction so that I can do
 * reliable selects on incoming ByteSwappedStreams.
 *
 * Revision 1.4  1995/05/05  22:40:58  geoff
 * Swap the endedness and protection arguments to the constructors.  Add
 * readEndedness and writeEndedness.
 *
 * Revision 1.3  1995/04/20  21:56:47  geoff
 * Remove the TRUST_INHERITANCE ifdef.
 *
 * Revision 1.2  1995/03/29  21:03:16  geoff
 * Numerous debugging changes.  It now works.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "cmdstream.hh"
#include <streambuf.h>

/*
 * A TestableStreambuf is a streambuf for which it is possible to test
 * whether the internal buffer currently contains read data.  This is
 * used to violate data encapsulation in ByteSwappedStream::dataAvailable.
 *  Ugly, but the only solution I know.
 */
class TestableStreambuf : public filebuf
    {
    public:
	inline bool	dataAvailable() const;
					// Test if input data available
    };

/*
 * A ByteSwappedStream provides the abstraction of a CompressedStream which
 * allows reading and writing of binary data in a
 * processor-independent order.  Either big- or little-endian I/O can
 * be chosen at constructor time.  The processor's ordering is
 * determined at run time, and all data types are swapped if
 * necessary.  Note that only primitive types can be read or written;
 * other types can be handled by deriving a class.
 *
 * LIMITATIONS: the caller is responsible for ordering and for
 * ensuring that typecasts are not used inappropriately.  Although
 * byte swapping of floats and doubles is supported, there is no
 * format conversion, so in practice these data types must be in IEEE
 * floating-point format.  Finally, because iostream.h declares
 * certain things as ints, attempts to read or write more than 2 GB at
 * one time may produce incorrect results.
 */

class ByteSwappedStream : public CompressedStream
    {
    public:
	enum Endedness			// Byte-ordering options for processors
	    {
	    bigEnded,			// Byte 0 is most significant
	    littleEnded,		// Byte 0 is least significant
	    naturalEndedness		// Use processor's endedness for file
	    };
			ByteSwappedStream(int fd,
			  Endedness desiredEndedness = littleEnded,
			  int childPID = 0);
			ByteSwappedStream(const char* name, int mode,
			  Endedness desiredEndedness = littleEnded,
			  int protection = 0664);
	inline		~ByteSwappedStream();
	static Endedness getProcessorEndedness();
					// Returns inferred endedness
	inline Endedness getFileEndedness() const;
					// Returns current file endedness
	inline bool	dataAvailable() const;
					// Test if input data available
	/*
	 * Functions to allow the default endedness to be encoded in
	 * the file.
	 */
	ByteSwappedStream&
			writeEndedness();
					// Write chosen endedness to stream
	ByteSwappedStream&
			readEndedness();
					// Read and set chosen endedness
	/*
	 * The following routines do not need to swap bytes, because
	 * they are dealing with abstract bytes, so they are just
	 * inline pass-throughs.  However, they must exist or C++ will
	 * attempt to convert the pointers to void*, which will fail.
	 *
	 * Incidentally, I wrestled a lot with the question of whether
	 * to default the second argument for character buffers.  The
	 * risk is that people will forget to specify a size and thus
	 * write only character of a string.  But I decided in favor
	 * of consistency.
	 */
	inline ByteSwappedStream&
			read(char* buffer, int numberOfBytes = sizeof(char));
	inline ByteSwappedStream&
			read(unsigned char* buffer,
			  int numberOfBytes = sizeof(unsigned char));
#ifndef _G_BROKEN_SIGNED_CHAR
	inline ByteSwappedStream&
			read(signed char* buffer,
			  int numberOfBytes = sizeof(signed char));
#endif // _G_BROKEN_SIGNED_CHAR
	inline ByteSwappedStream&
			write(const char* buffer,
			  int numberOfBytes = sizeof(char));
	inline ByteSwappedStream&
			write(const unsigned char* buffer,
			  int numberOfBytes = sizeof(unsigned char));
#ifndef _G_BROKEN_SIGNED_CHAR
	inline ByteSwappedStream&
			write(const signed char* buffer,
			  int numberOfBytes = sizeof(signed char));
#endif // _G_BROKEN_SIGNED_CHAR
	/*
	 * The following routines actually perform byte-swapping work.
	 * Note that we do *not* support access to void* pointers;
	 * this is very important (otherwise the compiler would
	 * silently cast any unknown type to void*, and we wouldn't
	 * get byte-swapping benefits for that type).  This is handled
	 * by making the void* routines be private.
	 */
	inline ByteSwappedStream&
			read(short* buffer, int numberOfBytes = sizeof(short));
	inline ByteSwappedStream&
			read(unsigned short* buffer,
			  int numberOfBytes = sizeof(unsigned short));
	inline ByteSwappedStream&
			read(int* buffer, int numberOfBytes = sizeof(int));
	inline ByteSwappedStream&
			read(unsigned int* buffer,
			  int numberOfBytes = sizeof(unsigned int));
	inline ByteSwappedStream&
			read(long* buffer, int numberOfBytes = sizeof(long));
	inline ByteSwappedStream&
			read(unsigned long* buffer,
			  int numberOfBytes = sizeof(unsigned long));
#ifndef bool
	inline ByteSwappedStream&
			read(bool* buffer, int numberOfBytes = sizeof(bool));
#endif // bool
	inline ByteSwappedStream&
			read(float* buffer, int numberOfBytes = sizeof(float));
	inline ByteSwappedStream&
			read(double* buffer,
			  int numberOfBytes = sizeof(double));
	inline ByteSwappedStream&
			write(const short* buffer,
			  int numberOfBytes = sizeof(short));
	inline ByteSwappedStream&
			write(const unsigned short* buffer,
			  int numberOfBytes = sizeof(unsigned short));
	inline ByteSwappedStream&
			write(const int* buffer,
			  int numberOfBytes = sizeof(int));
	inline ByteSwappedStream&
			write(const unsigned int* buffer,
			  int numberOfBytes = sizeof(unsigned int));
	inline ByteSwappedStream&
			write(const long* buffer,
			  int numberOfBytes = sizeof(long));
	inline ByteSwappedStream&
			write(const unsigned long* buffer,
			  int numberOfBytes = sizeof(unsigned long));
#ifndef bool
	inline ByteSwappedStream& write(const bool* buffer,
			  int numberOfBytes = sizeof(bool));
#endif // bool
	inline ByteSwappedStream&
			write(const float* buffer,
			  int numberOfBytes = sizeof(float));
	inline ByteSwappedStream&
			write(const double* buffer,
			  int numberOfBytes = sizeof(double));
    private:
	Endedness	fileEndedness;		// Endedness the file exists in
	static Endedness processorEndedness;	// Endedness the processor uses
	static void	findEndedness();	// Intuit processor endedness
	/*
	 * The following routines are the underlying functions that
	 * perform the real work of byte swapping.
	 */
	ByteSwappedStream& swappedRead(void* buffer, unsigned int typeSize,
			  int numberOfBytes);
	ByteSwappedStream& swappedWrite(const void* buffer,
			  unsigned int typeSize, int numberOfBytes);
    };

#define _BYTESWAPPEDSTREAM_ICC_NEEDED

#endif // _BYTESWAPPEDSTREAM_HH
