#ifndef _CLUSTER_HH
#define _CLUSTER_HH
/*
 * $Id: cluster.hh,v 1.23 1998/10/23 23:14:58 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: cluster.hh,v $
 * Revision 1.23  1998/10/23 23:14:58  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.22  1998/06/30 20:20:42  geoff
 * Preliminary changes to support separate tracking of overlapped and
 * nonoverlapped cluster members.  This code is only partially debugged
 * at this point.
 *
 * Revision 1.21  1998/05/14 01:15:23  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.20  1998/05/02 00:29:04  geoff
 * Put the element iterator in the cluster namespace, instead of using a
 * naming convention.
 *
 * Revision 1.19  1998/04/09 00:32:58  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.18  1998/02/19 02:09:55  geoff
 * Add destructiveCopyFrom and emptySelf.  Disable assignment and
 * nondestructive copying.
 *
 * Revision 1.17  1997/01/02 06:48:16  geoff
 * Add a removeCurrent operator for the cluster iterator.
 *
 * Revision 1.16  1996/09/11 22:59:32  geoff
 * Convert from dlists to slists.  Add combineDisjoint and findOverlap.
 *
 * Revision 1.15  1996/06/11  06:13:02  geoff
 * Changes to help support investigated clusters.  Add
 * ClusterElementIterator, newId, isEmpty, referenceTime, the
 * zero-argument updateReferenceTime, compareWeights, typeName, and name,
 * some of which are virtual.  Make the lastReference and hoardWeight
 * protected, so that the virtual computeWeight can work with them
 * conveniently.
 *
 * Revision 1.14  1996/05/03  01:11:34  geoff
 * Update the copyright.  Change the declaration of countMinimumStorage.
 *
 * Revision 1.13  1996/04/10  01:28:23  geoff
 * Change the parameter lists for countMaximumStorage and adjustStorage
 * to include the appropriate manager.
 *
 * Revision 1.12  1996/02/02  18:28:44  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.11  1995/11/22  00:19:45  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.10  1995/11/10  02:58:22  geoff
 * Include basictypes.hh (if new file relations).  Get rid of a spurious
 * "const" declaration.  Don't include the icc file, since
 * zzinlinefunctions now does that.
 *
 * Revision 1.9  1995/09/15  00:04:53  geoff
 * Globally change "cache" to "hoard".
 *
 * Revision 1.8  1995/06/30  19:04:24  geoff
 * Add updateReferenceTime, countMinimumStorage, and countMaximumStorage.
 *
 * Revision 1.7  1995/06/17  00:10:54  geoff
 * Make show be a parameterized routine, and operator<< just be a wrapper
 * around show.
 *
 * Revision 1.6  1995/05/23  00:39:02  geoff
 * Make addFile, removeFile, and combine virtual.
 *
 * Revision 1.5  1995/05/05  22:41:16  geoff
 * Add operator<<.
 *
 * Revision 1.4  1995/04/20  21:56:57  geoff
 * Add hasFile.
 *
 * Revision 1.3  1995/03/29  23:57:10  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:03:48  geoff
 * Add removeFile.  Use the ReferenceCount type for lastReference.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "basictypes.hh"
#include "list.hh"
#include "showoptions.hh"

/*
 * Class for describing a cluster of files.
 */

class ostream;
class File;
class HoardingManagerBase;

class Cluster
    {
    public:
	class		NonoverlappedIterator;
	class		OverlappedIterator;
	class		AllElementIterator;
	friend		NonoverlappedIterator;
	friend		OverlappedIterator;
	enum ClusterStorageState {STORED, NOT_STORED};
	inline		Cluster();
	virtual		~Cluster();
	virtual void	destructiveCopyFrom(Cluster& source);
					// Copy efficiently, destroying source
	static inline void
			resetIdNumbering(); // Reset ID numbering scheme
	inline void	newId();	// Set a new ID value
	inline unsigned long
			id() const;	// Return cluster's ID
	inline bool	isEmpty() const; // Test whether cluster is empty
	inline unsigned long
			totalCount() const; // Return no. of files in cluster
	inline unsigned long
			totalSize() const; // Return cluster size in blocks
	inline unsigned long
			overlappedCount() const;
					// Return no. overlapped files
	inline unsigned long
			overlappedSize() const;
					// Return size of overlapped files
	inline unsigned long
			nonoverlappedCount() const;
					// Return no. nonoverlapped files
	inline unsigned long
			nonoverlappedSize() const;
					// Return size of nonoverlapped files
	inline unsigned long
			weight() const;	// Return cluster weight
	virtual void	addNonoverlappedFile(File* newFile);
					// Add new file to cluster
    	virtual void	addOverlappedFile(File* newFile);
					// Add new overlapped file to cluster
	virtual void	prepareNonoverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
	virtual void	prepareOverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
	virtual void	removeNonoverlappedFile(File* oldFile);
					// Remove old file from cluster
	virtual void	removeOverlappedFile(File* oldFile);
					// Remove overlapped file from cluster
	inline void	emptySelf();	// Remove all cluster contents
	void		emptyNonoverlaps(); // Remove non-overlapped contents
	void		emptyOverlaps(); // Remove overlapped cluster contents
	virtual void	combineDisjoint(Cluster& that);
					// Combine two known-disjoint clusters,
					// emptying second
	virtual void	combine(Cluster& that); // Combine two clusters that
					// ..may overlap, emptying second
#if 0
	inline bool	hasFile(File* possibleMember)
			  const;	// Check for membership
#endif // 0
	inline bool	hasNonoverlappedFile(const File* possibleMember)
			  const;	// Check for membership
	inline bool	hasOverlappedFile(const File* possibleMember)
			  const;	// Check for membership
	unsigned long	findOverlap(const Cluster& that)
			  const;	// Find overlap between two clusters
	inline ReferenceCount
			referenceTime()
			  const;	// Return last reference time
	virtual void	updateReferenceTime();
					// Update last reference time
					// ..by scanning cluster members
	virtual void	updateReferenceTime(ReferenceCount newTime);
					// Update last reference time
					// ..from a known value
	virtual void	computeWeight(); // Calculate hoarding weight
	unsigned long	countMinimumStorage(const HoardingManagerBase& manager)
			  const;
					// Count min storage needed for cluster
	unsigned long	countMaximumStorage(const HoardingManagerBase& manager)
			  const;
					// Count max storage needed for cluster
	inline void	setStorageState(ClusterStorageState state);
					// Set preferred storage state
	virtual unsigned long
			adjustStorage(	// Cause files to be stored/unstored
			  const HoardingManagerBase& manager);
	static int	compareWeights(const Cluster** cluster1,
			  const Cluster** cluster2);
					// Used for qsort
	friend ostream&
			operator<<(ostream& stream, const Cluster& cluster);
					// Display cluster on a stream
	virtual ostream& show(ostream& outFile, ShowOptions options,
			  const char* managerName) const;
					// Display cluster on a stream,
					// ..with options
	virtual const char*
			typeName() const; // Return cluster type name (if any)
	virtual const char*
			encodedTypeName(ShowOptions options) const;
					// Return encoded cluster type name
	virtual const char*
			name() const;	// Return cluster name (if any)
	virtual const char*
			encodedName(ShowOptions options) const;
					// Return encoded cluster name (if any)
    protected:
	SList		nonoverlappedMembers; // List of unoverlapped files
	SList		overlappedMembers; // Members inserted by overlaps
					// ..Note that this might not
					// ..mean the members actually
					// ..belong to other clusters;
					// ..see Cluster::combine for
					// ..more information
	ClusterStorageState storage;	// Preferred storage state
	ReferenceCount	lastReference;	// Last reference to any file
	long		hoardWeight;	// Caching weight of the cluster
    private:
	/*
	 * It's not possible to copy a cluster, because of the
	 * intimate relationship between clusters and files.  (More
	 * accurately, copying a cluster would require that the new
	 * cluster be added to every file's cluster list.)  So copying
	 * and assignment are disabled.
	 */
			Cluster(const Cluster& source);
					// Disable copy constructor
	Cluster&	operator=(const Cluster& source);
					// Disable assignment
	static unsigned long
			nextID;		// Next ID to be assigned
	unsigned long	nonoverlappedBlocks; // Number of blocks in nonoverlaps
	unsigned long	overlappedBlocks; // Number of blocks in overlaps
	unsigned long	numNonoverlappedFiles;
					// Number of files in nonoverlaps
	unsigned long	numOverlappedFiles; // No. of files in overlaps
	unsigned long	identification;	// Cluster identification
    };

/*
 * The following two iterators walk through the nonoverlapped and
 * overlapped members of a cluster, respectively.
 */
class Cluster::NonoverlappedIterator : public SList::Iterator
    {
    public:
	inline		NonoverlappedIterator(Cluster& targetCluster);
	inline		~NonoverlappedIterator();
	inline File&	operator*();	// Return next file in cluster
	void		removeCurrent(); // Zap current w/o affecting iterator
    private:
	Cluster&	cluster;	// Cluster being iterated over
    };

class Cluster::OverlappedIterator : public SList::Iterator
    {
    public:
	inline		OverlappedIterator(Cluster& targetCluster);
	inline		~OverlappedIterator();
	inline File&	operator*();	// Return next file in cluster
	void		removeCurrent(); // Zap current w/o affecting iterator
    private:
	Cluster&	cluster;	// Cluster being iterated over
    };

class Cluster::AllElementIterator
    {
    public:
	inline		AllElementIterator(Cluster& targetCluster);
	inline		~AllElementIterator();
	inline		operator bool() const; // For completion testing
	inline void	reset();	// Reset to start of list
	inline File&	operator*();	// Return next file in cluster
	inline AllElementIterator& operator++();
					// Step to next entry
	void		removeCurrent(); // Zap current w/o affecting iterator
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			AllElementIterator(const AllElementIterator&);
	AllElementIterator&
			operator=(const AllElementIterator&);
	/*
	 * Private data
	 */
	Cluster&	cluster;	// Cluster being iterated over
	Cluster::NonoverlappedIterator
			nonoverlapIterator;
					// Subiterator for non-overlapped files
	Cluster::OverlappedIterator
			overlapIterator; // Subiterator for overlapped files
    };

#define _CLUSTER_ICC_NEEDED

#endif // _CLUSTER_HH
