#ifndef lint
static const char Rcs_Id[] =
    "$Id: clustering_mgr.cc,v 1.50 2005/04/13 21:03:01 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: clustering_mgr.cc,v $
 * Revision 1.50  2005/04/13 21:03:01  geoff
 * When marking a file dirty, also mark other members of the cluster.
 *
 * Revision 1.49  1999/02/04 21:04:00  geoff
 * More debugging of incremental-clustering support.  Also fix a bug in
 * the file-reading code.
 *
 * Revision 1.48  1998/10/23 23:14:59  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.47  1998/08/13 18:38:39  geoff
 * Integrate changes from the bug-fix branch:
 *     Remove a spurious return that prevented loose clustering.
 *
 *     Make the non-shared-neighbors code work again.
 *
 *     Consider the size of always-hoarded and often-referenced files when
 *     filling the hoard.  Add the meaningless argument to the
 *     reference-recording routines.
 *
 *     Many changes to make the last-reference time and similar things be
 *     manager-specific.
 *
 * Revision 1.46  1998/07/08 19:56:41  geoff
 * Make the non-shared-neighbors code work again
 *
 * Revision 1.45  1998/06/30 20:20:43  geoff
 * Preliminary changes to support separate tracking of overlapped and
 * nonoverlapped cluster members.  This code is only partially debugged
 * at this point.  Also get rid of a spurious return in clusterLooseFiles.
 *
 * Revision 1.44  1998/06/05 18:06:42  geoff
 * Add temporary debugging code to check for clusterless files.  Use new
 * dirty/clean functions instead of manipulating the flag bit directly.
 *
 * Revision 1.43  1998/05/14 01:15:24  geoff
 * Add (partially-tested) MIME-encoding support.  Get rid of a bogus
 * return statement from 1.39 that disabled all clustering.
 *
 * Revision 1.42  1998/05/02 00:29:05  geoff
 * Put the element iterator in the cluster namespace, instead of using a
 * naming convention.  Only report incremental statistics in incremental
 * mode.
 *
 * Revision 1.41  1998/04/09 00:32:59  geoff
 * Use the new notation for list iterators.  Turn off incremental
 * clustering by default.
 *
 * Revision 1.40  1998/03/05 03:18:14  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.39  1998/02/19 02:09:56  geoff
 * Add preliminary support for incremental clustering.  The code works,
 * but so far it doesn't provide much benefit.
 *
 * Revision 1.38  1997/12/18 01:24:15  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.37  1997/12/04 23:15:16  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.36  1997/05/16 21:26:19  geoff
 * Remove the debugging code used to find the LRU/clustering inconsistencies.
 *
 * Revision 1.35  1997/05/16 17:10:34  geoff
 * Remember to identify the oldestReference field in showParameters.  Add
 * temporary debugging code to catch discrepancies between the LRU and
 * clustering managers.  Use the wasOften routine to make the simulation
 * isOften decision based on old data rather than new.  This is the
 * version used to run the simulations in the dissertation.
 *
 * Revision 1.34  1997/02/12 02:52:02  geoff
 * Calculate attention shifts the same way that I did in the PCI paper,
 * by making it possible to exclude newly created files.
 *
 * Revision 1.33  1997/01/02 06:48:16  geoff
 * Add forgetCluster and requiredSize (including working-set
 * calculations), and support for the cluster iterator's removeCurrent
 * operator.  Make certain hoarding decisions more consistent with other
 * managers and with the calculation of working-set sizes.
 *
 * Revision 1.32  1996/12/03 02:57:50  geoff
 * Add support for reading and writing parameters.  Add code to calculate
 * the equivalent LRU depth needed.
 *
 * Revision 1.31  1996/10/11 22:08:52  geoff
 * Get rid of the workaround for the floating-output bug.
 *
 * Revision 1.30  1996/10/08 19:50:02  geoff
 * Add code to work around the floating-point output bug.
 *
 * Revision 1.29  1996/09/16 22:16:40  geoff
 * When calculating hoard weights, consider open files as having been
 * currently referenced.
 *
 * Revision 1.28  1996/09/11 22:59:35  geoff
 * Add combineDisjoint.  Make major improvements to the computation of
 * cluster weights.  Turn off timed management.  Make minimumReferences
 * be independed of sharedNeighbors.  Add code to suppress duplicate
 * clusters.  Convert from dlists to slists.
 *
 * Revision 1.27  1996/06/11  07:04:49  geoff
 * Get rid of some obsolete code that updated cluster reference times
 * excessively.
 *
 * Revision 1.26  1996/06/11  06:13:03  geoff
 * Many changes to support investigated clusters.  Many of these involve
 * inheriting or overriding virtual functions from the Cluster class;
 * others involve including the investigated clusters in the various
 * hoarding steps.  Change the shared-neighbors parameters to reflect
 * some of the parameter-setting runs.  Fix a bug in the detection of
 * cluster overruns.  Also make a number of changes which clean up the
 * handling of file references in various non-standard cases (such as
 * often files).  Move compareWeights to the base Cluster class.
 *
 * Revision 1.25  1996/05/03  01:11:41  geoff
 * Update the copyright.  Up the management interval to 100 days.  Make
 * the directory-distance and investigated-relation weights apply to the
 * shared-neighbor distance measure, and rename a lot of badly-name
 * variables in isSharedNearby.  Add support for always hoarding symlinks
 * and directories.  Be smarter about comparing cluster sizes to the
 * available hoard size.  Fix some stupid typos that made the previous
 * checkin uncompilable.  Rename the CD_xxx flags to include strategic
 * underscores.
 *
 * Revision 1.24  1996/04/10  06:09:26  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.23  1996/04/10  01:28:44  geoff
 * Add the new ManagerData::manager() routine.  Add support for
 * manager-specific storage statuses.
 *
 * Revision 1.22  1996/03/19  18:43:13  geoff
 * Inherit from HoardingManagerBase.  Add support for the
 * DERIVED_CLASS_BUG ifdef.
 *
 * Revision 1.21  1996/02/02  20:09:57  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.20  1996/02/02  18:29:20  geoff
 * Derive resetStatistics from the base class instead of redefining it.
 * Add the new derived data accessors.  Change the defaults for nearly
 * all parameters to reflect the new values gotten from running
 * multi_analyze.  Add directory-distance support.  Fix some minor
 * problems detected by the new C++ compiler.  Change the definition of
 * "tight" clustering to be a bit less lenient.
 *
 * Revision 1.19  1995/12/19  01:03:12  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.18  1995/12/19  00:34:57  geoff
 * Add some missing comments.  Add support for the
 * shared-nearest-neighbor nearness metric.  Add process-tracking
 * support.
 *
 * Revision 1.17  1995/11/22  00:22:29  geoff
 * Get rid of the TIME_SORTED_INPUT and NEW_FILE_RELATIONS ifdefs.
 *
 * Revision 1.16  1995/11/21  21:39:23  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support, including a complete rewrite of isNearby and isVeryNearby.
 *
 * Revision 1.15  1995/11/10  03:00:05  geoff
 * Add zzinlinefunctions.hh.  Add NEW_FILE_RELATIONS support, mostly by
 * moving code elsewhere.  Change the tight/loose clustering to be done
 * separately.
 *
 * Revision 1.14  1995/10/10  20:38:19  geoff
 * Get rid of the COMPARISON_SUPPORT and SIMULATION_SUPPORT ifdefs.
 * Don't insert a file into the previous table, nor track references to
 * it, if it's being forgotten.
 *
 * Revision 1.13  1995/09/15  00:06:26  geoff
 * Change the RCS ID to be a const.  Add statistics support.  Add
 * InternalTimer support.  Add proper bracketing to all switch
 * statements.  Lots of removals now that this manager isn't responsible
 * for deletion.  Add abortive TRACK_MEMORY_LEAKS support.
 *
 * Revision 1.12  1995/08/07  20:03:25  geoff
 * Add external-investigator support.  Modify the file-table iterations
 * to use the new and more sensible increment-and-remove scheme.  Comment
 * a lot of switch cases.
 *
 * Revision 1.11  1995/06/30  19:05:01  geoff
 * Get rid of adjustStorage; it's in cluster.cc now.  Change the default
 * clustering parameters to values that seem to work a bit better.  Put
 * out interesting statistics about the clusters to the log file.  Show
 * the cluster table after showing the file table at the end of
 * management.  Change to the new method of setting all files unstored,
 * then storing only the desired files.  Be a bit smarter about
 * estimating the hoard space needed by a cluster.  Only tightly-cluster
 * files if we have a reliable reference count.  Don't record destroy
 * references.
 *
 * Revision 1.10  1995/06/17  00:12:15  geoff
 * Add show options and clearManagerData.  Report time take to fill the
 * cache.  Flush cerr/cout before all aborts.
 *
 * Revision 1.9  1995/05/27  00:16:12  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.8  1995/05/26  23:06:17  geoff
 * Allow a null parent-file pointer.
 *
 * Revision 1.7  1995/05/23  00:40:20  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.6  1995/05/05  22:42:11  geoff
 * Add a lot of preliminary PRIVATE_MANAGEMENT_DATA support.
 * Add parameter descriptions to support the new setparams code.
 * Add USE_TIGHT_FRACTION support.
 *
 * Revision 1.5  1995/04/20  21:58:28  geoff
 * Move the maximum cache size to the global manager.  Add tight-fraction
 * support.  Report the completion of cluster creation.  Add support for
 * dumping to a stream after management.
 *
 * Revision 1.4  1995/03/30  00:44:31  geoff
 * Get rid of the OLD_PARAMETER_SETTING option.
 *
 * Revision 1.3  1995/03/29  23:57:39  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:05:04  geoff
 * Many changes to fix bugs in the new file-table code, and to support
 * the new parameter-setting styles.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "clustering_mgr.hh"
#include "communication.hh"
#include "correlator.hh"
#include "cluster.hh"
#include "file.hh"
#include "filereference.hh"
#include "globalmanager.hh"
#include "investigator.hh"
#include "internaltimer.hh"
#include "panic.hh"
#include "zzinlinefunctions.hh"
#include <fstream.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strstream.h>
#include <unistd.h>
#include <sys/stat.h>

/*
 * Define this to double-check every file free to be sure there are no
 * references left.
 */
#undef PARANOID_ABOUT_FREEING

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			ClusteringManager::ClusteringManager(
			  const char* managerName,
			  const char* managerDefaults);
			ClusteringManager::~ClusteringManager();
void			ClusteringManager::setParameter(
			  const char* description);
void			ClusteringManager::showParameters(
			  ostream& stream) const;
					// Display management parameters
long			ClusteringManager::writeParameters(
			    ByteSwappedStream& stream, bool reallyWrite)
			  const;
					// Write management params to stream
long			ClusteringManager::readParameters(
			  ByteSwappedStream& stream, long parameterBytes);
					// Read management params from stream
void			ClusteringManager::cleanGlobalData();
					// Clean out global file references
void			ClusteringManager::forgetCluster(Cluster * cluster);
					// Get rid of pointer to a disappearing
					// ..cluster
ostream&		ClusteringManager::showStatistics(ostream& stream)
			  const;
					// Display collected quality statistics
ostream&		ClusteringManager::requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const;
					// Display hoard size needed for
					// ..no misses
void			ClusteringManager::manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream,
			  bool encodeFilenames);
					// Run hoard management
unsigned long		ClusteringManager::prepareFile(File& fileent);
void			ClusteringManager::clusterTightFiles(File& fileent)
			  const;
					// Combine tightly-related files into
					// ..clusters
void			ClusteringManager::clusterLooseFiles(File& fileent)
			  const;
					// Add loosely-related files to
					// ..clusters
int			ClusteringManager::compareReferences(
			  const FileReference* reference1,
			  const FileReference* reference2);
void			ClusteringManager::fillHoard(unsigned long hoardSize);
					// Decide what goes in the hoard
void			ClusteringManager::stateChanged(
			  ClusteringManagerData* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Note that state has changed
			SortedCluster::~SortedCluster();
void			SortedCluster::destructiveCopyFrom(
			  SortedCluster& source);
					// Copy efficiently, destroying source
void			SortedCluster::addNonoverlappedFile(File* newFile);
					// Add new file to cluster
void			SortedCluster::addOverlappedFile(File* newFile);
					// Add new file to cluster
void			SortedCluster::prepareNonoverlappedRemove(
			  File* oldFile);
					// Prepare to remove file from cluster
void			SortedCluster::prepareOverlappedRemove(File* oldFile);
					// Prepare to remove file from cluster
void			SortedCluster::combineDisjoint(Cluster& that);
					// Combine two known-disjoint clusters,
					// emptying second
void			SortedCluster::combine(Cluster& that);
					// Combine two clusters that
					// ..may overlap, emptying second
bool			ClusteringManager::isNearby(
			    const File& referencingFile,
			    const RelatedFile& relatedFile)
			  const;
					// Decide if two files are close
bool			ClusteringManager::isVeryNearby(
			    const File& referencingFile,
			    const RelatedFile& relatedFile)
			  const;
					// Decide if two files are very close
bool			ClusteringManager::isSharedNearby(
			    const File& referencingFile,
			    const RelatedFile& relatedFile,
			    int sharedThreshold)
			  const;
					// Decide if two files are close, using
					// ..the shared-nearest-neighbor method
ClusteringManagerData&	ClusteringManager::managerData(File& file)
			  const;	// Provide data access
const ClusteringManagerData&		// Provide data access
			ClusteringManager::managerData(const File& file)
			  const;
void			SortedCluster::computeWeight();
					// Calculate hoarding weight
unsigned long		SortedCluster::adjustStorage(
			  const HoardingManagerBase& manager);
					// Cause files to be stored/unstored
void			SortedCluster::suppressDuplicates()
			  const;	// Get rid of any clusters that
					// ..duplicate this one
HoardingManagerBase&	ClusteringManagerData::manager() const;
					// Return pointer to our manager
			ClusteringManagerData::ClusteringManagerData(
			  File& containingFile);
			ClusteringManagerData::~ClusteringManagerData();
void			ClusteringManagerData::recordMissedReference(
			  File& file, ReferenceType referenceType,
			  bool isMeaningless);
					// Record that a file miss happened
void			ClusteringManagerData::recordReference(File& file,
			  ReferenceType type,
			  bool isMeaningless);
					// Keep track of a file reference
ByteSwappedStream&	ClusteringManagerData::read(
			  ByteSwappedStream& stream, File& file);
					// Read binary version from stream
ByteSwappedStream&	ClusteringManagerData::write(
			    ByteSwappedStream& stream, const File& file)
			  const;
					// Write binary version to stream
void			ClusteringManagerData::clear(const File& file);
					// Clear management data
ostream&		operator<<(class ostream& stream,
			  const ClusteringManagerData& data);
					// Write printable version
ostream&		ClusteringManagerData::show(ostream& stream,
			    ShowOptions options, const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
void			ClusteringManagerData::joinNonoverlappedCluster(
			  SortedCluster& newCluster);
					// Add cluster to file's cluster list
void			ClusteringManagerData::joinOverlappedCluster(
			  SortedCluster& newCluster);
					// Add cluster to file's cluster list
void			ClusteringManagerData::leaveNonoverlappedCluster(
			  SortedCluster& oldCluster);
					// Remove cluster from file's list
void			ClusteringManagerData::leaveOverlappedCluster(
			  SortedCluster& oldCluster);
					// Remove cluster from file's list
bool			ClusteringManagerData::isInCluster(
			  const SortedCluster* cluster) const;
					// See if file is in particular cluster
bool			ClusteringManagerData::inSameCluster(
			  const ClusteringManagerData& that)
			    const;
					// See if two files are in same cluster
void			ClusteringManagerData::prepare();
					// Prepare for clustering
void			ClusteringManagerData::cleanReferences();
					// Get rid of obsolete references
void			ClusteringManagerData::markDirty();
					// Set the dirty bit
void			ClusteringManagerData::relationsChanged();
					// Note that relations have changed
ByteSwappedStream&	ClusteringManagerData::readTotalReferences(
			  ByteSwappedStream& stream);
					// Read total refs from a stream
#endif // 0

ClusteringManager clusteringManager(
			  "Clustering",
			  "Clustering::alwaysClusterDistance=1.0\n"
			    "Clustering::closestK=1\n"
			    "Clustering::directoryDistanceWeight="
			      "1.0\n"
			    "Clustering::incremental=false\n"
			    "Clustering::investigatedRelationWeight="
			      "1.0\n"
			    "Clustering::managementInterval=0\n"
			    "Clustering::mediumReferences=4\n"
			    "Clustering::minimumReferences=3\n"
			    "Clustering::oldestReference=0\n"
			    "Clustering::sometimesClusterDistance="
			      "2.0\n"
			    "Clustering::sharedLooseThreshold=12\n"
			    "Clustering::sharedTightThreshold=14\n"
			    "Clustering::sharedNeighbors=true\n"
			    "Clustering::tightPercent=10\n"
			    );

unsigned int		ClusteringManager::mediumReferences = 0;
ReferenceCount		ClusteringManager::oldestReference = 0;
bool			ClusteringManager::sharedNeighbors = true;

extern void		correlatorPanic(const char* message);
extern void		nomem(const char* where, int line, unsigned int size);

ClusteringManager::ClusteringManager(
    const char*		managerName,	// String name of the manager
    const char*		managerDefaults) // Default parameter settings
    : HoardingManagerBase(managerName, managerDefaults),
      clusterTable(NULL),
      clusterNumber(0),
      clusterCount(0),
      sortedClusters(NULL),
      sortedClusterCount(0)
    {
    HoardingManagerBase::setManagerParameters(defaults);
    }

ClusteringManager::~ClusteringManager()
    {
    }

/*
 * Set a single management parameter.  This is called from
 * setManagerParameters.  The description string is of the form:
 *
 *	parameterName=value
 *
 * The format of the value depends on the parameter, but in general it
 * is numeric and is terminate by any invalid character.
 */
void ClusteringManager::setParameter(
    const char*		description)	// String with a parameter value
    {
    if (description == NULL)
	return;
    // NEEDSWORK: should use string class
    const char* valueStart = strchr(description, '=');
    if (valueStart == NULL)
	{
	// NEEDSWORK:  should throw exception
	cerr << "correlator:  bad ClusteringManager parameter format"
	    " in '"
	  << description << "'\n";
	return;
	}
    int nameLength = valueStart - description;
    valueStart++;
    // NEEDSWORK:  should use inline function
#define parameterIs(string)	nameLength == sizeof string - 1 \
				  &&  strncmp(description, string, \
				    sizeof string - 1) == 0
    if (parameterIs("alwaysClusterDistance"))
	alwaysClusterDistance = atof(valueStart);
    else if (parameterIs("closestK"))
	closestK = atoi(valueStart);
    else if (parameterIs("directoryDistanceWeight"))
	directoryDistanceWeight = atof(valueStart);
    else if (parameterIs("incremental"))
	incremental = valueStart[0] == 't';
    else if (parameterIs("investigatedRelationWeight"))
	investigatedRelationWeight = atof(valueStart);
    else if (parameterIs("managementInterval"))
	managementInterval = atol(valueStart);
    else if (parameterIs("mediumReferences"))
	mediumReferences = atoi(valueStart);
    else if (parameterIs("minimumReferences"))
	minimumReferences = atoi(valueStart);
    else if (parameterIs("sharedLooseThreshold"))
	sharedLooseThreshold = atoi(valueStart);
    else if (parameterIs("sharedNeighbors"))
	sharedNeighbors = valueStart[0] == 't';
    else if (parameterIs("sharedTightThreshold"))
	sharedTightThreshold = atoi(valueStart);
    else if (parameterIs("sometimesClusterDistance"))
	sometimesClusterDistance = atof(valueStart);
    else if (parameterIs("tightPercent"))
	tightPercent = atoi(valueStart);
    else if (parameterIs("oldestReference"))
	{
	long newValue = atoi(valueStart);
	if (verbose != 0)
	    cerr << "correlator:  " << (newValue ? "!!!" : "")
	      << "resetting Clustering::oldestReference from "
	      << oldestReference << " to " << newValue << '\n';
	if (newValue > 0)
	    oldestReference = (ReferenceCount)newValue;
	else
	    oldestReference = currentReferenceTime() + newValue;
	}
    else
	{
	// NEEDSWORK: should throw exception
	cerr
	  << "correlator: unrecognized ClusteringManager parameter '"
	  << description << "'\n";
	}
#undef parameterIs
    }

/*
 * Display management parameters.  The parameters are displayed
 * conditionally, in that only those relevant to the current mode are
 * shown.  Thus, to see the parameters for a different mode, the mode
 * must first be changed.  This may be a design error.
 */
void ClusteringManager::showParameters(
    ostream&		stream)		// Stream to display on
    const
    {
    stream << "Clustering::sharedNeighbors="
      << (sharedNeighbors ? "true" : "false")
      << "/false/true/1/Use Shared-Nearest-Neighbors Similarity Measure\n";
    stream << "Clustering::managementInterval="
      << managementInterval
      << "/1/infinity/60/Management Interval (Minutes)\n"
      << "Clustering::directoryDistanceWeight=" << directoryDistanceWeight
      << "/0.0/infinity/1/Weight Given to Directory Distances\n"
      << "Clustering::incremental="
      << (incremental ? "true" : "false")
      << "/false/true/1/Use Incremental Clustering Algorithm\n"
      << "Clustering::investigatedRelationWeight="
      << investigatedRelationWeight
      << "/0.0/infinity/1/Weight Given to Investigated Relations\n"
      << "Clustering::minimumReferences="
      << minimumReferences
      << "/0/infinity/1/Plausible-Distance Threshold (# Refs)\n"
      << "Clustering::oldestReference="
      << oldestReference
      << "/0/0/1/Oldest Equivalent-LRU Reference Number\n";
    if (sharedNeighbors)
	{
	stream << "Clustering::sharedLooseThreshold="
	  << sharedLooseThreshold
	  << "/0/" << GlobalManager::referencesToKeep()
	  << "/1/Shared Neighbors Needed for Loose Clustering\n";
	stream << "Clustering::sharedTightThreshold="
	  << sharedTightThreshold
	  << "/0/" << GlobalManager::referencesToKeep()
	  << "/1/Shared Neighbors Needed for Tight Clustering\n";
	}
    else
	{
	stream << "Clustering::alwaysClusterDistance="
	  << alwaysClusterDistance
	  << "/0.0/infinity/1/Always-Cluster Threshold\n";
	stream << "Clustering::closestK=" << closestK
	  << "/0/infinity/1/Closest Files to Cluster\n";
	stream << "Clustering::mediumReferences="
	  << mediumReferences
	  << "/0/infinity/1/Reliable-Distance Threshold (# Refs)\n";
	stream << "Clustering::sometimesClusterDistance="
	  << sometimesClusterDistance
	  << "/0.0/infinity/1/Conditional-Clustering Threshold\n";
	stream << "Clustering::tightPercent="
	  << tightPercent << "/0/100/1/Tight-Binding Threshold (Percent)\n";
	}
    }

/*
 * Write manager-specific parameters to a stream.  Returns the number
 * of bytes written, or zero if trouble occurs.  If reallyWrite is
 * false, the byte count is returned, but no writing is actually done.
 */
long ClusteringManager::writeParameters(
    ByteSwappedStream&	stream,		// Stream to write to
    bool		reallyWrite)	// True to actually do the writing
    const
    {
    long bytesWritten =
      HoardingManagerBase::writeParameters(stream, reallyWrite);
    if (reallyWrite  &&  !stream.write(&sharedNeighbors))
	return 0;
    bytesWritten += sizeof sharedNeighbors;
    if (reallyWrite  &&  !stream.write(&directoryDistanceWeight))
	return 0;
    bytesWritten += sizeof directoryDistanceWeight;
    if (reallyWrite  &&  !stream.write(&investigatedRelationWeight))
	return 0;
    bytesWritten += sizeof investigatedRelationWeight;
    if (reallyWrite  &&  !stream.write(&incremental))
	return 0;
    bytesWritten += sizeof incremental;
    if (reallyWrite  &&  !stream.write(&minimumReferences))
	return 0;
    bytesWritten += sizeof minimumReferences;
    if (reallyWrite  &&  !stream.write(&oldestReference))
	return 0;
    bytesWritten += sizeof oldestReference;
    if (reallyWrite  &&  !stream.write(&sharedLooseThreshold))
	return 0;
    bytesWritten += sizeof sharedLooseThreshold;
    if (reallyWrite  &&  !stream.write(&sharedTightThreshold))
	return 0;
    bytesWritten += sizeof sharedTightThreshold;
    if (reallyWrite  &&  !stream.write(&alwaysClusterDistance))
	return 0;
    bytesWritten += sizeof alwaysClusterDistance;
    if (reallyWrite  &&  !stream.write(&closestK))
	return 0;
    bytesWritten += sizeof closestK;
    if (reallyWrite  &&  !stream.write(&mediumReferences))
	return 0;
    bytesWritten += sizeof mediumReferences;
    if (reallyWrite  &&  !stream.write(&sometimesClusterDistance))
	return 0;
    bytesWritten += sizeof sometimesClusterDistance;
    if (reallyWrite  &&  !stream.write(&tightPercent))
	return 0;
    bytesWritten += sizeof tightPercent;
    return bytesWritten;
    }

/*
 * Read manager-specific parameters from a stream.  The second
 * argument gives the number of bytes available for consumption.
 * Returns the number of bytes remaining.
 */
long ClusteringManager::readParameters(
    ByteSwappedStream&	stream,		// Stream to read from
    long		parameterBytes)	// Number of bytes available to read
    {
    parameterBytes =
      HoardingManagerBase::readParameters(stream, parameterBytes);
    parameterBytes -= sizeof sharedNeighbors;
    if (parameterBytes >= 0)
	stream.read(&sharedNeighbors);
    parameterBytes -= sizeof directoryDistanceWeight;
    if (parameterBytes >= 0)
	stream.read(&directoryDistanceWeight);
    if (logVersion > 4)
	{
	parameterBytes -= sizeof incremental;
	if (parameterBytes >= 0)
	    stream.read(&incremental);
	}
    parameterBytes -= sizeof investigatedRelationWeight;
    if (parameterBytes >= 0)
	stream.read(&investigatedRelationWeight);
    parameterBytes -= sizeof minimumReferences;
    if (parameterBytes >= 0)
	stream.read(&minimumReferences);
    parameterBytes -= sizeof oldestReference;
    if (parameterBytes >= 0)
	stream.read(&oldestReference);
    parameterBytes -= sizeof sharedLooseThreshold;
    if (parameterBytes >= 0)
	stream.read(&sharedLooseThreshold);
    parameterBytes -= sizeof sharedTightThreshold;
    if (parameterBytes >= 0)
	stream.read(&sharedTightThreshold);
    parameterBytes -= sizeof alwaysClusterDistance;
    if (parameterBytes >= 0)
	stream.read(&alwaysClusterDistance);
    parameterBytes -= sizeof closestK;
    if (parameterBytes >= 0)
	stream.read(&closestK);
    parameterBytes -= sizeof mediumReferences;
    if (parameterBytes >= 0)
	stream.read(&mediumReferences);
    parameterBytes -= sizeof sometimesClusterDistance;
    if (parameterBytes >= 0)
	stream.read(&sometimesClusterDistance);
    parameterBytes -= sizeof tightPercent;
    if (parameterBytes >= 0)
	stream.read(&tightPercent);
    return parameterBytes;
    }

/*
 * Get rid of any references that point at files being forgotten.
 */
void ClusteringManager::cleanGlobalData()
    {
    if (clusterTable == NULL)
	return;
    for (unsigned long clusterNo = 0;  clusterNo < clusterCount;  clusterNo++)
	{
	for (Cluster::AllElementIterator element(clusterTable[clusterNo]);
	  element;
	  ++element)
	    {
	    File& file(*element);
	    if (file.isForgetting())
		element.removeCurrent();
	    }
	}
    }

/*
 * Get rid of any pointers to a cluster that is going to be deleted
 */
void ClusteringManager::forgetCluster(
    Cluster*		cluster)	// Cluster to be forgotten
    {
    for (unsigned long clusterNo = 0;
      clusterNo < sortedClusterCount;
      clusterNo++)
	{
	if (sortedClusters[clusterNo] == cluster)
	    sortedClusters[clusterNo] = NULL;
	}
    }


/*
 * Display collected quality statistics
 */
ostream& ClusteringManager::showStatistics(
    ostream&		stream)		// Stream to display on
    const
    {
    HoardingManagerBase::showStatistics(stream);
    return stream;
    }

/*
 * Calculate and display the hoard size needed to achieve zero misses.
 * For the clustering manager, this is a bit tricky.  First, we walk
 * the file table and unmark every file.  Then we walk the cluster
 * list in forward order, looking for unmarked files that ought to be
 * hoarded.  When we find one, we mark every file in the cluster and
 * add the relevant file sizes to the required size.
 *
 * For no particular reason except that it was a convenient place to
 * put the code, this routine also caculates the working set and the
 * attention-shift statistics.
 */
ostream& ClusteringManager::requiredSize(
    ostream&		stream,		// Stream to display result on
    ostream*		duplicateStream) // Stream to use for duplicate results
    const
    {
    /*
     * The following variables count up various versions of the most
     * recent period's working set (with the period being determined
     * by the global manager's base reference).  Variables ending in
     * "size" count blocks, while those ending in "count" count files.
     *
     * Variables prefixed with "total" count essentially all files.
     * Those prefixed with nothing count files that *could* be hoarded
     * if they were listed as controlled.  The "controlled"
     * prefix marks files that are actually controlled.  The
     * "cumulative" prefix marks all files that have been referenced
     * at least once since the correlator learned of them.
     */
    unsigned long controlledWorkingSetSize = 0;
					// Blocks in in this period's 
    unsigned long cumulativeWorkingSetSize = 0;
    unsigned long totalWorkingSetSize = 0;
    unsigned long workingSetSize = 0;
    unsigned long controlledWorkingSetCount = 0;
    unsigned long cumulativeWorkingSetCount = 0;
    unsigned long totalWorkingSetCount = 0;
    unsigned long workingSetCount = 0;
    /*
     * The following variables count attention-shift statistics.  We
     * divide the files into four classes, depending on the values of
     * two Boolean conditions: whether the file was referenced in the
     * previous period (defined by the wasPreviouslyReferenced()
     * condition) and whether it was referenced in the current period.
     * If the first bit refers to the previous period, then 00 files
     * aren't counted, 01 files are counted by currentOnlyCount, 10
     * files are counted by previousOnlyCount, and 11 files are
     * represented by bothCount.
     *
     * At the same time, for any file that was first referenced in the
     * current period by creating it, we increment newlyCreatedCount.
     *
     * As a side effect, we update the previous-reference and
     * newly-created bits for use in the following period.
     */
    unsigned long previousOnlyCount = 0;
    unsigned long currentOnlyCount = 0;
    unsigned long bothCount = 0;
    unsigned long newlyCreatedCount = 0;
    /*
     * The following variables count up the cluster hoard size needed
     * to get no misses in the most recent period.  The prefixes and
     * suffixes are as above.
     */
    unsigned long controlledSize = 0;	// Size of controlled files
    unsigned long totalSize = 0;	// Size of all files
    unsigned long controlledCount = 0;	// Count of controlled files
    unsigned long totalCount = 0;	// Count of all files
    for (FileTableIterator fileIterator;  fileIterator;  ++fileIterator)
	{
	File& file = *fileIterator;
	if (!file.includeInSimulation())
	    {
	    file.mark();
	    continue;
	    }
	file.unmark();
	bool currentReference =
	  managerData(file).lastReferenceTime() >= getBaseReferenceTime();
	unsigned long fileSize = file.size();
	if (!file.alwaysHoardableType())
	    {
	    if (currentReference)
		{
		if (file.wasPreviouslyReferenced())
		    bothCount++;
		else
		    currentOnlyCount++;
		file.markPreviouslyReferenced();
		}
	    else
		{
		if (file.wasPreviouslyReferenced())
		    previousOnlyCount++;
		file.markNotPreviouslyReferenced();
		}
	    if (file.wasNewlyCreated())
		newlyCreatedCount++;
	    file.markNotNewlyCreated();
	    }
	if (file.alwaysHoardableType()  ||  managerData(file).wasOften())
	    {
	    file.mark();
	    totalSize += fileSize;
	    totalCount++;
	    if (file.isControlled())
		{
		controlledSize += fileSize;
		controlledCount++;
		}
	    }
	if (managerData(file).lastReferenceTime() != 0  &&  file.couldHoard())
	    {
	    cumulativeWorkingSetSize += fileSize;
	    cumulativeWorkingSetCount++;
	    }
	if (currentReference)
	    {
	    totalWorkingSetSize += fileSize;
	    totalWorkingSetCount++;
	    if (file.couldHoard())
		{
		workingSetSize += fileSize;
		workingSetCount++;
		if (file.canHoard())
		    {
		    controlledWorkingSetSize += fileSize;
		    controlledWorkingSetCount++;
		    }
		}
	    }
	}
    stream << managerName() << " working set size is "
      << controlledWorkingSetSize << " controlled, " << workingSetSize
      << " controllable, " << totalWorkingSetSize << " total, "
      << cumulativeWorkingSetSize << " cumulative\n";
    stream << managerName() << " working set count is "
      << controlledWorkingSetCount << " controlled, " << workingSetCount
      << " controllable, " << totalWorkingSetCount << " total, "
      << cumulativeWorkingSetCount << " cumulative\n";
    stream << managerName() << " attention shift counts are "
      << previousOnlyCount << ' ' << currentOnlyCount << ' ' << bothCount
      << ' ' << newlyCreatedCount << '\n';
    if (duplicateStream != NULL)
	{
	*duplicateStream << managerName() << " working set size is "
	  << controlledWorkingSetSize << " controlled, " << workingSetSize
	  << " controllable, " << totalWorkingSetSize << " total\n";
	*duplicateStream << managerName() << " working set count is "
	  << controlledWorkingSetCount << " controlled, " << workingSetCount
	  << " controllable, " << totalWorkingSetCount << " total\n";
	*duplicateStream << managerName() << " attention shift counts are "
	  << previousOnlyCount << ' ' << currentOnlyCount << ' ' << bothCount
	  << ' ' << newlyCreatedCount << '\n';
	}
    if (sortedClusters == NULL)
	return stream;
    for (unsigned long clusterNo = 0;
      clusterNo < sortedClusterCount;
      clusterNo++)
	{
	if (sortedClusters[clusterNo] == NULL)
	    continue;
	for (Cluster::AllElementIterator element(*sortedClusters[clusterNo]);
	  element;
	  ++element)
	    {
	    File& file = *element;
	    if (!file.isMarked())
		{
		if (managerData(file).lastReferenceTime()
		    >= getBaseReferenceTime()
		  &&  file.couldHoard())
		    {
		    for (element.reset();  element;  ++element)
			{
			File& hoardedFile = *element;
			if (!hoardedFile.isMarked())
			    {
			    hoardedFile.mark();
			    if (hoardedFile.isControlled())
				{
				unsigned long blocks =
				  hoardedFile.hoardableSize();
				controlledSize += blocks;
				controlledCount++;
				totalSize += blocks;
				totalCount++;
				}
			    else if (hoardedFile.couldHoard())
				{
				totalSize += hoardedFile.size();
				totalCount++;
				}
			    }
			}
		    break;
		    }
		}
	    }
	}
    stream << managerName() << " hoard size needed is " << controlledSize
      << " controlled, " << totalSize << " total\n";
    stream << managerName() << " hoard count needed is " << controlledCount
      << " controlled, " << totalCount << " total\n";
    if (duplicateStream != NULL)
	{
	*duplicateStream
	  << managerName() << " hoard size needed is " << controlledSize
	  << " controlled, " << totalSize << " total\n";
	*duplicateStream
	  << managerName() << " hoard count needed is " << controlledCount
	  << " controlled, " << totalCount << " total\n";
	}
    return stream;
    }

void ClusteringManager::manage(		// Manage the file hoard
    const InternalTimer& now,		// Current time-of-day
    ByteSwappedStream*	dumpStream,	// Stream to dump results to
    bool		encodeFilenames) // True to encode filenames in dump
    {
    lastManagementTime = now;
    InternalTimer managementStartTime(InternalTimer::ACTUAL);
    InternalTimer phaseStartTime(managementStartTime);
    /*
     * Management takes place in several steps.  Most of the steps
     * involve a walk of the hash table.  The current steps are:
     * preparation, cluster creation, clustering, and
     * weighting.
     */
    if (verbose > 1)
	cerr << "correlator:  clustering management began at "
	  << managementStartTime.ctime() << '\n';
    /*
     * Preparation phase.
     *
     * We walk the hash table to ensure that every file has a known
     * state.  This phase also counts the number of files that will
     * participate in clustering, which will also be the initial
     * number of clusters, and counts up the equivalent LRU hoard size
     * that would be needed to avoid a miss during the last phase.
     */
    unsigned long oldClusterCount = incremental ? clusterCount : 0;
    clusterCount = 0;
    File::resetIdNumbering();
    FileTableIterator fileIterator;
    unsigned long lruSize = 0;
    unsigned long hoardSize = GlobalManager::hoardSize();
				// Number of blocks left in hoard
    for (  ;  fileIterator;  ++fileIterator)
	{
	File& file(*fileIterator);
	if (isLRU_Equivalent(file))
	    lruSize += file.hoardableSize();
	hoardSize -= ClusteringManager::prepareFile(file);
	ClusteringManagerData& data(managerData(file));
	if (incremental  &&  data.isDirty())
	    {
	    if (data.cluster != NULL)
		{
		data.cluster->emptyNonoverlaps();
		if (data.cluster != NULL)
		    correlatorPanic(
		      "correlator:  cluster failed to empty self");
		}
	    while (!data.overlappingClusters.isEmpty())
		((SortedCluster*)data.overlappingClusters.head())
		  ->removeOverlappedFile(&file);
	    }
	}
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  preparation took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	cerr << "correlator:  LRU age needed is " << equivalentLRU_Depth()
	  << ", LRU size is " << lruSize << '\n';
	}
    /*
     * Cluster-creation phase.
     *
     * Build an empty cluster table.  In incremental mode, copy all
     * non-dirty clusters into it from the old table.  Then put each
     * remaining non-FORGOTTEN file into a cluster by itself.
     */
    SortedCluster* oldClusterTable = NULL;
    if (incremental)
	oldClusterTable = clusterTable;
    else
	delete[] clusterTable;
    SortedCluster::resetIdNumbering();
    clusterTable = new SortedCluster[clusterCount];
    if (clusterTable == NULL)
	nomem("manage", __LINE__, clusterCount * sizeof(SortedCluster));
    clusterNumber = 0;
    for (unsigned long oldClusterNumber = 0;
      oldClusterNumber < oldClusterCount;
      ++oldClusterNumber)
	{
	if (!oldClusterTable[oldClusterNumber].isEmpty())
	    {
	    if (clusterNumber >= clusterCount)
		{
		cerr << "correlator:  too many clusters!\n";
		break;
		}
	    clusterTable[clusterNumber].destructiveCopyFrom(
	      oldClusterTable[oldClusterNumber]);
	    ++clusterNumber;
	    }
	}
    unsigned long clusterStartPoint = 0;
    if (incremental)
	{
	/*
	 * We're done with the old cluster table.  Keep track of how
	 * many pre-existing clusters there are, because we won't need
	 * to reprocess these in certain later phases.
	 */
	delete[] oldClusterTable;
	oldClusterTable = NULL;
	clusterStartPoint = clusterNumber;
	}
for (fileIterator.reset();  fileIterator;  ++fileIterator)
  (*fileIterator).unmark();
for (unsigned int i = 0;  i < clusterNumber;  i++)
  {
  for (Cluster::AllElementIterator element(clusterTable[i]);  element;  ++element)
    {
    File& file(*element);
    file.mark();
    if (!managerData(file).willCluster())
	cerr << "unclusterable clustered file!\n";
    }
  }
{
int clusteredFiles = 0;
int totalFiles = 0;
for (fileIterator.reset();  fileIterator;  ++fileIterator)
  {
  File& file(*fileIterator);
  if (managerData(file).willCluster())
    ++totalFiles;
  if (file.isMarked())
    ++clusteredFiles;
  }
if (incremental)
  cerr << "correlator:  " << clusterNumber << " returning clusters containing " << clusteredFiles << " files of " << totalFiles << " total, " << clusterCount - clusterNumber << " slots available.\n";
}
    for (fileIterator.reset();  fileIterator;  ++fileIterator)
	{
	File& file = *fileIterator;
	ClusteringManagerData& data(managerData(file));
	if (data.willCluster())
	    {
	    if (!incremental  ||  data.isDirty())
		{
		if (clusterNumber >= clusterCount)
		    {
		    cerr << "correlator:  too many clusters!\n";
		    break;
		    }
		clusterTable[clusterNumber].addNonoverlappedFile(&file);
		clusterTable[clusterNumber].updateReferenceTime(
		  managerData(file).lastReferenceTime());
		++clusterNumber;
		}
	    }
	}
for (fileIterator.reset();  fileIterator;  ++fileIterator)
  {
  ClusteringManagerData& data(managerData(*fileIterator));
  if (data.willCluster()  &&  data.cluster == NULL)
    correlatorPanic("correlator: file with no clusters found");
  }
    if (incremental  &&  verbose > 1)
	cerr << "correlator:  incremental clustering reused "
	  << clusterStartPoint << " clusters and must cluster "
	  << clusterNumber - clusterStartPoint << " of " << clusterCount
	  << " files\n";
    if (clusterNumber < clusterCount)
	{
	if (!incremental)
	    cerr << "correlator:  not enough clusters (got " << clusterNumber
	      << ", expected " << clusterCount << ")\n";
	clusterCount = clusterNumber;
	}
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  cluster creation took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    /*
     * Clustering phase.
     *
     * We walk the file table to collect files into clusters.  This is
     * done twice: once for tightly-related files, and again for loose
     * ones.
     */
    for (fileIterator.reset();  fileIterator;  ++fileIterator)
	clusterTightFiles(*fileIterator);
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  tight clustering took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    for (fileIterator.reset();  fileIterator;  ++fileIterator)
	clusterLooseFiles(*fileIterator);
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  loose clustering took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    for (clusterNumber = clusterStartPoint;
      clusterNumber < clusterCount;
      clusterNumber++)
	clusterTable[clusterNumber].suppressDuplicates();
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  duplicate suppression took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    /*
     * Weighting phase.
     *
     * For each cluster, the size, LRU time, and other cluster
     * parameters are integrated to get a single cluster weight.
     */
    for (clusterNumber = clusterStartPoint;
      clusterNumber < clusterCount;
      clusterNumber++)
	clusterTable[clusterNumber].computeWeight();
    /*
     * We must do the same for the investigated clusters.  Computing
     * their weight is a slightly more complicated, because we must
     * also update their LRU time and give them an ID.
     */
    for (InvestigatedClusterIterator investigatedCluster;
      investigatedCluster;
      ++investigatedCluster)
	{
	InvestigatedCluster& cluster(*investigatedCluster);
	cluster.newId();
	cluster.updateReferenceTime();
	cluster.computeWeight();
	}
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  weighting took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    /*
     * We now have a table of clusters that we would like to hoard.
     * Fill the hoard.
     */
    ClusteringManager::fillHoard(hoardSize);
    if (verbose > 1)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - phaseStartTime;
	phaseStartTime = phaseEndTime;
	cerr << "correlator:  hoard filling took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS) << '\n';
	}
    /*
     * Management is now complete.
     */
    if (verbose > 0)
	{
	InternalTimer phaseEndTime(InternalTimer::ACTUAL);
	InternalTimer elapsed = phaseEndTime - managementStartTime;
	cerr << "correlator:  clustering management took "
	  << elapsed.numericMicrotime(InternalTimer::MS_HMS_DHMS)
	  << ", complete at " << phaseEndTime.ctime() << '\n';
	if (verbose > 1)
	    {
	    cerr << "correlator:  file table ";
	    File::showHashStatistics(cerr);
	    cerr << '\n';
	    unsigned long maxClusterFiles = 0;
	    unsigned long maxClusterSize = 0;
	    unsigned long totalClusterFiles = 0;
	    unsigned long totalClusterSize = 0;
	    // NEEDSWORK: would be nice to calculate standard deviations
	    for (clusterNumber = 0;
	      clusterNumber < clusterCount;
	      clusterNumber++)
		{
		SortedCluster& cluster = clusterTable[clusterNumber];
		if (cluster.totalCount() > maxClusterFiles)
		    maxClusterFiles = cluster.totalCount();
		if (cluster.totalSize() > maxClusterSize)
		    maxClusterSize = cluster.totalSize();
		totalClusterFiles += cluster.totalCount();
		totalClusterSize += cluster.totalSize();
		if (verbose > 3)
		    cluster.show(cerr, SO_FULL_FILE_INFORMATION, NULL);
		}
	    if (clusterCount == 0)
		{
		cerr << "correlator:  0 clusters, average size = 0/0"
		  << ", max size = 0/0\n";
		}
	    else
		{
		cerr << "correlator:  "
		  << clusterCount << " clusters, average size = "
		  << (double)totalClusterFiles / clusterCount << '/'
		  << (double)totalClusterSize / clusterCount
		  << ", max size = " << maxClusterFiles << '/'
		  << maxClusterSize << '\n';
		}
	    }
	}
    ShowOptions options = SO_FULL_FILE_INFORMATION | SO_MANAGER;
    if (encodeFilenames)
	options |= SO_MIME_ENCODE;
    for (FileTableIterator iter;  iter;  ++iter)
	{
ClusteringManagerData& data(managerData(*iter));
if (data.willCluster()  &&  data.cluster == NULL)
    correlatorPanic("correlator: file with no clusters found");
	managerData(*iter).markClean();
	if (dumpStream != NULL)
	    (*iter).show(*dumpStream, options, name);
	}
    if (dumpStream != NULL)
	{
	options &= SO_MIME_ENCODE;
	for (clusterNumber = 0;
	  clusterNumber < clusterCount;
	  clusterNumber++)
	    clusterTable[clusterNumber].show(*dumpStream, options, NULL);
	for (InvestigatedClusterIterator cluster;  cluster;  ++cluster)
	    (*cluster).show(*dumpStream, options, NULL);
	}
    }

/*
 * Prepare a file to be managed
 */
unsigned long ClusteringManager::prepareFile(
    File&		file)		// File to manage
    {
    file.setId();
    bool couldCluster = managerData(file).willCluster();
    bool wasOften = managerData(file).wasOften();
    managerData(file).prepare();
    /*
     * If the file didn't participate in clustering last time, it
     * needs to be dirtied so it can potentially participate this
     * time.
     */
    if (!couldCluster  ||  wasOften)
	managerData(file).markDirty();
    /*
     * If the file is always hoarded, just set it for storing.  Also,
     * we need to mark it dirty in case it participated in clustering
     * last time.
     */
    if (file.alwaysHoardableType()
      ||  isOften(file, GlobalManager::getOftenPercent()))
	{
	managerData(file).store(file);
	managerData(file).setWasOften();
	managerData(file).markDirty();
	return file.hoardableSize();
	}
    else
	managerData(file).unstore(file);
    if (file.willForget())
	{
	/*
	 * We won't cluster this time.  We need to mark the file dirty
	 * in case it clustered last time.
	 */
	managerData(file).markDirty();
	return 0;
	}
    /*
     * Initialize clustering
     */
    ++clusterCount;
    managerData(file).okToCluster();
    file.discoverStatusIfUnknown();
    return 0;
    }

/*
 * Compare file references for sorting (called from qsort)
 */
int ClusteringManager::compareReferences(
    const FileReference* reference1,	// References to compare
    const FileReference* reference2)	// ..
    {
    /*
     * It turns out that qsort often calls us with identical pointers,
     * so this is worthwhile:
     */
    if (reference1 == reference2)
	return 0;
    /*
     * The next sort key is "oftenness."  Often-referenced files sort
     * to the end.
     */
    File& target1(reference1->targetRelatedFile().targetFile());
    File& target2(reference2->targetRelatedFile().targetFile());
    if (clusteringManager.isOften(target1, GlobalManager::getOftenPercent()))
	{
	if (clusteringManager.isOften(target2,
	  GlobalManager::getOftenPercent()))
	    return clusteringManager.managerData(target1).totalReferenceCount()
	      - clusteringManager.managerData(target2).totalReferenceCount();
	else
	    return 1;			// Reference 1 sorts last
	}
    else if (clusteringManager.isOften(target2,
      GlobalManager::getOftenPercent()))
	return -1;			// Reference 2 sorts last
    /*
     * The next sort key is distance, but only if there are at least
     * mediumReferences references to each file.  Smaller distances
     * sort first.
     */
    double distance1 = reference1->quickAverageDistance();
    double distance2 = reference2->quickAverageDistance();
    if (reference1->referenceCount() >= mediumReferences
      &&  reference2->referenceCount() >= mediumReferences)
	{
	double distanceDifference = distance1 - distance2;
	return distanceDifference < 0 ? -1 : distanceDifference > 0 ? 1 : 0;
	}
    /*
     * The next sort key is the number of references, if at least one
     * file has more than mediumReferences references.  Larger reference
     * counts sort first.
     */
    if (reference1->referenceCount() >= mediumReferences
      ||  reference2->referenceCount() >= mediumReferences)
	return reference2->referenceCount() - reference1->referenceCount();
    /*
     * The next sort key is distance, but only if there is the same
     * number of references to each file.  Smaller distances sort first.
     */
    if (reference1->referenceCount() == reference2->referenceCount())
	return distance1 < distance2 ? -1
	  : distance1 > distance2 ? 1
	  : 0;
    /*
     * The next sort key is the number of references.  Larger numbers of
     * references sort first.
     */
    return reference2->referenceCount() - reference1->referenceCount();
    }

/*
 * Combine closely-related files into clusters
 */
void ClusteringManager::clusterTightFiles(
    File&		file)		// File being clustered
    const
    {
    ClusteringManagerData& data(managerData(file));
    if (!data.willCluster())
	return;
    if (incremental  &&  !data.isDirty())
	return;				// File doesn't need clustering
    /*
     * Form files into clusters based on a similarity measurement.  The
     * algorithm is based on that described in R.A. Jarvis and Edward
     * A. Patrick, "Clustering Using a Similarity Measure Based on
     * Shared Near Neighbors," IEEE Transactions on Computers, C-22(11)
     * Nov. 1973, 1025-1034.  The similarity measurement is implemented
     * by the "isVeryNearby" routine.
     */
    for (RelatedFileIterator relatedFileIterator(file);
      relatedFileIterator;
      ++relatedFileIterator)
	{
	/*
	 * Combine this cluster with those of any nearby files.
	 */
	RelatedFile& relatedFile = *relatedFileIterator;
	File& targetFile = relatedFile.targetFile(); 
	if (managerData(targetFile).willCluster()
	  &&  !managerData(targetFile).inSameCluster(managerData(file))
	  &&  isVeryNearby(file, relatedFile))
	    {
	    /*
	     * Combine two closely-related clusters.  We make no
	     * attempt to preserve any particular order in the
	     * clusters.
	     */
	    Cluster* fileCluster(((ClusteringManagerData&)managerData(file))
		.primaryCluster());
	    Cluster* targetCluster(managerData(targetFile).primaryCluster());
	    ReferenceCount targetTime(targetCluster->referenceTime());
	    fileCluster->combineDisjoint(*targetCluster);
	    fileCluster->updateReferenceTime(targetTime);
	    }
	}
    }

/*
 * Add loosely-related files to clusters
 */
void ClusteringManager::clusterLooseFiles(
    File&		file)		// File being clustered
    const
    {
    ClusteringManagerData& data(managerData(file));
    if (!data.willCluster())
	return;
    if (incremental  &&  !data.isDirty())
	return;				// File doesn't need clustering
    for (RelatedFileIterator relatedFileIterator(file);
      relatedFileIterator;
      ++relatedFileIterator)
	{
	/*
	 * Overlap this cluster with those of any nearby files.
	 */
	RelatedFile& relatedFile = *relatedFileIterator;
	File& targetFile = relatedFile.targetFile();
	if (managerData(targetFile).willCluster()
	  &&  !managerData(targetFile).inSameCluster(managerData(file))
	  &&  isNearby(file, relatedFile))
	    {
	    /*
	     * The files aren't tightly-enough related to force their
	     * respective clusters to combine (otherwise, they would
	     * already be in the same cluster).  Instead, add each
	     * file to the other's clusters.  Also, make sure that the
	     * referencing file's clusters are at least as new as the
	     * referenced file.  This insures that, for example, a C
	     * compiler will be hoarded along with the C source file
	     * it references.
	     *
	     * NEEDSWORK: need to decide whether the inverse situation
	     * ever applies (e.g., source file hoarded along with
	     * compiler).
	     */
	    for (ClusteringOverlapIterator fileCluster(managerData(file));
	      fileCluster;
	      ++fileCluster)
		{
		SortedCluster& cluster(*fileCluster);
		if (!cluster.hasOverlappedFile(&targetFile))
		    cluster.addOverlappedFile(&targetFile);
		}
	    for (ClusteringOverlapIterator
		relatedCluster(managerData(targetFile));
	      relatedCluster;
	      ++relatedCluster)
		{
		SortedCluster& cluster(*relatedCluster);
		if (!cluster.hasOverlappedFile(&file))
		    cluster.addOverlappedFile(&file);
		}
	    }
	}
    }

/*
 * Fill the hoard with "good" files
 */
void ClusteringManager::fillHoard(
    unsigned long	availableSpace)	// Space available in the hoard
    {
    delete[] sortedClusters;
    sortedClusterCount = clusterCount + ClusterInvestigatorType::count();
    sortedClusters = new Cluster*[sortedClusterCount];
    if (sortedClusters == NULL)
	nomem("fillHoard", __LINE__, sortedClusterCount * sizeof(Cluster*));
    unsigned long clusternum;
    for (clusternum = 0;  clusternum < clusterCount;  clusternum++)
	sortedClusters[clusternum] = &clusterTable[clusternum];
    for (InvestigatedClusterIterator investigatedCluster;
      investigatedCluster;
      ++investigatedCluster)
	sortedClusters[clusternum++] = &(*investigatedCluster);
    qsort((void*) sortedClusters, sortedClusterCount, sizeof sortedClusters[0],
      (int (*)(const void*, const void*)) Cluster::compareWeights);
    /*
     * Scan the clusters, fitting in as many as possible.  Note that a
     * lower-priority cluster might be inserted if it will fit where the
     * higher-priority one will not.
     */
    unsigned long totblocks = 0;	// Total blocks in the hoard
    for (clusternum = 0;  clusternum < sortedClusterCount;  clusternum++)
	{
	Cluster* cluster = sortedClusters[clusternum];
	if (cluster->totalCount() == 0)
	    continue;
	unsigned long maximumClusterSize = cluster->countMaximumStorage(*this);
	if (totblocks + maximumClusterSize <= availableSpace)
	    {
	    cluster->setStorageState(Cluster::STORED);
	    totblocks += cluster->adjustStorage(*this);
	    }
	else
	    {
	    unsigned long minimumClusterSize =
	      cluster->countMinimumStorage(*this);
	    if (totblocks + minimumClusterSize <= availableSpace)
		{
		if (verbose > 2)
		    {
		    cerr << "correlator:  skipping cluster " << clusternum
		      << " (no. " << cluster->id()
		      << ") because it won't fit, size range "
		      << minimumClusterSize << " to " << maximumClusterSize
		      << '\n';
		    }
		}
	    else
		{
		if (verbose > 2  &&  totblocks < availableSpace
		  &&  maximumClusterSize >= availableSpace / 10)
		    {
		    cerr << "correlator:  skipping cluster " << clusternum
		      << " (no. " << cluster->id()
		      << ") because it won't fit, contents:\n";
		    cluster->show(cerr, SO_FULL_FILE_INFORMATION | SO_MANAGER,
		      name);
		    }
		}
	    }
	}
    if (verbose > 0)
	cerr << "correlator:  " << totblocks << " blocks in hoard\n";
    }

/*
 * Record that a change has been made in the file's storage status.
 */
void ClusteringManager::stateChanged(
    ClusteringManagerData* data,	// Manager-specific data for file
    FileState::State	,		// Original state of the file
    FileState::State	)		// New state of the file
    {
    if (data == NULL)
	correlatorPanic(
	  "ClusteringManager::stateChanged called with null data pointer");
    data->markDirty();
    }

SortedCluster::~SortedCluster()
    {
    clusteringManager.forgetCluster(this);
    while (!nonoverlappedMembers.isEmpty())
	{
	File* file = (File*)nonoverlappedMembers.popHead();
	clusteringManager.managerData(*file).leaveNonoverlappedCluster(*this);
	}
    while (!overlappedMembers.isEmpty())
	{
	File* file = (File*)overlappedMembers.popHead();
	clusteringManager.managerData(*file).leaveOverlappedCluster(*this);
	}
    }

/*
 * Efficiently copy a cluster, destroying the source cluster in the
 * process.
 */
void SortedCluster::destructiveCopyFrom(
    SortedCluster&	source)
    {
    Cluster::destructiveCopyFrom(source);
    /*
     * We must walk through all files in the cluster, replacing their
     * pointers to the source cluster with ones to the destination.
     */
    for (Cluster::NonoverlappedIterator file(*this);  file;  ++file)
	{
	ClusteringManagerData& data(clusteringManager.managerData(*file));
	data.leaveNonoverlappedCluster(source);
	data.joinNonoverlappedCluster(*this);
	}
    for (Cluster::OverlappedIterator file(*this);  file;  ++file)
	{
	ClusteringManagerData& data(clusteringManager.managerData(*file));
	data.leaveOverlappedCluster(source);
	data.joinOverlappedCluster(*this);
	}
    }

void SortedCluster::addNonoverlappedFile( // Add a file to a cluster
    File*		newFile)	// File to be added
    {
    Cluster::addNonoverlappedFile(newFile);
    clusteringManager.managerData(*newFile).joinNonoverlappedCluster(*this);
    }

void SortedCluster::addOverlappedFile(	// Add a file to a cluster
    File*		newFile)	// File to be added
    {
    Cluster::addOverlappedFile(newFile);
    clusteringManager.managerData(*newFile).joinOverlappedCluster(*this);
    }

/*
 * Prepare to remove a nonoverlapped file from a cluster
 */
void SortedCluster::prepareNonoverlappedRemove(
    File*		oldFile)	// File to be removed
    {
    Cluster::prepareNonoverlappedRemove(oldFile);
    clusteringManager.managerData(*oldFile).leaveNonoverlappedCluster(*this);
    }

/*
 * Prepare to remove an overlapped file from a cluster
 */
void SortedCluster::prepareOverlappedRemove(
    File*		oldFile)	// File to be removed
    {
    Cluster::prepareOverlappedRemove(oldFile);
    clusteringManager.managerData(*oldFile).leaveOverlappedCluster(*this);
    }

/*
 * Combine two clusters that are known to have no common members.
 * Unpleasant results may occur if common members actually exist.  The
 * second ("that") cluster is emptied by this operation.
 */
void SortedCluster::combineDisjoint(
    Cluster&		that)		// Cluster to combine with this one
    {
    if (this == &that)
	return;				// Don't combine cluster with itself!
    for (Cluster::NonoverlappedIterator thatElement(that);
      thatElement;
      ++thatElement)
	{
	File& file(*thatElement);
	clusteringManager.managerData(file).leaveNonoverlappedCluster(
	  (SortedCluster&)that);
	clusteringManager.managerData(file).joinNonoverlappedCluster(*this);
	}
    Cluster::combineDisjoint(that);
    }

/*
 * Combine two clusters that may have common members, eliminating
 * duplicates.  The second ("that") cluster is emptied by this
 * operation.
 */
void SortedCluster::combine(
    Cluster&		that)		// Cluster to combine with this one
    {
    if (this == &that)
	return;				// Don't combine cluster with itself!
    /*
     * Since a file can be a member of more than one cluster, it is
     * possible that it is already a member of "this".  We have to
     * check for and handle that possibility, avoiding joining the
     * cluster twice.  Note that Cluster::combine has similar code to
     * handle the general-case part of the problem.
     */
    Cluster::OverlappedIterator thatElement(that);
    for (  ;  thatElement;  ++thatElement)
	(*thatElement).mark();
    for (Cluster::OverlappedIterator thisElement(*this);
      thisElement;
      ++thisElement)
	(*thisElement).unmark();
    for (thatElement.reset();  thatElement;  ++thatElement)
	{
	File& file(*thatElement);
	clusteringManager.managerData(file).leaveOverlappedCluster(
	  (SortedCluster&)that);
	if (file.isMarked())
	    clusteringManager.managerData(file).joinOverlappedCluster(*this);
	}
    Cluster::combine(that);
    }

/*
 * Decide if a related file is nearby.  This is a complex decision,
 * partly because there are a lot of ways in which files are related,
 * and partly because the definition of "nearby" is itself complex.
 * The details of the decision are currently documented only by the
 * code, because I keep tinkering with it and it's not worth the
 * ongoing trouble of keeping the comments up to date.  Someday, when
 * this is stable, I'll stumble across this header and update the
 * comments appropriately.
 *
 * NEEDSWORK: add proper commentary in the header
 */
bool ClusteringManager::isNearby(
    const File&		referencingFile, // File that is source of relation
    const RelatedFile&	relatedFile)	// File that is target of relation
    const
    {
    if (sharedNeighbors)
	return isSharedNearby(referencingFile, relatedFile,
	  sharedLooseThreshold);
    // NEEDSWORK: This is very ad-hoc
    /*
     * For each relation, make a decision based on the relation type
     * and the closeness of the relation.
     */
    double distance(0);			// Avg. distance across all relations
    int numDistances(0);
    FileReference* reference(NULL);	// Reference relation, if one was found
    for (RelationIterator relationIterator((RelatedFile&)relatedFile);
      relationIterator;
      ++relationIterator)
	{
	/*
	 * Classify the relation according to type, and deal with it
	 * appropriately.  Unrecognized type classes are silently
	 * ignored.
	 */
	FileRelationBase& relation(*relationIterator);
	const char* relationClass(relation.type().className());
	if (relationClass == FileReferenceType::typeClass)
	    {
	    if (reference != NULL)
		correlatorPanic(
		  "correlator:  unexpected double reference to same file");
	    reference = (FileReference*)&relation;
	    if (reference->referenceCount() >= minimumReferences)
		{
		distance += reference->averageDistance();
		numDistances++;
		}
	    }
	else if (relationClass == RelationInvestigatorType::typeClass)
	    {
	    /*
	     * This is an investigated relation.  If the use of
	     * investigator-derived data is enabled, incorporate it
	     * into the average distance.
	     */
	    if (investigatedRelationWeight != 0)
		{
		RelationInvestigation&
		  investigation((RelationInvestigation&)relation);
		distance += investigation.weight()
		  * ((RelationInvestigatorType&)investigation.type()).weight();
		numDistances++;
		}
	    }
	}
    if (directoryDistanceWeight != 0.0)
	{
	distance +=
	  referencingFile.directoryDistance(relatedFile.targetFile())
	    * directoryDistanceWeight;
	numDistances++;
	}
    if (numDistances == 0)
	return false;			// No knowledge => not nearby
    distance /= numDistances;
    if (distance <= alwaysClusterDistance)
	return true;			// Very close files always cluster
    else if (reference == NULL)
	return false;			// If no usable inferred reference,
					// ..never cluster
    /*
     * Files which are not the closest K never cluster.  Find out if
     * this one is in that group.
     */
    int closerReferences = 0;
    if (closestK == 0)
	return false;
    for (FileReferenceIterator referenceIterator((File&)referencingFile);
      referenceIterator;
      ++referenceIterator)
	{
	FileReference& otherReference = *referenceIterator;
	// @h NEEDSWORK: should this be >=0?
	if (&otherReference != reference
	  &&  ClusteringManager::compareReferences(reference, &otherReference)
	    > 0)
	    {
	    closerReferences++;
	    if (closerReferences >= closestK)
		return false;
	    }
	}
    if (distance <= sometimesClusterDistance)
	return true;			// Semi-close files often cluster
    return false;			// Other files never cluster
    }

/*
 * Decide if a referenced file is "very" close to the referencing file
 * for purposes of tight clustering.
 */
bool ClusteringManager::isVeryNearby(
    const File&		referencingFile, // File that is source of relation
    const RelatedFile&	relatedFile)	// File that is target of relation
    const
    {
    if (sharedNeighbors)
	return isSharedNearby(referencingFile, relatedFile,
	  sharedTightThreshold);
    /*
     * If possible, we make the decision based on an inferred file
     * reference.  If this is not possible, we calculate an average
     * distance and consider ourselves "tight" if we fall below the
     * "always" cluster distance.
     */
    double distance(0);			// Avg. distance across all relations
    int numDistances(0);
    for (RelationIterator relationIterator((RelatedFile&)relatedFile);
      relationIterator;
      ++relationIterator)
	{
	/*
	 * Classify the relation according to type, and deal with it
	 * appropriately.  Unrecognized type classes are silently
	 * ignored.
	 */
	FileRelationBase& relation(*relationIterator);
	const char* relationClass(relation.type().className());
	if (relationClass == FileReferenceType::typeClass)
	    {
	    /*
	     * We will consider ourselves "tight" if the two files
	     * have been referenced more than a few times and are
	     * referenced as a pair more than "tightPercent" of the
	     * time.  Otherwise, we'll just accumulate distance and
	     * continue working, so that we can also become "tight" if
	     * the distance is small, even though the reference pair
	     * is relatively uncommon.
	     */
	    FileReference& reference((FileReference&)relation);
	    if (reference.referenceCount() >= mediumReferences
	      &&  (reference.referenceCount()
		>= (tightPercent
		  * managerData(referencingFile).totalReferenceCount()) / 100))
		return true;
	    if (reference.referenceCount() >= minimumReferences)
		{
		distance += reference.averageDistance();
		numDistances++;
		}
	    }
	else if (relationClass == RelationInvestigatorType::typeClass)
	    {
	    /*
	     * This is an investigated relation.  If the use of
	     * investigator-derived data is enabled, incorporate it
	     * into the average distance.
	     */
	    if (investigatedRelationWeight != 0)
		{
		RelationInvestigation&
		  investigation((RelationInvestigation&)relation);
		distance += investigation.weight()
		  * ((RelationInvestigatorType&)investigation.type()).weight();
		numDistances++;
		}
	    }
	}
    if (directoryDistanceWeight != 0.0)
	{
	distance +=
	  referencingFile.directoryDistance(relatedFile.targetFile())
	    * directoryDistanceWeight;
	numDistances++;
	}
    if (numDistances == 0)
	return false;
    else
	return distance / numDistances <= alwaysClusterDistance;
    }

/*
 * Decide if two files are close, using the shared-nearest-neighbor
 * method.  The similarity-measurement algorithm is based on that
 * described in R.A. Jarvis and Edward A. Patrick, "Clustering Using a
 * Similarity Measure Based on Shared Near Neighbors," IEEE
 * Transactions on Computers, C-22(11) Nov. 1993, 1025-1034.  The
 * difference here is that, since our relations are not symmetric, we
 * do not require the relatedFile to have the referencingFile in its
 * nearest-neighbor list.
 *
 * Modifications for directory distance and investigated relations:
 *
 * Directory distance is calculated as an integer between two files,
 * multiplied by a weight, and subtracted (rounded) from the
 * shared-neighbor count.  The result is that two files in the same
 * directory will have an unchanged count, while files farther apart
 * will have their counts reduced by a value proportional to their
 * distance, so that they have to be strong in other relationship
 * departments before we'll accept them as being nearby.
 *
 * Investigated relations are trickier, because there's no limit on
 * how many can be stored.  For example, the referencingFile might
 * have 50 investigated relations to 50 other files, while the
 * relatedFile might have only 1 relation, but it might be to the
 * single referencingFile and thus very important.  So instead of
 * using a shared-neighbor threshold, we consider only cases where one
 * file lists the other in an investigated relation.  If so, the given
 * weight is *added* to the shared-neighbor count.  (If each file
 * lists the other, two additions are done, on the assumption that the
 * relation must be more important in that case).
 */
bool ClusteringManager::isSharedNearby(
    const File&		referencingFile, // File doing the reference
    const RelatedFile&	maybeNearby,	// File that might be related
    int			sharedThreshold) // No. of shared neighbors needed to
					// ..achieve similarity
    const
    {
    /*
     * First walk through all related files.  If they're related by a
     * file reference, mark them as being neighbors.  Otherwise, if
     * they're investigated related files ignore them unless they
     * match the maybeNearby file (in which case we add them to the
     * investigatedFileWeight).
     */
    double investigatedFileWeight = 0.0;
    File& maybeNearbyFile = maybeNearby.targetFile();
    RelatedFileIterator referencingIterator((File&)referencingFile);
    for (  ;  referencingIterator;  ++referencingIterator)
	{
	RelatedFile& referencingRelatedFile(*referencingIterator);
	for (RelationIterator relationIterator(referencingRelatedFile);
	  relationIterator;
	  ++relationIterator)
	    {
	    FileRelationBase& referencingRelation(*relationIterator);
	    const char* relationClass(referencingRelation.type().className());
	    if (relationClass == FileReferenceType::typeClass)
		{
		if (((FileReference&)referencingRelation).referenceCount()
		  >= minimumReferences)
		    managerData(referencingRelatedFile.targetFile()).flags |=
		      ClusteringManagerData::CD_IS_NEIGHBOR;
		}
	    else if (relationClass == RelationInvestigatorType::typeClass
	      &&  &referencingRelatedFile == &maybeNearby)
		{
		/*
		 * This is an investigated relation to the
		 * maybeNearbyFile.  If the use of
		 * investigator-derived data is enabled, incorporate
		 * it into the investigatedFileWeight.
		 */
		if (investigatedRelationWeight != 0)
		    {
		    RelationInvestigation& investigation(
		      (RelationInvestigation&)referencingRelation);
		    investigatedFileWeight += investigation.weight()
		      * ((RelationInvestigatorType&)investigation.type())
			.weight();
		    }
		}
	    }
	}
    /*
     * Now walk the other file's related-file list, counting the ones
     * that have the neighbor flag set.  At the same time, if there's
     * an investigated relation pointing to the referencingFile, add
     * its weight into the investigatedFileWeight.
     */
    int sharedCount = 0;
    for (RelatedFileIterator maybeRelatedIterator(maybeNearbyFile);
      maybeRelatedIterator;
      ++maybeRelatedIterator)
	{
	RelatedFile& maybeRelated(*maybeRelatedIterator);
	File& maybeRelatedFile(maybeRelated.targetFile());
	for (RelationIterator relationIterator(maybeRelated);
	  relationIterator;
	  ++relationIterator)
	    {
	    FileRelationBase& relation(*relationIterator);
	    const char* relationClass(relation.type().className());
	    if (relationClass == FileReferenceType::typeClass)
		{
		if (((FileReference&)relation).referenceCount()
		  >= minimumReferences)
		    {
		    if (managerData(maybeRelatedFile).flags
		      & ClusteringManagerData::CD_IS_NEIGHBOR)
			++sharedCount;
		    }
		}
	    else if (relationClass == RelationInvestigatorType::typeClass
	      &&  &maybeRelatedFile == &referencingFile)
		{
		/*
		 * This is an investigated relation to the
		 * referencingFile.  If the use of
		 * investigator-derived data is enabled, incorporate
		 * it into the investigatedFileWeight.
		 */
		if (investigatedRelationWeight != 0)
		    {
		    RelationInvestigation& investigation(
		      (RelationInvestigation&)relation);
		    investigatedFileWeight += investigation.weight()
		      * ((RelationInvestigatorType&)investigation.type())
			.weight();
		    }
		}
	    }
	}
    /*
     * Before finishing, clear the neighbor flags so everything is clean.
     */
    for (referencingIterator.reset();
      referencingIterator;
      ++referencingIterator)
	managerData((*referencingIterator).targetFile()).flags &=
	  ~ClusteringManagerData::CD_IS_NEIGHBOR;
    /*
     * Handle directory distances.
     */
    if (directoryDistanceWeight != 0.0)
	{
	/*
	 * We subtract the (weighted) directory distance from the
	 * shared-neighbor count, so that files farther apart will be
	 * less likely to cluster.  The weighted distance is rounded
	 * (roughly) to the nearest integer.
	 */
	sharedCount -=
	  (int)((referencingFile.directoryDistance(maybeNearbyFile)
	      * directoryDistanceWeight)
	    + 0.5);
	}
    /*
     * We also add in the (rounded) weighted sum of any investigated
     * relations we found.
     */
    sharedCount +=
      (int)(investigatedFileWeight * investigatedRelationWeight + 0.5);
    /*
     * The final Boolean test is quite simple.
     */
    return sharedCount >= sharedThreshold;
    }

/*
 * Return a reference to the clustering manager's
 * file-private data, so that manager routines can access it.
 */
ClusteringManagerData& ClusteringManager::managerData(
    File&		file)		// File to provide access for
    const
    {
    return file.clusteringManagerData;
    }

const ClusteringManagerData& ClusteringManager::managerData(
    const File&		file)		// File to provide access for
    const
    {
    return file.clusteringManagerData;
    }

void SortedCluster::computeWeight()	// Calculate hoarding weight
    {
    /*
     * Compute the cluster weight.  First, we scan all files in the
     * cluster and determine a representative last-reference time.
     * Then we turn this into an age by subtracting it from the
     * current reference count.
     */
    lastReference = 0;
    unsigned long thisCount = totalCount();
    if (thisCount == 0)
	{
	hoardWeight = clusteringManager.currentReferenceTime();
	return;
	}
    Cluster::AllElementIterator element(*this);
    for (  ;  element;  ++element)
	{
	File& file(*element);
	ReferenceCount fileReferenceTime = file.isOpen()
	  ? clusteringManager.currentReferenceTime()
	  : clusteringManager.managerData(file).lastReferenceTime();
	if (fileReferenceTime > lastReference)
	    {
	    /*
	     * The file is the newest (so far) in the cluster.  See if
	     * it's one which should be allowed to update the
	     * reference time.  A file is allowed to update if:
	     *
	     * 1.   It is only a member of a single cluster, or
	     * 2.   All clusters containing it are subsets or
	     *	    supersets of this cluster, or
	     * 3.   NEEDSWORK: @h ???
	     */
	    ClusteringManagerData& data = clusteringManager.managerData(file);
	    if (data.clusterCount() == 1)
		lastReference = fileReferenceTime;
	    else
		{
		ClusteringAllClusterIterator fileCluster(data);
		for (  ;  fileCluster;  ++fileCluster)
		    {
		    SortedCluster& cluster(*fileCluster);
		    if (&cluster == this)
			continue;
		    unsigned long overlap = findOverlap(cluster);
		    if (overlap != thisCount
		      &&  overlap != cluster.totalCount())
			break;
		    }
		if (!fileCluster)
		    lastReference = fileReferenceTime;
		}
	    }
	}
    if (lastReference == 0)
	    {
	    /*
	     * We weren't able to find any "allowed" files to use to
	     * set the reference time.  Search the cluster again and
	     * set its reference time to that of the *oldest* element.
	     *
	     * NEEDSWORK: should this be the youngest?
	     */
	    for (element.reset();  element;  ++element)
		{
		ReferenceCount fileReferenceTime =
		  clusteringManager.managerData(*element).lastReferenceTime();
		if ((lastReference == 0  ||  fileReferenceTime < lastReference)
		  &&  fileReferenceTime != 0)
		    lastReference = fileReferenceTime;
		}
	    }
    hoardWeight = clusteringManager.currentReferenceTime() - lastReference;
    }

/*
 * Set the storage state of all files in a cluster so that it matches
 * the desired state.  Returns the number of blocks added to the
 * hoard.
 */
unsigned long SortedCluster::adjustStorage(
    const HoardingManagerBase&	)	// Manager we're hoarding for
    {
    if (storage == NOT_STORED)		// If not-stored, files are correct
	return 0;
    unsigned long blocksAdded = 0;
    for (Cluster::AllElementIterator clustiter(*this);
      clustiter;
      ++clustiter)
	{
	File& file = *clustiter;
	blocksAdded += clusteringManager.managerData(file).store(file);
	}
    return blocksAdded;
    }

/*
 * Get rid of any clusters that duplicate this one.  This is done by
 * searching the files in the cluster to find all clusters of which
 * they are a member.
 *
 * For efficiency, we first build a list of all clusters that hold any
 * file that is in this cluster.  (This can be relatively expensive,
 * because the no-duplicates code uses a linear search, but usually
 * the list is small.)  We then search this list to and compare each
 * cluster with our own.  If duplicates are found, we empty them.
 */
void SortedCluster::suppressDuplicates()
    const
    {
    SList clusters(false);
    /*
     * @h NEEDSWORK: this code will consider two clusters identical if
     * one contains "A" as an overlapped file and "B" as
     * nonoverlapped, and the other thinks the reverse.  Is this the
     * right thing to do?
     */
    for (Cluster::AllElementIterator element((SortedCluster&)*this);
      element;
      ++element)
	{
	File& file(*element);
	for (ClusteringAllClusterIterator
	    cluster(clusteringManager.managerData(file));
	  cluster;
	  ++cluster)
	    {
	    SortedCluster& that(*cluster);
	    if (this != &that  &&  !clusters.hasMember(&that))
		clusters.pushHead(&that);
	    }
	}
    unsigned long thisSize = totalCount();
    for (SList::Iterator listElement(clusters);  listElement;  ++listElement)
	{
	SortedCluster& that(*(SortedCluster*)(clusters[listElement]));
	if (&that == this)
	    correlatorPanic("correlator:  SortedCluster::suppressDuplicates"
	      " found 'this' in list");
	if (that.totalCount() == thisSize  &&  findOverlap(that) == thisSize)
	    {
	    /*
	     * 'that' is a duplicate.  Empty it.
	     */
	    that.emptySelf();
	    }
	}
    }

/*
 * Return pointer to our manager
 */
HoardingManagerBase& ClusteringManagerData::manager()
    const
    {
    return clusteringManager;
    }

ClusteringManagerData::ClusteringManagerData(
    File&		containingFile)	// File we are a member of
    : file(containingFile),
      flags(CD_DIRTY),
      cluster(NULL),
      overlappingClusters(),
      totalReferences(0)
    {
    }

ClusteringManagerData::~ClusteringManagerData()
    {
    }

/*
 * Track manager-specific aspects of a file miss
 */
void ClusteringManagerData::recordMissedReference(
    File&		file,		// File that missed
    ReferenceType	referenceType,	// Type of reference that missed
    bool		isMeaningless)	// True if reference is meaningless
    {
    if (isMeaningless)
	return;				// Ignore meaningless references
    ReferenceCount previousReferenceTime = lastReferenceTime();
    if (previousReferenceTime != 0
      &&  (ClusteringManager::oldestReference == 0
	||  ClusteringManager::oldestReference > previousReferenceTime))
	ClusteringManager::oldestReference = previousReferenceTime;
    if (previousReferenceTime < manager().getBaseReferenceTime())
	{
	switch (referenceType)
	    {
	    case SEER_REF_UNKNOWN:	// Unknown reference type
	    case SEER_REF_CONTROL:	// Fake reference for conn. control
	    case SEER_REF_NEWPROC:	// Report ID of a new process
	    case SEER_REF_PROCEXIT:	// Report exit of a process
	    case SEER_REF_READ:	    	// Read of a file
	    case SEER_REF_RW:	    	// Reads and/or writes of existing file
	    case SEER_REF_EXEC:	    	// Execution
	    case SEER_REF_DESTROY:	// Destruction (delete)
	    case SEER_REF_RENAME:	// Change of file name
	    case SEER_REF_OPENREAD:	// Open for read
	    case SEER_REF_OPENRW:	// Open for read/write
	    case SEER_REF_CLOSE:    	// Close of open file
	    case SEER_REF_EXIT:	    	// Process exit (or exec of new file)
	    case SEER_REF_PEEK:	    	// Peek at a file (stat or access)
	    case SEER_REF_PEEKLINK:	// Peek at a symbolic link's contents
	    case SEER_REF_FOLLOW:	// Follow a symbolic link
	    case SEER_REF_DUP:	    	// Duplicate an existing file reference
	    case SEER_REF_INTERPRET:	// Invoke a command interpreter
		{
		break;
		}
	    case SEER_REF_CREATE:	// Create or truncate/rewrite
	    case SEER_REF_OPENCREATE:   // Create or truncate/rewrite
	    case SEER_REF_MAKELINK:	// Make a symbolic link
		{
		file.markNewlyCreated();
		break;
		}
	    default:
		{
		cerr << "correlator: unexpected reference type "
		  << referenceType;
		correlatorPanic(
		  " in ClusteringManagerData::recordMissedReference");
		break;
		}
	    }
	}
    HoardingManagerDataBase::recordMissedReference(file, referenceType,
      isMeaningless);
    /*
     * We want to update the last-reference time of the file so that
     * it will be considered current for hoarding purposes.  However,
     * we don't increment the global reference counter, because it's
     * not a "real" reference.
     */
    lastReference = manager().currentReferenceTime();
    }

/*
 * Track manager-specific aspects of a file reference
 */
void ClusteringManagerData::recordReference(
    File&		file,		// File that was referenced
    ReferenceType	type,		// Type of reference
    bool		isMeaningless)	// True if reference is meaningless
    {
    if (isMeaningless)
	return;				// Ignore meaningless references
    ReferenceCount previousReferenceTime = lastReferenceTime();
    if (previousReferenceTime != 0
      &&  (ClusteringManager::oldestReference == 0
	||  ClusteringManager::oldestReference > previousReferenceTime))
	ClusteringManager::oldestReference = previousReferenceTime;
    if (previousReferenceTime < manager().getBaseReferenceTime())
	{
	switch (type)
	    {
	    case SEER_REF_UNKNOWN:	// Unknown reference type
	    case SEER_REF_CONTROL:	// Fake reference for conn. control
	    case SEER_REF_NEWPROC:	// Report ID of a new process
	    case SEER_REF_PROCEXIT:	// Report exit of a process
	    case SEER_REF_READ:	    	// Read of a file
	    case SEER_REF_RW:	    	// Reads and/or writes of existing file
	    case SEER_REF_EXEC:	    	// Execution
	    case SEER_REF_DESTROY:	// Destruction (delete)
	    case SEER_REF_RENAME:	// Change of file name
	    case SEER_REF_OPENREAD:	// Open for read
	    case SEER_REF_OPENRW:	// Open for read/write
	    case SEER_REF_CLOSE:    	// Close of open file
	    case SEER_REF_EXIT:	    	// Process exit (or exec of new file)
	    case SEER_REF_PEEK:	    	// Peek at a file (stat or access)
	    case SEER_REF_PEEKLINK:	// Peek at a symbolic link's contents
	    case SEER_REF_FOLLOW:	// Follow a symbolic link
	    case SEER_REF_DUP:	    	// Duplicate an existing file reference
	    case SEER_REF_INTERPRET:	// Invoke a command interpreter
		{
		break;
		}
	    case SEER_REF_CREATE:	// Create or truncate/rewrite
	    case SEER_REF_OPENCREATE:   // Create or truncate/rewrite
	    case SEER_REF_MAKELINK:	// Make a symbolic link
		{
		file.markNewlyCreated();
		break;
		}
	    default:
		{
		cerr << "correlator: unexpected reference type " << type;
		correlatorPanic(" in ClusteringManagerData::recordReference");
		break;
		}
	    }
	}
    HoardingManagerDataBase::recordReference(file, type, isMeaningless);
    switch (type)
	{
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	case SEER_REF_RW:		// Reads and/or writes of existing file
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	case SEER_REF_MAKELINK:		// Create a symbolic link
	case SEER_REF_EXEC:		// Execution
	case SEER_REF_OPENRW:		// Open for read/write
	case SEER_REF_OPENREAD:		// Open for read
	case SEER_REF_OPENCREATE:   	// Create or truncate/rewrite
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	case SEER_REF_PEEKLINK:		// Peek at a symbolic link's contents
	case SEER_REF_FOLLOW:		// Follow a symbolic link
	    {
	    lastReference = manager().nextReferenceTime();
	    totalReferences++;
	    break;
	    }
	case SEER_REF_DESTROY:		// Destruction (delete)
	    {
	    break;
	    }
	case SEER_REF_CLOSE:		// Close of open file
	case SEER_REF_EXIT:		// Process exit or re-exec
	    {
	    /*
	     * On closes/exits, we update the last reference time so
	     * that files/processes which have been open/running for a
	     * very long time (e.g., days) will show up as current.
	     * But we don't update the total references or bump the
	     * global reference counter, and the FileReference class
	     * won't update any references.
	     */
	    lastReference = manager().currentReferenceTime();
	    break;
	    }
	case SEER_REF_RENAME:		// Change of file name
	case SEER_REF_DUP:		// Duplicate existing open
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	default:
	    {
	    cerr << "correlator: unexpected reference type " << type;
	    correlatorPanic("");
	    break;
	    }
	}
    }

/*
 * Read binary version of manager data from stream
 */
ByteSwappedStream& ClusteringManagerData::read(
    ByteSwappedStream&	stream,		// Stream to read data from
    File&		file)		// File we are a member of
    {
    HoardingManagerDataBase::read(stream, file);
    /*
     * The flags aren't necessary, since they are dynamic, but they
     * were nevertheless included in earlier versions.
     */
    if (logVersion <= 5)
	{
	unsigned char dummyFlags;
	stream.read(&dummyFlags, sizeof dummyFlags);
	}
    else
	stream.read(&totalReferences, sizeof totalReferences);
    /*
     * There is no need to read the clusters, since they'll be
     * recalculated.
     */
    return stream;
    }

/* 
 * Write binary version of manager data to stream
 */
ByteSwappedStream& ClusteringManagerData::write(
    ByteSwappedStream&	stream,		// Stream to write data to
    const File&		file)		// File we are a member of
      const
    {
    HoardingManagerDataBase::write(stream, file);
    stream.write(&totalReferences, sizeof totalReferences);
    return stream;
    }

/*
 * Clear any saved management data
 */
void ClusteringManagerData::clear(
    const File&		clearedFile)	// File being cleared
    {
    if (&clearedFile != &file)
	correlatorPanic(
	  "correlator: file mismatch in ClusteringManagerData::clear");
    /*
     * In case the file survives until the next management pass, it
     * needs to be marked dirty so it can cluster again.
     */
    flags = 0;
    markDirty();
    if (cluster != NULL)
	cluster->removeNonoverlappedFile(&file);
    cluster = NULL;
    while (!overlappingClusters.isEmpty())
	((Cluster*)overlappingClusters.head())->removeOverlappedFile(&file);
    }

/*
 * Write a printable version of the clustering manager data
 * to a stream, with options
 *
 * Show options honored: none
 */
ostream& ClusteringManagerData::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    stream << "CMH\t    ClusteringManager: ";
    HoardingManagerDataBase::show(stream, options, managerName);
    stream << ' ' << totalReferences;
    stream << " |" << ((flags & ClusteringManagerData::CD_WEIGHTED)
	? "WEIGHTED|" : "")
      << ((flags & ClusteringManagerData::CD_WILL_CLUSTER)
	? "WILL_CLUSTER|" : "")
      << ((flags & ClusteringManagerData::CD_HAS_CLUSTERED)
	? "HAS_CLUSTERED|" : "")
      << ((flags & ClusteringManagerData::CD_IS_NEIGHBOR)
	? "IS_NEIGHBOR|" : "")
      << ((flags & ClusteringManagerData::CD_WAS_OFTEN)
	? "WAS_OFTEN|" : "")
      << ((flags & ClusteringManagerData::CD_DIRTY)
	? "DIRTY|" : "")
      << '\n';
    unsigned long numClusters = clusterCount();
    stream << "CCH\t\t" << numClusters << " cluster"
      << ((numClusters == 1) ? ":" : "s:");
    for (ClusteringAllClusterIterator cluster((ClusteringManagerData&)*this);
      cluster;
      ++cluster)
	stream << ' ' << (*cluster).id();
    stream << '\n';
    return stream;
    }

/*
 * Add a cluster to a file's cluster list
 */
void ClusteringManagerData::joinNonoverlappedCluster(
    SortedCluster&	newCluster)	// Cluster to join
    {
    if (cluster != NULL)
	correlatorPanic(
	  "correlator:  attempt to join two nonoverlapped clusters");
    cluster = &newCluster;
    }

/*
 * Add a cluster to a file's overlapping cluster list
 */
void ClusteringManagerData::joinOverlappedCluster(
    SortedCluster&	newCluster)	// Cluster to join
    {
    overlappingClusters.pushHead((void*)&newCluster);
    }

/*
 * Remove a nonoverlapped cluster from a file's cluster list
 */
void ClusteringManagerData::leaveNonoverlappedCluster(
    SortedCluster&	oldCluster)	// Cluster to leave
    {
    if (cluster != &oldCluster  &&  cluster != NULL)
	correlatorPanic("correlator:  attempt to leave non-member cluster");
    cluster = NULL;
    }

/*
 * Remove an overlapped cluster from a file's cluster list
 */
void ClusteringManagerData::leaveOverlappedCluster(
    SortedCluster&	oldCluster)	// Cluster to leave
    {
    overlappingClusters.remove((void*)&oldCluster);
    }

/*
 * See if file is in particular cluster
 */
bool ClusteringManagerData::isInCluster(
    const SortedCluster* testCluster)	// Cluster to test for
      const
    {
    for (ClusteringAllClusterIterator member((ClusteringManagerData&)*this);
      member;
      ++member)
	{
	if (&(*member) == testCluster)
	    return true;
	}
    return false;
    }

/*
 * See if two files are in same cluster
 */
bool ClusteringManagerData::inSameCluster(
    const ClusteringManagerData&	// Data for second file to be tested
			that)
      const
    {
    for (ClusteringAllClusterIterator
	thisMember((ClusteringManagerData&)*this);
      thisMember;
      ++thisMember)
	{
	SortedCluster* thisCluster = &(*thisMember);
	for (ClusteringAllClusterIterator
	    thatMember((ClusteringManagerData&)that);
	  thatMember;
	  ++thatMember)
	    {
	    if (&(*thatMember) == thisCluster)
		return true;
	    }
	}
    return false;
    }

/*
 * Prepare for clustering
 */
void ClusteringManagerData::prepare()
    {
    flags
      &= ~(CD_WEIGHTED | CD_WILL_CLUSTER | CD_HAS_CLUSTERED | CD_WAS_OFTEN);
    }

/*
 * Get rid of references to obsolete files
 */
void ClusteringManagerData::cleanReferences()
    {
    }

void ClusteringManagerData::markDirty() // Set the dirty bit
    {
    /*
     * Since marking something dirty is expensive, we don't do it
     * unless we have to.  This also prevents infinite recursion.
     */
    if ((flags & CD_DIRTY) == 0)
	{
	flags |= CD_DIRTY;
	for (ClusteringAllClusterIterator member(*this);  member;  ++member)
	    {
	    /*
	     * NEEDSWORK: can we get away with just setting the dirty
	     * bit here?  I don't think so, but I'm worried about the
	     * cascading effects of marking every file in every
	     * cluster.
	     */
	    for (Cluster::AllElementIterator file(*member);  file;  ++file)
		managerData(*file).markDirty();
	    }
	}
    }

/*
 * Take note of changed relations.  By default, managers don't care,
 * so this is a no-op unless overridden.
 */
void ClusteringManagerData::relationsChanged()
    {
    markDirty();
    }

/*
 * Read the total-references field from a stream.  This is needed for
 * backwards compatibility.
 */
ByteSwappedStream& ClusteringManagerData::readTotalReferences(
    ByteSwappedStream&	stream)		// Stream to read from
    {
    stream.read(&totalReferences, sizeof totalReferences);
    return stream;
    }

/*
 * Return a reference to the clustering manager's
 * file-private data, so that manager routines can access it
 * recursively.
 */
ClusteringManagerData& ClusteringManagerData::managerData(
    File&		file)
    const
    {
    return file.clusteringManagerData;
    }

const ClusteringManagerData& ClusteringManagerData::managerData(
    const File&		file)
    const
    {
    return file.clusteringManagerData;
    }
