#ifndef _COMMUNICATION_HH
#define _COMMUNICATION_HH
/*
 * $Id: communication.hh,v 1.20 1998/10/23 23:15:04 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: communication.hh,v $
 * Revision 1.20  1998/10/23 23:15:04  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.19  1998/07/15 21:43:32  geoff
 * Get rid of CONTROL_RESTART; it's too painful to keep supporting.
 *
 * Revision 1.18  1998/06/02 23:59:37  geoff
 * Add SS_SYMLINK.
 *
 * Revision 1.17  1997/07/11 17:40:55  geoff
 * Add preliminary support for including sizes in packets, and for Web
 * trace replay.
 *
 * Revision 1.16  1996/10/11 22:08:54  geoff
 * Minor comment fix.
 *
 * Revision 1.15  1996/10/04 07:11:57  geoff
 * Add support for reference-time setting, command-interpreter
 * references, and a magic number for the correlator database.
 *
 * Revision 1.14  1996/09/11 22:59:37  geoff
 * Add the StorageSetting enum.
 *
 * Revision 1.13  1996/05/03  01:11:47  geoff
 * Update the copyright.  Add CONTROL_READ_INSTRUCTIONS.
 *
 * Revision 1.12  1996/04/10  01:28:24  geoff
 * Add CONTROL_SET_STORAGE.
 *
 * Revision 1.11  1995/12/19  01:02:31  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.10  1995/12/19  00:34:05  geoff
 * Add a lot of process-tracking support.
 *
 * Revision 1.9  1995/11/22  00:19:50  geoff
 * Get rid of the TIME_SORTED_INPUT ifdef.
 *
 * Revision 1.8  1995/11/10  02:58:25  geoff
 * Document the reasoning behind MAX_PACKET_LENGTH, and add it.  Add the
 * SEER_REF_CONTROL constant and the ConnectionControl type.
 *
 * Revision 1.7  1995/09/15  00:04:59  geoff
 * Move the seer reference types here from seer.h, and make them an enum.
 *
 * Revision 1.6  1995/08/07  20:02:24  geoff
 * Add CONTROL_SET_INVESTIGATION.
 *
 * Revision 1.5  1995/05/27  00:15:19  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.4  1995/05/23  00:39:12  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.3  1995/04/20  21:57:15  geoff
 * Add a control command for running management and dumping the results.
 * Make the arguments to connectServer be enums.
 *
 * Revision 1.2  1995/03/29  21:03:52  geoff
 * Add the reportFailures argument to connectServer.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

/*
 * This file contains definitions that need to be shared among various
 * communicating processes.
 */

#include <sys/param.h>

/*
 * Connection types, sent as bytes upon connection establishment
 */
const int OBSERVER_CONNECTION =	0;	// Connection made from observer
const int CONTROL_CONNECTION =	1;	// Connection made from controller

/*
 * Control commands sent to the correlator
 */
const int CONTROL_EXIT =	'x';	// Exit correlator
const int CONTROL_GETPARMS =	'g';	// Get correlator parameters
const int CONTROL_READ_INSTRUCTIONS = 'I'; // (Re)read instruction file
const int CONTROL_SET_INVESTIGATION = 'i'; // Set investigator data
const int CONTROL_MANAGE_DUMP =	'M';	// Run management and dump results
const int CONTROL_MANAGE =	'm';	// Run file management
const int CONTROL_SET_REFERENCE = 'R';	// Set particular reference times
const int CONTROL_SETPARMS =	'p';	// Set correlator parameters
const int CONTROL_SET_STORAGE =	'S';	// Set file storage status
const int CONTROL_SAVE =	's';	// Save correlation table in a file

/*
 * Storage settings, sent by CONTROL_SET_STORAGE
 */
enum StorageSetting
    {
    SS_STORE,				// Store the file locally
    SS_STORE_AND_REFERENCE,		// Store file and set reference time
    SS_UNSTORE,				// Don't store the file locally
    SS_REFERENCE,			// Set the file's reference time
    SS_SYMLINK				// Set the file's symlink status
    };

enum ConnectionBlocking
    {
    blockingConnection,			// Use a blocking connection
    nonBlockingConnection		// Open in non-blocking mode
    };

enum ConnectionFailureReporting
    {
    silentFailures,			// Keep silent about failures
    reportFailures			// Report failures to cerr
    };

/*
 * Packets are formatted as follows:
 *
 *	Byte(s)		Contents
 *	-------		--------
 *	0		Reference type
 *	1-8		Timestamp, seconds/usec, little-endian
 *	9-12		Process ID, little-endian
 *	13-14		Length of path 1, little-endian, not including null
 *	15-15+n1	Path 1, null-terminated
 *	15+n1-16+n1	Length of path 2, little-endian, not including null
 *	17+n1-17+n1+n2	Path 2, null-terminated
 *
 * Path 2 is included only if the reference type is SEER_REF_RENAME.
 * Each path is limited to a maximum size of MAXPATHLEN.
 */
const unsigned int MAX_PACKET_LENGTH = 1 + 8 + 4 + (MAXPATHLEN + 2 + 1) * 2;

enum ReferenceType
    {
    SEER_REF_UNKNOWN,			// Unknown reference type
    SEER_REF_CONTROL,			// Fake reference for conn. control
    SEER_REF_NEWPROC,			// Report ID of a new process
    SEER_REF_PROCEXIT,			// Report exit of a process
    SEER_REF_READ,		    	// Read of a file
    SEER_REF_RW,		    	// Reads and/or writes of existing file
    SEER_REF_CREATE,			// Create or truncate/rewrite
    SEER_REF_EXEC,		    	// Execution
    SEER_REF_DESTROY,			// Destruction (delete)
    SEER_REF_RENAME,			// Change of file name
    SEER_REF_OPENREAD,			// Open for read
    SEER_REF_OPENRW,			// Open for read/write
    SEER_REF_OPENCREATE,	    	// Create or truncate/rewrite
    SEER_REF_CLOSE,		    	// Close of open file
    SEER_REF_EXIT,		    	// Process exit (or exec of new file)
    SEER_REF_PEEK,		    	// Peek at a file (stat or access)
    SEER_REF_MAKELINK,			// Make a symbolic link
    SEER_REF_PEEKLINK,			// Peek at a symbolic link's contents
    SEER_REF_FOLLOW,			// Follow a symbolic link
    SEER_REF_DUP,		    	// Duplicate an existing file reference
    SEER_REF_INTERPRET			// Invoke a command interpreter
    };

/*
 * Control commands (for SEER_REF_CONTROL).  Must be nonzero and
 * single-byte.
 */
enum ConnectionControl
    {
    replayMode = 'r',			// A save file is being replayed
    realTimeMode = 't',			// Observations are real-time
    sizeMode = 's',			// Sizes are included in packets
    webMode = 'w'			// Web trace is being replayed
    };

/*
 * Magic numbers for save files.
 */
const int CORRELATOR_DATABASE_MAGIC = 0x10221553;
					// Magic written to database save file

extern int		connectServer(char* serverid, int type,
			  ConnectionBlocking blockingType,
			  ConnectionFailureReporting reportingMethod);

#endif // _COMMUNICATION_HH
