#ifndef lint
static const char Rcs_Id[] =
    "$Id: dumpdatabase.cc,v 1.24 1998/10/23 23:15:08 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: dumpdatabase.cc,v $
 * Revision 1.24  1998/10/23 23:15:08  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.23  1998/07/15 21:43:35  geoff
 * Add dumping support for the latest 2 database formats.
 *
 * Revision 1.22  1997/02/12 02:52:04  geoff
 * Handle a floating-point oftenPercent.
 *
 * Revision 1.21  1996/12/03 02:58:00  geoff
 * Add support for databases that record manager parameters.
 *
 * Revision 1.20  1996/10/11 22:09:01  geoff
 * Get rid of the workaround for the floating-output bug.  Add
 * geometric-mean support.
 *
 * Revision 1.19  1996/10/08 19:50:04  geoff
 * Add code to work around the floating-point output bug.
 *
 * Revision 1.18  1996/10/04 07:12:04  geoff
 * Add support for a magic number and different versions in the
 * correlator database.  Add support for command-interpreter references.
 *
 * Revision 1.17  1996/09/23 23:41:23  geoff
 * Fix a couple of missing blanks in messages.  Fix a place where the
 * wrong path was reported in an error message.
 *
 * Revision 1.16  1996/06/11 07:04:52  geoff
 * Get rid of READ/WRITE EXTRAS.
 *
 * Revision 1.15  1996/06/11  06:13:09  geoff
 * Add code to dump the reference flags field, investigated clusters, the
 * file type byte, and the peek counts.
 *
 * Revision 1.14  1996/05/03  01:11:59  geoff
 * Update the copyright.  Change what's supported in the save file
 * (READ/WRITE_EXTRAS).
 *
 * Revision 1.13  1996/04/10  06:09:29  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.12  1996/04/10  01:28:28  geoff
 * Add support for the new UNSTORED file state.  Get rid of the obsolete
 * all-managers file storage state.
 *
 * Revision 1.11  1996/03/19  18:42:52  geoff
 * Add code to dump the bounded and weighted LRU manager's data.
 *
 * Revision 1.10  1996/02/02  18:28:51  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.9  1995/12/19  01:02:50  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.8  1995/12/19  00:34:15  geoff
 * Add process-tracking support.
 *
 * Revision 1.7  1995/11/22  00:20:07  geoff
 * Fix the RCS comment leader.  Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.6  1995/11/21  21:37:53  geoff
 * Add support for dumping investigator data in the database.
 *
 * Revision 1.5  1995/11/10  02:58:57  geoff
 * Add zzinlinefunctions.hh.  Add NEW_FILE_RELATIONS and SEER_REF_CONTROL
 * support.
 *
 * Revision 1.4  1995/09/15  00:05:15  geoff
 * Change the RCS ID to be a const.  Add support for compressed input
 * files.
 *
 * Revision 1.3  1995/08/07  20:02:42  geoff
 * Add SEER_REF_DUP support.
 *
 * Revision 1.2  1995/06/30  19:04:31  geoff
 * Give file offsets on most error messages, and in front of every output
 * line.  Support MAKELINK and PEEKLINK references.
 *
 * Revision 1.1  1995/06/17  00:12:59  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "communication.hh"
#include "zzinlinefunctions.hh"
#include <errno.h>
#include <string.h>
#include <sys/param.h>

int			main(int argc, char* argv[]);
static void		dumpCorrelatorDatabase(ByteSwappedStream& file);
static void		dumpManagerParameters(ByteSwappedStream& file);

/*
 * Values stolen from various header files.  This is done because the
 * header files are deeply intertwined with the correlator, and can't
 * be easily disentangled.
 */

/*
 * Values from file.hh
 */

/*
 * File types (that we care about)
 */
enum FileType
    {
    FT_FILE,			// Plain file or other "thing" we hoard
    FT_DIR,			// Directory
    FT_FIFO,			// FIFO (named pipe)
    FT_CHARACTER_SPECIAL,	// Character device
    FT_BLOCK_SPECIAL,		// Block device
    FT_SYMLINK,			// Symbolic link
    FT_SOCK			// BSD-style socket
    };

/*
 * Possible file states
 */
enum FileState
    {
    FS_UNKNOWN,			// File state is unknown
    FS_KNOWN,			// We know where the file is stored
    FS_UNSTORED,		// The file exists, but not locally
    FS_MISSING,			// File no longer exists
    FS_FORGETTING,		// We want to forget the file
    FS_FORGOTTEN		// We've forgotten the file, can delete
    };

/*
 * Values from filereference.hh
 */
typedef unsigned long ReferenceCount;

const int		RF_CORRECTED = 0x01; // NPREV corr has been applied


/*
 * Values from clustering_mgr.hh
 */
const int		CD_WEIGHTED = 0x01; // Weighting has been done on file
const int		CD_WILL_CLUSTER = 0x02; // This file can be clustered
const int		CD_HAS_CLUSTERED = 0x04; // This file's been clustered
const int		CD_IS_NEIGHBOR = 0x08; // This file is a nearest
					// ..neighbor (see isSharedNearby)
const int		CD_WAS_OFTEN = 0x10; // This file was isOften at last
					// ..clustering (see requiredSize)
const int		CD_DIRTY = 0x20; // This file will cluster differently
					// ..the next time clustering is run

static int		fileVersion = 0; // Database version number

extern "C" char*	strerr(int errnum);

int main(				// Dump an observer trace file
    int			argc,		// Argument count
    char*		argv[])		// Argument vector
    {
    int trouble = 0;
    if (argc == 1)
	{
	ByteSwappedStream stdinStream(0, ByteSwappedStream::bigEnded);
	dumpCorrelatorDatabase(stdinStream);
	}
    else
	{
	for (int argno = 1;  argno < argc;  argno++)
	    {
	    ByteSwappedStream* file = new ByteSwappedStream(argv[argno],
	      ios::in, ByteSwappedStream::bigEnded);
	    if (!*file)
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  couldn't open '" << argv[argno]
		  << "': " << strerr(saveErrno) << '\n';
		trouble = 1;
		}
	    else
		dumpCorrelatorDatabase(*file);
	    delete file;
	    file = NULL;
	    }
	}
    return trouble;
    }

static void dumpCorrelatorDatabase(	// Dump a correlator database
    ByteSwappedStream&		file)	// File to dump from
    {
    file.readEndedness();
    /*
     * Try to find a magic number and a version.  This may not exist,
     * so be flexible.
     */
    unsigned long magic;
    if (file.read(&magic))
	{
	if (magic == CORRELATOR_DATABASE_MAGIC)
	    file.read(&fileVersion);
	else
	    {
	    /*
	     * Old database files don't have a magic number.  Rewind
	     * and try again.
	     */
	    file.seekg(0, ios::beg);
	    file.readEndedness();
	    }
	}
    if (fileVersion > 6)
	{
	cerr << "dumpdatabase:  unrecognized file version " << fileVersion
	  << '\n';
	return;
	}
    cerr << "0: File version " << fileVersion << '\n';
    if (fileVersion > 2)
	dumpManagerParameters(file);
    bool geometricMeans;
    if (fileVersion <= 1)
	geometricMeans = false;
    else
	{
	char flag;
	file.read(&flag, sizeof flag);
	geometricMeans = flag;
	}
    /*
     * Rather than using the file object's I/O routines, which would
     * pull in all the managers and give me all sorts of linking
     * problems, as well as invoking File::findCanonical, this routine
     * just uses ad-hoc knowledge of the database structure.  That
     * introduces a maintenance problem, of course, but I consider it
     * the lesser of two evils.
     */
    unsigned char relationTypeCount;
    if (!file.read(&relationTypeCount, sizeof relationTypeCount))
	{
	int saveErrno = errno;
	cerr << "dumpdatabase:  trouble looking for "
	  << "relation type count: " << strerr(saveErrno) << '\n';
	return;
	}
    if (relationTypeCount <= 0)
	{
	cerr << "dumpdatabase:  bad relation type count "
	  << (int)relationTypeCount << '\n';
	return;
	}
    char** relationTypeNames = new char*[relationTypeCount];
    unsigned char* relationIds = new unsigned char[relationTypeCount];
    int relationType;
    for (relationType = 0;  relationType < relationTypeCount;  relationType++)
	{
	unsigned long recordPosition = file.bytesRead();
	if (!file.read(&relationIds[relationType], sizeof relationIds[0]))
	    {
	    int saveErrno = errno;
	    cerr << "dumpdatabase:  trouble looking for relation ID "
	      << "for relation " << relationType << ": " << strerr(saveErrno)
	      << '\n';
	    delete[] relationTypeNames;
	    delete[] relationIds;
	    return;
	    }
	for (int duplicateType = 0;
	  duplicateType < relationType;
	  duplicateType++)
	    {
	    if (relationIds[duplicateType] == relationIds[relationType])
		{
		cerr << "dumpdatabase:  relation ID "
		  << (int)relationIds[relationType]
		  << " already used for relation '"
		  << relationTypeNames[duplicateType] << "'\n";
		delete[] relationTypeNames;
		delete[] relationIds;
		return;
		}
	    }
	short relationTypeNameLength;
	if (!file.read(&relationTypeNameLength, sizeof relationTypeNameLength))
	    {
	    int saveErrno = errno;
	    cerr << "dumpdatabase:  trouble looking for relation type length "
	      << "for relation " << relationType << ": " << strerr(saveErrno)
	      << '\n';
	    delete[] relationTypeNames;
	    delete[] relationIds;
	    return;
	    }
	if (relationTypeNameLength <= 0
	  ||  relationTypeNameLength >= MAXPATHLEN)
	    {
	    cerr << "dumpdatabase:  bad relation type name length "
	      << relationTypeNameLength << '\n';
	    delete[] relationTypeNames;
	    delete[] relationIds;
	    return;
	    }
	relationTypeNames[relationType] = new char[relationTypeNameLength + 1];
	if (!file.read(relationTypeNames[relationType],
	  relationTypeNameLength + 1))
	    {
	    int saveErrno = errno;
	    cerr << "dumpdatabase:  trouble looking for relation type length "
	      << "for relation " << relationType << ": " << strerr(saveErrno)
	      << '\n';
	    delete relationTypeNames[relationType];
	    delete[] relationTypeNames;
	    delete[] relationIds;
	    return;
	    }
	if ((int)strlen(relationTypeNames[relationType])
	  != relationTypeNameLength)
	    {
	    cerr << "dumpdatabase:  relation type length "
	      << relationTypeNameLength << " doesn't match name length "
	      << strlen(relationTypeNames[relationType]) << " for relation '"
	      << (int)relationIds[relationType] << '\n';
	    delete relationTypeNames[relationType];
	    delete[] relationTypeNames;
	    delete[] relationIds;
	    return;
	    }
	cout << recordPosition << " Relation "
	  << (int)relationIds[relationType] << ' '
	  << relationTypeNames[relationType];
	if (strchr(relationTypeNames[relationType], ':') == NULL)
	    cout << '\n';
	else
	    {
	    double overallWeight;
	    if (!file.read(&overallWeight))
		{
		cerr
		  << "dumpdatabase:  trouble reading overall weight at offset "
		  << file.bytesRead() << '\n';
		return;
		}
	    cout << ' ' << overallWeight << '\n';
	    if (strncmp(relationTypeNames[relationType], "Cluster:", 8) == 0)
		{
		while (1)
		    {
		    unsigned int clusterNameLength;
		    recordPosition = file.bytesRead();
		    if (!file.read(&clusterNameLength))
			{
			cerr << "dumpdatabase: trouble reading cluster name"
			    " length at offset "
			  << file.bytesRead() << '\n';
			return;
			}
		    char clusterName[MAXPATHLEN];
		    if (clusterNameLength == 0)
			break;
		    else if (clusterNameLength >= sizeof clusterName)
			{
			cerr << "dumpdatabase:  bad cluster name length "
			  << clusterNameLength << " at offset "
			  << file.bytesRead() << '\n';
			}
		    if (!file.read(clusterName, clusterNameLength + 1))
			{
			cerr << "dumpdatabase:  trouble reading cluster name"
			    " at offset "
			  << file.bytesRead() << '\n';
			return;
			}
		    if (clusterNameLength != strlen(clusterName))
			{
			cerr << "dumpdatabase:  cluster name length mismatch ("
			  << clusterNameLength << '/' << strlen(clusterName)
			  << ") at offset " << file.bytesRead() << '\n';
			return;
			}
		    cout << '\t' << recordPosition << " Cluster "
		      << clusterName << ":\n";
		    while (1)
			{
			unsigned int fileNameLength;
			recordPosition = file.bytesRead();
			if (!file.read(&fileNameLength))
			    {
			    cerr << "dumpdatabase: trouble reading file name"
				" length at offset "
			      << file.bytesRead() << '\n';
			    return;
			    }
			char fileName[MAXPATHLEN];
			if (fileNameLength == 0)
			    break;
			else if (fileNameLength >= sizeof fileName)
			    {
			    cerr << "dumpdatabase:  bad file name length "
			      << fileNameLength << " at offset "
			      << file.bytesRead() << '\n';
			    }
			if (!file.read(fileName, fileNameLength + 1))
			    {
			    cerr << "dumpdatabase:  trouble reading file name"
				" at offset "
			      << file.bytesRead() << '\n';
			    return;
			    }
			if (fileNameLength != strlen(fileName))
			    {
			    cerr
			      << "dumpdatabase:  file name length mismatch ("
			      << fileNameLength << '/' << strlen(fileName)
			      << ") at offset " << file.bytesRead() << '\n';
			    return;
			    }
			cout << "\t    " << recordPosition << ' ' << fileName
			  << '\n';
			}
		    }
		}
	    }
	}
    unsigned int pathLength;
    while (file.read(&pathLength, sizeof pathLength))
	{
	unsigned long recordPosition = file.bytesRead() - sizeof pathLength;
	char pathBuffer[MAXPATHLEN];
	if (pathLength <= 0  ||  pathLength >= sizeof pathBuffer)
	    {
	    cerr << "dumpdatabase:  bad path length " << pathLength
	      << " at offset " << file.bytesRead() - sizeof pathLength << '\n';
	    return;
	    }
	if (!file.read(pathBuffer, pathLength + 1))
	    {
	    cerr << "dumpdatabase:  trouble reading file path at offset "
	      << file.bytesRead() << '\n';
	    return;
	    }
	if (strlen(pathBuffer) != pathLength)
	    {
	    cerr << "dumpdatabase: path length " << pathLength
	      << " doesn't match name length " << strlen(pathBuffer)
	      << " for file '" << pathBuffer << "' at offset "
	      << file.bytesRead() - pathLength - 1 - sizeof pathLength << '\n';
	    return;
	    }
	char typeByte;
	if (!file.read(&typeByte, sizeof typeByte))
	    {
	    cerr << "dumpdatabase: trouble reading type byte for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	char stateByte;
	if (!file.read(&stateByte, sizeof stateByte))
	    {
	    cerr << "dumpdatabase: trouble reading state byte for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	unsigned long blocks;
	if (!file.read(&blocks, sizeof blocks))
	    {
	    cerr << "dumpdatabase: trouble reading file size for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	ReferenceCount totalReferences;
	ReferenceCount lastReference;
	if (fileVersion <= 5)
	    {
	    if (!file.read(&totalReferences, sizeof totalReferences))
		{
		cerr << "dumpdatabase: trouble reading total references"
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    if (!file.read(&lastReference, sizeof lastReference))
		{
		cerr << "dumpdatabase: trouble reading last reference"
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	unsigned char charReferenceType;
	if (!file.read(&charReferenceType, sizeof charReferenceType))
	    {
	    cerr
	      << "dumpdatabase: trouble reading last reference type for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	ReferenceType lastReferenceType = (ReferenceType)charReferenceType;
	ReferenceCount peekCount;
	if (!file.read(&peekCount, sizeof peekCount))
	    {
	    cerr << "dumpdatabase: trouble reading peek count for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	ReferenceCount potentialPeeks;
	if (!file.read(&potentialPeeks, sizeof potentialPeeks))
	    {
	    cerr << "dumpdatabase: trouble reading potential peeks for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	/*
	 * We've read the constant non-manager part of the file.  Display it.
	 */
	cout << recordPosition << ' ' << pathBuffer
	  << ' '
	  << (typeByte == (char) FT_FILE ? "FT_FILE"
	    : typeByte == (char) FT_DIR ? "FT_DIR"
	    : typeByte == (char) FT_FIFO ? "FT_FIFO"
	    : typeByte == (char) FT_CHARACTER_SPECIAL ? "FT_CHARACTER_SPECIAL"
	    : typeByte == (char) FT_BLOCK_SPECIAL ? "FT_BLOCK_SPECIAL"
	    : typeByte == (char) FT_SYMLINK ? "FT_SYMLINK"
	    : typeByte == (char) FT_SOCK ? "FT_SOCK"
	    : "BAD_FILE_TYPE")
	  << ' '
	  << (stateByte == (char) FS_UNKNOWN ? "UNKNOWN"
	    : stateByte == (char) FS_KNOWN ? "KNOWN"
	    : stateByte == (char) FS_UNSTORED ? "UNSTORED"
	    : stateByte == (char) FS_MISSING ? "MISSING"
	    : stateByte == (char) FS_FORGETTING ? "FORGETTING"
	    : stateByte == (char) FS_FORGOTTEN ? "FORGOTTEN"
	    : "BAD_FILE_STATE")
	  << ' ' << blocks;
	if (fileVersion <= 5)
	    cout << ' ' << totalReferences << ' ' << lastReference;
	cout << ' '
	  << (lastReferenceType == SEER_REF_UNKNOWN ? "SEER_REF_UNKNOWN"
	    : lastReferenceType == SEER_REF_CONTROL ? "SEER_REF_CONTROL"
	    : lastReferenceType == SEER_REF_NEWPROC ? "SEER_REF_NEWPROC"
	    : lastReferenceType == SEER_REF_PROCEXIT ? "SEER_REF_PROCEXIT"
	    : lastReferenceType == SEER_REF_READ ? "SEER_REF_READ"
	    : lastReferenceType == SEER_REF_RW ? "SEER_REF_RW"
	    : lastReferenceType == SEER_REF_CREATE ? "SEER_REF_CREATE"
	    : lastReferenceType == SEER_REF_EXEC ? "SEER_REF_EXEC"
	    : lastReferenceType == SEER_REF_DESTROY ? "SEER_REF_DESTROY"
	    : lastReferenceType == SEER_REF_RENAME ? "SEER_REF_RENAME"
	    : lastReferenceType == SEER_REF_OPENREAD ? "SEER_REF_OPENREAD"
	    : lastReferenceType == SEER_REF_OPENRW ? "SEER_REF_OPENRW"
	    : lastReferenceType == SEER_REF_OPENCREATE
		? "SEER_REF_OPENCREATE"
	    : lastReferenceType == SEER_REF_CLOSE ? "SEER_REF_CLOSE"
	    : lastReferenceType == SEER_REF_EXIT ? "SEER_REF_EXIT"
	    : lastReferenceType == SEER_REF_PEEK ? "SEER_REF_PEEK"
	    : lastReferenceType == SEER_REF_MAKELINK ? "SEER_REF_MAKELINK"
	    : lastReferenceType == SEER_REF_PEEKLINK ? "SEER_REF_PEEKLINK"
	    : lastReferenceType == SEER_REF_FOLLOW ? "SEER_REF_FOLLOW"
	    : lastReferenceType == SEER_REF_DUP ? "SEER_REF_DUP"
	    : lastReferenceType == SEER_REF_INTERPRET ? "SEER_REF_INTERPRET"
 	    : "BAD_REF_TYPE")
	  << ' ' << peekCount << ' ' << potentialPeeks
	  << '\n';
	/*
	 * Dump relation data
	 */
	short relatedFileCount;
	if (!file.read(&relatedFileCount, sizeof relatedFileCount))
	    {
	    cerr
	      << "dumpdatabase: trouble reading related-file count for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	for (short relatedFileNumber = 0;
	  relatedFileNumber < relatedFileCount;
	  relatedFileNumber++)
	    {
	    unsigned short relationCount;
	    if (!file.read(&relationCount, sizeof relationCount))
		{
		cerr
		  << "dumpdatabase: trouble reading relation Count for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    recordPosition = file.bytesRead();
	    unsigned int targetPathLength;
	    if (!file.read(&targetPathLength, sizeof targetPathLength))
		{
		cerr << "dumpdatabase: trouble reading target path length for"
		  << " related file " << relatedFileNumber << " of file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    char targetPathBuffer[MAXPATHLEN];
	    if (targetPathLength <= 0
	      ||  targetPathLength >= sizeof targetPathBuffer)
		{
		cerr << "dumpdatabase: bad path length " << targetPathLength
		  << " for related file " << relatedFileNumber << " of file '"
		  << pathBuffer << "' at offset "
		  << file.bytesRead() - sizeof targetPathLength<< '\n';
		return;
		}
	    if (!file.read(targetPathBuffer, targetPathLength + 1))
		{
		cerr << "dumpdatabase: trouble reading target path"
		    " for related file "
		  << relatedFileNumber << " of file '" << pathBuffer
		  << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    if (strlen(targetPathBuffer) != targetPathLength)
		{
		cerr << "dumpdatabase: target path length "
		  << targetPathLength << " doesn't match name length "
		  << strlen(targetPathBuffer) << " for file '"
		  << targetPathBuffer << "' at offset "
		  << file.bytesRead() - targetPathLength - 1
		    - sizeof targetPathLength
		  << '\n';
		return;
		}
	    cout << "    " << recordPosition << ' ' << targetPathBuffer
	      << '\n';
	    for (int relationNumber = 0;
	      relationNumber < relationCount;
	      relationNumber++)
		{
		recordPosition = file.bytesRead();
		unsigned char relationId;
		if (!file.read(&relationId, sizeof relationId))
		    {
		    cerr << "dumpdatabase: trouble reading type ID for file '"
		      << pathBuffer << "' at offset " << file.bytesRead()
		      << '\n';
		    return;
		    }
		int typeId;
		for (typeId = 0;  typeId < relationTypeCount;  typeId++)
		    {
		    if (relationIds[typeId] == relationId)
			break;
		    }
		/*
		 * NEEDSWORK:  The following is *exceedingly* ad-hoc!
		 */
		cout << '\t' << recordPosition << ' '
		  << relationTypeNames[typeId] << ' ';
		if (strcmp(relationTypeNames[typeId], "FileReference") == 0)
		    {
		    ReferenceCount numReferences;
		    if (!file.read(&numReferences, sizeof numReferences))
			{
			cerr << "dumpdatabase: trouble reading number "
			  << "of references for file '" << pathBuffer
			  << "' at offset " << file.bytesRead() << '\n';
			return;
			}
		    ReferenceCount totalDistance;
		    double totalLogDistance;
		    if (geometricMeans)
			{
			if (!file.read(&totalLogDistance,
			  sizeof totalLogDistance))
			    {
			    cerr
			      << "dumpdatabase: trouble reading total log"
			      << " distance for file '" << pathBuffer
			      << "' at offset " << file.bytesRead() << '\n';
			    return;
			    }
			}
		    else
			{
			if (!file.read(&totalDistance, sizeof totalDistance))
			    {
			    cerr
			      << "dumpdatabase: trouble reading total distance"
			      << " for file '" << pathBuffer
			      << "' at offset " << file.bytesRead() << '\n';
			    return;
			    }
			}
		    char flags;
		    if (!file.read(&flags, sizeof flags))
			{
			cerr << "dumpdatabase: trouble reading flags"
			  << " for file '" << pathBuffer
			  << "' at offset " << file.bytesRead() << '\n';
			return;
			}
		    cout << numReferences << ' ';
		    if (geometricMeans)
			{
			cout << totalLogDistance;
			}
		    else
			cout << totalDistance;
		    cout << ' '
		      << (flags & RF_CORRECTED ? "|RF_CORRECTED|" : "|");
		    }
		else if (strncmp(relationTypeNames[typeId], "Relation:", 9)
		  == 0)
		    {
		    double weight;
		    if (!file.read(&weight, sizeof weight))
			{
			cerr << "dumpdatabase:  trouble reading weight "
			    "for file '"
			  << pathBuffer << "' at offset " << file.bytesRead()
			  << '\n';
			return;
			}
		    cout << weight;
		    }
		else if (strncmp(relationTypeNames[typeId], "Cluster:", 8)
		  == 0)
		    {
		    // Clusters have no extra information to display
		    }
		else
		    {
		    cout << '\n';
		    cerr << "dumpdatabase:  unknown type name "
		      << relationTypeNames[typeId] << " for file '"
		      << pathBuffer << "' at offset " << file.bytesRead()
		      << '\n';
		    return;
		    }
		cout << '\n';
		}
	    }
	/*
	 * Dump clustering manager data
	 */
	recordPosition = file.bytesRead();
	unsigned char clusteringFlags;
	if (fileVersion <= 5)
	    {
	    if (!file.read(&clusteringFlags, sizeof clusteringFlags))
		{
		cerr << "dumpdatabase: trouble reading clustering flags"
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	else
	    {
	    if (!file.read(&lastReference, sizeof lastReference))
		{
		cerr << "dumpdatabase: trouble reading last reference "
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	if (fileVersion > 5)
	    {
	    if (!file.read(&totalReferences, sizeof totalReferences))
		{
		cerr << "dumpdatabase: trouble reading total references "
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	cout << "    " << recordPosition << " ClusteringManager:";
	if (fileVersion <= 5)
	    {
	    cout << " |"
	      << ((clusteringFlags & CD_WEIGHTED) ? "WEIGHTED|" : "")
	      << ((clusteringFlags & CD_WILL_CLUSTER) ? "WILL_CLUSTER|" : "")
	      << ((clusteringFlags & CD_HAS_CLUSTERED) ? "HAS_CLUSTERED|" : "")
	      << ((clusteringFlags & CD_IS_NEIGHBOR) ? "IS_NEIGHBOR|" : "")
	      << ((clusteringFlags & CD_WAS_OFTEN) ? "WAS_OFTEN|" : "")
	      << ((clusteringFlags & CD_DIRTY) ? "DIRTY|" : "");
	    }
	else
	    cout << " lastReference " << lastReference;
	if (fileVersion > 5)
	    cout << " totalReferences " << totalReferences;
	cout << '\n';
	/*
	 * Dump LRU manager data
	 */
	recordPosition = file.bytesRead();
	if (fileVersion > 5)
	    {
	    if (!file.read(&lastReference, sizeof lastReference))
		{
		cerr << "dumpdatabase: trouble reading last reference "
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    cout << "    " << recordPosition << " LRU_Manager:";
	    cout << " lastReference " << lastReference;
	    cout << '\n';
	    }
	/*
	 * Dump bounded LRU manager data
	 */
	recordPosition = file.bytesRead();
	if (fileVersion > 5)
	    {
	    if (!file.read(&lastReference, sizeof lastReference))
		{
		cerr << "dumpdatabase: trouble reading last reference "
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	unsigned long boundedLRU_Bound;
	if (!file.read(&boundedLRU_Bound, sizeof boundedLRU_Bound))
	    {
	    cerr << "dumpdatabase: trouble reading LRU bound"
		" for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	cout << "    " << recordPosition << " BoundedLRU_Manager:";
	if (fileVersion > 5)
	    cout << " lastReference " << lastReference;
	cout << " bound " <<  boundedLRU_Bound;
	cout << '\n';
	if (fileVersion > 0)
	    {
	    /*
	     * Dump linear LRU manager data
	     */
	    recordPosition = file.bytesRead();
	    if (fileVersion > 5)
		{
		if (!file.read(&lastReference, sizeof lastReference))
		    {
		    cerr << "dumpdatabase: trouble reading last reference "
			" for file '"
		      << pathBuffer << "' at offset " << file.bytesRead()
		      << '\n';
		    return;
		    }
		}
	    double linearLRU_Intercept;
	    double linearLRU_Slope;
	    double linearLRU_Bound;
	    if (!file.read(&linearLRU_Intercept, sizeof linearLRU_Intercept)
	      ||  !file.read(&linearLRU_Slope, sizeof linearLRU_Slope)
	      ||  !file.read(&linearLRU_Bound, sizeof linearLRU_Bound))
		{
		cerr << "dumpdatabase: trouble reading linear LRU parameters"
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    cout << "    " << recordPosition << " LinearLRU_Manager:";
	    if (fileVersion > 5)
		cout << " lastReference " << lastReference;
	    cout << " intercept " << linearLRU_Intercept
	      << " slope " << linearLRU_Slope
	      << " bound " << linearLRU_Bound;
	    cout << '\n';
	    }
	/*
	 * Dump weighted LRU manager data
	 */
	recordPosition = file.bytesRead();
	if (fileVersion > 5)
	    {
	    if (!file.read(&lastReference, sizeof lastReference))
		{
		cerr << "dumpdatabase: trouble reading last reference "
		    " for file '"
		  << pathBuffer << "' at offset " << file.bytesRead() << '\n';
		return;
		}
	    }
	double weightedLRU_Weight;
	if (!file.read(&weightedLRU_Weight, sizeof weightedLRU_Weight))
	    {
	    cerr << "dumpdatabase: trouble reading LRU weight"
		" for file '"
	      << pathBuffer << "' at offset " << file.bytesRead() << '\n';
	    return;
	    }
	cout << "    " << recordPosition << " WeightedLRU_Manager:";
	if (fileVersion > 5)
	    cout << " lastReference " << lastReference;
	cout << " weight " << weightedLRU_Weight;
	cout << '\n';
	}
    }

static void dumpManagerParameters(	// Dump various managers' parameters
    ByteSwappedStream&		file)	// File to dump from
    {
    int nameLength;
    for (unsigned long recordPosition = file.bytesRead();
      file.read(&nameLength);
      recordPosition = file.bytesRead())
	{
	char managerName[MAXPATHLEN];
	if (nameLength == 0)
	    return;
	else if (nameLength < 0  ||  nameLength > (int)sizeof managerName - 1)
	    {
	    cerr << "dumpdatabase: bad manager name length " << nameLength
	      << " at offset " << recordPosition << '\n';
	    return;
	    }
	if (!file.read(managerName, nameLength + 1))
	    {
	    cerr << "dumpdatabase: trouble reading manager name\n";
	    return;
	    }
	if ((int)strlen(managerName) != nameLength)
	    {
	    cerr << "dumpdatabase: bad name length " << strlen(managerName)
	      << " for '" << managerName << "', expected "
	      << nameLength << '\n';
	    return;
	    }
	long parameterBytes;
	if (!file.read(&parameterBytes))
	    {
	    int saveErrno = errno;
	    cerr << "dumpdatabase:  trouble looking for parameter length "
	      << "for manager " << managerName << ": " << strerr(saveErrno)
	      << '\n';
	    return;
	    }
	/* 
	 * Read parameters that are common to all managers.
	 */
	unsigned long managementInterval;
	unsigned long firstMissReference;
	unsigned long missCount;
	if (fileVersion > 5)
	    {
	    parameterBytes -= sizeof managementInterval;
	    if (parameterBytes < 0)
		{
		cerr << "dumpdatabase:  insufficient parameters for "
		  << managerName << " manager at offset " << recordPosition
		  << '\n';
		return;
		}
	    if (!file.read(&managementInterval))
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  trouble getting common parameters for "
		  << managerName << " manager: " << strerr(saveErrno)
		  << '\n';
		return;
		}
	    cout << recordPosition << ": " << managerName
	      << " managementInterval " << managementInterval;
	    }
	else
	    {
	    parameterBytes -= sizeof managementInterval
	      + sizeof firstMissReference + sizeof missCount;
	    if (parameterBytes < 0)
		{
		cerr << "dumpdatabase:  insufficient parameters for "
		  << managerName << " manager at offset " << recordPosition
		  << '\n';
		return;
		}
	    if (!file.read(&managementInterval)
	      ||  !file.read(&firstMissReference)
	      ||  !file.read(&missCount))
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  trouble getting common parameters for "
		  << managerName << " manager: " << strerr(saveErrno)
		  << '\n';
		return;
		}
	    cout << recordPosition << ": " << managerName
	      << " managementInterval " << managementInterval
	      << " firstMissReference " << firstMissReference
	      << " missCount " << missCount;
	    }
	if (strcmp(managerName, "Global") == 0)
	    {
	    unsigned long referenceBase;
	    double ageWeightFactor;
	    unsigned long agingThreshold;
	    double oftenPercent;
	    int peekThreshold;
	    long maximumHoardBlocks;
	    int keepReferences;
	    bool processTree;
	    if (fileVersion <= 5)
		parameterBytes -= sizeof referenceBase;
	    parameterBytes -= sizeof ageWeightFactor + sizeof agingThreshold
	      + sizeof peekThreshold + sizeof maximumHoardBlocks
	      + sizeof keepReferences + sizeof processTree;
	    parameterBytes -=
	      fileVersion > 3 ? sizeof oftenPercent : sizeof (int);
	    if (parameterBytes < 0)
		{
		cerr << "dumpdatabase:  insufficient parameters for Global "
		  << " manager at offset " << recordPosition << '\n';
		return;
		}
	    if (fileVersion <= 5)
		{
		if (!file.read(&referenceBase))
		    {
		    int saveErrno = errno;
		    cerr
		      << "dumpdatabase:  trouble getting specific parameters "
		      << "for Global manager: " << strerr(saveErrno) << '\n';
		    return;
		    }
		}
	    if (!file.read(&ageWeightFactor)  ||  !file.read(&agingThreshold))
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  trouble getting specific parameters "
		  << "for Global manager: " << strerr(saveErrno) << '\n';
		return;
		}
	    bool failure = false;
	    if (fileVersion > 3)
	        failure = !file.read(&oftenPercent);
	    else
		{
		int intOftenPercent;
		failure = !file.read(&intOftenPercent);
		oftenPercent = intOftenPercent;
		}
	    if (failure
	      ||  !file.read(&peekThreshold)
	      ||  !file.read(&maximumHoardBlocks)
	      ||  !file.read(&keepReferences)  ||  !file.read(&processTree))
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  trouble getting specific parameters "
		  << "for Global manager: " << strerr(saveErrno) << '\n';
		return;
		}
	    if (fileVersion <= 5)
		cout << " referenceBase " << referenceBase;
	    cout
	      << " ageWeightFactor " << ageWeightFactor
	      << " agingThreshold " << agingThreshold
	      << " oftenPercent " << oftenPercent
	      << " peekThreshold " << peekThreshold
	      << " maximumHoardBlocks " << maximumHoardBlocks
	      << " keepReferences " << keepReferences
	      << " processTree " << (processTree ? "true" : "false");
	    }
	else if (strcmp(managerName, "Deletion") == 0)
	    {
	    unsigned long deletionTrigger;
	    parameterBytes -= sizeof deletionTrigger;
	    if (parameterBytes < 0)
		{
		cerr << "dumpdatabase:  insufficient parameters for Deletion "
		  << " manager at offset " << recordPosition << '\n';
		return;
		}
	    if (!file.read(&deletionTrigger))
		{
		int saveErrno = errno;
		cerr << "dumpdatabase:  trouble getting reference base "
		  << "for Deletion manager: " << strerr(saveErrno) << '\n';
		return;
		}
	    cout << " deletionTrigger " << deletionTrigger;
	    }
	else if (strcmp(managerName, "Save") == 0)
	    {
	    /*
	     * Save manager has no additional parameters
	     */
	    }
	else
	    {
	    /*
	     * Remaining managers are hoarding managers
	     */
	    if (fileVersion > 5)
		{
		unsigned long currentReference;
		unsigned long referenceBase;
		parameterBytes -= sizeof currentReference
		  + sizeof firstMissReference + sizeof missCount
		  + sizeof referenceBase;
		if (parameterBytes < 0)
		    {
		    cerr << "dumpdatabase:  insufficient parameters for "
		      << managerName << " manager at offset " << recordPosition
		      << '\n';
		    return;
		    }
		if (!file.read(&currentReference)
		  ||  !file.read(&firstMissReference)
		  ||  !file.read(&missCount)
		  ||  !file.read(&referenceBase))
		    {
		    int saveErrno = errno;
		    cerr
		      << "dumpdatabase:  trouble getting hoarding parameters"
			" for "
		      << managerName << " manager: " << strerr(saveErrno)
		      << '\n';
		    return;
		    }
		cout
		  << " currentReference " << currentReference
		  << " firstMissReference " << firstMissReference
		  << " missCount " << missCount
		  << " referenceBase " << referenceBase;
		}
	    if (strcmp(managerName, "Clustering") == 0)
		{
		bool sharedNeighbors;
		double directoryDistanceWeight;
		bool incremental;
		double investigatedRelationWeight;
		unsigned int minimumReferences;
		unsigned long oldestReference;
		int sharedLooseThreshold;
		int sharedTightThreshold;
		double alwaysClusterDistance;
		int closestK;
		unsigned int mediumReferences;
		double sometimesClusterDistance;
		int tightPercent;
		parameterBytes -= sizeof sharedNeighbors
		  + sizeof directoryDistanceWeight
		  + sizeof investigatedRelationWeight
		  + sizeof minimumReferences
		  + sizeof oldestReference + sizeof sharedLooseThreshold
		  + sizeof sharedTightThreshold + sizeof alwaysClusterDistance
		  + sizeof closestK + sizeof mediumReferences
		  + sizeof sometimesClusterDistance + sizeof tightPercent;
		if (fileVersion > 4)
		    parameterBytes -= sizeof incremental;
		if (parameterBytes < 0)
		    {
		    cerr << "dumpdatabase:  insufficient parameters for "
		      << "Clustering manager at offset " << recordPosition
		      << '\n';
		    return;
		    }
		if (!file.read(&sharedNeighbors)
		  ||  !file.read(&directoryDistanceWeight)
		  ||  (fileVersion > 4  &&  !file.read(&incremental))
		  ||  !file.read(&investigatedRelationWeight)
		  ||  !file.read(&minimumReferences)
		  ||  !file.read(&oldestReference)
		  ||  !file.read(&sharedLooseThreshold)
		  ||  !file.read(&sharedTightThreshold)
		  ||  !file.read(&alwaysClusterDistance)
		  ||  !file.read(&closestK)  ||  !file.read(&mediumReferences)
		  ||  !file.read(&sometimesClusterDistance)
		  ||  !file.read(&tightPercent))
		    {
		    int saveErrno = errno;
		    cerr
		      << "dumpdatabase:  trouble getting specific parameters "
		      << "for Clustering manager: " << strerr(saveErrno)
		      << '\n';
		    return;
		    }
		cout
		  << " sharedNeighbors "
		  <<  (sharedNeighbors ? "true" : "false")
		  << " directoryDistanceWeight " <<  directoryDistanceWeight;
		if (fileVersion > 4)
		    cout
		      << " incremental " << (incremental ? "true" : "false");
		cout
		  << " investigatedRelationWeight "
		  <<  investigatedRelationWeight
		  << " minimumReferences " <<  minimumReferences
		  << " oldestReference " <<  oldestReference
		  << " sharedLooseThreshold " <<  sharedLooseThreshold
		  << " sharedTightThreshold " <<  sharedTightThreshold
		  << " alwaysClusterDistance " <<  alwaysClusterDistance
		  << " closestK " <<  closestK
		  << " mediumReferences " <<  mediumReferences
		  << " sometimesClusterDistance " <<  sometimesClusterDistance
		  << " tightPercent " <<  tightPercent;
		}
	    }
	cout << '\n';
	if (parameterBytes > 0)
	    {
	    file.rdbuf()->seekoff(parameterBytes, ios::cur, ios::in);
	    cerr << "dumpdatabase: " << parameterBytes
	      << " extra bytes of parameters for " << managerName
	      << " manager ignored\n";
	    }
	}
    }
