#ifndef lint
static const char Rcs_Id[] =
    "$Id: dumpobs.cc,v 1.11 1998/01/31 05:38:52 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: dumpobs.cc,v $
 * Revision 1.11  1998/01/31 05:38:52  geoff
 * Don't include the inline functions, since there are no class includes.
 * Return an exit status of 2 on usage errors.
 *
 * Revision 1.10  1997/12/18 01:24:43  geoff
 * Make some things const.
 *
 * Revision 1.9  1996/10/11 22:09:02  geoff
 * Use seeropen/close to open and close the trace file.
 *
 * Revision 1.8  1996/09/11 22:59:46  geoff
 * Add the -n option to show filenames while dumping.
 *
 * Revision 1.7  1996/05/03  01:12:00  geoff
 * Update the copyright.  Add code to keep track of process names.
 *
 * Revision 1.6  1995/11/10  02:58:59  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.5  1995/10/10  20:37:56  geoff
 * Include stdlib.h to get rid of some Linux warnings.
 *
 * Revision 1.4  1995/09/15  00:05:17  geoff
 * Change the RCS ID to be a const.  Add starting-offset support and
 * support for compressed input files.
 *
 * Revision 1.3  1995/06/30  19:04:32  geoff
 * Put a file offset in front of every output line.
 *
 * Revision 1.2  1995/06/17  00:11:09  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include <errno.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <seer/seer.h>
#include <sys/param.h>

int			main(int argc, char* argv[]);
static void		usage();
static void		dumpfile(const char* fileName, FILE* infile);
void			nomem(const char* where, int line, unsigned int size);

#define MAXPIDS		32768		// Maximum no. of PIDs to support

static const char*	procnames[MAXPIDS]; // Table of process names
char*			progname;	// Name we were run under
static unsigned long	startOffset = 0; // Where to start dumping

extern seerbuf_t*	readent(FILE* replayfile);
extern int		seerclose(FILE* replayFile);
extern FILE*		seeropen(const char* fileName);
extern unsigned long	seertell(FILE* replayFile);
extern void		showent(seerbuf_t* seerp, const char* procNames[],
			  int maxNames);
extern "C" char*	strerr(int errnum);

int main(				// Dump an observer trace file
    int			argc,		// Argument count
    char*		argv[])		// Argument vector
    {
    bool		showName = false; // True to put filename on every line
    progname = argv[0];
    int trouble = 0;
    while (argc > 1  &&  argv[1][0] == '-')
	{
	argc--;
	argv++;
	switch (argv[0][1])
	    {
	    case 'n':
		{
		showName = true;
		break;
		}
	    case 'o':
		{
		if (--argc < 1)
		    usage();
		argv++;
		startOffset = atol(argv[0]);
		break;
		}
	    default:
		usage();
		break;
	    }
	}
    if (argc == 1)
	{
	FILE* infile = seeropen((char*)NULL);
	dumpfile(showName ? "<stdin>" : (const char*)NULL, infile);
	}
    else
	{
	for (int argno = 1;  argno < argc;  argno++)
	    {
	    FILE* infile = seeropen(argv[argno]);
	    if (infile == NULL)
		{
		int saveErrno = errno;
		cerr << "dumpobs:  couldn't open '" << argv[argno] << "': "
		  << strerr(saveErrno) << '\n';
		trouble = 1;
		}
	    else
		{
		dumpfile(showName ? argv[argno] : (const char*)NULL, infile);
		(void)seerclose(infile);
		}
	    }
	}
    return trouble;
    }

static void usage()			// Issue a usage message and exit
    {
    (void)fprintf(stderr, "Usage: dumpobs [-n] [-o offset] [files]\n");
    exit(2);
    }

static void dumpfile(			// Dump an observer trace
    const char*		fileName,	// Name of file dumped from
    FILE*		infile)		// File to dump from
    {
    unsigned long offset = seertell(infile);
    seerbuf_t* seerp = NULL;
    while ((seerp = readent(infile)) != NULL)
	{
	if (offset >= startOffset)
	    {
	    if (fileName != NULL)
		cout << fileName << ':';
	    cout << offset << ' ';
	    showent(seerp, procnames, MAXPIDS);
	    }
	offset = seertell(infile);
	}
    if (offset >= startOffset)
	startOffset = 0;
    else
	startOffset -= offset;
    }

void nomem(			// Issue an out-of-memory message
    const char*		where,	// Routine that failed to allocate
    int			line,	// Line number of failure
    unsigned int	size)	// Size that was wanted
    {
    cout.flush();
    cerr << "dumpobs:  out of memory at " << line << " in " << where;
    if (size)
	cerr << ", couldn't get " << size << " bytes";
    cerr << "!\n";
    cerr.flush();
    abort();
    }
