#ifndef lint
static const char Rcs_Id[] =
    "$Id: file.cc,v 1.52 1998/10/23 23:15:09 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: file.cc,v $
 * Revision 1.52  1998/10/23 23:15:09  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.51  1998/07/28 23:47:41  geoff
 * Add support for a separate distribution for directory sizes.  Fix the
 * storage-setting code so that it actually sets some sizes in simulation
 * mode.
 *
 * Revision 1.50  1998/07/15 21:43:36  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.49  1998/07/09 21:02:01  geoff
 * Add the meaningless argument to the reference-recording routines.  Use
 * canHoard when counting storage.
 *
 * Revision 1.48  1998/07/08 19:59:02  geoff
 * In simulation mode, be willing to change the type of a "file" to a
 * directory later.  This is needed because directories may first be
 * referenced as the tail of a path, and only later referenced in a
 * directory context.
 *
 * Revision 1.47  1998/06/02 23:59:41  geoff
 * Add support for symlink initialization, and don't read symlinks in
 * simulation mode.  Fix a longstanding bug that could cause the root
 * directory to become missing, and add some panics for that case.
 * Always infer the file state in simulation mode.  Never run
 * discoverDirectorySize in simulation mode.
 *
 * Revision 1.46  1998/05/14 01:15:29  geoff
 * Add (partially-tested) MIME-encoding support.
 *
 * Revision 1.45  1998/05/02 00:29:06  geoff
 * Use unsigned ints for path lengths.  Don't do mount-forcing lstats in
 * simulation mode.  Add some temporary code to suppress readlinks in
 * simulation mode.  Fix a bug in renameDescendants that failed to set
 * the correct path length, add some more validity checking and
 * commentary, and improve some error messages.  Use unsigned shorts
 * rather than unsigned chars for the strip length.
 *
 * Revision 1.44  1998/04/09 00:33:00  geoff
 * Use the new notation for hash and list iterators.  Don't ever try to
 * do stats if we're running in simulation mode with a statistically
 * generated file distribution.
 *
 * Revision 1.43  1998/03/05 03:18:15  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.42  1998/02/19 02:10:02  geoff
 * Break out the filestate class into a separate file.  Add an argument
 * to the setState call, and add stateChanged and relationsChanged, plus
 * a bunch of calls to the latter, to support incremental clustering.
 *
 * Revision 1.41  1997/12/18 01:24:44  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.40  1997/12/04 23:15:38  geoff
 * Use correlatorPanic to do aborts.  Fix a bug that inverted
 * stays-with-name flag manipulation in file renaming.  Don't mark a
 * directory for deletion if any child is marked as unstored.  Fix a bug
 * that caused files to get assigned random sizes when they were marked
 * as unstored.  When handling creates, rediscover the status regardless
 * of whether the file is currently marked for deletion; this allows
 * creates to override the "unstored" status.  This might turn out to be
 * a bug in simulation mode.
 *
 * Revision 1.39  1997/07/11 17:40:55  geoff
 * Fix a bug that caused size fakery to always produce zero sizes.
 *
 * Revision 1.38  1997/05/16 21:26:20  geoff
 * Remove the debugging code used to find the LRU/clustering inconsistencies.
 *
 * Revision 1.37  1997/05/16 17:10:35  geoff
 * Add temporary debugging code to catch discrepancies between the LRU
 * and clustering managers.  Make sure parent directories are set to
 * directory type in simulation mode.  This is the version used to run
 * the simulations in the dissertation.
 *
 * Revision 1.36  1997/03/13 23:00:15  geoff
 * More fixes to renaming, in particular fixing the path-copying code so
 * it doesn't clobber anything and updates the length properly.  When
 * renaming descendants, properly update the parent directory's children
 * list.
 *
 * Revision 1.35  1997/02/12 02:52:05  geoff
 * Break inheritFromParent into inheritFlagsFromParent and
 * inheritManagerDataFromParent.  Change a flag output ordering, for no
 * discernible reason.  Remember to display the storage-set,
 * previously-referenced, and newly-created flags in the show routine.
 *
 * Revision 1.34  1997/01/02 06:48:19  geoff
 * Add support for size fakery.  Allow DESTROY references to delete
 * unstored files in simulation mode if FF_STORAGE_SET is clear; this is
 * important for correct simulation and hoard-size calculation.
 *
 * Revision 1.33  1996/12/03 02:58:01  geoff
 * Change the name of the routine that records missed references.  Get
 * rid of a spurious markDeleted call that caused important information
 * about unstored files to be lost.  In markDeleted, Never set files to
 * FORGETTING during a replay.  Preserve the STORAGE_SET flag when
 * clearing a file's data.
 *
 * Revision 1.32  1996/10/04 07:12:06  geoff
 * Add support for reference-time setting.  Add support for
 * command-interpreter references.  Improve a couple of error messages.
 *
 * Revision 1.31  1996/09/23 23:41:25  geoff
 * Don't ever rename directories by deleting.  Fix a number of bugs in
 * the descendant-renaming code.
 *
 * Revision 1.30  1996/09/16 22:16:45  geoff
 * On closes and process exits, update the last reference time to the
 * current time, so that long-running processes and long-open files will
 * show up as recently used.
 *
 * Revision 1.29  1996/09/13 18:01:26  geoff
 * When MUST_SAVE files wind up in the deletion process, set their state
 * to UNSTORED rather than UNKNOWN, to save stats and keep them from
 * reentering deletion.
 *
 * Revision 1.28  1996/09/11 22:59:46  geoff
 * Add renameDescendants and finally fix the rename code to handle
 * directory renames properly.  Add REPORT_SELECTED_FAILURES support.
 * Add setStorage.  Make all files uncontrolled by default.  Don't turn
 * missing into unstored for uncontrolled files.  When marking files
 * deleted, set them unstored if they have had their storage set and
 * there is an observer replaying data.  Don't complain about parents
 * going away if the file is unstored; instead just switch to the unknown
 * state so the file will eventually get deleted.  Convert from dlists to
 * slists.  Add setStorage.  When discovering directory size, don't
 * depend on d_namlen, which doesn't exist on some systems.  When
 * handling must-save files, set the state directly to unknown rather
 * than using forgetStatus, since the latter now checks the observer
 * replay count.
 *
 * Revision 1.27  1996/06/11  07:04:54  geoff
 * Get rid of an obsolete root-directory test.  Eliminate READ/WRITE
 * EXTRAS.  Drop an obsolete peek test, USE_PROBABLY_MEANINGLESS, and an
 * outdated meaninglessness test.
 *
 * Revision 1.26  1996/06/11  06:13:11  geoff
 * Add writePathTerminator (rather than letting callers do it in an
 * ad-hoc fashion).  Add discoverDirectorySize and
 * potential-peek-counting support, plus associated meaninglessness
 * inference.  Add code to dump/restore the file type byte, plus peek
 * fields.  Finally add code to walk the file table after renames to
 * retarget symlinks to the new name.  Do a lot of surgery on the rename
 * code to make it faster, more robust, and to do the right thing in more
 * cases.  Add automatic getcwd detection.  Do a lot of rethinking of how
 * references are recorded.  Make a lot of changes in how peeks are
 * handled.  Get rid of some wired-in constants (shame on me!).  Use
 * hoardableSize(), rather than the plain size, in countStorage().
 *
 * Revision 1.25  1996/05/03  01:12:02  geoff
 * Update the copyright.  Add tildeExpand to make instruction files
 * easier to write.  Add inheritFromParent.  Always make dot files
 * uncontrolled by default.  Add code to limit the number of hops on
 * symlinks, and fix some serious bugs in symlink tracking. in
 * findCanonical.  Add some special (and ugly) code to deal with
 * stripping in relation to the automounter.  Add a printout to detect
 * whether some obsolete code ever gets executed.  Add (mostly-untested)
 * support for rereading the instructions file dynamically.  Add support
 * for talking about either symlinks or their targets in the instructions
 * file.  Add READ/WRITE_EXTRAS and MISSING_BECOMES_UNSTORED support.
 * Remove support for believeSavedState.  Fix (I hope) the MISSINGDIR
 * bug, caused by bad symlink following on destructions.  If a process
 * has a directory open for read, mark it meaningless.
 *
 * Revision 1.24  1996/04/10  06:09:31  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.23  1996/04/10  01:28:31  geoff
 * Add a << operator for writing file states.  Change countStorage,
 * store, and unstore to be manager-specific.  Get rid of the old
 * obsolete all-managers file storage state ("where" field).  Add support
 * for the CONTROLLED flag.  Add support for the new UNSTORED file state.
 * Add debugging code to try to catch the cause of the parent-going-away
 * problems under Ficus.  Change the meaningless-process code to let
 * exec's, destroys, and exits pass through.
 *
 * Revision 1.22  1996/03/19  18:42:54  geoff
 * Display the MEANINGLESS, CONTROLLED, FAKE_SIZE, and MARKED flags when
 * showing a file.  Use the fields package to read the instruction file.
 * Many changes to support the new class inheritance structure for
 * managers and manager data.  Fix a longstanding bug in findCanonical
 * where following ".." could cause a null-pointer dereference.  Add the
 * control/nocontrol and fakesize commands, plus other necessary
 * fake-size support.  Don't drop NEVERHOARD files even if the MUST_SAVE
 * bit is clear.
 *
 * Revision 1.21  1996/02/02  20:09:43  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.20  1996/02/02  18:28:52  geoff
 * Add directoryDistance() to evaluate the directory-distance measure.
 * Move some virtual routines here from the inline file, since inline
 * virtual makes little or no sense.  Don't call recordStateChange unless
 * the state is really changing (for efficiency).  Use count() routines
 * rather than counting loops.  Fix some minor problems detected by the
 * new C++ compiler.  Fix a very bad bug that caused parent pointers to
 * be set incorrectly when a ".." component was followed.  Allow EAGAIN
 * as an acceptable reason for readlink failures.  Allow blank lines in
 * the instruction file (as the commentary claims).  Support the
 * MEANINGLESS instruction line, and manager-specific lines.  Add
 * meaningless-process support.  Use the new forgetStatus routine when
 * status becomes unknown.
 *
 * Revision 1.19  1995/12/19  01:02:52  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.18  1995/12/19  00:34:16  geoff
 * Add extensive process-tracking support.  Add FinalFileDestructor to
 * help in detecting unfreed memory (by arranging that there shouldn't be
 * any).  Fix some bugs in the symlink-following code.  Don't use
 * fstream::get when reading instructions, because it screws up the
 * byte-offset tracking.  Add FF_MUST_SAVE support, and add code to get
 * rid of file-table files that aren't really important.  Get rid of some
 * repeated calls to inlined functions, since they aren't necessarily
 * inlined when optimization is off (as with the -g switch).
 *
 * Revision 1.17  1995/11/22  00:20:15  geoff
 * Get rid of the TIME_SORTED_INPUT and NEW_FILE_RELATIONS ifdefs.
 *
 * Revision 1.16  1995/11/21  21:38:07  geoff
 * Add support for doing investigations through NEW_FILE_RELATIONS.
 *
 * Revision 1.15  1995/11/10  02:59:04  geoff
 * Add zzinlinefunctions.hh.  Add extensive NEW_FILE_RELATIONS support.
 * Other minor fixes.  Support SEER_REF_CONTROL and
 * LENGTH_IN_STAT_RECORD.  Remember to set currentReference when reading
 * from the save file.  Explicitly set the size of missing files to zero.
 * Support FICUS_STAT_BUG and FICUS_MISSING_DIR_BUG.
 *
 * Revision 1.14  1995/10/10  20:37:58  geoff
 * Get rid of the obsolete TRACK_MEMORY_LEAKS stuff.  Handle missing
 * files differently in simulation mode.
 *
 * Revision 1.13  1995/09/15  00:05:28  geoff
 * Change the RCS ID to be a const.  Add proper bracketing to all switch
 * statements.  Add the FileState class.  Many changes to add
 * deletion-manager support.  Add partial memory-leak detection (which is
 * already made obsolete by the dmalloc library).
 *
 * Revision 1.12  1995/08/07  20:02:48  geoff
 * Add investigator support.  Add the readPathname and findFromStream
 * routines.  Add ASSIGNABLE_FILE_IDS support.  Add code to handle the
 * DUP system call specially.  Comment a number of switch cases.
 *
 * Revision 1.11  1995/06/30  19:04:37  geoff
 * Add support for working on symlinks, rather than their targets, when
 * an operation is intended for the link (e.g., make link, read link).
 * Allow writePath to write either stripped or unstripped paths.  Add
 * believeSavedState support.  Add countStorage, and change store() to
 * return the amount of space added to the hoard.  Add some new show
 * options to the file showing.  Report file offsets in most error
 * messages.  Fix a bug that caused the save file to contatin bad data if
 * a readlink failed.  Rewrite the rename code to handle most things
 * properly (but directory and symlink-target renames are still bogus).
 * When storing files, make sure the link target is also stored.
 *
 * Revision 1.10  1995/06/17  00:18:13  geoff
 * Don't treat CREATE as an open
 *
 * Revision 1.9  1995/06/17  00:11:11  geoff
 * Save the errno before reporting errors, to be sure it doesn't get
 * clobbered.  Flush cerr/cout before all aborts.  Get rid of find, which
 * is obsolete.  Add show, which is parameterized, and make operator<<
 * just be a wrapper.  Rewrite rename to be more careful about renaming
 * things properly (this fixes some major problems).  Move markDeleted
 * from file.icc.  Improve readlink error handling.  Add support for
 * SEER_REF_FOLLOW.  Don't call underlyingFind twice with the same
 * pathname from findCanonical.  Add code in readInstructions to skip
 * ST_STAT records in the replay file when necessary.  Fix a bug that
 * could write a null path when the strip length was equal to the name
 * length.  Add FF_WRITING support.
 *
 * Revision 1.8  1995/05/27  00:15:54  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.7  1995/05/26  23:05:36  geoff
 * Allow a null parent-file pointer.  Don't set the parent pointer if the
 * parent is going to go away.
 *
 * Revision 1.6  1995/05/23  00:39:35  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.  Take out FF_STRIP and add
 * stripLength support.  Put tags on management output dump lines.
 * Change a bunch or exits to aborts.  Allow the save file to record
 * failed readlinks so we can handle dangling symlinks.  Save the
 * instruction file in the save file, or be able to replay from there, as
 * appropriate.
 *
 * Revision 1.5  1995/05/05  22:41:32  geoff
 * Add a lot of preliminary PRIVATE_MANAGEMENT_DATA support, including
 * File::recordReference.  Add support for the new
 * SS_UFS_STORED/NOT_STORED states.  Add ST_READLINK support in the save
 * file.
 *
 * Revision 1.4  1995/04/20  21:57:35  geoff
 * Add findCanonical and readInstructions, plus the support routines
 * underlyingFind and setParent.  Improve some error detection and
 * messages.  Handle most file types in setState.
 *
 * Revision 1.3  1995/03/29  23:57:26  geoff
 * Get rid of the OLD_FILE_TABLE ifdef.
 *
 * Revision 1.2  1995/03/29  21:04:14  geoff
 * Switch to using byte-swapped streams for all I/O.  Add support for the
 * NAME_IN_STAT_RECORD debugging option.  Fix many bugs in the new file
 * table code.  Add a destructor that departs all clusters upon
 * destruction.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "clustering_mgr.hh"
#include "communication.hh"
#include "correlator.hh"
#include "deletion_mgr.hh"
#include "fields.hh"
#include "file.hh"
#include "filerelation.hh"
#include "filestate.hh"
#include "investigator.hh"
#include "list.hh"
#include "managerbase.hh"
#include "panic.hh"
#include "process.hh"
#include "zzinlinefunctions.hh"
#include <dirent.h>
#include <errno.h>
#include <fstream.h>
#include <math.h>
#include <pwd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			FinalFileDestructor::FinalFileDestructor();
			FinalFileDestructor::~FinalFileDestructor();
			File::File();
			File::~File();
int			File::operator==(const File& that) const;
ostream&		operator<<(ostream& stream, const File& file);
					// Write a file in printable fashion
ostream&		File::show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Write a file in printable fashion,
					// ..with options.  Appends \n

const char*		File::tildeExpand(const char* path);
					// Expand ~user sequences
File*			File::underlyingFind(const char* name,
			  unsigned int nameLength, bool create, bool& isNew);
					// Find a file entry in file table,
					// ..telling if it's a new one
File*			File::findCanonical(const char* name, bool wantLink,
			  bool create, Process* process);
					// Find a file entry in file table,
					// ..canonicalizing the name
bool			File::readPathname(ByteSwappedStream& stream,
			  char pathname[MAXPATHLEN], unsigned int* pathLength);
					// Read a pathname from a stream,
					// ..verifying the format
File*			File::findFromStream(ByteSwappedStream& stream,
			  bool wantLink);
					// Read a pathname from a stream and
					// ..locate the file-table entry for it
void			File::readInstructions(const char* fileName);
					// Read an instruction file
unsigned int		File::hash(File* file);
					// Hash file pathname, for file table
int			File::compare(File* a, File* b);
					// Compare file entries for hash table
void			File::write(ByteSwappedStream& stream) const;
					// Write a file entry to a stream
void			File::writePath(ByteSwappedStream& stream,
			  bool stripped)
			    const;
					// Write a file's path to a stream
void			File::writePathTerminator(ByteSwappedStream& stream);
					// Write a file-list terminator
bool			File::read(ByteSwappedStream& stream);
					// Read a file entry from a stream
void			File::rename(const char* oldName, const char* newName,
			  Process* process);
					// Change the name of a file
void			File::renameDescendants(const char* oldName,
			 const char* newName, File* newParent);
					// Rename file's descendants
#ifdef FICUS_MISSING_DIR_BUG
bool			File::markDeleted(bool clearAll);
					// Mark file to be deleted
#else // FICUS_MISSING_DIR_BUG
void			File::markDeleted(bool clearAll);
					// Mark file to be deleted
#endif // FICUS_MISSING_DIR_BUG
void			File::loseForgettingPointers(); // Get rid of pointers
					// ..that will disappear
void			File::mustSave(); // Mark that file must be saved
void			File::addRelation(FileRelationBase* relation,
			  File& target);
					// Add a new relation to file
void			File::removeRelation(FileRelationBase* relation);
					// Remove a relation from file
					// *** May invalidate iterators! ***
void			File::recordReference(const char* path,
			  Process* process, ReferenceType type);
					// Record a reference to a named file
void			File::recordReference(Process* process,
			  ReferenceType type);
					// Record a reference to this file
void			File::recordMissedReference(const char* path,
			  Process* process, ReferenceType type);
					// Record a missed reference
void			File::recordMissedReference(Process* process,
			  ReferenceType type, bool isMeaningless);
					// Record a missed reference
const char*		File::encodedName(ShowOptions options)
			  const;	// Return encoded name of the file
void			File::setDirectorySizeDistribution(const char* name,
			  int nParameters, const double* parameters);
					// Set a particular size distribution
					// ..for use during simulation
void			File::setFileSizeDistribution(const char* name,
			  int nParameters, const double* parameters);
					// Set a particular size distribution
					// ..for use during simulation
void			File::discoverStatus();
					// Figure out the status of a file
void			File::setStorage(StorageSetting stored,
			  unsigned long size, File* targetFile);
					// Set file storage status
void			File::setParent(File* parent);
					 // Set file's parent and inherit flags
void			File::inheritFlagsFromParent();
					// Inherit general info from parent
void			File::inheritManagerDataFromParent(
			  const HoardingManagerBase* targetManager);
					// Inherit manager-specific information
					// ..from parent, for one or all
					// ..managers
void			File::discoverDirectorySize();
					// Discover size of a directory
bool			File::deleteIfObsolete();
					// Arrange deletion if obsolete file
unsigned long		File::countStorage(const HoardingManagerBase& manager)
			  const;
					// Count storage needed by file
void			File::relationsChanged();
					// Inform managers that relations have
					// ..changed
void			File::stateChanged(FileState* stateAddress,
			  FileState::State oldState,
			  FileState::State newState);
					// Inform managers that state has
					// ..changed
unsigned long		File::directoryDistance(const File& that) const;
					// Find directory-path distance to
					// ..another file
void			File::showHashStatistics(ostream& stream);
					// Display hash table statistics
void			File::cleanRelationData();
					// Clean relations to forgetting files
void			File::clearRelationData();
					// Get rid of all relations
void			File::cleanManagerData();
					// Clean references to forgetting files
FileRelationBase*	File::chosenRelation(File& targetFile,
			  const FileRelationTypeBase& relationType,
			  bool create);
					// Find or create a relationship
					// ..of a chosen type
void			File::destroyTable(); // Destroy the entire file table
					// ..(for memory-leak detection)
RelatedFileIterator&	RelatedFileIterator::operator++();
					// Step to next relation, cleaning up
			SelectedRelatedFileIteratorBase::
			  ~SelectedRelatedFileIteratorBase();
void			SelectedRelatedFileIteratorBase::findSelectedType();
					// Point iterator at a relation
#endif // 0

/*
 * Static class variables
 */
HashTable		File::fileTable(5000,
			  (unsigned int (*)(const void*)) File::hash,
			  (int (*)(const void*, const void*)) File::compare);
FinalFileDestructor	File::finalFileDestructor;
File*			File::newEntry = NULL;
					// Scratch entry for lookups/inserts
#ifdef ASSIGNABLE_FILE_IDS
unsigned long		File::nextID = 0; // Next ID to be assigned
#endif // ASSIGNABLE_FILE_IDS
double			File::logDirectoryGeometricP = 0.0;
					// Parameter of geometric distribution
double			File::logFileGeometricP = 0.0;
					// Parameter of geometric distribution

extern void		correlatorPanic(const char* message);
extern int		ficus_extended_stat(const char* path, int iscanonical,
			  struct stat* statbuf);
extern "C" char*	mimeEncode(const char* source,
			  const char* whatToEncode);
					// MIME-encode a string
extern void		nomem(const char* where, int line, unsigned int size);
extern char*		strdup(const char* string, int size);
extern "C" char*	strerr(int errnum);

FinalFileDestructor::FinalFileDestructor()
    {
    }

FinalFileDestructor::~FinalFileDestructor()
    {
    if (File::newEntry != NULL)
	{
	delete File::newEntry;
	File::newEntry = NULL;
	}
    }

File::File()
    : path(NULL),
      pathLength(0),
      flags(0),
      stripLength(0),
      type(FT_FILE),
      state(FileState::UNKNOWN),
      blocks(0),
      parentDirectory(NULL),
      children(true),
      unknownChildren(0),
      linkTargetFile(NULL),
      openCount(0),
#ifdef ASSIGNABLE_FILE_IDS
      identification(nextID++),
#endif // ASSIGNABLE_FILE_IDS
      lastReferenceType(SEER_REF_READ),
      peekCount(0),
      potentialPeeks(0),
      boundedLRU_ManagerData(),
      clusteringManagerData(*this),
      lruManagerData(),
      weightedLRU_ManagerData()
    {
    }

File::~File()
    {
    // NEEDSWORK: should use new/delete
    if (path != NULL)
	free((void*)path);
    path = NULL;
    setParent(NULL);
    linkTargetFile = NULL;
    state.setState(*this, FileState::UNKNOWN);	// Deletion manager needs this
    clearRelationData();
    }

/*
 * Compare File items for pathname equality.  Note that this ignores
 * all other fields.
 */
int File::operator==(
    const File&	that)			// File to compare this against
    const
    {
    if (pathLength != that.pathLength)
	return 0;
    else
	return strcmp(path, that.path) == 0;
    }

ostream& operator<<(			// Write a file in printable fashion
    ostream&		stream,		// Stream to write to
    const File&		file)		// File to be written
    {
    return file.show(stream, SO_EVERYTHING, NULL);
    }

/*
 * Show file in printable form, with options
 *
 * Show options honored: SO_MANAGER, SO_ALL_MANAGERS, SO_FULL_FILE_INFORMATION
 *
 * Note that if SO_FULL_FILE_INFORMATION is not specified, no newline
 * is provided.
 */
ostream& File::show(
    ostream&		stream,		// Stream to print to
    ShowOptions		options,	// Desired display options
    const char*		managerName)	// Manager to show for, if SO_MANAGER
    const
    {
    if ((options & SO_FULL_FILE_INFORMATION) == 0)
	{
	stream << encodedName(options) << ' ' << id();
	return stream;
	}
    stream << "FH\t";
    if (options & SO_MIME_ENCODE)
	{
	const char* encoding = mimeEncode(path, NULL);
	stream << encoding;
	free((void *)encoding);
	}
    else
      stream << path;
    stream << ' '
      << id()
      << ' '
      << (type == (char) File::FT_FILE ? "FT_FILE"
	: type == (char) File::FT_DIR ? "FT_DIR"
	: type == (char) File::FT_FIFO ? "FT_FIFO"
	: type == (char) File::FT_CHARACTER_SPECIAL
	  ? "FT_CHARACTER_SPECIAL"
	: type == (char) File::FT_BLOCK_SPECIAL ? "FT_BLOCK_SPECIAL"
	: type == (char) File::FT_SYMLINK ? "FT_SYMLINK"
	: type == (char) File::FT_SOCK ? "FT_SOCK"
	: "BAD_FILE_TYPE")
      << ' ' << state
      << ' ' << blocks
      << ' ' << openCount
      << ' '
      << (lastReferenceType == SEER_REF_UNKNOWN ? "SEER_REF_UNKNOWN"
	: lastReferenceType == SEER_REF_READ ? "SEER_REF_READ"
	: lastReferenceType == SEER_REF_RW ? "SEER_REF_RW"
	: lastReferenceType == SEER_REF_CREATE ? "SEER_REF_CREATE"
	: lastReferenceType == SEER_REF_EXEC ? "SEER_REF_EXEC"
	: lastReferenceType == SEER_REF_DESTROY ? "SEER_REF_DESTROY"
	: lastReferenceType == SEER_REF_EXIT ? "SEER_REF_EXIT"
	: lastReferenceType == SEER_REF_OPENCREATE
	    ? "SEER_REF_OPENCREATE"
	: lastReferenceType == SEER_REF_OPENREAD ? "SEER_REF_OPENREAD"
	: lastReferenceType == SEER_REF_OPENRW ? "SEER_REF_OPENRW"
	: lastReferenceType == SEER_REF_CLOSE ? "SEER_REF_CLOSE"
	: lastReferenceType == SEER_REF_RENAME ? "SEER_REF_RENAME"
	: lastReferenceType == SEER_REF_PEEK ? "SEER_REF_PEEK"
	: lastReferenceType == SEER_REF_MAKELINK ? "SEER_REF_MAKELINK"
	: lastReferenceType == SEER_REF_PEEKLINK ? "SEER_REF_PEEKLINK"
	: lastReferenceType == SEER_REF_FOLLOW ? "SEER_REF_FOLLOW"
	: lastReferenceType == SEER_REF_DUP ? "SEER_REF_DUP"
	: lastReferenceType == SEER_REF_INTERPRET ? "SEER_REF_INTERPRET"
	: lastReferenceType == SEER_REF_CONTROL ? "SEER_REF_CONTROL"
	: lastReferenceType == SEER_REF_NEWPROC ? "SEER_REF_NEWPROC"
	: lastReferenceType == SEER_REF_PROCEXIT ? "SEER_REF_PROCEXIT"
	: "BAD_REF_TYPE")
      << ' ' << peekCount << ' ' << potentialPeeks
      << " |"
      << ((flags & File::FF_NEVERHOARD) ? "NEVERHOARD|" : "")
      << ((flags & File::FF_IGNORE) ? "IGNORE|" : "")
      << ((flags & File::FF_MEANINGLESS) ? "MEANINGLESS|" : "")
      << ((flags & File::FF_CONTROLLED) ? "CONTROLLED|" : "")
      << ((flags & File::FF_MUST_SAVE) ? "MUST_SAVE|" : "")
      << ((flags & File::FF_WRITING) ? "WRITING|" : "")
      << ((flags & File::FF_FAKE_SIZE) ? "FAKE_SIZE|" : "")
      << ((flags & File::FF_SIZE_DISCOVERED) ? "SIZE_DISCOVERED|" : "")
      << ((flags & File::FF_STORAGE_SET) ? "STORAGE_SET|" : "")
      << ((flags & File::FF_PREVIOUSLY_REFERENCED)
	? "PREVIOUSLY_REFERENCED|"
	: "")
      << ((flags & File::FF_NEWLY_CREATED) ? "NEWLY_CREATED|" : "")
      << ((flags & File::FF_SIZE_GUESSED) ? "SIZE_GUESSED|" : "")
      << ((flags & File::FF_MARKED) ? "MARKED|" : "")
      << '\n';
    unsigned long numRelatedFiles = relatedFileCount();
    stream << "RH\t\t" << numRelatedFiles << " related file"
      << ((numRelatedFiles == 1) ? ":\n" : "s:\n");
    for (RelatedFileIterator relatedFile((File&)*this);
      relatedFile;
      ++relatedFile)
	(*relatedFile).show(stream, options, managerName);
    if (options & SO_ALL_MANAGERS)
	{
	for (HoardingManagerIterator manager; manager; ++manager)
	    (*manager).managerData(*this).show(stream, options, NULL);
	}
    else if (options & SO_MANAGER)
	{
	ManagerBase* manager = ManagerBase::findNamedManager(managerName);
	if (manager != NULL  &&  manager->hasData())
	    manager->managerData(*this).show(stream, options, managerName);
	}
    return stream;
    }

/*
 * Expand ~user sequences.  Handles only the ~user notation, since the
 * correlator normally runs as root.  Note that this routine returns a
 * pointer to a static buffer, so the usual restrictions apply.
 *
 * Returns NULL if there is trouble with the expansion
 */
const char* File::tildeExpand(
    const char*		path)		// Path to expand
    {
    if (path[0] != '~')
	return path;
    static char expandedPath[MAXPATHLEN];
    path++;
    char* cp;
    for (cp = expandedPath;  *path != '\0'  &&  *path != '/';  )
	{
	if (cp - expandedPath >= MAXPATHLEN)
	    return NULL;
	*cp++ = *path++;
	}
    if (cp == expandedPath)
	return NULL;
    *cp = '\0';
    struct passwd* pwEntry = getpwnam(expandedPath);
    if (pwEntry == NULL)
	return NULL;
    (void)strcpy(expandedPath, pwEntry->pw_dir);
    (void)strncat(expandedPath, path, MAXPATHLEN - strlen(expandedPath) - 1);
    expandedPath[MAXPATHLEN - 1] = '\0';
    return expandedPath;
    }

/*
 * Find a file entry in the file table.  If "create" is true, and the
 * entry does not exist, a new entry is created and entered in the
 * table.  In this case, "isNew" is set to true, otherwise it is set
 * false.  The new entry is initialized with the given path name, but
 * otherwise with unknown information (in particular, the type is
 * assumed to be "file".)
 */
File* File::underlyingFind(
    const char*		name,		// Pathname of file to find
    unsigned int	nameLength,	// Length of file name, zero = unknown
    bool		create,		// True to create if not found
    bool&		isNew)		// OUT: Set true if newly-created
    {
    if (newEntry == NULL)
	{
	newEntry = new File;
	if (newEntry == NULL)
	    nomem("File::find", __LINE__, sizeof(File));
	}
    newEntry->path = name;
    if (nameLength == 0)
	nameLength = strlen(name);
    if (nameLength >= MAXPATHLEN)
	{
	// NEEDSWORK:  should handle more gracefully
	cerr << "correlator:  file name '" << name << "' is too long!\n";
	if (!create)
	    return NULL;
	correlatorPanic("");
	}
    newEntry->pathLength = nameLength;
    File* foundEntry = (File*) fileTable.find(newEntry,
      create ? HashTable::INSERT : HashTable::FAIL);
    if (foundEntry != newEntry)
	{
	isNew = false;
	newEntry->path = NULL;		// So final destruction won't fail
	newEntry->pathLength = 0;
	}
    else
	{
	/*
	 * We just inserted a new entry.  Finish initializing it.
	 */
	isNew = true;
	newEntry = NULL;		// This has been used up
	// NEEDSWORK: should use new/delete
	foundEntry->path = strdup(name, nameLength);
	}
    return foundEntry;
    }

/*
 * Locate a file entry in the file table, canonicalizing the pathname
 * in the process.  Regardless of the value of "create," file-table
 * entries will be made for all directories in the file's path,
 * including symlinks.  However, if a nonexistent path element is
 * encountered anywhere in the path, it and subsequent elements will
 * not be entered in the file table.
 */
File* File::findCanonical(
    const char*		name,		// Pathname of file to find
    bool		wantLink,	// True if want symlink, not target
    bool		create,		// True to create if not found
    Process*		process)	// Process for whom find is done
    {
    static File*	rootDirectory = NULL; // Entry for "/"
    if (rootDirectory == NULL)
	{
	bool rootWasCreated;
	rootDirectory = underlyingFind(rootName, 1, true, rootWasCreated);
	rootDirectory->parentDirectory = rootDirectory; // Own parent
	rootDirectory->discoverStatus();
	if (simulationMode)
	    {
	    rootDirectory->type = FT_DIR;
	    if (logDirectoryGeometricP != 0.0
	      &&  rootDirectory->sizeWasGuessed())
		{
		long byteSize =
		  (long) (log(drand48()) / logDirectoryGeometricP);
		rootDirectory->blocks =
		  (byteSize + STAT_BYTES_PER_BLOCK - 1)
		    / STAT_BYTES_PER_BLOCK;
		}
	    }
	}
    /*
     * We special-case the root, because it's the only place where we
     * allow a trailing slash, and it's easier to special-case than to
     * write code that handles it in the loop.
     */
    while (*name == '/')
	name++;				// Skip irrelevant slashes
    if (*name == '\0')
	return rootDirectory;
    char truePath[MAXPATHLEN];		// Canonicalized path
    truePath[0] = '\0';
    char* trueEnd = truePath;
    File* lastParent = rootDirectory;
    File* component = NULL;
    while (*name != '\0')
	{
	/*
	 * Loop invariants:
	 *
	 * 1.	truePath contains the canonical absolute path so far.
	 * 2.	trueEnd points to the null byte that terminates
	 *	truePath.
	 * 3.	name points to the next element in the input path,
	 *	with no leading slashes.
	 * 4.	lastParent points to the last directory encountered in
	 *	truePath, or rootDirectory if truePath is null (which
	 *	means that the proper parent is the root).  Note that
	 *	this means that most of the time, lastParent points to
	 *	the file referenced by truePath itself.
	 */
	/*
	 * Handle '.' and '..'.  Note that, since we have a budding
	 * absolute path in truePath, we don't have to go through the
	 * gyrations of getcwd() to go upwards.
	 */
	if (isDot(name))
	    {
	    name++;
	    while (*name == '/')
		name++;
	    continue;
	    }
	else if (isDotDot(name))
	    {
	    name += 2;
	    while (*name == '/')
		name++;
	    trueEnd = strrchr(truePath, '/');
	    if (trueEnd == NULL)
		trueEnd = truePath;
	    *trueEnd = '\0';
	    if (lastParent == NULL)
		component = NULL;
	    else
		component = lastParent->parentDirectory;
	    /*
	     * It is possible that the lastParent has a null
	     * parentDirectory (this happens if it is in the process
	     * of being deleted, as for example a missing file).  We
	     * handle that case by just returning to the hash table;
	     * the use of lastParent->parentDirectory is just an
	     * optimization anyway.
	     */
	    if (component == NULL)
		{
		bool componentIsNew;
		component = underlyingFind(truePath, trueEnd - truePath, true,
		  componentIsNew);
		}
	    lastParent = component;
	    continue;
	    }
	/*
	 * Append next component of name to truePath
	 */
	if (trueEnd < &truePath[MAXPATHLEN - 1])
	    *trueEnd++ = '/';
	bool dotFile = name[0] == '.';
	while (*name != '\0'  &&  *name != '/'
	  &&  trueEnd < &truePath[MAXPATHLEN - 1])
	    *trueEnd++ = *name++;
	while (*name == '/')
	    name++;			// Skip irrelevant slashes
	/*
	 * Look for a file-table entry for this file, because if we
	 * have one, we can avoid doing a stat.  If there is no entry,
	 * create one to hold the stat information, but keep track of
	 * whether it's new because if it is, we will want to delete
	 * it if the stat fails due to ENOENT.
	 */
	*trueEnd = '\0';
	bool componentIsNew;
	component =
	  underlyingFind(truePath, trueEnd - truePath, true, componentIsNew);
	if (component->parentDirectory == NULL)
	    component->setParent(lastParent);
	/*
	 * Because dot files tend to be important and small (e.g.,
	 * login scripts, configuration control, etc.), we'll
	 * arbitrarily mark all dot files as uncontrolled.
	 *
	 * NEEDSWORK: this is far too system-specific.
	 */
	if (dotFile  &&  componentIsNew)
	    component->flags &= ~FF_CONTROLLED;
	/*
	 * Set the component's state, if possible.
	 */
	if (!component->isKnown())
	    {
	    component->discoverStatus();
	    if (simulationMode
	      ||  component->state.state() == FileState::MISSING)
		{
		/*
		 * The file (or a component in the path) doesn't
		 * exist.  If we are supposed to be creating a file
		 * table entry, we will continue, but we may want to
		 * treat it as a directory if the pathname continues
		 * beyond this point.  This will cause us to do the
		 * "right" thing for pathnames that are temporarily
		 * inaccessible due to missing network mounts.
		 */
		if (create)
		    {
		    if (*name != '\0'  &&  component->type == FT_FILE)
			{
			component->type = FT_DIR;
			if (simulationMode  &&  logDirectoryGeometricP != 0.0
			  &&  component->sizeWasGuessed())
			    {
			    long byteSize =
			      (long) (log(drand48()) / logDirectoryGeometricP);
			    component->blocks =
			      (byteSize + STAT_BYTES_PER_BLOCK - 1)
				/ STAT_BYTES_PER_BLOCK;
			    }
			}
		    /*
		     * In simulation mode, having a file be missing is
		     * not a big deal.  For example, the file in
		     * question might have been renamed later, or
		     * referenced for a few days and then deleted.  So
		     * we'll force the state to UNSTORED to prevent
		     * future deletions.  Note that this is less than
		     * perfect, because the file size is still listed
		     * as being zero blocks, so it'll take up no space
		     * in the hoard.
		     */
		    if (simulationMode)
			component->state.setState(*component,
			  FileState::UNSTORED);
		    }
		else if (componentIsNew)
		    {
		    /*
		     * The file (or a component in the path) doesn't
		     * exist, the file table entry was just created, and
		     * the caller doesn't want it created.  Delete what we
		     * just created and fail.
		     */
		    (void) fileTable.remove(component);
		    delete component;
		    component = NULL;
		    return NULL;
		    }
		}
	    }
	else if (simulationMode)
	    {
	    /*
	     * Even if the component was previously listed as a file,
	     * it's probably a directory if we're trying to look up
	     * something beneath it.  This is because simulation mode
	     * doesn't stat files, and we can get a sequence like:
	     *
	     *	reference to /foo/bar (e.g., readdir)
	     *		-- sets type to FT_FILE, even though it's a directory
	     *	reference to /foo/bar/baz
	     *		-- /foo/bar needs to be changed to FT_DIR
	     */
	    if (*name != '\0'  &&  component->type == FT_FILE)
		{
		component->type = FT_DIR;
		if (simulationMode  &&  logDirectoryGeometricP != 0.0
		  &&  component->sizeWasGuessed())
		    {
		    long byteSize =
		      (long) (log(drand48()) / logDirectoryGeometricP);
		    component->blocks =
		      (byteSize + STAT_BYTES_PER_BLOCK - 1)
			/ STAT_BYTES_PER_BLOCK;
		    }
		}
	    }
	/*
	 * We now know whether the component exists, and what sort of
	 * component it is.
	 *
	 * If it is a directory, we just record it as the last parent.
	 * Except for symlinks, all other components are trivial to
	 * handle.
	 *
	 * If the component is a symlink, we must chase it (unless the
	 * caller wants the link itself and it's at the end of the
	 * path).  We do this by substituting the symlink's value for
	 * the current component in the name.  That's handled by
	 * prepending truePath and appending the rest of the name,
	 * being careful to avoid buffer overflows.  We must also set
	 * lastParent properly in this case.
	 */
	if (component->type == FT_DIR)
	    lastParent = component;
	else
	    {
	    int linkLimit = MAX_LINKS;
	    while (component->type == FT_SYMLINK  &&  --linkLimit >= 0)
		{
		if (wantLink  &&  *name == '\0')
		    {
		    if (componentIsNew  &&  !create)
			{
			/*
			 * The file (or a component in the path)
			 * doesn't exist, the file table entry was
			 * just created, and the caller doesn't
			 * want it created.  Delete what we just
			 * created and fail.
			 */
			(void) fileTable.remove(component);
			delete component;
			component = NULL;
			return NULL;
			}
		    else
			return component;
		    }
		/*
		 * If the component is a symlink, make sure we know
		 * the target.  However, in simulation mode, don't
		 * attempt to read the target because there's no
		 * guarantee that the simulation is being run on a
		 * machine that actually stores the filesystem in
		 * question.  Instead, in simulation mode we'll depend
		 * on the controller to pre-load link targets.
		 */
		if (component->type == FT_SYMLINK
		  &&  component->linkTargetFile == NULL
		  &&  !simulationMode)
		    {
		    char linkValue[MAXPATHLEN];
		    if (replayFile != NULL  &&  replaymode
		      &&  !replayedRestart)
			{
			char saveType;
			if (!replayFile->read(&saveType, sizeof saveType))
			    {
			    cerr << "correlator:  error reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for savetype at offset "
			      << replayFile->bytesRead();
			    correlatorPanic("");
			    }
			if (saveType != ST_READLINK)
			    {
			    cerr << "correlator:  unexpected savetype "
			      << (int) saveType << " at offset "
			      << replayFile->bytesRead() - sizeof saveType;
			    correlatorPanic(
			      " of replay file in File::findCanonical");
			    }
			int nameLength;
#ifdef NAME_IN_STAT_RECORD
			if (!replayFile->read(&nameLength, sizeof nameLength))
			    {
			    cerr << "correlator:  error reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink name length at offset "
			      << replayFile->bytesRead();
			    correlatorPanic("");
			    }
			char savedName[MAXPATHLEN];
			if (nameLength >= (int)sizeof(savedName))
			    {
			    cerr << "correlator:  bad name length "
			      << nameLength << " reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink name at offset "
			      << replayFile->bytesRead() - sizeof nameLength
			      << '\n';
			    }
			if (!replayFile->read(savedName, nameLength + 1))
			    {
			    cerr << "correlator:  error reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink name at offset "
			      << replayFile->bytesRead();
			    correlatorPanic("");
			    }
			if (strcmp(component->path, savedName) != 0)
			    {
			    cerr << "correlator:  looking for readlink record"
				" for '"
			      << truePath << "', got one for '" << savedName
			      << "' at offset "
			      << replayFile->bytesRead() - nameLength - 1;
			    correlatorPanic(
			      " of replay file in File::findCanonical");
			    }
#endif // NAME_IN_STAT_RECORD
			if (!replayFile->read(&nameLength, sizeof nameLength))
			    {
			    cerr << "correlator:  error reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink target length at offset "
			      << replayFile->bytesRead();
			    correlatorPanic("");
			    }
			if (nameLength >= (signed)sizeof(linkValue))
			    {
			    cerr << "correlator:  bad name length "
			      << nameLength << " reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink target at offset "
			      << replayFile->bytesRead() - sizeof nameLength;
			    correlatorPanic("");
			    }
			if (nameLength <= 0)
			    {
			    /*
			     * A zero length indicates that the
			     * readlink failed.
			     */
			    if (!component->willForget())
				{
				cerr << "correlator:  unexpected readlink"
				    " error for '"
				  << truePath << "'\n";
				component->state.setState(*component,
				  FileState::MISSING);
				}
			    if (componentIsNew  &&  !create)
				{
				/*
				 * The file (or a component in the path)
				 * doesn't exist, the file table entry was
				 * just created, and the caller doesn't
				 * want it created.  Delete what we just
				 * created and fail.
				 */
				(void) fileTable.remove(component);
				delete component;
				component = NULL;
				return NULL;
				}
			    else
				// NEEDSWORK: we should return NULL here,
				// I think
				return component;
			    }
			else if (!replayFile->read(linkValue, nameLength + 1))
			    {
			    cerr << "correlator:  error reading fd "
			      << replayFile->rdbuf()->fd()
			      << " for readlink target at offset "
			      << replayFile->bytesRead();
			    correlatorPanic("");
			    }
			else if ((int)strlen(linkValue) != nameLength)
			    {
			    cerr << "correlator:  actual link length of "
			      << strlen(linkValue)
			      << " doesn't match saved length of "
			      << nameLength << " at offset "
			      << replayFile->bytesRead() - nameLength - 1;
			    correlatorPanic("");
			    }
			}
		    else
			{
			if (component->stripLength != 0)
			    {
			    /*
			     * Files with stripping are usually caused
			     * by the automounter, and we have to do
			     * some special gymnastics to get the
			     * right information.  See discoverStatus
			     * for a more complete explanation.  I'm
			     * not sure that this code is necessary in
			     * this place, because I think that
			     * discoverStatus will have already taken
			     * care of the automounting, but better
			     * safe than sorry...
			     */
			    char dummyBuffer[MAXPATHLEN];
			    (void)strncpy(dummyBuffer, component->name(),
			      MAXPATHLEN);
			    dummyBuffer[MAXPATHLEN - 1] = '\0';
			    (void)strncat(dummyBuffer, "/.",
			      MAXPATHLEN - strlen(dummyBuffer) - 1);
			    if (!simulationMode)
				{
				struct stat statBuffer;
				(void)lstat(dummyBuffer, &statBuffer);
				}
			    }
			int linkSize =
			  readlink(component->path, linkValue, MAXPATHLEN);
			if (linkSize >= MAXPATHLEN)
			    {
			    cerr << "correlator:  bad readlink size "
			      << linkSize;
			    correlatorPanic("");
			    }
			else if (linkSize >= 0)
			    linkValue[linkSize] = '\0';
			int linkErrno = errno;
			if (saveFile != NULL)
			    {
			    char saveType = ST_READLINK;
#ifdef NAME_IN_STAT_RECORD
			    int nameLength = trueEnd - truePath;
#endif // NAME_IN_STAT_RECORD
			    if (!saveFile->write(&saveType, sizeof saveType)
#ifdef NAME_IN_STAT_RECORD
			      ||  !saveFile->write(&nameLength,
				sizeof nameLength)
			      ||  !saveFile->write(truePath, nameLength + 1)
#endif // NAME_IN_STAT_RECORD
			      ||  !saveFile->write(&linkSize, sizeof linkSize)
			      ||  (linkSize >= 0
				&&  !saveFile->write(linkValue, linkSize + 1))
			      ||  !saveFile->flush())
				{
				int saveErrno = errno;
				cerr << "correlator:  trouble writing"
				    " link target to save file: "
				  << strerr(saveErrno);
				correlatorPanic("");
				}
			    }
			if (linkSize < 0)
			    {
			    if (!component->willForget())
				{
				cerr << "correlator:  unexpected readlink"
				    " error for '"
				  << truePath << "': " << strerr(linkErrno)
				  << '\n';
				if (linkErrno != ENOENT
				  &&  linkErrno != EACCES
				  &&  linkErrno != EINVAL
				  &&  linkErrno != EAGAIN)
				    correlatorPanic("");
						// This is a serious error!
				component->state.setState(*component,
				  FileState::MISSING);
				}
			    if (componentIsNew  &&  !create)
				{
				/*
				 * The file (or a component in the path)
				 * doesn't exist, the file table entry was
				 * just created, and the caller doesn't
				 * want it created.  Delete what we just
				 * created and fail.
				 */
				(void) fileTable.remove(component);
				delete component;
				component = NULL;
				return NULL;
				}
			    else
				// NEEDSWORK: we should return NULL here,
				// I think
				return component;
			    }
			}
		    if (linkValue[0] == '/')
			component->linkTargetFile
			  = findCanonical(linkValue, true, true, process);
		    else
			{
			// NEEDSWORK: this could be made more
			// efficient by carefully building a new
			// truePath instead of recursing
			char expandedName[MAXPATHLEN];
			strcpy(expandedName, truePath);
			char* nameStart = strrchr(expandedName, '/');
			if (nameStart == NULL)
			    correlatorPanic(
			      "correlator:  missing slash in absolute path!");
			*++nameStart = '\0';
			strncat(expandedName, linkValue, MAXPATHLEN - 1);
			File* target =
			  findCanonical(expandedName, true, true, process);
			target->discoverStatus();
			component->linkTargetFile = target;
			}
		    }
		if (process != NULL)
		    component->recordReference(process, SEER_REF_FOLLOW);
		File* target = component->linkTargetFile;
		if (target == NULL)
		    {
		    /*
		     * The target is unknown.  If we're at the end of
		     * the pathname, return the symlink itself as
		     * being the best substitute.  Otherwise, pretend
		     * the symlink is a valid directory.
		     */
		    if (name[0] == '\0')
			return component;
		    else
			break;
		    }
		/*
		 * We now know the true absolute pathname up to this
		 * point.  If the link target is a directory, it becomes
		 * lastParent.  Otherwise, the link target's parent
		 * becomes lastParent (and either the name should be
		 * finished, or the target should be a symlink, but we
		 * don't check that).
		 */
		strcpy(truePath, target->path);
		trueEnd = truePath + strlen(truePath);
		if (target->type == FT_DIR)
		    lastParent = target;
		else
		    lastParent = target->parentDirectory;
		component->loseForgettingPointers();
		component = target;
		}
	    }
	}
    /*
     * We've finally found the true absolute pathname of the file.
     * Look it up.
     */
    if (trueEnd == truePath)
	return rootDirectory;
    else if (component == NULL)
	{
/*
 * I think this code is obsolete, so I'm going to instrument it.
 */
cerr << "correlator: OLDCODE: executed non-obsolete code in findCanonical while looking for '" << truePath << "'\n";
	bool componentIsNew;
	component =
	  underlyingFind(truePath, trueEnd - truePath, create, componentIsNew);
	if (component != NULL)
	    component->discoverStatus();
	}
    if (component->parentDirectory == NULL)
	component->setParent(lastParent);
    return component;
    }

/*
 * Read a pathname from a stream, performing validity checks.  The
 * input stream should provide an integer length of the pathname,
 * followed by the zero-terminated pathname itself.  A length of zero
 * is legal, and in that case no following pathname is expected.  The
 * length given is checked for reasonableness, and also compared
 * against the actual length of the pathname read.
 *
 * The length of the pathname is returned in the "pathLength"
 * parameter, if it is non-null.
 *
 * Returns true if the pathname was read and verified successfully,
 * false otherwise.  (Failure includes difficulties with the input
 * stream.)
 */
bool File::readPathname(
    ByteSwappedStream&	stream,		// Stream to read from
    char		pathname[MAXPATHLEN], // Where to put the path
    unsigned int*	pathLength)	// Where to put the length
    {
    int			length;
    if (!stream.read(&length, sizeof length))
	{
	cerr << "trouble getting path length in readPathname\n";
	return false;
	}
    if (pathLength != NULL)
	*pathLength = length;
    if (length == 0)
	return true;			// Zero-length path is a terminator
    else if (length < 0  ||  length > MAXPATHLEN)
	{
	cerr << "correlator:  bad path length "
	  << length << " in readPathname\n";
	return false;
	}
    if (!stream.read(pathname, length + 1))
	{
	cerr << "correlator:  trouble reading path in readPathname\n";
	return false;
	}
    if ((int)strlen(pathname) != length)
	{
	cerr << "correlator:  bad path length "
	  << strlen(pathname) << " for '" << pathname << "', expected "
	  << length << " in readPathname\n";
	return false;
	}
    return true;
    }

/*
 * Read a pathname from a stream, and return a pointer to the
 * file-table entry for that file, creating it if necessary.
 *
 * Returns NULL if trouble occurred, or if a zero-length filename was
 * received.  If trouble occurred, the stream is also closed.
 */
File* File::findFromStream(
    ByteSwappedStream&	stream,		// Stream to read pathname from
    bool		wantLink)	// True if want symlink, not target
    {
    char path[MAXPATHLEN];
    unsigned int pathLength;
    if (!readPathname(stream, path, &pathLength))
	return NULL;
    else if (pathLength == 0)
	return NULL;
    else
	return findCanonical(path, wantLink, true, NULL);
    }

/*
 * Read a file containing special instructions about certain files.
 * In general, the instructions are implemented by setting appropriate
 * flags in the file table entry.
 *
 * The general format of the instruction file is one or more lines of
 * the form:
 *
 *	<command> <filename>
 *
 * Where "filename" is the full path name of a file (not necessarily
 * canonical) and "command" is one of:
 *
 *	neverhoard	Never hoard the named file, even if related
 *			files are hoarded.  Applies to all descendents
 *			if file is a directory.
 *	ignore		Completely ignore all references to the named
 *			file.  Applies to all descendents if file is a
 *			directory.
 *	strip		When expanding path names, strip this file
 *			prefix from any path to get the canonical
 *			version.  Applies only to this file.
 *
 * In addition, blank lines (consisting of newlines without embedded
 * whitespace) and lines beginning with '#' are ignored as comments.
 */
void File::readInstructions(
    const char*		fileName)	// File to read instructions from
    {
    ByteSwappedStream*	instructionFile = NULL; // Access to instruction file
    if (replayFile != NULL)
	instructionFile = replayFile;
    else
	{
	if (fileName == NULL)
	    return;
	instructionFile = new ByteSwappedStream(fileName, ios::in,
	  ByteSwappedStream::bigEnded);
	if (instructionFile == NULL  ||  !*instructionFile)
	    {
	    int saveErrno = errno;
	    cerr << "Couldn't open instruction file '" << fileName << "': "
	      << strerr(saveErrno) << '\n';
	    return;
	    }
	}
    /*
     * Before reading the instructions, clear out any current
     * instruction information in all files.
     */
    for (FileTableIterator fileIterator;  fileIterator;  ++fileIterator)
	{
	File& fileToFix(*fileIterator);
	fileToFix.flags &= ~(FF_NEVERHOARD | FF_IGNORE | FF_MEANINGLESS
	  | FF_FAKE_SIZE | FF_SIZE_GUESSED);
	fileToFix.flags |= FF_CONTROLLED;
	fileToFix.stripLength = 0;
	for (ManagerIterator manager;  manager;  ++manager)
	    (*manager).clearInstructions(fileToFix);
	}
    setpwent();				// Ensure that endpwent will work...
    ParsedFields* fields = NULL;
    while (1)
	{
	delete fields;
	if (instructionFile == replayFile)
	    {
	    char saveType;
	    if (!replayFile->read(&saveType, sizeof saveType))
		{
		cerr << "correlator:  trouble reading save type at offset "
		  << replayFile->bytesRead();
		correlatorPanic(" in File::readInstructions");
		}
	    if (!replaymode  &&  saveType == ST_STAT)
		{
		// NEEDSWORK: this incorporates too much knowledge!
		/*
		 * On catchup-type replays, we ignore stat records.
		 */
		unsigned int nameLength;
#ifdef NAME_IN_STAT_RECORD
		if (!replayFile->read(&nameLength, sizeof nameLength))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " for stat name length at offset "
		      << replayFile->bytesRead();
		    correlatorPanic("");
		    }
		char statFileName[MAXPATHLEN];
		if (nameLength >= sizeof(statFileName))
		    {
		    cerr << "correlator:  bad name length " << nameLength
		      << " at offset "
		      << replayFile->bytesRead() - sizeof nameLength
		      << " reading fd " << replayFile->rdbuf()->fd();
		    correlatorPanic(" for stat name");
		    }
		if (!replayFile->read(statFileName, nameLength + 1))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " at offset " << replayFile->bytesRead();
		    correlatorPanic(" for stat name");
		    }
#endif // NAME_IN_STAT_RECORD
		char charDummy;
		if (!replayFile->read(&charDummy, sizeof charDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " at offset " << replayFile->bytesRead();
		    correlatorPanic(" for state");
		    }
		if (!replayFile->read(&charDummy, sizeof charDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " at offset " << replayFile->bytesRead();
		    correlatorPanic(" for type");
		    }
		unsigned long longDummy;
#ifdef LENGTH_IN_STAT_RECORD
		if (!replayFile->read(&longDummy, sizeof longDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " at offset " << replayFile->bytesRead();
		    correlatorPanic(" for file size");
		    }
#endif // LENGTH_IN_STAT_RECORD
		if (!replayFile->read(&longDummy, sizeof longDummy))
		    {
		    cerr << "correlator:  error reading fd "
		      << replayFile->rdbuf()->fd()
		      << " at offset " << replayFile->bytesRead();
		    correlatorPanic(" for blocks");
		    }
		continue;
		}
	    else if (saveType == ST_INSTRUCTION_END)
		break;
	    else if (saveType != ST_INSTRUCTION)
		{
		cerr << "correlator:  unexpected save type " << (int)saveType
		  << " at offset "
		  << replayFile->bytesRead() - sizeof saveType;
		correlatorPanic(" in File::readInstructions");
		}
	    fields = new ParsedFields(*replayFile, " \t", ParsedFields::RUNS);
	    if (!*fields)
		{
		cerr << "correlator:  early EOF on replay file at offset "
		  << replayFile->bytesRead() << " in File::readInstructions\n";
		break;
		}
	    replayFile->reportBytesRead(fields->inputLineLength());
	    }
	else
	    {
	    fields =
	      new ParsedFields(*instructionFile, " \t", ParsedFields::RUNS);
	    if (!*fields)
		{
		delete fields;
		break;
		}
	    }
	if (saveFile != NULL)
	    {
	    char saveType = ST_INSTRUCTION;
	    if (!saveFile->write(&saveType, sizeof saveType)
	      ||  !fields->write(*saveFile, ' '))
		correlatorPanic("correlator:  trouble writing to save file");
	    }
	if (fields->count() == 0  ||  (*fields)[0][0] == '#')
	    continue;
	const char* instructionPath = (*fields)[1];
	if (fields->count() < 2)
	    cerr << "correlator:  instruction must mention path name in '"
	      << *fields << "'\n";
	else if ((instructionPath = tildeExpand(instructionPath)) == NULL)
	    cerr << "correlator:  can't expand ~user notation in '"
	      << *fields << "'\n";
	else if (instructionPath[0] != '/')
	    cerr << "correlator:  instruction path name must be absolute in '"
	      << *fields << "'\n";
	else
	    {
	    /*
	     * With the exception of manager commands, all
	     * instructions have the same general form:
	     *
	     * <command> <path> [link|target]
	     *
	     * where <command> is the command to be executed and
	     * <path> is the path it applies to.  If a third argument
	     * is present, "link" specifies that if <path> refers to a
	     * symbolic link, the command refers to the link itself,
	     * while "target" specifies that the command refers to the
	     * target of the link (and any intervening links).  The
	     * default is "target".
	     */
	    File* file = NULL;
	    bool wantLink = false;
	    if (fields->count() > 2)
		{
		if (strcmp((*fields)[2], "link") == 0)
		    wantLink = true;
		else if (strcmp((*fields)[2], "target") == 0)
		    wantLink = false;
		else
		    cerr << "correlator: invalid link/target selection in '"
		      << *fields << "'\n";
		}
	    if (strcmp((*fields)[0], "neverhoard") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->flags |= FF_NEVERHOARD;
		// NEEDSWORK: don't really need to set must-save if
		// parent has the same bits, but it's not very
		// expensive to have it set
		file->mustSave();
		}
	    else if (strcmp((*fields)[0], "ignore") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->flags |= FF_IGNORE;
		// NEEDSWORK: don't really need to set must-save if
		// parent has the same bits, but it's not very
		// expensive to have it set
		file->mustSave();
		}
	    else if (strcmp((*fields)[0], "meaningless") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->flags |= FF_MEANINGLESS;
		// NEEDSWORK: don't really need to set must-save if
		// parent has the same bits, but it's not very
		// expensive to have it set
		file->mustSave();
		}
	    else if (strcmp((*fields)[0], "control") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->flags |= FF_CONTROLLED;
		// NEEDSWORK: don't really need to set must-save if
		// parent has the same bits, but it's not very
		// expensive to have it set
		file->mustSave();
		}
	    else if (strcmp((*fields)[0], "nocontrol") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->flags &= ~FF_CONTROLLED;
		// NEEDSWORK: don't really need to set must-save if
		// parent has the same bits, but it's not very
		// expensive to have it set
		file->mustSave();
		}
	    else if (strcmp((*fields)[0], "fakesize") == 0)
		{
		if (fields->count() != 3)
		    cerr << "correlator:  fakesize command must have exactly"
		      << " 2 arguments in '" << *fields << "'\n";
		else
		    {
		    file =
		      findCanonical(instructionPath, wantLink, true, NULL);
		    file->flags |= FF_FAKE_SIZE;
		    file->flags &= ~FF_SIZE_GUESSED;
		    file->blocks = atoi((*fields)[2]);
		    file->mustSave();
		    }
		}
	    else if (strcmp((*fields)[0], "strip") == 0)
		{
		file = findCanonical(instructionPath, wantLink, true, NULL);
		file->stripLength = file->pathLength;
		}
	    else
		{
		/*
		 * Unrecognized instructions might be the names of
		 * files to feed to individual managers.
		 */
		ManagerBase* manager =
		 ManagerBase::findNamedManager((*fields)[0]);
		if (manager == NULL)
		    cerr << "bad instruction line '" << *fields << "'\n";
		else
		    manager->readInstructions(instructionPath);
		}
	    if (file != NULL)
		{
		for (FileDescendantsIterator descendant(*file);
		  descendant;
		  ++descendant)
		    (*descendant).inheritFlagsFromParent();
		}
	    }
	}
    if (saveFile != NULL)
	{
	char saveType = ST_INSTRUCTION_END;
	if (!saveFile->write(&saveType, sizeof saveType)
	  ||  !saveFile->flush())
	    correlatorPanic("correlator:  trouble writing to save file");
	}
    if (instructionFile != replayFile)
	delete instructionFile;
    instructionFile = NULL;
    endpwent();
    }

unsigned int File::hash(		// Calculate hash code for a file
    File*		file)		// File to figure code for
    {
    return hashString(file->path);
    }

int File::compare(			// Compare files for hash purposes
    File*		a,		// Files to compare
    File*		b)		// ...
    {
    return *a != *b;
    }

void File::write(			// Write a file entry to a stream
    ByteSwappedStream&	stream)		// Stream to dump to
    const
    {
    if (pathLength == 0)
	{
	cerr << "correlator:  zero path length!\n";
	return;
	}
    if (state.state() == FileState::FORGETTING
      ||  state.state() == FileState::FORGOTTEN)
	return;				// Don't dump what we're deleting

    /*
     * Note that we don't write the flags.  All flag bits are either
     * dynamic or are gotten from the instructions file.
     */
    writePath(stream, false);
    unsigned char dumpByte;
    dumpByte = (unsigned char)type;
    stream.write(&dumpByte, sizeof dumpByte);
    dumpByte = (unsigned char)state.state();
    stream.write(&dumpByte, sizeof dumpByte);
    stream.write(&blocks, sizeof blocks);
    dumpByte = (unsigned char)lastReferenceType;
    stream.write(&dumpByte, sizeof dumpByte);
    stream.write(&peekCount, sizeof peekCount);
    stream.write(&potentialPeeks, sizeof potentialPeeks);
    short relatedFileCount = 0;
    RelatedFileIterator relatedFile((File&)*this);
    for (  ;  relatedFile;  ++relatedFile)
	{
	RelatedFile& possibleFile(*relatedFile);
	if (!possibleFile.isEmpty()
	  &&  !possibleFile.targetFile().isForgetting())
	    relatedFileCount++;
	}
    stream.write(&relatedFileCount, sizeof relatedFileCount);
    for (relatedFile.reset();  relatedFile;  ++relatedFile)
	{
	RelatedFile& writableFile(*relatedFile);
	if (!writableFile.isEmpty()
	  &&  !writableFile.targetFile().isForgetting())
	    writableFile.write(stream);
	}
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).managerData(*this).write(stream, *this);
    }

void File::writePath(			// Write a file's path to a stream
    ByteSwappedStream&	stream,		// Stream to write to
    bool		stripped)	// True to write stripped,
					// ..semi-canonical version of the path
    const
    {
    if (stripped  &&  pathLength > stripLength)
	{
	unsigned int writtenPathLength = pathLength - stripLength;
	stream.write(&writtenPathLength, sizeof writtenPathLength);
	stream.write(path + stripLength, writtenPathLength + 1);
	}
    else
	{
	stream.write(&pathLength, sizeof pathLength);
	stream.write(path, pathLength + 1);
	}
    }

void File::writePathTerminator(		// Write a file-list terminator
    ByteSwappedStream&	stream)		// Stream to write to
    {
    static unsigned int	nullLength = 0;
    stream.write(&nullLength);
    }

/*
 * Read a file entry from a given stream.  If believeSavedState is
 * true, the file table entry is filled in from the information
 * provided in the stream.  If false, most of this information is not
 * trusted, and the file is left uninitialized.  The File::discoverStatus()
 * routine will be called later to stat the file and fill in the
 * uninitialized fields.
 *
 * Note that this routine knows far too much about the types of
 * various file-table fields.
 *
 * Returns true if the stream is still valid.
 */
bool File::read(			// Read a file entry from a stream
    ByteSwappedStream&	stream)		// Stream to read from
    {
    unsigned int pathLength;		// Length of next path
    if (!stream.read(&pathLength, sizeof pathLength))
	return false;			// EOF, give up
    char pathBuffer[MAXPATHLEN];	// Buffer for pathname
    if (pathLength <= 0  ||  pathLength >= (int)sizeof pathBuffer)
	{
	cerr << "File::read:  bad path length " << pathLength
	  << " at offset " << stream.bytesRead() - sizeof pathLength << '\n';
	return false;
	}
    stream.read(pathBuffer, pathLength + 1);
    if (strlen(pathBuffer) != (unsigned)pathLength)
	{
	cerr << "File::read: path length " << pathLength
	  << " doesn't match name length " << strlen(pathBuffer)
	  << " at offset "
	  << stream.bytesRead() - pathLength - 1 - sizeof pathLength
	  << " for file '" << pathBuffer
	  << "'\n";
	return false;
	}
    /*
     * We use findCanonical instead of underlyingFind for a couple of
     * reasons.  One is to protect ourselves against the possibility
     * that part of the path has changed from a regular directory to a
     * symlink.  Another is to ensure that things like
     * parent-directory pointers are filled in correctly.
     */
    File* file = findCanonical(pathBuffer, true, true, NULL);
    if (file == NULL)
	{
	cerr << "File::read:  unexpected failure to find file '" << pathBuffer
	  << "'\n";
	return false;
	}
    /*
     * We've got the proper file table entry.  Fill it in.  Note that
     * we don't read the flags.  All flag bits are either dynamic or
     * are gotten from the instructions file via findCanonical.
     * Similarly, findCanonical sets the file type, block size, and
     * strip length.  However, if the file is currently unstored, we
     * must set the type and size from the save file.
     */
    unsigned char typeByte;
    stream.read(&typeByte, sizeof typeByte);
    unsigned char inputByte;
    /*
     * If the file was marked as UNSTORED and is currently marked as
     * anything other than KNOWN, we preserve the UNSTORED status.
     *
     * Also, if the file is currently missing, was marked as anything
     * other than missing in the save file and is a controlled file
     * (for hoarding purposes), we mark it as unstored on the
     * assumption that it was dropped while we were not running.  This
     * might have the unpleasant side effect of causing a lot of
     * truly-missing files to get preserved in the tables.
     */
    stream.read(&inputByte, sizeof inputByte);
    if ((FileState::State)inputByte == FileState::UNSTORED
      &&  !file->isKnown())
	file->state.setState(*file, FileState::UNSTORED);
    else if ((FileState::State)inputByte != FileState::MISSING
      &&  file->currentState() == FileState::MISSING
      &&  file->isControlled())
	file->state.setState(*file, FileState::UNSTORED);
    /*
     * We read the saved block size only if the file isn't marked KNOWN.
     */
    if (file->currentState() == FileState::KNOWN)
	stream.rdbuf()->seekoff(sizeof file->blocks, ios::cur, ios::in);
    else
	{
	file->type = (FileType)typeByte;
	stream.read(&file->blocks, sizeof file->blocks);
	}
    if (logVersion <= 5)
	{
	/*
	 * Older correlators kept one total-references field for all managers
	 */
	clusteringManager.managerData(*file).readTotalReferences(stream);
	/*
	 * Older correlators kept one last-reference field for all managers
	 */
	ReferenceCount lastReference;
	stream.read(&lastReference, sizeof lastReference);
	for (HoardingManagerIterator manager;  manager;  ++manager)
	    {
	    HoardingManagerBase& thisManager(*manager);
	    thisManager.setReferenceTime(*file, lastReference, 0);
	    thisManager.increaseCurrentReferenceTime(lastReference);
	    }
	}
    stream.read(&inputByte, sizeof inputByte);
    file->lastReferenceType = (ReferenceType)inputByte;
    stream.read(&file->peekCount, sizeof file->peekCount);
    stream.read(&file->potentialPeeks, sizeof file->potentialPeeks);
    short relatedFileCount;
    stream.read(&relatedFileCount, sizeof relatedFileCount);
    for (  ;  relatedFileCount > 0;  relatedFileCount--)
	RelatedFile::read(stream, *file);
    if (!stream)
	{
	cerr << "File::read:  unexpected error reading entry for file '"
	  << file->path << "' at offset " << stream.bytesRead() << '\n';
	return false;
	}
    for (HoardingManagerIterator manager;  manager;  ++manager)
	{
	if (!(*manager).managerData(*file).read(stream, *file))
	    return false;
	}
    return true;
    }

/*
 * Change the name of an existing file.  Operates by finding the file
 * under its old name, and replacing the name-related parts of its
 * entry.  If the new file existed, it is given the old name (just as
 * a convenience, because it needs to have some sort of name) and
 * marked for deletion.
 *
 * An important point to remember here is that we see the rename after
 * it has already happened.  This means that (barring further system
 * calls relating to the two files) the new file already exists, and
 * the old file does not.
 *
 * If the old file didn't exist in our tables, nothing is done.  The
 * caller is then responsible for creating an entry for the new file.
 *
 * NEEDSWORK: on directory renames, we must walk the file table and
 * change the path names of all children.
 */
void File::rename(			// Change the name of a file
    const char*		oldName,	// Name file used to have
    const char*		newName,	// Name file will have now
    Process*		process)	// Process doing the rename
    {
    /*
     * Find the old and new files, and rename the old to the new.  If
     * the new file doesn't exist, and the old one does, we'll create
     * the new file as a way to canonicalize its name.
     * If the new file already existed, we probably want to keep its
     * relations rather than those from the old file.  But if the new
     * file didn't already exist, then the old file's relations are
     * much more likely to be relevant than the empty ones from the
     * new file.  In the first case (new already existed), we'll just
     * mark the old file for deletion, and reset the new file to an
     * unknown status so that its size, etc., will be rediscovered.
     * In the second case, we'll swap the old and new names, and then
     * mark the undesired one for deletion.
     */
    File* oldFile = findCanonical(oldName, true, false, process);
    if (oldFile != NULL)
	{
	File* newFile = findCanonical(newName, true, false, process);
	// @h NEEDSWORK: shouldn't depend on clustering manager
	if (newFile != NULL
	  &&  !newFile->isDirectory()
	  &&  (clusteringManager.managerData(*newFile).totalReferenceCount()
	      == 0
	    ||  newFile->relatedFiles.isEmpty()))
	    {
	    /*
	     * The new file has nothing worth preserving, so ignore it
	     * and arrange to rename the old file instead.
	     */
	    newFile = NULL;
	    }
	if (newFile != NULL)
	    {
	    oldFile->markDeleted(false);
	    oldFile->renameDescendants(oldFile->path, newFile->path, newFile);
	    newFile->forgetStatus();
	    }
	else
	    {
	    newFile = findCanonical(newName, true, true, process);
	    /*
	     * Give the old file the new file's name.  This is
	     * trickier than it seems, because a number of fields
	     * (e.g., parent directory, strip information, flags)
	     * depend on the file name.
	     *
	     * This is done so that any file-related data, such as
	     * references, is kept properly.
	     */
	    (void) fileTable.remove(oldFile);
	    (void) fileTable.remove(newFile);
	    const char* tcp = newFile->path;
	    newFile->path = oldFile->path;
	    oldFile->path = tcp;
	    unsigned int tempLength = newFile->pathLength;
	    newFile->pathLength = oldFile->pathLength;
	    oldFile->pathLength = tempLength;
	    /*
	     * Since we swapped names, the stays-with-name flags must
	     * follow the names, which means that they must also be
	     * swapped.  The others (the ones that don't stay with the
	     * name) should stay where they are.
	     */
	    unsigned short tempFlags = newFile->flags & FF_STAYS_WITH_NAME;
	    newFile->flags &= ~FF_STAYS_WITH_NAME;
	    newFile->flags |= oldFile->flags & FF_STAYS_WITH_NAME;
	    oldFile->flags &= ~FF_STAYS_WITH_NAME;
	    oldFile->flags |= tempFlags;
	    File* tempParent = newFile->parentDirectory;
	    newFile->setParent(oldFile->parentDirectory);
	    oldFile->setParent(tempParent);
	    unsigned short tempStrip = newFile->stripLength;
	    newFile->stripLength = oldFile->stripLength;
	    oldFile->stripLength = tempStrip;
	    if (fileTable.find(oldFile, HashTable::INSERT) != oldFile
	      ||  fileTable.find(newFile, HashTable::INSERT) != newFile)
		correlatorPanic(
		  "correlator:  unexpected file reinsertion failure "
		  "in rename");
	    /*
	     * Note that the paths have already been exchanged here
	     */
	    oldFile->renameDescendants(newFile->path, oldFile->path, NULL);
	    /*
	     * The file pointed to by "newFile" has now been given the old
	     * name, and is now obsolete.  Mark it for deletion.
	     */
#ifdef FICUS_MISSING_DIR_BUG
	    if (newFile->markDeleted(true))
		{
		cerr << "MISSINGDIR:  marked file '" << newFile->path
		  << "' deleted because '" << oldName << "' was renamed to '"
		  << newName << "'\n";
		}
#else // FICUS_MISSING_DIR_BUG
	    newFile->markDeleted(true);
#endif // FICUS_MISSING_DIR_BUG
	    /*
	     * Walk the file table to find all symbolic links pointing
	     * at either file.  If any are found, replace the target
	     * pointers.
	     *
	     * NEEDSWORK: this is a very expensive bit of code.
	     */
	    for (FileTableIterator iter;  iter;  ++iter)
		{
		File& possibleLink(*iter);
		if (possibleLink.linkTargetFile == oldFile)
		    possibleLink.linkTargetFile = newFile;
		else if (possibleLink.linkTargetFile == newFile)
		    possibleLink.linkTargetFile = oldFile;
		}
	    }
	}
    }

/*
 * Walk through all of a directory's descendants, correcting their
 * pathnames to reflect their new parentage.
 */
void File::renameDescendants(
    const char*		oldName,	// Previous name of the directory
    const char*		newName,	// New name of the directory
    File*		newParent)	// New parent directory for
					// ..descendants, if needed
    {
    if (!isDirectory())
	return;
    unsigned int oldLength = strlen(oldName);
    unsigned int newLength = strlen(newName);
    int difference = newLength - oldLength;
    for (FileDescendantsIterator descendants(*this);
      descendants;
      ++descendants)
	{
	/*
	 * The renaming code here is tricky.  "File" is the file that
	 * we want to rename.  "NewPath" is the name that it will
	 * eventually get.  "OldPath" is the name it started with.  We
	 * start by setting up these variables.  Also note that we
	 * remove "file" from the file table.  That means that there
	 * should be no file-table entry for "oldPath".
	 */
	File& file = *descendants;
	if (fileTable.remove(&file))
	    correlatorPanic("correlator: couldn't remove file known to exist "
	      "in renameDescendants");
	char* newPath = (char*)malloc(file.pathLength + difference + 1);
	if (newPath == NULL)
	    nomem("File::renameDescendants", __LINE__,
	      file.pathLength + difference + 1);
	memcpy(newPath, newName, newLength + 1);
	if (file.pathLength < oldLength)
	    correlatorPanic("correlator: unreasonably short path length "
	      "in renameDescendants");
	memcpy(newPath + newLength, file.path + oldLength,
	  file.pathLength + 1 - oldLength);
	const char* oldPath = file.path;
	unsigned int oldPathLength = file.pathLength;
	file.path = newPath;
	file.pathLength += difference;
	/*
	 * "File.path" is now the same as "newPath", the new name of
	 * the file that we want to rename.  We look up the new path
	 * in the file table.  There are two possibilities.  First and
	 * simplest, the new path might not be in the file table.  In
	 * that case, because we use HashTable::INSERT, "file" will be
	 * inserted into the table under the new name, "foundFile"
	 * will then be the same as "file", and all we have to do is
	 * free a bit of memory.
	 *
	 * The complicated case, handled in the "else" clause, is when
	 * there was already a file with the same name as "newPath".
	 */
	File* foundFile = (File*)fileTable.find(&file, HashTable::INSERT);
	if (foundFile == &file)
	    free((void*)oldPath);
	else
	    {
	    /*
	     * There was already a file named "newPath" in the file
	     * table.  "FoundFile" is the file-table entry for this
	     * file.  We would rather have all the data from the
	     * renamed file ("file") rather than from the one that
	     * currently exists in the file table.  So we need to
	     * delete "foundFile".
	     *
	     * Because of the way garbage collection works, we can't
	     * just delete "foundFile".  Instead we must mark it
	     * deleted and let it go away later.  That means that it
	     * must exist in the file table under some name.  There's
	     * only one name that we can be absolutely certain isn't
	     * in the table (without generating a unique one), and
	     * that's "oldPath", because we removed it from the table
	     * in the code above.  So we'll delete "foundFile" from
	     * the file table, give it the old name, and then reinsert
	     * "file" (under the its new name) and "foundFile" (under
	     * the name "file" used to have).
	     */
	    if (fileTable.remove(foundFile))
		correlatorPanic("correlator: couldn't remove file known to "
		  "exist in renameDescendants");
	    free((void*)foundFile->path);
	    foundFile->path = oldPath;
	    foundFile->pathLength = oldPathLength;
	    if (fileTable.find(&file, HashTable::INSERT) != &file
	      ||  fileTable.find(foundFile, HashTable::INSERT) != foundFile)
		correlatorPanic(
		  "correlator: unexpected file reinsertion failure "
		  "in renameDescendants");
	    foundFile->markDeleted(true);
	    }
	}
    if (newParent != NULL)
	{
	/*
	 * The physical location of the new directory has changed, so
	 * we must update parent pointers.  This is only necessary for
	 * direct children.  The code is very similar to that in
	 * setParent(), with the exception that removeCurrent() is
	 * used to remove the child from its old parent.
	 */
	if (newParent->willForget())
	    newParent = NULL;
	for (FileChildrenIterator child(*this);  child;  ++child)
	    {
	    /*
	     * This code is similar to setParent
	     */
	    File& file(*child);
	    if (newParent == file.parentDirectory)
		continue;
	    if (file.parentDirectory != NULL)
		{
		if (file.parentDirectory == this)
		    child.removeCurrent();
		else
		    file.parentDirectory->children.remove(&file);
		}
	    file.parentDirectory = newParent;
	    if (newParent != NULL)
		{
		newParent->children.pushHead(&file);
		if (newParent->unknownChildren > 0)
		    newParent->unknownChildren--;
		}
	    file.inheritFlagsFromParent();
	    /*
	     * We don't inherit the manager data, on the assumption
	     * that it belongs with the object rather than the name.
	     */
	    }
	}
    }

#ifdef FICUS_MISSING_DIR_BUG
bool File::markDeleted(			// Mark file to be deleted
    bool		clearAll)	// True to clear all management data
#else // FICUS_MISSING_DIR_BUG
void File::markDeleted(			// Mark file to be deleted
    bool		clearAll)	// True to clear all management data
#endif // FICUS_MISSING_DIR_BUG
    {
    /*
     * We delete a file from the table by setting its state to
     * "forgetting".  Later management passes will remove it.
     */
#ifdef FICUS_MISSING_DIR_BUG
    bool debugged = false;
#endif // FICUS_MISSING_DIR_BUG
    /*
     * If the file is a directory, and some child is marked is
     * unstored, then we change the state to unstored rather than
     * marking the file for deletion.  This is necessary because some
     * outside force might set a child to unstored and then
     * recursively remove the tree, causing the child to stay in the
     * file table, in which case the directory must also remain.
     */
    if (type == FT_DIR)
	{
	if (path[0] == '/'  &&  path[1] == '\0')
	    correlatorPanic("correlator: attempt to delete root directory");
	for (FileChildrenIterator child(*this);  child;  ++child)
	    {
	    if ((*child).state.state() == FileState::UNSTORED)
		{
		state.setState(*this, FileState::UNSTORED);
		return;
		}
	    }
	}
    if (state.state() != FileState::FORGOTTEN)
	{
#ifdef FICUS_MISSING_DIR_BUG
	if (type == FT_DIR  &&  state.state() == FileState::KNOWN)
	    {
	    debugged = true;
	    cerr << "MISSINGDIR:  Changing file state for directory '" << path
	      << "' at " << (void*)this << " from " << state
	      << " to FORGETTING in markDeleted\n";
	    }
#endif // FICUS_MISSING_DIR_BUG
	if (observerReplayCount != 0  &&  (flags & FF_STORAGE_SET) != 0)
	    state.setState(*this, FileState::UNSTORED);
	else
	    state.setState(*this, FileState::FORGETTING);
	}
    if (clearAll)
	{
	/*
	 * The caller has decided that all currently-recorded
	 * management data is invalid.  Ask the manager and
	 * investigators to clear things out, in case this file entry
	 * gets re-used before it is deleted.
	 */
	clearRelationData();
	for (HoardingManagerIterator manager;  manager;  ++manager)
	    (*manager).managerData(*this).clear(*this);
	/*
	 * Now re-initialize nearly everything we know about this
	 * file.
	 */
	flags &= FF_INHERIT_LIST | FF_MUST_SAVE | FF_STORAGE_SET;
	openCount = 0;
	lastReferenceType = SEER_REF_READ;
	}
#ifdef FICUS_MISSING_DIR_BUG
    return debugged;
#endif // FICUS_MISSING_DIR_BUG
    }

/*
 * Check a file for various pointers to other files that are going to
 * disappear, and get rid of the pointers so they won't wind up
 * dangling.
 */
void File::loseForgettingPointers()
    {
    if (parentDirectory != NULL  &&  parentDirectory->isForgetting())
	{
	/*
	 * If a file is unstored, it's possible that the replication
	 * system or the user will decide to destroy its parent
	 * without destroying the file itself.  So if a parent is
	 * going away and we're unstored, we will force our state to
	 * UNKNOWN so that we'll eventually disappear ourselves.  We'd
	 * rather go directly to MISSING, but that would intefere with
	 * the deletion manager's sanity checks.
	 */
	if (state.state() == FileState::UNSTORED)
	    state.setState(*this, FileState::UNKNOWN);
	else if (!willForget())
	    {
	    cerr << "correlator:  parent directory of '" << path;
	    correlatorPanic("' unexpectedly going away!");
	    }
	setParent(NULL);
	}
    if (linkTargetFile != NULL  &&  linkTargetFile->willForget())
	linkTargetFile = NULL;		// We will become a dangling symlink
    }

/*
 * Mark that a file must be preserved in the file table because it has
 * some information (such as flags) that can't be reconstructed from
 * other sources.  This is recursive: all of the file's parents are
 * also marked the same way.
 */
void File::mustSave()
    {
    for (File* file = this;  file != NULL;  file = file->parent())
	{
	file->flags |= FF_MUST_SAVE;
	if (file->parent() == file)
	    break;
	}
    }

void File::addRelation(			// Add a new relation to file
    FileRelationBase*	relation,	// Relation to be added
    File&		target)		// Target file of the relation
    {
    for (RelatedFileIterator relatedFile(*this);  relatedFile;  ++relatedFile)
	{
	RelatedFile& possibleTarget(*relatedFile);
	if (&possibleTarget.targetFile() == &target)
	    {
	    possibleTarget.addRelation(relation);
	    relationsChanged();
	    return;
	    }
	}
    RelatedFile* newRelatedFile = new RelatedFile(target);
    relatedFiles.pushHead(newRelatedFile);
    newRelatedFile->addRelation(relation);
    relationsChanged();
    }

/*
 * Remove a relation from a file by searching for it.  If the relation
 * is the last one to the given target file, the entire target file
 * will also be removed.  This will invalidate any RelatedFileIterator
 * that is active on the file, so be forewarned!
 *
 * The removed relation is also deleted, so the caller can't refer to
 * it after this routine is called.  This may be a design error, but
 * that's how I decided to do it.
 */
void File::removeRelation(
    FileRelationBase*	relation)	// Relation to be removed
    {
    if (relation == NULL)		// Ignore deletes of null relations
	return;
    for (RelatedFileIterator iterator(*this);  iterator;  ++iterator)
	{
	RelatedFile* relatedFile(&(*iterator));
	if (&relation->targetRelatedFile() == relatedFile)
	    {
	    relatedFile->removeRelation(relation);
	    delete relation;
	    if (relatedFile->isEmpty())
		{
		iterator.removeCurrent();
		delete relatedFile;
		}
	    relationsChanged();
	    break;
	    }
	}
    }

void File::recordReference(		// Record a reference to a named file
    const char*		path,		// Pathname of the file to record
    Process*		process,	// Process doing the reference
    ReferenceType	type)		// Type of reference
    {
    /*
     * Decide how we want to process the reference.  This involves
     * deciding whether we want the link or its target (if the path
     * turns out to refer to a symlink), and deciding whether to
     * discard the reference because the process is known to be
     * meaningless.  Note that we always process some types of
     * references, such as creates, even if the process is
     * meaningless, because it's important to know about major changes
     * in the file tree.
     */
    bool wantLink;			// True if we want symlink, not target
    bool makeNewEntry = true;		// True to create entry if missing
    switch (type)
	{
	case SEER_REF_EXEC:		// Execution
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = false;
	    break;
	    }
	case SEER_REF_EXIT:		// Process exit or re-exec
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = false;
	    makeNewEntry = false;
	    break;
	    }
	case SEER_REF_OPENRW:		// Open for read/write
	case SEER_REF_OPENCREATE:	// Create or truncate/rewrite
	    {
	    /*
	     * @h NEEDSWORK: for opens by alwaysMeaningless processes,
	     * must be sure we don't bump the open count.  This is
	     * tough because we also want to mark the status unknown
	     * at close time, and I'm not sure how to arrange that.
	     * Probably I'm going to need to arrange to look up the
	     * file on a no-creation basis.  But for now, I'll just
	     * fudge the reference type.
	     */
	    if (process != NULL)
		{
		if (process->alwaysMeaningless())
		    type = SEER_REF_CREATE;
		process->setInGetcwd(false);
		}
	    wantLink = false;
	    break;
	    }
	case SEER_REF_OPENREAD:		// Open for read
	case SEER_REF_CLOSE:		// Close of open file
	    {
	    if (process != NULL  &&  process->currentlyMeaningless())
		makeNewEntry = false;
	    wantLink = false;
	    break;
	    }
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_RW:		// Reads and/or writes of existing file
	case SEER_REF_DUP:		// Duplicate an open descriptor
	    {
	    if (process != NULL)
		{
		process->setInGetcwd(false);
		if (process->currentlyMeaningless())
		    makeNewEntry = false;
		}
	    wantLink = false;
	    break;
	    }
	case SEER_REF_DESTROY:		// Destruction (delete)
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = true;
	    makeNewEntry = false;
	    break;
	    }
	case SEER_REF_MAKELINK:		// Make a symbolic link
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = true;
	    break;
	    }
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	case SEER_REF_PEEKLINK:		// Peek at a symbolic link's contents
	    {
	    if (process != NULL)
		{
		if (process->currentlyMeaningless())
		    makeNewEntry = false;
		File* program = process->program();
		if (program != NULL  &&  !process->inGetcwd()
		  &&  !process->hasDirectoryOpen)
		    program->peekCount++;
		}
	    wantLink = type == SEER_REF_PEEKLINK;
	    break;
	    }
	case SEER_REF_FOLLOW:		// Follow a symbolic link
	    {
	    if (process != NULL  &&  process->currentlyMeaningless())
		makeNewEntry = false;
	    wantLink = true;
	    break;
	    }
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	case SEER_REF_RENAME:		// Change of file name
	default:
	    {
	    cerr << "correlator: unexpected reference type " << type;
	    correlatorPanic(" in File::recordReference");
	    break;
	    }
	}
    File* file = findCanonical(path, wantLink, makeNewEntry, process);
    if (file == NULL)
	return;				// Destroy of nonexistent file
    if (process != NULL  &&  type == SEER_REF_OPENREAD)
	{
	if (file->isDirectory())
	    {
	    unsigned int originalPathLength = strlen(path);
	    if (isDotDot(&path[originalPathLength] - 2))
		{
		/*
		 * The process is opening a directory for read.  If the raw
		 * (not the canonicalized) path name ends in "..", assume that
		 * it's because the process is doing a getcwd or similar
		 * operation.  The relevant flag will be cleared when the
		 * process does something that getcwd doesn't do (like
		 * creating a file) or opens a plain file or a
		 * directory not ending in "..".
		 */
		process->setInGetcwd(true);
		}
	    }
	else
	    process->setInGetcwd(false);
	}
    file->recordReference(process, type); // Let underlying routine do the work
    }

void File::recordReference(		// Record a reference to this file
    Process*		process,	// Process doing the reference
    ReferenceType	type)		// Type of reference
    {
    if (type == SEER_REF_EXEC  &&  !isForgotten())
	process->exec(this);		// Keep track of what's running
    if (flags & FF_IGNORE)
	return;				// Pretend this file doesn't exist
    /*
     * If this reference is a dup, just bump the open count and ignore
     * the reference for all other purposes.
     */
    if (type == SEER_REF_DUP)
	{
	recordOpen(process);
	return;
	}
    /*
     * If the immediately previous reference was a peek, and this
     * reference is an open-type reference, change the type of the
     * previous reference, count the open, and otherwise ignore it.
     * NEEDSWORK: This assumes that all managers treat peeks and opens
     * the same.
     */
#if 0
    if ((lastReferenceType == SEER_REF_PEEK
	||  lastReferenceType == SEER_REF_PEEKLINK)
      &&  lastReference == currentReference)
#else // 0
    if (lastReferenceType == SEER_REF_PEEK
	||  lastReferenceType == SEER_REF_PEEKLINK)
#endif // 0
	{
	switch (type)
	    {
	    case SEER_REF_OPENREAD:	// Open for read
		{
		lastReferenceType = type;
		recordOpen(process);
		return;
		}
	    case SEER_REF_OPENRW:	// Open for read/write
	    case SEER_REF_OPENCREATE:	// Create or truncate/rewrite
		{
		flags |= FF_WRITING;
		lastReferenceType = type;
		recordOpen(process);
		return;
		}
	    case SEER_REF_UNKNOWN:	// Unknown reference type
	    case SEER_REF_CREATE:	// Create or truncate/rewrite
	    case SEER_REF_EXEC:		// Execution
	    case SEER_REF_INTERPRET:	// Execution of script interpreter
	    case SEER_REF_READ:		// Read of a file
	    case SEER_REF_RW:		// Reads and/or writes of existing file
	    case SEER_REF_DESTROY:	// Destruction (delete)
	    case SEER_REF_CLOSE:	// Close of open file
	    case SEER_REF_EXIT:		// Process exit or re-exec
	    case SEER_REF_MAKELINK:	// Create symbolic link
		{
		break;
		}
	    case SEER_REF_FOLLOW:	// Follow symbolic link
	    case SEER_REF_PEEK:		// Peek at a file (stat or access)
	    case SEER_REF_PEEKLINK:	// Peek at a link (lstat or readlink)
		{
		lastReferenceType = type;
		return;
		}
	    case SEER_REF_RENAME:	// Change of file name
	    case SEER_REF_DUP:		// Duplicate existing open
	    case SEER_REF_CONTROL:	// Control command from observer
	    case SEER_REF_NEWPROC:	// Fork of new process
	    case SEER_REF_PROCEXIT:	// Actual process exit
	    default:
		{
		cerr << "correlator: unexpected reference type " << type;
		correlatorPanic("");
		break;
		}
	    }
	}
    if (type == SEER_REF_OPENREAD  &&  process != NULL  &&  isDirectory()
      &&  !process->inGetcwd())
	{
	/*
	 * The process has opened a directory for read.  That might
	 * mean that the process is of the sort whose future activity
	 * isn't meaningful.  To help us figure this out, we'll keep
	 * track of how many files the process might have peeked at,
	 * versus how many actually get peeked.
	 */
	File* program = process->program();
	if (program != NULL  &&  !process->hasDirectoryOpen())
	    {
	    discoverDirectorySize();
	    if (program->potentialPeeks == 0)
		program->peekCount = 0;
	    program->potentialPeeks += children.count() + unknownChildren;
	    }
	}
    /*
     * We only arrive here if the reference is an "important" one.  We
     * start by recording manager-independent information.  Then we
     * let the managers grab whatever they would like.
     *
     * Note that we don't count close-type references in the reference
     * counters.
     */
    lastReferenceType = type;
    bool meaningless = process != NULL  &&  process->currentlyMeaningless();
    switch (type)
	{
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	    {
	    break;
	    }
	case SEER_REF_RW:		// Reads and/or writes of existing file
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	case SEER_REF_MAKELINK:		// Create a symbolic link
	    {
	    /*
	     * NEEDSWORK: should status be forgotten for unstored
	     * files when simulating?
	     */
	    forgetStatus();		// Arrange to rediscover status
	    break;
	    }
	case SEER_REF_EXEC:		// Execution
	    {
	    recordOpen(process);
	    break;
	    }
	case SEER_REF_DESTROY:		// Destruction (delete)
	    {
#ifdef FICUS_MISSING_DIR_BUG
	    if (simulationMode  ||  state.state() != FileState::UNSTORED)
		{
		if (markDeleted(false))
		    cerr << "MISSINGDIR:  marked file '" << path
		      << "' deleted because of a DESTROY reference\n";
		}
#else // FICUS_MISSING_DIR_BUG
	    if (simulationMode  ||  state.state() != FileState::UNSTORED)
		markDeleted(false);
#endif // FICUS_MISSING_DIR_BUG
	    break;
	    }
	case SEER_REF_OPENRW:		// Open for read/write
	    {
	    if (willForget())
		forgetStatus();		// Be sure file won't go away
	    flags |= FF_WRITING;	// Remember that file's being written
	    recordOpen(process);
	    break;
	    }
	case SEER_REF_OPENREAD:		// Open for read
	    {
	    recordOpen(process);
	    break;
	    }
	case SEER_REF_OPENCREATE:   	// Create or truncate/rewrite
	    {
	    if (willForget())
		forgetStatus();		// Be sure file won't go away
	    flags |= FF_WRITING;	// Remember that file's being written
	    recordOpen(process);
	    break;
	    }
	case SEER_REF_CLOSE:		// Close of open file
	    {
	    recordClose(process);
	    if (flags & FF_WRITING)
		{
		forgetStatus();		// Force reevaluation of file params
		if (openCount == 0)
		    flags &= ~FF_WRITING;
		}
	    break;
	    }
	case SEER_REF_EXIT:		// Process exit or re-exec
	    {
	    recordClose(process);
	    /*
	     * It may seem that checking for WRITING on an exit is
	     * pointless.  However, we must do this because the
	     * WRITING flag is only cleared on the final close, which
	     * might be an exit.
	     */
	    if (flags & FF_WRITING)
		{
		forgetStatus();		// Force reevaluation of file params
		if (openCount == 0)
		    flags &= ~FF_WRITING;
		}
	    if (process != NULL)
		process->exec(NULL);
	    break;
	    }
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	case SEER_REF_PEEKLINK:		// Peek at a symbolic link's contents
	case SEER_REF_FOLLOW:		// Follow a symbolic link
	    {
	    break;
	    }
	case SEER_REF_RENAME:		// Change of file name
	case SEER_REF_DUP:		// Duplicate existing open
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	default:
	    {
	    cerr << "correlator: unexpected reference type " << type;
	    correlatorPanic("");
	    break;
	    }
	}
    for (FileRelationTypeIterator relationType;  relationType;  ++relationType)
	(*relationType).recordReference(*this, type, process, meaningless);
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).managerData(*this).recordReference(*this, type,
	  meaningless);
    }

void File::recordMissedReference(	// Record a missed reference
    const char*		path,		// Pathname of the file to record
    Process*		process,	// Process doing the reference
    ReferenceType	type)		// Type of reference that missed
    {
    /*
     * Decide how we want to process the reference.  This involves
     * deciding whether we want the link or its target (if the path
     * turns out to refer to a symlink), and deciding whether to
     * discard the reference because the process is known to be
     * meaningless.
     */
    bool wantLink;			// True if we want symlink, not target
    bool meaningless = process != NULL  &&  process->currentlyMeaningless();
    switch (type)
	{
	case SEER_REF_EXEC:		// Execution
	case SEER_REF_INTERPRET:	// Execution of script interpreter
	case SEER_REF_OPENRW:		// Open for read/write
	case SEER_REF_CREATE:		// Create or truncate/rewrite
	case SEER_REF_OPENCREATE:	// Create or truncate/rewrite
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = false;
	    break;
	    }
	case SEER_REF_OPENREAD:		// Open for read
	    {
	    wantLink = false;
	    break;
	    }
	case SEER_REF_READ:		// Read of a file
	case SEER_REF_RW:		// Reads and/or writes of existing file
	    {
	    if (process != NULL)
		{
		process->setInGetcwd(false);
		}
	    wantLink = false;
	    break;
	    }
	case SEER_REF_DESTROY:		// Destruction (delete)
	case SEER_REF_MAKELINK:		// Make a symbolic link
	    {
	    if (process != NULL)
		process->setInGetcwd(false);
	    wantLink = true;
	    break;
	    }
	case SEER_REF_PEEK:		// Peek at a file (stat or access)
	case SEER_REF_PEEKLINK:		// Peek at a symbolic link's contents
	    {
	    /*
	     * In simulation mode, a failed peek might not indicate a
	     * real miss, so we'll ignore it.
	     */
	    if (simulationMode)
		return;
	    wantLink = type == SEER_REF_PEEKLINK;
	    break;
	    }
	case SEER_REF_UNKNOWN:		// Unknown reference type
	case SEER_REF_CONTROL:		// Control command from observer
	case SEER_REF_NEWPROC:		// Fork of new process
	case SEER_REF_PROCEXIT:		// Actual process exit
	case SEER_REF_RENAME:		// Change of file name
	case SEER_REF_CLOSE:		// Close of open file
	case SEER_REF_EXIT:		// Process exit or re-exec
	case SEER_REF_FOLLOW:		// Follow a symbolic link
	case SEER_REF_DUP:		// Duplicate an open descriptor
	default:
	    {
	    cerr << "correlator: unexpected reference type " << type;
	    correlatorPanic(" in File::recordMissedReference");
	    break;
	    }
	}
    File* file = findCanonical(path, wantLink, false, process);
    if (file == NULL)
	return;				// Truly nonexistent file
    file->recordMissedReference(process, type, meaningless);
					// Let underlying routine do the work
    }

void File::recordMissedReference(	// Record a missed reference
    Process*		process,	// Process doing the reference
    ReferenceType	referenceType,	// Type of reference that missed
    bool		isMeaningless)	// True if reference is meaningless
    {
    if (referenceType == SEER_REF_DESTROY)
	{
	/*
	 * NEEDSWORK: perhaps we shouldn't remove files from the table
	 * quite so easily on a failed destroy.  Perhaps, like with a
	 * successful destroy, we shouldn't remove UNSTORED files.
	 * But then we'd need another way to clean out truly
	 * nonexistent files.
	 */
#ifdef FICUS_MISSING_DIR_BUG
	if (markDeleted(false))
	    cerr << "MISSINGDIR:  marked file '" << path
	      << "' deleted because of a DESTROY reference\n";
#else // FICUS_MISSING_DIR_BUG
	markDeleted(false);
#endif // FICUS_MISSING_DIR_BUG
	}
    if (flags & FF_IGNORE)
	return;				// Pretend this file doesn't exist
    for (FileRelationTypeIterator relationType;  relationType;  ++relationType)
	(*relationType).recordMissedReference(*this, referenceType, process,
	 isMeaningless);
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).managerData(*this)
	  .recordMissedReference(*this, referenceType, isMeaningless);
    }

const char* File::encodedName(		// Return encoded name of the file
    ShowOptions		options)	// Display options
    const
    {
    if (options & SO_MIME_ENCODE)
	{
	static const char* encoding = NULL;
	if (encoding != NULL)
	    free((void*)encoding);
	encoding = mimeEncode(name(), NULL);
	return encoding;
	}
    else
	return name();
    }

/*
 * Select a particular size distribution for use during simulation.
 * The distribution is used to select sizes for directories that are
 * otherwise unknown.  he named distribution must be chosen
 * from the following list, and the correct number of parameters must
 * be provided.  The parameter array will be copied into internal
 * variables for use during the simulation run.
 *
 * Supported distributions:
 *
 *	geometric	Geometric with parameter "p" equal to parameters[0]
 *
 * No other distributions are currently supported.
 */
void File::setDirectorySizeDistribution(
    const char*		name,		// Name of distribution to use
    int			nParameters,	// Number of parameters provided
    const double*	parameters)	// List of distribution parameters
    {
    if (strcmp(name, "geometric") == 0)
	{
	if (nParameters != 1)
	    correlatorPanic("correlator: bad number of parameters in"
	      " setDirectorySizeDistribution");
	if (parameters[0] < 0.0  ||  parameters[0] >= 1.0)
	    correlatorPanic("correlator: bad geometric parameter in"
	      " setDirectorySizeDistribution");
	logDirectoryGeometricP = log(1.0 - parameters[0]);
	}
    else
	correlatorPanic("correlator: Unsupported distribution selected"
	  " in setDirectorySizeDistribution");
    }

/*
 * Select a particular size distribution for use during simulation.
 * The distribution is used to select sizes for files that are
 * otherwise unknown.  he named distribution must be chosen
 * from the following list, and the correct number of parameters must
 * be provided.  The parameter array will be copied into internal
 * variables for use during the simulation run.
 *
 * Supported distributions:
 *
 *	geometric	Geometric with parameter "p" equal to parameters[0]
 *
 * No other distributions are currently supported.
 */
void File::setFileSizeDistribution(
    const char*		name,		// Name of distribution to use
    int			nParameters,	// Number of parameters provided
    const double*	parameters)	// List of distribution parameters
    {
    if (strcmp(name, "geometric") == 0)
	{
	if (nParameters != 1)
	    correlatorPanic("correlator: bad number of parameters in"
	      " setFileSizeDistribution");
	if (parameters[0] < 0.0  ||  parameters[0] >= 1.0)
	    correlatorPanic("correlator: bad geometric parameter in"
	      " setFileSizeDistribution");
	logFileGeometricP = log(1.0 - parameters[0]);
	}
    else
	correlatorPanic("correlator: Unsupported distribution selected"
	  " in setFileSizeDistribution");
    }

/*
 * Figure out the status of a file.  If it hasn't previously been
 * done, this routine sets the following fields:
 *
 *	state
 *	type
 *	blocks
 *
 * Depending on the replay mode, this information might be extracted
 * from the replay file rather than actually getting the status of the
 * file from the operating system.  Also, if there is a save file
 * active, the information will be written to an ST_STAT record in the
 * save file.
 */
void File::discoverStatus()
    {
    static char		saveType = ST_STAT; // Type for savefile
    struct stat		statBuffer;	// For getting file status

    if (isKnown())
	return;
    /*
     * In certain situations, discoverStatus may reset a file's state
     * to KNOWN from one of UNKNOWN, MISSING, FORGETTING, or
     * FORGOTTEN.  If so, we need to be sure that none of the file's
     * parent directories is in one of those three states, because if
     * it is, it will get removed from the file table, which will
     * eventually cause a sanity-check abort.  So we need to make sure
     * that all of our parents are removed from these states.  It
     * turns out that it's sufficient to check our immediate parent,
     * because if it needs to be changed, we'll recurse, and if it
     * doesn't, we can be sure that its parents are OK (otherwise we'd
     * have executed this little recursion on them sometime in the
     * past).
     */
    if (parentDirectory != NULL  &&  parentDirectory != this
      &&  !parentDirectory->isKnown())
	parentDirectory->discoverStatus();
    if (replayFile != NULL  &&  replaymode  &&  !replayedRestart)
	{
	if (!replayFile->read(&saveType, sizeof saveType))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for savetype at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	if (saveType != ST_STAT)
	    {
	    cerr << "correlator:  unexpected savetype " << (int) saveType
	      << " at offset " << replayFile->bytesRead() - sizeof saveType;
	    correlatorPanic(" of replay file in File::discoverStatus");
	    }
#ifdef NAME_IN_STAT_RECORD
	int nameLength;
	if (!replayFile->read(&nameLength, sizeof nameLength))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd()
	      << " for stat name length at offset " << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	char savedName[MAXPATHLEN];
	if ((unsigned)nameLength >= sizeof(savedName))
	    {
	    cerr << "correlator:  bad name length " << nameLength
	      << " reading fd " << replayFile->rdbuf()->fd();
	    correlatorPanic(" for stat name");
	    }
	if (!replayFile->read(savedName, nameLength + 1))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for stat name at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	if (strcmp(path, savedName) != 0)
	    {
	    cerr << "correlator:  looking for stat record for '" << path
	      << "', got one for '" << savedName << "' at offset "
	      << replayFile->bytesRead() - nameLength - 1;
	    correlatorPanic(" of replay file in File::discoverStatus");
	    }
#endif // NAME_IN_STAT_RECORD
	char charState;
	if (!replayFile->read(&charState, sizeof charState))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for state at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	state.setState(*this, (FileState::State)charState);
	char charType;
	if (!replayFile->read(&charType, sizeof charType))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for type at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	type = (FileType)charType;
#ifdef LENGTH_IN_STAT_RECORD
	unsigned long longDummy;
	if (!replayFile->read(&longDummy, sizeof longDummy))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for file size at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	statBuffer.st_size = longDummy;
#endif // LENGTH_IN_STAT_RECORD
	if (!replayFile->read(&blocks, sizeof blocks))
	    {
	    cerr << "correlator:  error reading fd "
	      << replayFile->rdbuf()->fd() << " for blocks at offset "
	      << replayFile->bytesRead();
	    correlatorPanic("");
	    }
	}
    else
	{
	if (stripLength != 0)
	    {
	    /*
	     * Files with stripping are usually caused by the
	     * automounter, and we have to do some special gymnastics
	     * to get the right information.  For example, assume
	     * "/home/foo" is an automounted directory.  This will
	     * show up as a symbolic link to "/tmp_mnt/home/foo".
	     * When we stat the entry for the latter file, we want to
	     * stat the real thing, but there's no guarantee that it
	     * is actually mounted right now.  So we need to trick the
	     * automounter into mounting the path before we mount it.
	     * We do this by statting a dummy entry, "/home/foo/.",
	     * which will force the automount to happen at the right
	     * time.
	     */
	    char dummyBuffer[MAXPATHLEN];
	    (void)strncpy(dummyBuffer, name(), MAXPATHLEN);
	    dummyBuffer[MAXPATHLEN - 1] = '\0';
	    (void)strncat(dummyBuffer, "/.",
	      MAXPATHLEN - strlen(dummyBuffer) - 1);
	    if (!simulationMode)
		(void)lstat(dummyBuffer, &statBuffer);
	    }
	if ((simulationMode  &&  logFileGeometricP != 0.0)
	  ||  ficus_extended_stat(path, 1, &statBuffer) < 0)
	    {
	    /*
	     * Assume all stat errors are due to missing files.  This
	     * isn't always true; in particular, we can get EPERM if we
	     * don't have access to a path component.  But it hardly
	     * matters from our point of view, since in either case the
	     * file might as well not be there.
	     *
	     * In this case, we leave the block size unchanged.  If
	     * the file has disappeared for a reason such as
	     * replication changes, this will allow us to make future
	     * hoarding decisions correctly.  It also allows the
	     * fake-size feature to be used for nonexistent files.
	     */
	    if (state.state() == FileState::UNKNOWN)
		state.setState(*this,
		  simulationMode ? FileState::KNOWN : FileState::MISSING);
	    type = FT_FILE;
	    if (simulationMode  &&  logFileGeometricP != 0.0)
		{
		statBuffer.st_size =
		  (long) (log(drand48()) / logFileGeometricP);
		statBuffer.st_blocks =
		  (statBuffer.st_size + STAT_BYTES_PER_BLOCK - 1)
		    / STAT_BYTES_PER_BLOCK;
		flags |= FF_SIZE_GUESSED;
		}
	    else
		{
		statBuffer.st_blocks = 0;
		statBuffer.st_size = 0;
		}
	    }
	else
	    {
#ifdef FICUS_MISSING_DIR_BUG
	    if (type != FT_DIR  &&  parentDirectory != NULL
	      &&  !parentDirectory->isKnown())
		cerr << "MISSINGDIR:  Setting state to KNOWN for '" << path
		  << "' even though parent '" << parentDirectory->path
		  << "' has state " << parentDirectory->state << '\n';
#endif // FICUS_MISSING_DIR_BUG
	    state.setState(*this, FileState::KNOWN);
	    switch (statBuffer.st_mode & S_IFMT)
		{
		case S_IFIFO:
		    {
		    type = FT_FIFO;
		    break;
		    }
		case S_IFCHR:
		    {
		    type = FT_CHARACTER_SPECIAL;
		    break;
		    }
		case S_IFDIR:
		    {
		    type = FT_DIR;
		    break;
		    }
		case S_IFBLK:
		    {
		    type = FT_BLOCK_SPECIAL;
		    break;
		    }
		case S_IFLNK:
		    {
		    type = FT_SYMLINK;
		    break;
		    }
		case S_IFSOCK:
		    {
		    type = FT_SOCK;
		    break;
		    }
		case S_IFREG:
		default:
		    {
		    type = FT_FILE;
		    break;
		    }
		}
#ifdef FICUS_STAT_BUG
	    /*
	     * Ficus has a bug in which stat reports a zero block count
	     * for files that haven't propagated yet.  We'll kludge around
	     * this by converting the file length into a 512-byte block
	     * count.  This is imperfect, since it ignores holes and
	     * indirect blocks, but it's the best we can do without
	     * knowing details about the underlying file system (such
	     * as block size).
	     */
	    if (statBuffer.st_blocks == 0  &&  (flags & FF_FAKE_SIZE) == 0
	      &&  (flags & FF_SIZE_GUESSED) == 0)
		statBuffer.st_blocks =
		  (statBuffer.st_size + STAT_BYTES_PER_BLOCK - 1)
		    / STAT_BYTES_PER_BLOCK;
#endif // FICUS_STAT_BUG
	    }
	if ((flags & FF_FAKE_SIZE) != 0)
	    flags &= ~FF_SIZE_GUESSED;
	else
	    blocks = statBuffer.st_blocks;
	}
#ifdef FICUS_MISSING_DIR_BUG
    /*
     * Under some conditions, Ficus directories appear to go away for
     * brief periods, even though they really exist.  So whenever we
     * successfully stat a file, we go through the parents and
     * hand-correct any missing directory back to unknown, which will
     * force it to be re-stat'ed.
     */
    File* currentParent = parent();
    while (currentParent != NULL)
	{
	if (currentParent->state.state() == FileState::MISSING)
	    {
	    if (verbose > 1)
		cerr << "correlator:  resetting state of '"
		  << currentParent->path << "' from "
		  << currentParent->state.state() << " to UNKNOWN\n";
	    currentParent->state.setState(*currentParent, FileState::UNKNOWN);
	    }
	if (currentParent->parent() == currentParent)
	    break;
	else
	    currentParent = currentParent->parent();
	}
#endif // FICUS_MISSING_DIR_BUG
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).interpretStatus(*this);
    if (saveFile != NULL)
	{
	char charState = (char)state.state();
	char charType = (char)type;
#ifdef LENGTH_IN_STAT_RECORD
	unsigned long fileSize = statBuffer.st_size;
#endif // LENGTH_IN_STAT_RECORD
	if (!saveFile->write(&saveType, sizeof saveType)
#ifdef NAME_IN_STAT_RECORD
	  ||  !saveFile->write(&pathLength, sizeof pathLength)
	  ||  !saveFile->write(path, pathLength + 1)
#endif // NAME_IN_STAT_RECORD
	  ||  !saveFile->write(&charState, sizeof charState)
	  ||  !saveFile->write(&charType, sizeof charType)
#ifdef LENGTH_IN_STAT_RECORD
	  ||  !saveFile->write(&fileSize, sizeof fileSize)
#endif // LENGTH_IN_STAT_RECORD
	  ||  !saveFile->write(&blocks, sizeof blocks)
	  ||  !saveFile->flush())
	    {
	    int saveErrno = errno;
	    long now = time((long*) NULL);
	    cerr << "correlator:  difficulty writing file state to save file: "
	      << strerr(saveErrno) << " at " << ctime(&now);
	    }
	}
    }

void File::setStorage(			// Set file storage status
    StorageSetting	setting,	// New storage setting desired
    unsigned long	size,		// File size, if setting unstored
    File*		targetFile)	// Target of possible symlink
    {
    /*
     * "Stored" is true if the file is stored, zero otherwise.
     * Change the state appropriately, depending on the current
     * state.
     */
    flags |= FF_STORAGE_SET;
    switch (setting)
	{
	case SS_STORE_AND_REFERENCE:
	case SS_REFERENCE:
	    {
	    /*
	     * Update the last-reference time, but don't increment the
	     * reference counter or do any normal reference recording
	     * (e.g., updating the semantic distance).  This will
	     * ensure that the file remains in the hoard, but won't
	     * disturb the clustering algorithm.
	     */
	    if (setting == SS_STORE_AND_REFERENCE)
		{
		for (HoardingManagerIterator manager;  manager;  ++manager)
		    (*manager).setReferenceTime(*this, 0, 2);
		}
	    break;
	    }
	default:
	    {
	    break;
	    }
	}
    switch (setting)
	{
	case SS_STORE:
	case SS_STORE_AND_REFERENCE:
	    {
	    switch (currentState())
		{
		case FileState::UNKNOWN:
		    {
		    discoverStatus();
		    break;
		    }
		case FileState::KNOWN:
		    {
		    break;		// Already in the correct state
		    }
		case FileState::UNSTORED:
		    {
#if 0
		    /*
		     * I think the following two lines are bogus.  Why
		     * should the reference time be changed just
		     * because a file goes from unstored to stored?
		     * In any case, the code above already did this
		     * thing for the STORE_AND_REFERENCE case.
		     *
		     * GHK 7-28-98.
		     */
		    for (HoardingManagerIterator manager;  manager;  ++manager)
			(*manager).setReferenceTime(*this, 0, 2);
#endif // 0
		    /*
		     * See the MISSING case for further comments.
		     */
		    forgetStatus();
		    discoverStatus();
		    break;
		    }
		case FileState::MISSING:
		case FileState::FORGETTING:
		case FileState::FORGOTTEN:
		    {
		    /*
		     * The easiest (and most reliable) way to reset a
		     * file to the KNOWN state is to forget and
		     * rediscover the status.  This protects us
		     * against possible size changes that may have
		     * happened while it was unstored.  It also
		     * correctly handles files that are still
		     * missing.
		     */
		    forgetStatus();
		    discoverStatus();
		    break;
		    }
		default:
		    {
		    correlatorPanic("Unknown file state in setStorage");
		    break;
		    }
		}
	    break;
	    }
	case SS_UNSTORE:
	    {
	    discoverStatusIfUnknown();
	    /*
	     * If the file was previously unknown, the caller may
	     * have told us the size.  We only believe them if we
	     * don't have plausible information of our own.  Note
	     * that in a few cases we will discard the information
	     * incorrectly: specifically, if the unstoring party
	     * knows of a size change that occurred while the file
	     * was still unstored.  This is only an approximation
	     * because it doesn't account for indirect blocks.
	     */
	    if (currentState() != FileState::KNOWN
	      &&  (blocks == 0  ||  (flags & FF_SIZE_GUESSED) != 0)
	      &&  (flags & FF_FAKE_SIZE) == 0)
		{
		blocks = (size + STAT_BYTES_PER_BLOCK - 1)
		  / STAT_BYTES_PER_BLOCK;
		flags &= ~FF_SIZE_GUESSED;
		}
	    state.setState(*this, FileState::UNSTORED);
	    break;
	    }
	case SS_REFERENCE:
	    {
	    /*
	     * Reference updating was done above.
	     */
	    break;
	    }
	case SS_SYMLINK:
	    {
	    /*
	     * Symlink-setting commands only take effect in simulation
	     * mode; otherwise we'd rather believe what we can learn
	     * from the filesystem.
	     */
	    if (simulationMode)
		{
		if (linkTargetFile == NULL)
		    {
		    type = FT_SYMLINK;
		    linkTargetFile = targetFile;
		    /*
		     * The size may have been set randomly.  Symlinks
		     * are assumed to always take only one block.
		     */
		    blocks = 1;
		    flags |= FF_FAKE_SIZE;
		    flags &= ~FF_SIZE_GUESSED;
		    }
		}
	    break;
	    }
	default:
	    {
	    correlatorPanic("Unknown storage setting in setStorage");
	    break;
	    }
	}
    }

void File::setParent(			// Set file's parent and inherit flags
    File*		parent)		// This file's parent directory
    {
    /*
     * WARNING: if you change this routine, also change renameDescendants
     */
    if (parent != NULL  &&  parent->willForget())
	parent = NULL;
    if (parent == parentDirectory)
	return;
    else
	{
	if (parentDirectory != NULL)
	    parentDirectory->children.remove(this);
	parentDirectory = parent;
	if (parent != NULL)
	    {
	    parent->children.pushHead(this);
	    if (parent->unknownChildren > 0)
		parent->unknownChildren--;
	    if (parent->type != FT_DIR)
		{
		parent->type = FT_DIR;	// Ensure parent is marked as directory
		if (simulationMode  &&  logDirectoryGeometricP != 0.0
		  &&  parent->sizeWasGuessed())
		    {
		    long byteSize =
		      (long) (log(drand48()) / logDirectoryGeometricP);
		    parent->blocks =
		      (byteSize + STAT_BYTES_PER_BLOCK - 1)
			/ STAT_BYTES_PER_BLOCK;
		    }
		}
	    }
	inheritFlagsFromParent();
	inheritManagerDataFromParent(NULL);
	}
    }

/*
 * Inherit flags and other general information from the parent
 * directory.
 */
void File::inheritFlagsFromParent()
    {
    if (parentDirectory == NULL)
	return;
    /*
     * Some commentary is in order regarding inheritance.  Some flags
     * are only inherited if set; others are inherited regardless of
     * setting.
     *
     * If our flags differ from our parent's after inheritance, we set
     * the must-save bit.
     *
     * This all works very nicely for the INHERIT_IF_SET flags.
     * However, for INHERIT_VERBATIM, we want to inherit only if the
     * flags weren't already explicitly set.  For the moment, this is
     * enforced by ordering restrictions in the instructions file:
     * parent directories must be mentioned before any children.
     */
    flags &= ~FF_INHERIT_VERBATIM;
    flags |=
      parentDirectory->flags & (FF_INHERIT_IF_SET | FF_INHERIT_VERBATIM);
    if (((parentDirectory->flags ^ flags) & FF_INHERIT_LIST) != 0)
	mustSave();
    if (stripLength == 0)
	stripLength = parentDirectory->stripLength;
    }

/*
 * Inherit manager-specific information from the parent directory.  If
 * targetManager is set, inherit only for that manager; otherwise
 * inherit for all managers.
 *
 * If manager data is set for more than one file in the same tree, the
 * parent must be set before the child, so that the child can override
 * the data set by the parent.
 */
void File::inheritManagerDataFromParent(
    const HoardingManagerBase*	targetManager)
    {
    if (parentDirectory == NULL)
	return;
    /*
     * Some commentary is in order regarding inheritance.  Some flags
     * are only inherited if set; others are inherited regardless of
     * setting.
     *
     * If our flags differ from our parent's after inheritance, we set
     * the must-save bit.
     *
     * This all works very nicely for the INHERIT_IF_SET flags.
     * However, for INHERIT_VERBATIM, we want to inherit only if the
     * flags weren't already explicitly set.  For the moment, this is
     * enforced by ordering restrictions in the instructions file:
     * parent directories must be mentioned before any children.
     */
    for (HoardingManagerIterator manager;  manager;  ++manager)
	{
	HoardingManagerBase& thisManager(*manager);
	if (targetManager == NULL  ||  targetManager == &thisManager)
	    thisManager.managerData(*this).inherit(*this, *parentDirectory);
	}
    }

/*
 * Count the number of children in a directory that are not explicitly
 * listed in the file table.
 */
void File::discoverDirectorySize()
    {
    if (flags & FF_SIZE_DISCOVERED)
	return;
    flags |= FF_SIZE_DISCOVERED;
    if (simulationMode)
	return;
    char pathName[MAXPATHLEN];
    if (pathLength >= sizeof pathName - 1)
	return;
    (void)strcpy(pathName, name());
    DIR* directory = opendir(pathName);
    if (directory == NULL)
	{
	int saveErrno = errno;
	cerr << "correlator: couldn't open directory '" << pathName << "':  "
	  <<  strerr(saveErrno) << '\n';
	return;
	}
    struct dirent* entry;
    pathName[pathLength] = '/';
    pathName[sizeof pathName - 1] = '\0';
    /*
     * Read the directory, and for each child see if it's in the file
     * table.  If not, increment the unknown-children count.  We can
     * use underlyingFind because we know that the directory's
     * pathname is already canonicalized, and we're not following
     * symbolic links.
     */
    while ((entry = readdir(directory)) != NULL)
	{
	if (!isDot(entry->d_name)  &&  !isDotDot(entry->d_name))
	    {
	    (void)strncpy(&pathName[pathLength + 1], entry->d_name,
	      sizeof pathName - pathLength - 1 - 1);
	    bool isNew;
	    File* child = underlyingFind(pathName, 0, false, isNew);
	    if (child == NULL)
		unknownChildren++;
	    }
	}
    (void)closedir(directory);
    }

bool File::deleteIfObsolete()		// Arrange deletion if obsolete file
    {
    switch (state.state())
	{
	case FileState::UNKNOWN:
	    {
	    break;
	    }
	case FileState::KNOWN:
	case FileState::UNSTORED:
	    {
	    /*
	     * If a file has the IGNORE flag set, but not the
	     * MUST_SAVE bit, we don't need to keep it in the table.
	     * Furthermore, such files are often very long-lived, so
	     * they will take up a lot of unnecessary file-table
	     * space.  We will change the state of such a file to
	     * MISSING, which is a lie, but won't interfere with the
	     * deletion manager's consistency checks.
	     */
	    if ((flags & (FF_MUST_SAVE | FF_IGNORE)) == FF_IGNORE)
		{
#ifdef FICUS_MISSING_DIR_BUG
		if (type == FT_DIR  &&  state.state() == FileState::KNOWN)
		    {
		    cerr << "MISSINGDIR:  Changing file state for directory '"
		      << path
		      << "' from KNOWN TO MISSING in deleteIfObsolete\n";
		    }
#endif // FICUS_MISSING_DIR_BUG
		if (type == FT_DIR  &&  path[0] == '/'  &&  path[1] == '\0')
		    correlatorPanic(
		      "correlator: setting root directory to MISSING");
		state.setState(*this, FileState::MISSING);
		}
	    break;
	    }
	case FileState::MISSING:
	    {
	    if (type == FT_DIR  &&  path[0] == '/'  &&  path[1] == '\0')
		correlatorPanic(
		  "correlator: setting root directory to FORGETTING");
	    state.setState(*this, FileState::FORGETTING);
	    break;
	    }
	case FileState::FORGETTING:
	    {
	    if (type == FT_DIR  &&  path[0] == '/'  &&  path[1] == '\0')
		correlatorPanic(
		  "correlator: setting root directory to FORGOTTEN");
	    state.setState(*this, FileState::FORGOTTEN);
	    break;
	    }
	case FileState::FORGOTTEN:
	    {
#ifdef PARANOID_ABOUT_FREEING
	    /*
	     * Be sure previous-file table doesn't reference this one.
	     */
	    for (refno = MAX_NPREV;  --refno >= 0;  )
		{
		if (previous[refno] == fileent)
		    correlatorPanic(
		      "freed pointer found in previous-file table\n");
		}
	    for (HashTable::Iterator fileiter(filetable);
	      fileiter;
	      ++fileiter)
		checkfree((file_t*) filetable[fileiter], fileent);
#endif // PARANOID_ABOUT_FREEING
	    /*
	     * If this file's MUST_SAVE bit is set, then we cannot
	     * delete it.  We will reset its state to UNSTORED, which
	     * will cause us to leave it alone in the future.
	     */
	    if (flags & FF_MUST_SAVE)
		{
		state.setState(*this, FileState::UNSTORED);
		return false;
		}
	    delete this;
	    return true;		// Remove this file from table
	    }
	default:
	    {
	    cerr << "correlator:  unexpected file state " << state.state();
	    correlatorPanic(" in File::deleteIfObsolete");
	    break;
	    }
	}
    return false;
    }

/*
 * Count up the amount of hoard storage that would be occupied by
 * storing a file.  This includes the storage needed by the file
 * itself, as well as that needed by parent directories and link
 * targets.
 */
unsigned long File::countStorage(	// Count storage needed by file
    const HoardingManagerBase& 
			manager)	// Manager we're counting for
    const
    {
    if (!canHoard())
	return 0;
    if (manager.managerData(*this).isStored())
	return 0;
    unsigned long blocksNeeded = hoardableSize();
    /*
     * Add in the sizes need for parent directories and link targets.
     */
    if (parentDirectory != NULL  &&  parentDirectory != this)
	blocksNeeded += parentDirectory->countStorage(manager);
    if (linkTargetFile != NULL  &&  linkTargetFile != this)
	blocksNeeded += linkTargetFile->countStorage(manager);
    return blocksNeeded;
    }

/*
 * Tell all interested managers that the file relations have changed,
 * in case they might want to take an interest in this fact.
 */
void File::relationsChanged()
    {
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).managerData(*this).relationsChanged();
    }

/*
 * Tell all interested managers that the file state has changed,
 * in case they might want to take an interest in this fact.
 */
void File::stateChanged(
    FileState*		stateAddress,	// Address of state field
    FileState::State	oldState,	// Original state of the file
    FileState::State	newState)	// New state of the file
    {
    if (stateAddress != &state)
	correlatorPanic("File::stateChanged called for somebody else's state");
    for (ManagerIterator manager;  manager;  ++manager)
	{
	ManagerBase& thisManager(*manager);
	if (thisManager.hasData())
	    thisManager.stateChanged(&(thisManager.managerData(*this)),
	      oldState, newState);
	else
	    thisManager.stateChanged(NULL, oldState, newState);
	}
    }

/*
 * Find the directory-path distance between two files.  This distance
 * is defined as follows: two files in the same directory have a
 * distance of zero.  Otherwise, the distance is the number of
 * directories that have to be traveled from each file to reach a
 * common parent.  Thus, files in a parent-child directory
 * relationship have a distance of 1, siblings are 2 apart, and so
 * forth.
 */
unsigned long File::directoryDistance(
    const File&		that)		// File to calculate distance from
      const
    {
    File* thisParent = parent();
    File* thatParent = that.parent();
    /*
     * Members of the same directory are common enough to be worth
     * special-casing.  Note that as a side effect, this catches cases
     * where both parents are NULL.
     */
    if (thisParent == thatParent)
	return 0;
    /*
     * If one parent is NULL, arbitrarily set the distance to be the
     * depth of the other file.  We set the depth of the root to 1,
     * rather than zero, primarily as a heuristic to keep orphaned
     * files from clustering with files in the root.
     */
    if (thisParent == NULL  ||  thatParent == NULL)
	{
	File* chosenParent = thisParent;
	if (chosenParent == NULL)
	    chosenParent = thatParent;
	for (unsigned long depth = 1;  chosenParent != NULL;  depth++)
	    {
	    File* nextParent = chosenParent->parent();
	    if (nextParent == chosenParent)
		break;
	    chosenParent = nextParent;
	    }
	}
    /*
     * Both parents are non-NULL.  We unmark all of the parents of
     * "this", then mark the parents of "that".  Then we return to
     * "this" and count unmarked parents; this determines the common
     * parent and the depth of "this".  Finally, we return to "that"
     * and count the depth to the now-known parent.
     */
    File* currentParent;
    for (currentParent = thisParent;  currentParent != NULL;  )
	{
	currentParent->unmark();
	File* nextParent = currentParent->parent();
	if (nextParent == currentParent)
	    break;
	currentParent = nextParent;
	}
    for (currentParent = thatParent;  currentParent != NULL;  )
	{
	currentParent->mark();
	File* nextParent = currentParent->parent();
	if (nextParent == currentParent)
	    break;
	currentParent = nextParent;
	}
    unsigned long distance = 0;
    for (currentParent = thisParent;
      currentParent != NULL  &&  !currentParent->isMarked();
      distance++)
	{
	File* nextParent = currentParent->parent();
	if (nextParent == currentParent)
	    break;
	currentParent = nextParent;
	}
    File* commonParent = currentParent;
    for (currentParent = thatParent;
      currentParent != NULL  &&  currentParent != commonParent;
      distance++)
	{
	File* nextParent = currentParent->parent();
	if (nextParent == currentParent)
	    break;
	currentParent = nextParent;
	}
    return distance;
    }

void File::showHashStatistics(		// Display hash table statistics
    ostream&		stream)		// Stream to display on
    {
    fileTable.showStatistics(stream);
    }

void File::cleanRelationData()		// Clean relations to forgetting files
    {
    /*
     * Go through the list of relations, looking for files that are
     * being forgotten, and delete all relations to such files,
     * regardless of relation type.
     */
    for (RelatedFileIterator relatedFile(*this);  relatedFile;  ++relatedFile)
	{
	RelatedFile& testedFile(*relatedFile);
	if (testedFile.targetFile().isForgetting())
	    {
	    delete &testedFile;
	    relatedFile.removeCurrent();
	    }
	}
    }

void File::clearRelationData()		// Clear out relations to all files
    {
    /*
     * Go rid of all relations to all files.
     */
    for (RelatedFileIterator relatedFile(*this);  relatedFile;  ++relatedFile)
	{
	delete &(*relatedFile);
	relatedFile.removeCurrent();
	}
    }

/*
 * Ask all managers to clean out references to files that are being
 * forgotten.
 */
void File::cleanManagerData()
    {
    for (HoardingManagerIterator manager;  manager;  ++manager)
	(*manager).managerData(*this).cleanReferences();
    }

/*
 * Return a relation (to the target file) of a chosen type.  If no
 * such relation is found, returns NULL if create is false, or creates
 * one if create is true.
 */
FileRelationBase* File::chosenRelation(
    File&		targetFile,	// File that's target of the relation
    const FileRelationTypeBase&
			relationType,	// Type desired
    bool		create)		// True to create if not found
    {
    for (RelatedFileIterator relatedFile(*this);  relatedFile;  ++relatedFile)
	{
	RelatedFile& testedFile(*relatedFile);
	if (&testedFile.targetFile() == &targetFile)
	    return testedFile.chosenRelation(relationType, create);
	}
    /*
     * There is currently no relation to the target file.  If we are
     * creating, make a new target related file, add it, and then put
     * the new relation on it.
     */
    if (!create)
	return NULL;
    RelatedFile* newRelatedFile = new RelatedFile(targetFile);
    relatedFiles.pushHead(newRelatedFile);
    return newRelatedFile->chosenRelation(relationType, true);
    }

/*
 * Get rid of every entry in the file table, even if it would normally
 * be kept.  This is used to help detect memory leaks.  Note that
 * since this is an exit-time operation, there is no attempt to find
 * and destroy all dangling pointers to file-table entries, only to
 * return all memory to the allocation system.
 */
void File::destroyTable()
    {
    for (FileTableIterator iter;  iter;  ++iter)
	{
	File* fileToZap(&(*iter));
	while (!fileToZap->children.isEmpty())
	    {
	    File* child((File*)(fileToZap->children.popHead()));
	    child->setParent(NULL);
	    }
	delete fileToZap;
	iter.removeCurrent();
	}
    }

/*
 * Step to the next file that has a relation.  If any files with empty
 * relations are found, silently remove them.
 */
RelatedFileIterator& RelatedFileIterator::operator++()
    {
    if (!source.relatedFiles.isEmpty()  &&  *this)
	{
	RelatedFile& relatedFile(**this);
	if (relatedFile.isEmpty())
	    {
	    delete &relatedFile;
	    removeCurrent();
	    }
	}
    SList::Iterator::operator++();
    while (*this)
	{
	RelatedFile& relatedFile(**this);
	if (!relatedFile.isEmpty())
	    break;
	delete &relatedFile;
	removeCurrent();
	SList::Iterator::operator++();
	}
    return *this;
    }

SelectedRelatedFileIteratorBase::~SelectedRelatedFileIteratorBase()
    {
    if (subIterator != NULL)
	delete subIterator;
    }

/*
 * Make sure the iterator is pointing at a relation of the selected
 * type.  If not, step the subsidiary iterators until we reach such a
 * relation or complete.
 *
 * This code implicitly assumes that a give relation type can appear
 * only once for a give target related file.
 */
void SelectedRelatedFileIteratorBase::findSelectedType()
    {
    if (subIterator == NULL)
	{
	if (RelatedFileIterator::operator bool())
	    subIterator =
	      new RelationIterator(RelatedFileIterator::operator*());
	else
	    return;
	}
    for (  ;
      RelatedFileIterator::operator bool();
      RelatedFileIterator::operator++())
	{
	for (subIterator->retarget(RelatedFileIterator::operator*());
	  *subIterator;
	  ++(*subIterator))
	    {
	    if (&(**subIterator).type() == &selectedType)
		return;
	    }
	}
    }
