#ifndef _FILE_HH
#define _FILE_HH
/*
 * $Id: file.hh,v 1.40 1998/10/23 23:15:13 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: file.hh,v $
 * Revision 1.40  1998/10/23 23:15:13  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.39  1998/07/28 23:47:44  geoff
 * Add support for a separate distribution for directory sizes.
 *
 * Revision 1.38  1998/07/15 21:43:39  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.37  1998/07/09 21:02:01  geoff
 * Add the meaningless argument to recordMissedReference.
 *
 * Revision 1.36  1998/06/02 23:59:44  geoff
 * Add support for symlink initialization.
 *
 * Revision 1.35  1998/05/14 01:15:30  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.34  1998/05/02 00:29:08  geoff
 * Use unsigned ints for path lengths.  Use unsigned shorts rather than
 * unsigned chars for the strip length.
 *
 * Revision 1.33  1998/04/09 00:33:00  geoff
 * Use the new notation for hash and list iterators.
 *
 * Revision 1.32  1998/02/19 02:10:05  geoff
 * Break out the filestate class into a separate file.  Add declarations needed to support incremental clustering.
 *
 * Revision 1.31  1997/12/18 01:25:05  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.30  1997/05/16 21:26:21  geoff
 * Remove the debugging code used to find the LRU/clustering inconsistencies.
 *
 * Revision 1.29  1997/05/16 17:10:36  geoff
 * Add includeInSimulation.  Add temporary debugging code to catch
 * discrepancies between the LRU and clustering managers.  This is the
 * version used to run the simulations in the dissertation.
 *
 * Revision 1.28  1997/02/12 02:52:06  geoff
 * Change isOften to accept a double.  Add support for tracking newly
 * created files.
 *
 * Revision 1.27  1997/01/02 06:48:20  geoff
 * Add couldHoard, support for size fakery, and support for the
 * previously-referenced bit, which is used in detecting attention
 * shifts.
 *
 * Revision 1.26  1996/12/03 02:58:03  geoff
 * Change the name of the routine that records missed references.
 *
 * Revision 1.25  1996/10/04 07:12:11  geoff
 * Add support for the new linear LRU manager.  Add support for
 * reference-time setting.
 *
 * Revision 1.24  1996/09/11 22:59:49  geoff
 * Convert from dlists to slists.  Add REPORT_SELECTED_FAILURES support.
 * Remove markUnstored and add setStorage.  Add FF_STORAGE_SET.  Make
 * FF_WRITING stay with the file name (this is probably a bug).
 *
 * Revision 1.23  1996/06/11  06:13:14  geoff
 * Add a number of new functions to support the latest variations in how
 * we track things, plus the latest cleanups.  Expand the flags field,
 * and add some new ones to support those variations.  Get rid of the
 * investigatorData list (it was never used) and the iterator associated
 * with it.  Add some new flags (FF_SIZE_DISCOVERED and
 * FF_STAYS_WITH_NAME).  Add STAT_BYTES_PER_BLOCK.
 *
 * Revision 1.22  1996/05/03  01:12:06  geoff
 * Update the copyright.  Add various forms of support for tracking the
 * children of a directory.  Add alwaysHoardableType, hoardableSize,
 * tildeExpand, and inheritFromParent.
 *
 * Revision 1.21  1996/04/10  06:09:33  geoff
 * The great renaming (change the clunky sorted agglomeration manager to
 * become the catchy clustering manager).
 *
 * Revision 1.20  1996/04/10  01:28:33  geoff
 * Add a << operator for writing file states.  Add the UNSTORED file
 * state, and get rid of the obsolete "where" field.  Add isKnown and the
 * CONTROLLED flag.
 *
 * Revision 1.19  1996/03/19  18:42:56  geoff
 * Add isControlled() and isLocallyStored().  Add the CONTROLLED and
 * FAKE_SIZE flags.
 *
 * Revision 1.18  1996/02/02  18:28:53  geoff
 * Support manager-data accessors that are part of the data class; this
 * is part of a manager-data cleanup involving deriving the classes.
 * Declare inline functions as such in the class definition.  Add the
 * FF_MEANINGLESS flag.
 *
 * Revision 1.17  1995/12/19  01:02:34  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.16  1995/12/19  00:34:19  geoff
 * Get rid of NUMOF (it's in basictypes.hh) and move NULL to
 * basictypes.hh.  Add FinalFileDestructor.  Add process-tracking
 * support.  Add FF_MUST_SAVE.  Make FileTableIterator be a derived
 * class.
 *
 * Revision 1.15  1995/11/22  00:20:35  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.14  1995/11/21  21:38:27  geoff
 * More changes to support NEW_FILE_RELATIONS, in particular changing
 * investigations to work through that mechanism.
 *
 * Revision 1.13  1995/11/10  02:59:07  geoff
 * Support NEW_FILE_RELATIONS.  Include basictypes.hh.  Don't include the
 * icc files, since zzinlinefunctions now does that.
 *
 * Revision 1.12  1995/10/10  20:38:00  geoff
 * Get rid of the obsolete TRACK_MEMORY_LEAKS stuff.
 *
 * Revision 1.11  1995/09/15  00:05:32  geoff
 * Add the FileState class.  Globally change "cache" to "hoard".
 *
 * Revision 1.10  1995/08/07  20:02:52  geoff
 * Add investigator support and declarations for a couple of new
 * routines.
 *
 * Revision 1.9  1995/06/30  19:04:39  geoff
 * Change declarations of various routines to match their new
 * implementations.  Add id() and countStorage().
 *
 * Revision 1.8  1995/06/17  00:11:15  geoff
 * Add show.  Get rid of a bunch of "class" keywords.  Get rid of the
 * obsolete "find" routine.  Shorten the flags, and rename the bits a
 * little.  Add FF_WRITING.
 *
 * Revision 1.7  1995/05/27  00:15:57  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.6  1995/05/26  23:06:08  geoff
 * Change the return type of parent() to be a pointer, so that NULL can
 * be returned if there is no known parent.
 *
 * Revision 1.5  1995/05/23  00:39:39  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.  Take out FF_STRIP and add
 * stripLength.
 *
 * Revision 1.4  1995/05/05  22:41:36  geoff
 * Add a lot of preliminary PRIVATE_MANAGEMENT_DATA support.
 * Move MAX_K to compileoptions.hh.  Make parent() return a const.
 *
 * Revision 1.3  1995/04/20  21:57:38  geoff
 * Add a number of file types, and declarations of findCanonical and
 * support routines, plus fields for their use.  Add some new flags to
 * support stripping, ignoring, and subtree inheritance of flags.  Add
 * FileClusterIterator.
 *
 * Revision 1.2  1995/03/29  21:04:22  geoff
 * Use the new ReferenceCount type.  Change declarations to expect
 * byte-swapped streams.  Add a couple of new functions needed by the
 * new file table.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

/*
 * Class used to keep track of files known to the correlator
 */

#include "basictypes.hh"
#include "clustering_mgr.hh"
#include "communication.hh"
#include "correlator.hh"
#include "filerelation.hh"
#include "filestate.hh"
#include "hashtable.hh"
#include "list.hh"
#include "lrumanager.hh"
#include "modlrumanager.hh"
#include "process.hh"
#include "showoptions.hh"
#include <sys/param.h>

/*
 * Name of the root directory, just to make it easy to move to DOS
 */
const char*const	rootName = "/";

class ByteSwappedStream;
class FileChildrenIterator;
class FileDescendantsIterator;
class FileRelationBase;
class FileTableIterator;
class FinalFileDestructor;
class Investigator;
class InvestigatorFileDataHead;
class Process;
class RelatedFileIterator;
class SelectedRelatedFileIteratorBase;
class ostream;

/*
 * Dummy class used solely to provide an on-exit destructor for
 * File.
 */
class FinalFileDestructor
    {
	friend		File;
    public:
			~FinalFileDestructor();
    private:
			FinalFileDestructor();
    };

/*
 * Now the File class itself
 */
class File
    {
	friend		FileTableIterator;
	friend		FinalFileDestructor;
	/*
	 * The following functions are friends purely so that they can
	 * provide managers with access to their own private data.
	 * They are prohibited (by convention) from doing anything
	 * else.
	 */
	friend BoundedLRU_ManagerData&
			BoundedLRU_Manager::managerData(File& file)
			  const;
	friend const BoundedLRU_ManagerData&
			BoundedLRU_Manager::managerData(const File& file)
			  const;
	friend BoundedLRU_ManagerData&
			BoundedLRU_ManagerData::managerData(File& file)
			  const;
	friend const BoundedLRU_ManagerData&
			BoundedLRU_ManagerData::managerData(const File& file)
			  const;
	friend LRU_ManagerData&
			LRU_Manager::managerData(File& file)
			  const;
	friend const LRU_ManagerData&
			LRU_Manager::managerData(const File& file)
			  const;
	friend LRU_ManagerData&
			LRU_ManagerData::managerData(File& file)
			  const;
	friend const LRU_ManagerData&
			LRU_ManagerData::managerData(const File& file)
			  const;
	friend ClusteringManagerData&
			ClusteringManager::managerData(File& file)
			  const;
	friend const ClusteringManagerData&
			ClusteringManager::managerData(
			    const File& file)
			  const;
	friend ClusteringManagerData&
			ClusteringManagerData::managerData(File& file)
			  const;
	friend const ClusteringManagerData&
			ClusteringManagerData::managerData(
			    const File& file)
			  const;
	friend LinearLRU_ManagerData&
			LinearLRU_Manager::managerData(File& file)
			  const;
	friend const LinearLRU_ManagerData&
			LinearLRU_Manager::managerData(const File& file)
			  const;
	friend LinearLRU_ManagerData&
			LinearLRU_ManagerData::managerData(File& file)
			  const;
	friend const LinearLRU_ManagerData&
			LinearLRU_ManagerData::managerData(const File& file)
			  const;
	friend WeightedLRU_ManagerData&
			WeightedLRU_Manager::managerData(File& file)
			  const;
	friend const WeightedLRU_ManagerData&
			WeightedLRU_Manager::managerData(const File& file)
			  const;
	friend WeightedLRU_ManagerData&
			WeightedLRU_ManagerData::managerData(File& file)
			  const;
	friend const WeightedLRU_ManagerData&
			WeightedLRU_ManagerData::managerData(const File& file)
			  const;
	/*
	 * Other friends
	 */
	friend		FileChildrenIterator;
					// So iterator can access children
	friend		FileDescendantsIterator;
					// So iterator can access children
	friend		RelatedFileIterator;
					// So iterator can access related files
	friend		SelectedRelatedFileIteratorBase;
	friend ByteSwappedStream&
			RelatedFile::read(ByteSwappedStream& stream,
			  File& sourceFile);
					// So read can access related files
    public:
	/*
	 * File types (that we care about)
	 */
	enum FileType
	    {
	    FT_FILE,			// Plain file or other "thing" we hoard
	    FT_DIR,			// Directory
	    FT_FIFO,			// FIFO (named pipe)
	    FT_CHARACTER_SPECIAL,	// Character device
	    FT_BLOCK_SPECIAL,		// Block device
	    FT_SYMLINK,			// Symbolic link
	    FT_SOCK			// BSD-style socket
	    };
	/*
	 * Possible file states returned from "ficus_extended_stat".
	 */
			~File();
	int		operator==(const File& that) const;
	inline int	operator!=(const File& that) const;
	friend ostream&	operator<<(ostream& stream, const File& file);
					// Write a file in printable fashion
					// ..appends \n
	ostream&	show(ostream& stream, ShowOptions options,
			    const char* managerName) const;
					// Write a file in printable fashion,
					// ..with options.  Appends \n

	static File*	findCanonical(const char* name, bool wantLink,
			  bool create, Process* process);
					// Find a file entry in file table,
					// ..canonicalizing the name
	static bool	readPathname(ByteSwappedStream& stream,
			  char pathname[MAXPATHLEN], unsigned int* pathLength);
					// Read a pathname from a stream,
					// ..verifying the format
	static File*	findFromStream(ByteSwappedStream& stream,
			  bool wantLink);
					// Read a pathname from a stream and
					// ..locate the file-table entry for it
	static void	readInstructions(const char* fileName);
					// Read an instruction file
	inline File*	parent() const;	// Return parent of the file
	inline const File*
			linkTarget() const; // Return target of link
	static inline unsigned long
			count();	// Return number of files known
	void		write(ByteSwappedStream& stream) const;
					// Write a file entry to a stream
	void		writePath(ByteSwappedStream& stream, bool stripped)
			  const;
					// Write file's pathname to a stream
	static void	writePathTerminator(ByteSwappedStream& stream);
					// Write a file-list terminator
	static bool	read(ByteSwappedStream& stream);
					// Read a file entry from a stream
	static void	rename(const char* oldName, const char* newName,
			  Process* process);
					// Change the name of a file
	void		renameDescendants(const char* oldName,
			  const char* newName, File* newParent);
					// Rename file's descendants
	inline bool	couldHoard() const; // Test if file could be hoarded
					// ..if it were controlled
	inline bool	isControlled() const; // Test if file can be controlled
	inline bool	canHoard() const; // Test if file can be hoarded
	inline bool	isDirectory() const; // Test if file is a directory
	inline bool	alwaysHoardableType() const;
					// Test for always-hoarded types
	inline FileType	fileType() const; // Return file's type field
	inline bool	isOpen() const;	// Test if file is currently open
	inline FileState::State
			currentState() const; // Return current state field
					// ..use following tests when possible
	inline bool	isKnown() const; // Test if file status is known
	inline bool	willForget() const; // Test if file will go away
	inline bool	isForgetting() const; // Test for forgetting/forgotten
	inline bool	isForgotten() const; // Test for forgotten
	inline bool	heavyPeeker(int threshold) const;
					// Test for process that peeks lots
	inline bool	includeInSimulation() const;
					// Test whether simulations would
					// ..consider this file
	inline bool	sizeWasGuessed() const;
					// Test whether size came from a
					// ..random distribution
#ifdef FICUS_MISSING_DIR_BUG
	bool		markDeleted(bool clearAll); // Set file to be deleted
#else // FICUS_MISSING_DIR_BUG
	void		markDeleted(bool clearAll); // Set file to be deleted
#endif // FICUS_MISSING_DIR_BUG
	void		loseForgettingPointers(); // Get rid of pointers
					// ..that will disappear
	void		mustSave();	// Mark that file must be saved
	void		addRelation(FileRelationBase* relation, File& target);
					// Add a new relation to file
	void		removeRelation(FileRelationBase* relation);
					// Remove a relation from file
					// *** May invalidate iterators! ***
	static void	recordReference(const char* path,
			  Process* referencingProcess, ReferenceType type);
					// Record a reference to a named file
	void		recordReference(Process* referencingProcess,
			  ReferenceType type);
					// Record a reference to this file
	static void	recordMissedReference(const char* path,
			  Process* referencingProcess, ReferenceType type);
					// Record a missed reference
	void		recordMissedReference(Process* referencingProcess,
			  ReferenceType type, bool isMeaningless);
					// Record a missed reference
	inline void	recordOpen(Process* openingProcess);
					// Record an open of this file
	inline void	recordClose(Process* closingProcess);
					// Record a close of this file
	inline const char*
			name() const;	// Return name of the file
	const char*	encodedName(ShowOptions options)
			  const;	// Return encoded name of the file
	inline unsigned long
			id() const;	// Return an arbitrary file identifier
	inline void	setId();	// Choose a new identifier for file
	static inline void
			resetIdNumbering(); // Reset numbering scheme
					// ..may be no-op depending on
					// ..compilation options
	static void	setDirectorySizeDistribution(const char* name,
			  int nParameters, const double* parameters);
					// Set a particular directory size
					// ..distribution for use during
					// ..simulation
	static void	setFileSizeDistribution(const char* name,
			  int nParameters, const double* parameters);
					// Set a particular file size
					// ..distribution for use during
					// ..simulation
	inline unsigned long
			size() const;	// Return file size in blocks
	inline unsigned long
			hoardableSize() const;
					// Return file size in blocks or zero
	inline unsigned long
			relatedFileCount() const;
					// Give count of related files
	void		discoverStatus(); // Figure out the status of a file
	inline void	forgetStatus();	// Arrange to forget current status
	inline void	discoverStatusIfUnknown();
					// Discover status only if unknown
	void		setStorage(StorageSetting setting, unsigned long size,
			  File* targetFile);
					// Set file storage status
	bool		deleteIfObsolete();
					// Arrange deletion if obsolete file
	unsigned long	countStorage(const HoardingManagerBase& manager) const;
					// See how much storage a file needs
	inline bool	isMeaningless() const;
					// Test for currently meaningless
					// ..executable
	inline void	mark();		// Mark the file (see FF_MARKED)
	inline void	unmark();	// Unmark the file
	inline bool	isMarked() const; // Test for marked file
	inline void	markPreviouslyReferenced();
					// Set the previous-reference bit
	inline void	markNotPreviouslyReferenced();
					// Unset the previous-reference bit
	inline bool	wasPreviouslyReferenced() const;
					// Test the previous-reference bit
	inline void	markNewlyCreated();
					// Set the newly-created bit
	inline void	markNotNewlyCreated();
					// Unset the newly-created bit
	inline bool	wasNewlyCreated() const;
					// Test the newly-created bit
	void		relationsChanged();
					// Inform managers that relations have
					// ..changed
	void		stateChanged(FileState* stateAddress,
			  FileState::State oldState,
			  FileState::State newState);
					// Inform managers that state has
					// ..changed
	unsigned long	directoryDistance(const File& that) const;
					// Find directory-path distance to
					// ..another file
	static inline long
			numberOfFiles(); // Return count of known files
	static void	showHashStatistics(ostream& stream);
					// Display hash table statistics
	void		cleanRelationData();
					// Clean relations to forgetting files
	void		clearRelationData();
					// Get rid of all relations
	void		cleanManagerData();
					// Clean references to forgetting files
	FileRelationBase* chosenRelation(File& targetFile,
			  const FileRelationTypeBase& relationType,
			  bool create = false);
					// Find or create a relationship
					// ..of a chosen type
	static void	destroyTable();	// Destroy the entire file table
	static const char*
			tildeExpand(const char* path);
					// Expand ~user sequences
	/*
	 * C++ won't let me declare a friend until the entire class has
	 * been declared, so I have to declare underlyingFind as public.
	 * I think that this is a bug.
	 */
	static File*	underlyingFind(const char* name,
			  unsigned int nameLength, bool create, bool& isNew);
					// Find a file entry in file table,
					// ..telling if it's a new one
    private:
			File();		// Private construction only!
			File(File& that); // Disable assignment
	File&		operator=(const File& that); // Disable copy
	void		setParent(File* parent);
					 // Set file's parent and inherit flags
	void		inheritFlagsFromParent();
					// Inherit general info from parent
	void		inheritManagerDataFromParent(
			  const HoardingManagerBase* targetManager);
					// Inherit manager-specific information
					// ..from parent, for one or all
					// ..managers
	void		discoverDirectorySize();
					// Discover size of a directory
	static unsigned int hash(File* file);
					// Hash name, for file table hashing
	static int	compare(File* a, File* b);
					// Compare names for hash table
	/*
	 * Static data
	 */
	static HashTable fileTable;	// Table of all files
	static FinalFileDestructor
			finalFileDestructor;
	static File*	newEntry;	// Scratch entry for lookups/inserts
#ifdef ASSIGNABLE_FILE_IDS
	static unsigned long
			nextID;		// Next ID to be assigned
#endif // ASSIGNABLE_FILE_IDS
	static double	logDirectoryGeometricP;
					// Log of geometric distribution
					// ..parameter, for assigning fake
					// ..directory sizes during simulation
	static double	logFileGeometricP; // Log of geometric distribution
					// ..parameter, for assigning fake
					// ..file sizes during simulation
	/*
	 * Private data.  If you add new pointer fields, don't forget
	 * to modify the rename and deletion processing to handle
	 * them.
	 */
	const char*	path;		// Full pathname of the file
	unsigned int	pathLength;	// Length of the path
	unsigned short	flags;		// Various flag bits, see below
	unsigned short	stripLength;	// Amount of path to strip
	FileType	type;		// Type of the file
	FileState	state;		// Current state of the file
	unsigned long	blocks;		// No. of blocks needed to store file
	File*		parentDirectory; // Entry for directory we're in
	SList		children;	// List of children, if directory
	int		unknownChildren; // Number of children whose identities
					// ..are unknown
	File*		linkTargetFile;	// Target, if symbolic link
	unsigned int	openCount;	// Number of descriptors currently open
#ifdef ASSIGNABLE_FILE_IDS
	unsigned long	identification;	// Identification assigned to file
#endif // ASSIGNABLE_FILE_IDS
	ReferenceType	lastReferenceType; // Type of last reference
	ReferenceCount	peekCount;	// Peeks done by this file as a process
	ReferenceCount	potentialPeeks;	// Count of files that this process
					// ..could have peeked at as a
					// ..result of opening a
					// ..directory
	SList		relatedFiles;	// Files that we are related to
	BoundedLRU_ManagerData
			boundedLRU_ManagerData;
					// Management data for bounded LRU mgr
	ClusteringManagerData		// Management data for clustering mgr
			clusteringManagerData;
	LRU_ManagerData	lruManagerData;	// Management data for LRU manager
	LinearLRU_ManagerData
			linearLRU_ManagerData;
					// Management data for linear LRU mgr
	WeightedLRU_ManagerData
			weightedLRU_ManagerData;
					// Management data for weighted LRU mgr
	/*
	 * File table flags.  These should be const ints, but ANSI C++
	 * prevents this, so I'll use an enum to kludge it.
	 */
	enum
	    {
	    FF_NEVERHOARD = 0x01,	// Never hoard this file
	    FF_IGNORE = 0x02,		// Ignore this file completely
	    FF_MEANINGLESS = 0x04,	// References from this executable
					// ..are meaningless
	    FF_CONTROLLED = 0x08,	// Hoard managers control this file
					// ..if clear, hoard managers don't
					// ..consider it in hoarding decisions
					// ..and don't change its storage state
	    FF_INHERIT_IF_SET = FF_NEVERHOARD | FF_IGNORE | FF_MEANINGLESS,
					// Flags inherited from parents if set
	    FF_INHERIT_VERBATIM = FF_CONTROLLED,
					// Flags inherited verbatim
	    FF_INHERIT_LIST = FF_INHERIT_IF_SET | FF_INHERIT_VERBATIM,
					// All inherited flags
	    FF_MUST_SAVE = 0x10,	// This file can't be removed
					// ..from the file table
	    FF_WRITING = 0x20,		// This file is open for write
	    FF_FAKE_SIZE = 0x40,	// Size field was falsified
					// ..(used in demos)
	    FF_SIZE_DISCOVERED = 0x80,	// discoverDirectorySize has run
	    FF_STORAGE_SET = 0x100,	// Storage status was once set
					// ..externally
	    FF_PREVIOUSLY_REFERENCED = 0x200,
					// File was referenced in previous
					// ..period
	    FF_NEWLY_CREATED = 0x400,	// First reference in current period
					// ..was a creation
	    FF_SIZE_GUESSED = 0x800,	// Size was generated randomly
	    /*
	     * FF_MARKED is intended for temporary use by any routine
	     * that needs to divide files into two classes.  When
	     * entered, routines may not assume that FF_MARKED has any
	     * particular state, and may not assume that it will
	     * retain any particular value past their exit.
	     * Furthermore, routines that use this flag may not call
	     * "complex" subroutines that might also use the flag.
	     */
	    FF_MARKED = 0x1000,		// File is marked for temp use
	    /*
	     * Flags marked as STAYS_WITH_NAME are originally set
	     * based (fundamentally) on the name of the file, so in a
	     * rename, the flags from the *new* name are assigned.
	     * Other flags are based on the file itself, and so stay
	     * with the underlying file regardless of renaming.
	     */
	    FF_STAYS_WITH_NAME = FF_NEVERHOARD | FF_IGNORE | FF_MEANINGLESS
	      | FF_CONTROLLED | FF_MUST_SAVE | FF_WRITING | FF_FAKE_SIZE
	      | FF_PREVIOUSLY_REFERENCED | FF_NEWLY_CREATED
	    };
	/*
	 * Number of bytes stored in a block, according to stat system
	 * call.  This should be a const int, but ANSI C++ forbids
	 * this, so I'll use an enum to kludge it.
	 */
	enum
	    {
	    STAT_BYTES_PER_BLOCK = 512
	    };
    };

/*
 * Simple iterator class for walking the file table.  This is a
 * trivial derivation from the HashTable::Iterator class.
 */
class FileTableIterator : public HashTable::Iterator
    {
    public:
	inline		FileTableIterator();
	inline File&	operator*() const; // Return element of table
    };

/*
 * Simple iterator class for walking a file's child list.  This is a
 * trivial derivation from the SList::Iterator class.
 */
class FileChildrenIterator : public SList::Iterator
    {
    public:
	inline		FileChildrenIterator(const File& targetFile);
	inline		~FileChildrenIterator();
	inline File&	operator*() const; // Return a child
    private:
	const File&	target;		// File being iterated
    };

/*
 * Iterator for walking all of a file's descendants.  This is a
 * based on the SList::Iterator class, but it's not so trivial because we
 * have to track all unprocessed subdirectories.
 */
class FileDescendantsIterator
    {
    public:
	inline		FileDescendantsIterator(const File& targetFile);
	inline		~FileDescendantsIterator();
	inline File&	operator*() const; // Return a descendant
	inline		operator bool() const; // For completion testing
	inline FileDescendantsIterator&
			operator++();	// Step to next descendant
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			FileDescendantsIterator(
			  const FileDescendantsIterator&);
	FileDescendantsIterator&
			operator=(const FileDescendantsIterator&);
	void		removeCurrent(); // Disable removeCurrent because it's
					// ..tricky to handle properly
	/*
	 * Private data
	 */
	const File*	currentTarget; // File being iterated now
	SList::Iterator	subIterator;	// Iterator for current sublist
	SList		pendingDirectories; // Unprocessed directories
    };

/*
 * Iterator for walking a file's related-file list.
 *
 * As a side effect, this iterator cleans up any related-file entries
 * that have no actual relations.
 */
class RelatedFileIterator : public SList::Iterator
    {
    public:
	inline		RelatedFileIterator(File& sourceFile);
	inline RelatedFile&
			operator*() const; // Return a related file
	RelatedFileIterator&
			operator++();	// Step to next relation, cleaning up
    protected:
	File&		source;		// File whose relations we're doing
    };

/*
 * Iterator base for use in walking one relation type in a file's
 * related-file list.  Each relation type should derive an iterator
 * from this base for use in walking its type.
 */
class SelectedRelatedFileIteratorBase : private RelatedFileIterator
    {
	friend		FileRelationTypeBase;
    public:
	virtual		~SelectedRelatedFileIteratorBase();
	inline SelectedRelatedFileIteratorBase&
			operator++();	// Step to next selected file
	inline		operator bool(); // For completion testing
	inline void	reset();	// Reset iterator
	inline void	removeCurrent(); // Remove current element
    protected:
	inline		SelectedRelatedFileIteratorBase(File& sourceFile,
			  const FileRelationTypeBase& desiredType);
					// Derived construction only
	inline FileRelationBase&
			operator*() const; // Return current relation
    private:
	void		findSelectedType(); // Point iterator at a relation
	const FileRelationTypeBase&
			selectedType; // Type we're looking for
	RelationIterator* subIterator;	// Iterator for individual relation
    };

#define _FILE_ICC_NEEDED

#endif // _FILE_HH
