#ifndef lint
static const char Rcs_Id[] =
    "$Id: filehistory.cc,v 1.6 1998/10/23 23:15:15 geoff Exp $";
#endif

/*
 * $Id: filehistory.cc,v 1.6 1998/10/23 23:15:15 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filehistory.cc,v $
 * Revision 1.6  1998/10/23 23:15:15  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.5  1998/07/15 21:43:40  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.4  1996/05/03 01:12:10  geoff
 * Update the copyright.
 *
 * Revision 1.3  1996/02/02  18:28:56  geoff
 * Add an assignment operator.  Fix some minor problems detected by the
 * new C++ compiler.
 *
 * Revision 1.2  1995/12/19  01:02:54  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.1  1995/12/19  00:02:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "filehistory.hh"
#include "filereference.hh"
#include "managerbase.hh"
#include "zzinlinefunctions.hh"

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			FileHistory::FileHistory(int initialSize);
FileHistory&		FileHistory::operator=(const FileHistory& that);
			FileHistory::~FileHistory();
void			FileHistory::newFile(HoardingManagerBase& manager,
			  File& file);
					// Add a new file to the table
void			FileHistory::loseForgettingPointers();
					// Clean refs to forgetting files
FileHistory&		FileHistory::combine(FileHistory& that);
					// Add that's references into this
#endif // 0

FileHistory::FileHistory(
    int			initialSize)	// Size to make the history table
    : tableSize(initialSize),
      first(0)
    {
    elements = new HistoryElement[initialSize];
    }

FileHistory& FileHistory::operator=(	// Perform assignment
    const FileHistory&	that)		// History to copy from
    {
    if (this == &that)
	return *this;
    int thisElement = tableSize - 1;
    for (FileHistoryIterator thatElement((FileHistory&)that,
	FileHistoryIterator::REVERSE, true);
      thisElement >= 0  &&  thatElement;
      --thisElement, --thatElement)
	elements[thisElement] = *thatElement;
    first = thisElement + 1;
    for (  ;  thisElement >= 0;  --thisElement)
	{
	elements[thisElement].referencedFile = NULL;
	elements[thisElement].referenceTime = 0;
	}
    return *this;
    }

FileHistory::~FileHistory()
    {
    delete[] elements;
    }

void FileHistory::newFile(		// Add a new file to the table
    HoardingManagerBase& manager,	// Manager we're working on behalf of
    File&		file)		// File to be added
    {
    /*
     * PERFORM MAX_NPREV CORRECTION:
     *
     * For the file that is going to be dropped from "previous", scan
     * all references to see if they are older than the file.  If so,
     * we will assume that the next reference to that file will be at
     * least MAX_NPREV+1 references away, so we add that much to the total
     * distance.  This assumption will be verified or corrected later,
     * in one of two places.  Call the file being dropped now "D", and
     * the file in its reference table (which we are correcting) "R".
     *
     * At the next reference to "D", there either has or has not been
     * an intervening reference to "R".  If there has, then the total
     * distance (incorporating the MAX_NPREV correction) is correct, but
     * we must increment the number of references.  If there has not,
     * then we must remove the correction.  The reason is the
     * possibility of a sequence like this:
     *
     *	    D, R, D, <many references>, D, R
     *
     * In this case, the proper distance from D to R is (1+1)/2.
     * Contrast this with the similar case:
     *
     *	    D, R, D, <many references>, R, D
     *
     * In this case, the best estimate of the proper distance is
     * (MAX_NPREV+1+1)/2.
     *
     * At file management time, a similar situation occurs, except
     * that there has not been a reference to D.  It is more accurate
     * to remove the correction in this case.  Consider the sequence:
     *
     *	    D, R, D, R, D, R, D, <many references>, <management>
     *
     * In this case, the best estimate of the correct distance from D
     * to R is (1+1+1)/3, rather than (MAX_NPREV+1+3)/4.
     */
    File* previousFile = elements[first].referencedFile;
    if (previousFile != NULL  &&  previousFile != &file)
	{
	for (FileReferenceIterator reference(*previousFile);
	  reference;
	  ++reference)
	    (*reference).correct(
	      manager.managerData(*previousFile).lastReferenceTime(),
	      manager);
	}
    elements[first].referencedFile = &file;
    elements[first].referenceTime =
      manager.managerData(file).lastReferenceTime();
    if (++first >= tableSize)
	first = 0;
    }

void FileHistory::loseForgettingPointers() // Clean refs to forgetting files
    {
    for (FileHistoryIterator element(*this);  element;  ++element)
	{
	HistoryElement& history = *element;
	File* previous = history.file();
	if (previous->isForgetting)
	    history.clear();
	}
    }

/*
 * Combine two history tables into one.  The result is achieved by
 * modifying the current table.  The "that" table becomes obsolete.
 */
FileHistory& FileHistory::combine(
    FileHistory&	that)
    {
    /*
     * Unmark all files in both history lists.  We will use the
     * presence of a mark to detect files that are in both lists; in
     * this case, we'll choose the entry with the latest reference
     * time.
     */
    FileHistoryIterator thisIterator(*this, FileHistoryIterator::REVERSE);
    for (  ;  thisIterator;  --thisIterator)
	(*thisIterator).file()->unmark();
    FileHistoryIterator thatIterator(that, FileHistoryIterator::REVERSE);
    for (  ;  thatIterator;  --thatIterator)
	(*thatIterator).file()->unmark();
    /*
     * Merge the two histories into a single new one.
     */
    HistoryElement* newElements = new HistoryElement[tableSize];
    int elementIndex = tableSize - 1;
    thisIterator.reset(FileHistoryIterator::REVERSE, true);
    thatIterator.reset(FileHistoryIterator::REVERSE, true);
    while (elementIndex >= 0)
	{
	HistoryElement* thisElement = thisIterator ? &(*thisIterator) : NULL;
	HistoryElement* thatElement = thatIterator ? &(*thatIterator) : NULL;
	if (thisElement == NULL  &&  thatElement == NULL)
	    break;
	if (thisElement != NULL  &&  thisElement->referencedFile != NULL
	  &&  thisElement->referencedFile->isMarked())
	    {
	    /*
	     * The file in "this" is marked, so it was already found
	     * in "that".  We'll skip it.
	     */
	    --thisIterator;
	    continue;
	    }
	if (thatElement != NULL  &&  thatElement->referencedFile != NULL
	  &&  thatElement->referencedFile->isMarked())
	    {
	    /*
	     * The file in "that" is marked, so it was already found
	     * in "this".  We'll skip it.
	     */
	    --thatIterator;
	    continue;
	    }
	if (thisElement != NULL  &&  thatElement != NULL
	  &&  thisElement->referenceTime > thatElement->referenceTime)
	    thatElement = NULL;		// Force choice from thisElement
	if (thatElement != NULL)
	    {
	    if (thatElement->referencedFile != NULL)
		thatElement->referencedFile->mark();
	    newElements[elementIndex] = *thatElement;
	    --thatIterator;
	    }
	else
	    {
	    if (thisElement->referencedFile != NULL)
		thisElement->referencedFile->mark();
	    newElements[elementIndex] = *thisElement;
	    --thisIterator;
	    }
	elementIndex--;
	}
    /*
     * Replace the old history table in "this" with the new one.
     */
    delete[] elements;
    elements = newElements;
    first = 0;
    return *this;
    }
