#ifndef _FILEREFERENCE_HH
#define _FILEREFERENCE_HH
/*
 * $Id: filereference.hh,v 1.19 1998/10/23 23:15:17 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filereference.hh,v $
 * Revision 1.19  1998/10/23 23:15:17  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.18  1998/07/15 21:43:42  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.17  1998/07/09 21:02:02  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.16  1998/07/08 19:59:04  geoff
 * Add quickAverageDistance.
 *
 * Revision 1.15  1996/10/11 22:09:03  geoff
 * Add geometric-mean support.
 *
 * Revision 1.14  1996/09/11 22:59:52  geoff
 * Use a type-specific allocator.
 *
 * Revision 1.13  1996/06/11  06:13:18  geoff
 * Shorten the flags field.  Add a new parameter to newReference and
 * updateDistance, to support aging of old references.
 *
 * Revision 1.12  1996/05/03  01:12:15  geoff
 * Update the copyright.
 *
 * Revision 1.11  1996/02/02  18:28:59  geoff
 * Declare inline functions as such in the class definition.  Declare all
 * virtual functions as such in derived classes.
 *
 * Revision 1.10  1995/12/19  01:02:37  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.9  1995/12/19  00:34:24  geoff
 * Add process-tracking support.  Add FinalFileReferenceDestructor.
 *
 * Revision 1.8  1995/11/22  00:20:47  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.7  1995/11/21  21:38:33  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support.  Add new/delete operators in an attempt to improve
 * efficiency.
 *
 * Revision 1.6  1995/11/10  02:59:12  geoff
 * Add NEW_FILE_RELATIONS support.
 *
 * Revision 1.5  1995/09/15  00:05:38  geoff
 * Rename clearForgotten to clearForgetting to reflect its new
 * functionality.
 *
 * Revision 1.4  1995/06/17  00:11:32  geoff
 * Add show and clear.  Get rid of a bunch of "class" keywords.
 *
 * Revision 1.3  1995/05/23  00:39:53  geoff
 * Add a comment about why filereference.icc isn't included.
 *
 * Revision 1.2  1995/03/29  21:04:33  geoff
 * Switch to byte-swapped streams.  Add the new ReferenceCount typedef.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "basictypes.hh"
#include "correlator.hh"
#include "file.hh"
#include "filehistory.hh"
#include "filerelation.hh"
#include "showoptions.hh"
#ifdef USE_TYPE_ALLOCATOR
#include "typealloc.hh"
#endif // USE_TYPE_ALLOCATOR

/*
 * Class used to describe a reference to a file
 */

#ifndef NULL
#define NULL		0
#endif // NULL

class ByteSwappedStream;
class File;
class FileReference;
class FinalFileReferenceDestructor;
class HoardingManagerBase;
class Process;
class ostream;

class FileReferenceType : public FileRelationTypeBase
    {
    public:
	inline		FileReferenceType(const char* name);
	virtual		~FileReferenceType();
	virtual FileRelationBase&
			newRelation() const;
					// Create new relation instance
	virtual FileRelationBase*
			readRelation(ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
	virtual void	recordReference(File& file, ReferenceType type,
			    Process* process, bool isMeaningless)
			  const;
					// Record a reference to a given file
	virtual void	cleanGlobalData() const;
					// Clean out type's global data
	/*
	 * Public static data
	 */
	static const char* typeClass;	// Master type class name
    };

/*
 * Dummy class used solely to provide an on-exit destructor for
 * FileReference.
 */
class FinalFileReferenceDestructor
    {
	friend		FileReference;
    public:
			~FinalFileReferenceDestructor();
    private:
			FinalFileReferenceDestructor();
    };

class FileReference : public FileRelationBase
    {
	friend		FileReferenceType::readRelation(
			    ByteSwappedStream& stream)
			  const;
	friend		FinalFileReferenceDestructor;
	friend void	FileReferenceType::recordReference(File& file,
			    ReferenceType type, Process* process,
			    bool isMeaningless)
			  const;
	friend void	FileReferenceType::cleanGlobalData()
			  const;
    public:
	inline		FileReference();
	virtual		~FileReference();
	inline void*	operator new(size_t size);
					// Allocate a reference
	inline void	operator delete(void* reference);
					// Free a reference
	virtual const FileRelationTypeBase&
			type() const;	// Return type of the relation
	friend ostream&	operator<<(ostream& stream,
			  const FileReference& reference);
					// Show reference in printable form
					// ..appends \n
	virtual ostream& show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show reference in printable form,
					// ..with options.  Appends \n
	void		newReference(File* target, ReferenceCount distance,
			  ReferenceCount previousReferenceTime);
					// Set file being referenced
	ByteSwappedStream& write(ByteSwappedStream& stream) const;
					// Write a reference entry to a stream
	inline bool	isActive() const; // Test if a ref entry has valid data
	inline bool	referenceIsTo(File* that) const;
					// Test if ref is to a particular file
	static void	updateDistance(File& sourceFile, File& targetFile,
			  ReferenceCount localDistance,
			  ReferenceCount previousReferenceTime,
			  HoardingManagerBase& manager);
					// Update distance between two files
	void		correct(ReferenceCount lastReference,
			  HoardingManagerBase& manager);
					// Perform NPREV correction
	void		uncorrect(ReferenceCount referenceTime,
			  HoardingManagerBase& manager);
					// Undo NPREV correction if necessary
	double		averageDistance() const;
					// Return avg distance of reference
	double		quickAverageDistance() const;
					// Return avg distance of reference,
					// ..calculated quickly and only useful
					// ..for relative comparisons
	inline ReferenceCount
			referenceCount() const;
					// Return number of references
	static void	setGeometricMeans(bool useGeometric);
					// Set whether to use geometric means
	static void	readGeometricMeans(ByteSwappedStream& stream);
					// Read geometric-mean flag from stream
	static void	writeGeometricMeans(ByteSwappedStream& stream);
					// Write geometric-mean flag to stream
	/*
	 * Public static data
	 */
	static const FileReferenceType	// Master data for file-ref relations
			typeInformation;
    private:
			FileReference(FileReference& that);
					// Disable assignment
	FileReference&	operator=(const FileReference& that);
					// Disable copy
	/*
	 * Parameters of the implementation.  This should be defined
	 * using 'const int', but ANSI C++ forbids it, so I'll use an
	 * enum to kludge it.
	 */
	enum
	    {
	    FILE_REFERENCE_MAX_FREE = 30 // Maximum items on free list
	    };
	/*
	 * Private data
	 */
	ReferenceCount	numReferences;	// Number of references to this file
	union
	    {
	    ReferenceCount totalDistance; // Total distance of all refs
	    double	totalLogDistance; // Total of log of distance
	    }
			accumulator;	// Accumulator for calculating mean
	char		flags;		// Flags for this entry
	/*
	 * Static data
	 */
	static FinalFileReferenceDestructor
			finalFileReferenceDestructor;
	static FileReference* freeReferences; // SList of items for allocation
	static int	freeCount;	// Number of items on free list
	static bool	geometricMeans;	// True to use geometric means
	static FileHistory globalHistory; // History of previous references
	static ReferenceCount
			tieBreaker;	// Counter used to break reference ties
#ifdef USE_TYPE_ALLOCATOR
	static TypeAllocator
			allocator;	// Type-specific allocator for elements
#endif // USE_TYPE_ALLOCATOR
	/*
	 * Flags for a FileReference.  These should be defined using
	 * 'const int', but C++ forbids it, so I'll kludge it with an
	 * enum.
	 */
	enum
	    {
	    RF_CORRECTED = 0x01		 // NPREV corr has been applied
	    };
    };


/*
 * Iterator for walking a file's reference list
 */
class FileReferenceIterator : public SelectedRelatedFileIteratorBase
    {
    public:
	inline		FileReferenceIterator(File& sourceFile);
	virtual		~FileReferenceIterator();
	inline FileReference&
			operator*();	// Return next reference
    };

#define _FILEREFERENCE_ICC_NEEDED

#endif // _FILEREFERENCE_HH
