#ifndef _FILERELATION_HH
#define _FILERELATION_HH
/*
 * $Id: filerelation.hh,v 1.16 1998/10/23 23:15:20 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: filerelation.hh,v $
 * Revision 1.16  1998/10/23 23:15:20  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.15  1998/07/15 21:43:43  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.14  1998/07/09 21:02:02  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.13  1998/06/02 23:59:45  geoff
 * Add a couple of minor descriptive comments.
 *
 * Revision 1.12  1998/05/14 01:15:32  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.11  1998/04/09 00:33:01  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.10  1997/12/18 01:25:10  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.9  1996/12/03 02:58:06  geoff
 * Change the name of the routine that records missed references.
 *
 * Revision 1.8  1996/09/11 22:59:56  geoff
 * Add REPORT_SELECTED_FAILURES support.  Convert from dlists to slists.
 * Use a type-specific allocator.
 *
 * Revision 1.7  1996/05/03  01:12:19  geoff
 * Update the copyright.
 *
 * Revision 1.6  1996/02/02  18:29:02  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.5  1995/12/19  01:02:38  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.4  1995/12/19  00:34:31  geoff
 * Add FinalRelatedFileDestructor.  Add process-tracking stuff.
 *
 * Revision 1.3  1995/11/22  00:20:59  geoff
 * Fix the RCS comment leader.  Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.2  1995/11/21  21:38:44  geoff
 * More NEW_FILE_RELATIONS changes engendered by adding investigation
 * support.  Add new/delete operators in an attempt to improve
 * efficiency.
 *
 * Revision 1.1  1995/11/10  02:09:40  geoff
 * Initial revision
 *
 */

/*
 * Classes used to describe arbitrary relationships among files.
 */

/*
 * A given source file contains a list of zero or more RelatedFiles.
 * Each RelatedFile gives a target file and one or more relationships
 * between the source and target.  Each relationship is described by a
 * FileRelationBase.
 */
#include "list.hh"
#include "showoptions.hh"
#ifdef USE_TYPE_ALLOCATOR
#include "typealloc.hh"
#endif // USE_TYPE_ALLOCATOR

class ByteSwappedStream;
class File;
class FileRelationBase;
class FileRelationTypeIterator;
class FinalRelatedFileDestructor;
class Process;
class RelatedFile;
class RelationIterator;
class ostream;

/*
 * Types of file relations.  Note that the className must be a pointer
 * into constant storage, so that clients can compare addresses rather
 * than doing expensive string comparisons.
 */
class FileRelationTypeBase
    {
    friend		FileRelationTypeIterator;
    public:
			FileRelationTypeBase(const char* name,
			  const char* className);
	virtual		~FileRelationTypeBase();
	static FileRelationTypeBase*
			findByName(const char* name);
					// Locate type entry for named relation
	static FileRelationTypeBase*
			findById(int id);
					// Locate relation by ID number
	inline const char*
			name() const;	// Return name of the relation type
	const char*	encodedName(ShowOptions options)
			  const;	// Return encoded name of relation type
	inline const char*
			className() const; // Return name of relation's class
	const char*	encodedClassName(ShowOptions options)
			  const;	// Return encoded class name
	inline int	id() const;	// Return ID of the relation type
	unsigned long	count(const File& file) const;
					// Return count of the number of
					// relations of this type that a given
					// ..file has
	virtual FileRelationBase&
			newRelation() const = 0;
					// Create new relation instance
	virtual FileRelationBase*
			readRelation(ByteSwappedStream& stream)
			  const = 0;
					// Read a new relation from a file
	static ByteSwappedStream&
			writeRelationTypes(ByteSwappedStream& stream);
					// Write relation-type table to stream
	static ByteSwappedStream&
			readRelationTypes(ByteSwappedStream& stream);
					// Read relation-type table from stream
	virtual ByteSwappedStream&
			write(ByteSwappedStream& stream)
			  const;	// Write type-specific info to stream
	virtual ByteSwappedStream&
			read(ByteSwappedStream& stream);
					// Read type-specific info from stream
	virtual void	recordMissedReference(File& file, ReferenceType type,
			    Process* process, bool isMeaningless)
			  const;
					// Record a missed reference to a file
	virtual void	recordReference(File& file, ReferenceType type,
			    Process* process, bool isMeaningless)
			  const = 0;
					// Record a reference to a file
	virtual void	cleanGlobalData() const = 0;
					// Clean out type's global data
    private:
	const char*	typeName;	// Name given to this type
	const char*	typeClassName;	// Name given to this type class
	mutable unsigned char
			typeId;		// ID given to type when writing
	static SList	allTypes;	// List of all known types
	static unsigned char
			nextId;		// Next ID available for assignment
    };

class FileRelationTypeIterator : public SList::Iterator
    {
    public:
	inline		FileRelationTypeIterator();
	inline		~FileRelationTypeIterator();
	inline FileRelationTypeBase&
			operator*();	// Return target of iterator
    };

class FileRelationBase
    {
	friend		RelatedFile;
    public:
	inline		FileRelationBase();
	virtual		~FileRelationBase();
	virtual const FileRelationTypeBase&
			type() const = 0; // Return type of the relation
	inline RelatedFile&
			targetRelatedFile() const; // Return target of relation
	virtual ByteSwappedStream&
			write(ByteSwappedStream& stream) const;
					// Write relation to a stream
	static FileRelationBase*
			read(ByteSwappedStream& stream);
					// Read relation from a stream
	virtual ostream& show(ostream& stream, ShowOptions options,
			    const char* managerName)
			  const = 0;	// Show relation in printable form

    private:
			FileRelationBase(const FileRelationBase&);
	FileRelationBase& operator=(const FileRelationBase&);
	/*
	 * Private data
	 */
	RelatedFile*	target;		// File we're related to, kept 
					// ..because it's used so much
    };

/*
 * Dummy class used solely to provide an on-exit destructor for
 * RelatedFile.
 */
class FinalRelatedFileDestructor
    {
	friend		RelatedFile;
    public:
			~FinalRelatedFileDestructor();
    private:
			FinalRelatedFileDestructor();
    };

class RelatedFile
    {
	friend		FinalRelatedFileDestructor;
	friend		RelationIterator;
    public:
	inline		RelatedFile(File& targetFile);
			~RelatedFile();
	inline void*	operator new(size_t size);
					// Allocate a related file
	inline void	operator delete(void* relatedFile);
					// Free a related file
	bool		hasType(const FileRelationTypeBase& type) const;
					// See if file has a given rel. type
	FileRelationBase*
			chosenRelation(const FileRelationTypeBase& type,
			  bool create = false);
					// Return relation of chosen type
	inline File&	targetFile() const; // Return target of relation
	inline bool	isEmpty() const; // Test for empty relation list
	inline void	addRelation(FileRelationBase* relation);
					// Add a new relation to the list
	inline void	removeRelation(FileRelationBase* relation);
					// Remove a relation from the list
	ByteSwappedStream& write(ByteSwappedStream& stream) const;
					// Write a related file's relations
	static ByteSwappedStream&
			read(ByteSwappedStream& stream, File& targetFile);
					// Read a related file's relations
	ostream&	show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show related file in printable form
    private:
	/*
	 * Implementation constants.  This should be defined using
	 * 'const int', but ANSI C++ forbids it, so I'll use an enum
	 * to kludge it.
	 */
	enum
	    {
	    RELATED_FILE_MAX_FREE = 30	// Maximum items on free list

	    };

	File&		target;		// File we are related to
	SList		relations;	// Relations between us and this target
	static RelatedFile* freeRelatedFiles; // List of items for allocation
	static int	freeCount;	// Number of items on free list
	static FinalRelatedFileDestructor
			finalRelatedFileDestructor;
#ifdef USE_TYPE_ALLOCATOR
	static TypeAllocator
			allocator;	// Type-specific allocator for elements
#endif // USE_TYPE_ALLOCATOR
    };

/*
 * Trivial iterator to step through the relations for a given target file.
 */
class RelationIterator : public SList::Iterator
    {
    public:
	inline		RelationIterator(RelatedFile& targetFile);
	inline void	retarget(RelatedFile& newTargetFile);
					// Point iterator at new list
	inline FileRelationBase&
			operator*() const; // Return a relation
    private:
	RelatedFile*	target;		// Related file we're iterating through
    };

#define _FILERELATION_ICC_NEEDED

#endif // _FILERELATION_HH
