#ifndef _GLOBALMANAGER_HH
#define _GLOBALMANAGER_HH
/*
 * $Id: globalmanager.hh,v 1.18 1998/10/23 23:15:25 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: globalmanager.hh,v $
 * Revision 1.18  1998/10/23 23:15:25  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.17  1998/07/15 21:43:47  geoff
 * Get rid of baseReference; it's now manager-specific.
 *
 * Revision 1.16  1998/05/14 01:15:33  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.15  1997/02/12 02:52:07  geoff
 * Make oftenPercent be a double.
 *
 * Revision 1.14  1997/01/02 06:48:22  geoff
 * Add resetStatistics and support for the base reference.
 *
 * Revision 1.13  1996/12/03 02:58:07  geoff
 * Add support for reading and writing parameters.
 *
 * Revision 1.12  1996/06/11 06:13:21  geoff
 * Add support for the peek threshold, age weight factor, and aging
 * threshold.
 *
 * Revision 1.11  1996/05/03  01:12:24  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/03/19  18:43:00  geoff
 * Get rid of the show/read/write/clear-manager-data routines now that
 * they can be inherited.  Inherit from EmptyManagerBase.
 *
 * Revision 1.9  1996/02/02  18:29:04  geoff
 * Declare all virtual functions as such in derived classes.  Declare
 * inline functions as such in the class definition.
 *
 * Revision 1.8  1995/12/19  00:34:36  geoff
 * Add process-tree support.
 *
 * Revision 1.7  1995/11/22  00:21:22  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.6  1995/11/10  02:59:18  geoff
 * Don't include the icc file, since zzinlinefunctions now does that.
 * Get rid of a spurious "const".  Add NEW_FILE_RELATIONS support
 * (keepReferences).
 *
 * Revision 1.5  1995/09/15  00:05:42  geoff
 * Globally change "cache" to "hoard".  Add internal-timer and statistics
 * support.
 *
 * Revision 1.4  1995/06/17  00:11:37  geoff
 * Add show options and clearManagerData.  Add a couple of const
 * keywords.
 *
 * Revision 1.3  1995/05/23  00:40:00  geoff
 * Modify some declarations to use the "class" keyword a bit more
 * cleanly, and declare some new I/O routines required by
 * PRIVATE_MANAGEMENT_DATA.
 *
 * Revision 1.2  1995/04/20  21:58:07  geoff
 * Make maximum cache size be a global parameter.
 *
 * Revision 1.1  1995/03/29  21:06:45  geoff
 * Initial revision
 *
 */

#include "managerbase.hh"

class ByteSwappedStream;
class File;
class InternalTimer;
class ostream;

/*
 * The global manager is a dummy used to support parameters which are
 * common to all managers, such as the interval between management
 * runs.  The management routines are dummies, but the parameter
 * setting is real.
 */
class GlobalManager : public EmptyManagerBase
    {
    public:
			GlobalManager(const char* managerName,
			  const char* managerDefaults);
	virtual		~GlobalManager();
	virtual bool	shouldManage(const InternalTimer& now,
			  InternalTimer& nextManagementTime);
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream,
			    bool encodeFilenames);
					// Run dummy hoard management
	virtual void	setParameter(const char* description);
	virtual void	showParameters(ostream& stream) const;
					// Display management parameters
	virtual long	writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
	virtual long	readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
	virtual ostream& showStatistics(ostream& stream) const;
					// Display collected quality statistics
	virtual void	resetStatistics();
					// Reset quality statistics
	static inline double
			ageWeight();	// Return weight to use in aging
	static inline unsigned long
			ageThreshold();	// Return threshold to use in aging
	static inline double
			getOftenPercent();
					// Return "often" percentage
	static inline int
			getPeekThreshold(); // Return peek threshold
	static inline unsigned long
			hoardSize();	// Return hoard size in blocks
	static inline int
			referencesToKeep(); // Return keepReferences
	static inline bool
			useProcessTree(); // Return processTree
    private:
	static double	ageWeightFactor; // Multiplier for total distance to
					// ..use in reference aging
	static unsigned long
			agingThreshold;	// Threshold to use in reference aging
	static double	oftenPercent;	// Percent refs defined as "often"
	static int	peekThreshold;	// Percent peeks to be meaningless
	static long	maximumHoardBlocks; // Size of the hoard
	static int	keepReferences; // Max. number of references to keep
					// ..used to be MAX_K
	static bool	processTree;	// True to use process trees in
					// ..calculating semantic distance
    };

#define _GLOBALMANAGER_ICC_NEEDED

#endif // _GLOBALMANAGER_HH
