#ifndef _INTERNALTIMER_HH
#define _INTERNALTIMER_HH
/*
 * $Id: internaltimer.hh,v 1.5 1998/10/23 23:15:30 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: internaltimer.hh,v $
 * Revision 1.5  1998/10/23 23:15:30  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.4  1996/05/03 01:12:28  geoff
 * Update the copyright.
 *
 * Revision 1.3  1996/02/02  18:29:05  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.2  1995/11/10  02:59:36  geoff
 * Don't include the icc file, since zzinlinefunctions now does that.
 *
 * Revision 1.1  1995/09/14  23:32:25  geoff
 * Initial revision
 *
 */

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

/*
 * Classes supporting the tracking of real and simulated time.
 */
class InternalTimer
    {
    public:
	enum TimerType			// Types of internal timers
	    {
	    ACTUAL,			// Timer stores actual time of day
	    VIRTUAL			// Timer stores simulated time of day
	    };
			InternalTimer(TimerType desiredType);
	inline		InternalTimer(const InternalTimer& source);
			~InternalTimer();
	enum Conversion			// Ways to convert times
	    {
	    S,				// Seconds only
	    MS,				// Minutes:seconds only
	    S_MS,			// Seconds if < 60, else MS
	    HMS,			// Hours:minutes:seconds only
	    MS_HMS,			// Minutes:seconds if < 60, else HMS
	    S_MS_HMS,			// Most compact form HMS possible
	    DHMS,			// Days hour:minutes:seconds only
	    HMS_DHMS,			// DHMS if >= 24 hours, else HMS
	    MS_HMS_DHMS,		// DHMS if >= 24 hours, else MS_HMS
	    S_MS_HMS_DHMS		// DHMS if >= 24 hours, else S_MS_HMS
	    };
	/*
	 * Assignment and casting
	 */
	const InternalTimer&
			operator=(const InternalTimer& source);
	inline		operator const struct timeval&() const;
	/*
	 * Arithmetic
	 */
	inline InternalTimer
			operator+(const struct timeval& addend) const;
	inline InternalTimer
			operator+(time_t increment) const;
	inline InternalTimer&
			operator+=(const struct timeval& addend);
	inline InternalTimer&
			operator+=(time_t increment);
	inline InternalTimer
			operator-(const struct timeval& subtrahend) const;
	inline InternalTimer
			operator-(time_t decrement) const;
	inline InternalTimer&
			operator-=(const struct timeval& subtrahend);
	inline InternalTimer&
			operator-=(time_t decrement);
	/*
	 * Comparison and testing.  Note that timevals can be compared
	 * as to InternalTimes, so long aas they are on the right-hand
	 * side.
	 */
	inline bool	operator==(const struct timeval& comparand) const;
	inline bool	operator!=(const struct timeval& comparand) const;
	inline bool	operator<(const struct timeval& comparand) const;
	inline bool	operator<=(const struct timeval& comparand) const;
	inline bool	operator>(const struct timeval& comparand) const;
	inline bool	operator>=(const struct timeval& comparand) const;
	inline bool	isValid() const; // Test for validity (nonzero)
	/*
	 * String conversion
	 */
	const char*	ctime() const;	// Return timer as English string
	const char*	cmicrotime() const; // Return timer as string
	const char*	numericTime(Conversion conversion = S_MS_HMS_DHMS)
			  const;
					// Return timer as numeric string
	const char*	numericMicrotime(Conversion conversion = S_MS_HMS_DHMS)
			  const;
					// Return timer w/ microseconds
					// ..as numeric string
	/*
	 * Control
	 */
	static inline void
			setVirtualTime(const struct timeval& when);
					// Change current virtual time
	void		updateTime();	// Update timer's recorded time
	inline void	makeInvalid();	// Make timer contents invalid
	static void	timerMode(TimerType desiredMode);
					// Set global mode to real/virtual
    private:
	inline void	markChanged();	// Mark that time has changed
	TimerType	type;		// Type of this timer
	struct timeval	currentTime;	// Current time value
	mutable char	seconds[
			  sizeof "Wed Jun 30 21:49:08 1993"];
					// String version in seconds
	mutable char	microseconds[
			  sizeof "Wed Jun 30 21:49:08.123456 1993"];
					// String version in microseconds
	mutable char	numericMicroseconds[
			  sizeof "1234567890d 00:59:59.123456"];
					// Numeric version in microseconds
	/*
	 * Static data
	 */
	static TimerType mode;		// Current mode for virtual timers
	static struct timeval
			simulatedTime;	// Current simulated time
    };

#define _INTERNALTIMER_ICC_NEEDED

#endif // _INTERNALTIMER_HH
