#ifndef _INVESTIGATOR_HH
#define _INVESTIGATOR_HH
/*
 * $Id: investigator.hh,v 1.17 1998/10/23 23:15:35 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This file describes classes used to support external investigators.
 */

/*
 * $Log: investigator.hh,v $
 * Revision 1.17  1998/10/23 23:15:35  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.16  1998/07/15 21:43:48  geoff
 * Changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.15  1998/07/09 21:02:03  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.14  1998/05/14 01:15:34  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.13  1998/04/09 00:33:02  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.12  1996/09/11 23:00:02  geoff
 * Convert from dlists to slists.
 *
 * Revision 1.11  1996/06/11  06:13:22  geoff
 * Major changes to support cluster investigators.  Add learnFromStream
 * routines, all necessary cluster-investigator stuff, new virtual
 * routines required by cluster-investigator support,
 * InvestigatedClusterIterator.
 *
 * Revision 1.10  1996/05/03  01:12:32  geoff
 * Update the copyright.
 *
 * Revision 1.9  1996/02/02  18:29:08  geoff
 * Declare inline functions as such in the class definition.
 * Declare all virtual functions as such in derived classes.
 *
 * Revision 1.8  1995/12/19  01:02:40  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.7  1995/12/19  00:34:41  geoff
 * Add process-tracking support.
 *
 * Revision 1.6  1995/11/22  00:21:33  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.5  1995/11/21  21:39:11  geoff
 * Rewrite the investigator support to make use of NEW_FILE_RELATIONS.
 *
 * Revision 1.4  1995/11/10  02:59:43  geoff
 * Don't include the icc file, since zzinlinefunctions now does that.
 * Fix a formatting error.
 *
 * Revision 1.3  1995/10/10  20:38:03  geoff
 * Get rid of the obsolete TRACK_MEMORY_LEAKS stuff.
 *
 * Revision 1.2  1995/09/15  00:05:49  geoff
 * Add abortive TRACK_MEMORY_LEAKS support.
 *
 * Revision 1.1  1995/08/07  19:14:49  geoff
 * Initial revision
 *
 */

#include "cluster.hh"
#include "file.hh"
#include "filerelation.hh"
#include "list.hh"

class ByteSwappedStream;
class File;
class InvestigatedClusterIterator;
class Process;

/*
 * Class used to maintain general information about investigators.
 * There is one data structure per investigator.  This is a base
 * class; classes for particular types of investigators are derived
 * from it.
 *
 * Note: although the type class is a string, many routines assume
 * that a given class keeps only one global copy of the string, so
 * that they can check for a particular class with a simple address
 * comparison, rather than string comparison.  This invariant must be
 * maintained for correct operation.
 *
 * Note: if you add a new investigator type, you must change the
 * static table in InvestigatorTypeBase::find() to know about it.  You
 * will also probably have to change setInvestigation() in
 * controller.cc, and possibly similar code in correlator.cc.
 */
class InvestigatorTypeBase : public FileRelationTypeBase
    {
    public:
	typedef InvestigatorTypeBase*
			InvestigatorTypeMaker(const char* name);
	inline		InvestigatorTypeBase(const char* name,
			  const char* className);
	virtual		~InvestigatorTypeBase() = 0;
	inline double	weight() const; // Return overall weight to give
					// ..to investigator's suggestions
	void		setParameter(const char* description);
					// Set a changeable parameter
	virtual bool	learnFromStream(ByteSwappedStream& controlStream) = 0;
					// Collect investigated data from an
					// ..external stream
	static InvestigatorTypeBase*
			find(const char* nameSought, bool create);
					// Find a named investigator
	virtual ByteSwappedStream&
			write(ByteSwappedStream& stream)
			  const;
					// Write type-specific info to stream
	virtual ByteSwappedStream&
			read(ByteSwappedStream& stream);
					// Read type-specific info from stream
	virtual void	recordReference(File& file, ReferenceType type,
			    Process* process, bool isMeaningless)
			  const;
					// Record a reference 
	virtual void	cleanGlobalData() const;
					// Clean out type's global data
    private:
			InvestigatorTypeBase(InvestigatorTypeBase& that);
					// Disable assignment
	InvestigatorTypeBase& operator=(const InvestigatorTypeBase& that);
					// Disable copy
	/*
	 * Private data
	 */
	double		overallWeight;	// Weight to give to this
					// ..investigator's suggestions
    };

class ClusterInvestigatorType : public InvestigatorTypeBase
    {
	friend		InvestigatedClusterIterator;
    public:
	inline		ClusterInvestigatorType(const char* name);
	virtual		~ClusterInvestigatorType();
	static InvestigatorTypeBase*
			newInvestigatorType(const char* name);
					// Create a new investigator type
	virtual FileRelationBase&
			newRelation() const;
					// Create new relation instance
	virtual FileRelationBase* 
			readRelation(ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
	virtual bool	learnFromStream(ByteSwappedStream& controlStream);
					// Collect investigated data from an
					// ..external stream
	virtual ByteSwappedStream&
			write(ByteSwappedStream& stream)
			  const;
					// Write type-specific info to stream
	virtual ByteSwappedStream&
			read(ByteSwappedStream& stream);
					// Read type-specific info from stream
	virtual void	cleanGlobalData() const;
					// Clean out type's global data
	static inline unsigned long
			count();	// Return count of clusters
	/*
	 * Public static data
	 */
	static const char* typeClass;	// Master type class name
    private:
	static SList	investigatedClusters;
					// Clusters provided by investigators
    };

class InvestigatedCluster : public Cluster
    {
    public:
	inline		InvestigatedCluster(
			  const ClusterInvestigatorType& type,
			  const char* name);
			~InvestigatedCluster();
	inline const ClusterInvestigatorType&
			type() const;	// Return cluster's type
	virtual void	computeWeight(); // Calculate hoarding weight
	virtual const char*
			typeName() const; // Return cluster's type name
	virtual const char*
			encodedTypeName(ShowOptions options)
			  const;	// Return cluster's encoded type name
	virtual const char*
			name() const;	// Return cluster's name
	virtual const char*
			encodedName(ShowOptions options)
			  const;	// Return cluster's encoded name
    private:
	const ClusterInvestigatorType&
			clusterType;	// Type of this cluster
	const char*	clusterName;	// Name of this cluster
    };

class InvestigatedClusterIterator : public SList::Iterator
    {
    public:
	inline		InvestigatedClusterIterator();
	inline		~InvestigatedClusterIterator();
	inline InvestigatedCluster&
			operator*();	// Return next iterator element
    };


class RelationInvestigatorType : public InvestigatorTypeBase
    {
    public:
	inline		RelationInvestigatorType(const char* name);
	virtual		~RelationInvestigatorType();
	static InvestigatorTypeBase*
			newInvestigatorType(const char* name);
					// Create a new investigator type
	virtual FileRelationBase&
			newRelation() const;
					// Create new relation instance
	virtual FileRelationBase*
			readRelation(ByteSwappedStream& stream)
			  const;
					// Read a new relation from a file
	virtual bool	learnFromStream(ByteSwappedStream& controlStream);
					// Collect investigated data from an
					// ..external stream
	/*
	 * Public static data
	 */
	static const char* typeClass;	// Master type class name
    };

/*
 * Class used to describe the per-file investigator data for
 * relation investigators.
 */
class RelationInvestigation : public FileRelationBase
    {
	friend FileRelationBase*
			RelationInvestigatorType::readRelation(
			    ByteSwappedStream& stream)
			  const;
    public:
	inline		RelationInvestigation(
			  const FileRelationTypeBase& chosenType);
	virtual		~RelationInvestigation();
	virtual const FileRelationTypeBase&
			type() const;	// Return type of the relation
	friend ostream&	operator<<(ostream& stream,
			  const RelationInvestigation& relation);
					// Show relation in printable form
					// ..appends \n
	virtual ostream& show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show relation in printable form,
					// ..with options.  Appends \n
	virtual ByteSwappedStream&
			write(ByteSwappedStream& stream) const;
					// Write a relation entry to a stream
	inline void	setWeight(double newWeight);
					// Set the weight for a relation
	inline double	weight() const;	// Return relation's weight
    private:
			RelationInvestigation(RelationInvestigation& that);
					// Disable assignment
	RelationInvestigation&
			operator=(const RelationInvestigation& that);
					// Disable copy
	/*
	 * Private data
	 */
	const FileRelationTypeBase&
			relationType;	// Type of this investigated relation
	double		relationWeight; // Weight given to relationship
    };

#define _INVESTIGATOR_ICC_NEEDED

#endif // _INVESTIGATOR_HH
