#ifndef _LRUMANAGER_HH
#define _LRUMANAGER_HH
/*
 * $Id: lrumanager.hh,v 1.19 1998/10/23 23:15:41 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: lrumanager.hh,v $
 * Revision 1.19  1998/10/23 23:15:41  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.18  1998/07/15 21:43:51  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.
 *
 * Revision 1.17  1998/07/09 21:02:03  geoff
 * Keep our own internal reference counter.  Add our own
 * reference-recording routines.
 *
 * Revision 1.16  1998/05/14 01:15:35  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.15  1998/03/05 03:18:16  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.14  1997/05/16 17:10:38  geoff
 * Add a clear() routine.  This is the version used to run the
 * simulations in the dissertation.
 *
 * Revision 1.13  1997/01/02 06:48:24  geoff
 * Add cleanGlobalData and support for the requiredSize routine.
 *
 * Revision 1.12  1996/05/03 01:12:36  geoff
 * Update the copyright.
 *
 * Revision 1.11  1996/04/10  01:28:37  geoff
 * Add the new ManagerData::manager() routine.
 *
 * Revision 1.10  1996/03/19  18:43:02  geoff
 * Get rid of the show/read/write/clear-manager-data, store/unstore, and
 * record-reference routines now that they can be inherited.  Inherit
 * from HoardingManagerBase.  Add support for the DERIVED_CLASS_BUG
 * ifdef.
 *
 * Revision 1.9  1996/02/02  18:29:10  geoff
 * Derive resetStatistics from the base class instead of redefining it.
 * Derive data accessors from the base class.  Declare inline functions
 * as such in the class definition.  Declare all virtual functions as
 * such in derived classes.
 *
 * Revision 1.8  1995/11/22  00:21:45  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.7  1995/11/10  02:59:48  geoff
 * Add NEW_FILE_RELATIONS support.  Don't include the icc file, since
 * zzinlinefunctions now does that.  Get rid of a spurious "const".
 *
 * Revision 1.6  1995/10/10  20:38:07  geoff
 * Get rid of the COMPARISON_SUPPORT, SIMULATION_SUPPORT, and
 * TRACK_MEMORY_LEAKS ifdefs.
 *
 * Revision 1.5  1995/09/15  00:05:55  geoff
 * Add InternalTimer support.  Add abortive TRACK_MEMORY_LEAKS support.
 *
 * Revision 1.4  1995/06/17  00:11:42  geoff
 * Add show options and clearManagerData.  Add a couple of const
 * keywords.
 *
 * Revision 1.3  1995/05/27  00:16:04  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.2  1995/05/23  00:40:04  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.1  1995/05/05  21:47:04  geoff
 * Initial revision
 *
 */

#include "basictypes.hh"
#include "communication.hh"
#include "managerbase.hh"
class ostream;
class ByteSwappedStream;
class File;
class InternalTimer;
class LRU_Manager;

/*
 * Class to hold LRU management data for files
 */
class LRU_ManagerData : public HoardingManagerDataBase
    {
    public:
	inline		LRU_ManagerData();
	virtual		~LRU_ManagerData();
	/*
	 * This should return LRU_Manager&, but that requires a
	 * cross-reference to a derived class, which at least g++
	 * can't handle.
	 */
	virtual HoardingManagerBase&
			manager() const; // Return pointer to our manager
	virtual void	recordMissedReference(File& file,
			  ReferenceType referenceType, bool isMeaningless);
					// Record that a file miss happened
	virtual void	recordReference(File& file, ReferenceType type,
			  bool isMeaningless);
					// Keep track of a file reference
	virtual void	clear(const File& file);
					// Clear management data
	virtual ostream& show(ostream& stream, ShowOptions options,
			    const char* managerName)
			  const;
					// Show manager data in printable form,
					// ..with options
	void		cleanReferences();
					// Get rid of obsolete references
    private:
			LRU_ManagerData(const LRU_ManagerData& that);
					// Disable copy
	LRU_ManagerData& operator=(const LRU_ManagerData& that);
					// Disable assignment
	virtual LRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide recursive data access
	virtual const LRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide recursive data access
    };

class LRU_Manager : public HoardingManagerBase
    {
    friend LRU_ManagerData;
    public:
			LRU_Manager(const char* managerName,
			  const char* managerDefaults);
	virtual		~LRU_Manager();
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
	virtual void	showParameters(ostream& stream) const;
					// Display management parameters
	virtual void	cleanGlobalData(); // Clean out global file references
	virtual ostream& showStatistics(ostream& stream) const;
					// Display collected quality statistics
	virtual ostream& requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const;
					// Display hoard size needed for
					// ..no misses
    protected:
	File**		lruFileTable;	// Table of files in priority order
	unsigned long	lruFileTableSize; // Number of entries in file table
	unsigned long	lruActiveCount;	// Number of active entries in table
    private:
			LRU_Manager(const LRU_Manager& that);
					// Disable assignment
	LRU_Manager&	operator=(const LRU_Manager& that); // Disable copy
	virtual LRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const LRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide data access
	static int	compareLRU(File** file1, File** file2);
    };

#define _LRUMANAGER_ICC_NEEDED

#endif // _LRUMANAGER_HH
