#ifndef _MANAGERBASE_HH
#define _MANAGERBASE_HH
/*
 * $Id: managerbase.hh,v 1.25 1998/10/23 23:15:44 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: managerbase.hh,v $
 * Revision 1.25  1998/10/23 23:15:44  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.24  1998/07/15 21:43:52  geoff
 * Many changes to make the last-reference time and similar things be
 * manager-specific.  Also add HoardingManagerIterator and move certain
 * other things to the hoarding manager base class.
 *
 * Revision 1.23  1998/07/09 21:02:04  geoff
 * Add the meaningless argument to the reference-recording routines.
 *
 * Revision 1.22  1998/05/14 01:15:37  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.21  1998/03/05 03:18:16  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.20  1998/02/19 02:10:12  geoff
 * Add stateChanged and relationsChanged as virtual functions.
 *
 * Revision 1.19  1997/12/18 01:25:14  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.18  1997/05/16 17:10:39  geoff
 * Fix the clear() routine to take an argument.  This is the version used
 * to run the simulations in the dissertation.
 *
 * Revision 1.17  1997/01/02 06:48:26  geoff
 * Add support for the requiredSize routines.
 *
 * Revision 1.16  1996/12/03 02:58:10  geoff
 * Add support for reading and writing parameters.  Change the name of
 * the routines that record miss statistics and missed references.  Bump
 * MAX_PARAMETER_LENGTH, just to be safe.
 *
 * Revision 1.15  1996/09/11 23:00:09  geoff
 * Add REPORT_SELECTED_FAILURES support.
 *
 * Revision 1.14  1996/05/03  01:12:40  geoff
 * Update the copyright.  Add some appropriate const declarations.
 *
 * Revision 1.13  1996/04/10  01:28:39  geoff
 * Add the new ManagerData::manager() routine.  Add support for
 * manager-specific storage statuses.
 *
 * Revision 1.12  1996/03/19  18:43:06  geoff
 * Add support for the new manager class inheritance structure.
 *
 * Revision 1.11  1996/02/02  18:29:12  geoff
 * Add new data accessors to the base class so that they can be derived.
 * Add the new ManagerDataBase class.  Add inheritManagerData,
 * readInstructions, and managerCount.  Declare inline functions as such
 * in the class definition.  Declare all virtual functions as such in
 * derived classes.
 *
 * Revision 1.10  1995/11/22  00:22:02  geoff
 * Get rid of the NEW_FILE_RELATIONS ifdef.
 *
 * Revision 1.9  1995/11/10  02:59:53  geoff
 * Add NEW_FILE_RELATIONS support.  Don't include the icc file, since
 * zzinlinefunctions now does that.
 *
 * Revision 1.8  1995/10/10  20:38:12  geoff
 * Get rid of the COMPARISON_SUPPORT and SIMULATION_SUPPORT ifdefs.
 *
 * Revision 1.7  1995/09/15  00:06:06  geoff
 * Add InternalTimer support.  Add statistics support.
 *
 * Revision 1.6  1995/06/17  00:11:48  geoff
 * Add show options and clearManagerData.  Add a couple of const
 * keywords.  Make the manager name protected.
 *
 * Revision 1.5  1995/05/27  00:16:10  geoff
 * Remove PRIVATE_MANAGEMENT_DATA, KEEP_CLUSTER_TABLE, and
 * NEW_MANAGEMENT_DISPLAYS options.
 *
 * Revision 1.4  1995/05/23  00:40:10  geoff
 * Add more PRIVATE_MANAGEMENT_DATA support.
 *
 * Revision 1.3  1995/04/20  21:58:16  geoff
 * Add support for dumping management results to a stream.
 *
 * Revision 1.2  1995/03/29  21:04:43  geoff
 * Many changes to get working new parameter-setting support.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

/*
 * Abstract base class for hoard managers.
 */

#include "basictypes.hh"
#include "communication.hh"
#include "filestate.hh"
#include "internaltimer.hh"
#include "showoptions.hh"

#ifndef NULL
#define NULL	0
#endif // NULL

class ByteSwappedStream;
class File;
enum FileState::State;
class ManagerBase;
class ManagerIterator;
class ManagerDataBase;
class EmptyManagerBase;
class HoardingManagerDataBase;
class HoardingManagerBase;
class HoardingManagerIterator;
class ostream;

class ManagerBase
    {
	friend		ManagerIterator;
    public:
			ManagerBase(const char* managerName,
			  const char* managerDefaults,
			  bool dataStoringManager);
	virtual		~ManagerBase();
	static inline unsigned long
			managerCount();	// Return count of known managers
	inline const char*
			managerName() const; // Return name of manager
	inline bool	hasData() const;
					// Tell whether manager keeps file data
	virtual bool	shouldManage(const InternalTimer& now,
			    InternalTimer& nextManagementTime);
					// Decide if manager should run
	virtual void	manage(const InternalTimer& now,
			    ByteSwappedStream* dumpStream,
			    bool encodeFilenames) = 0;
					// Run hoard management
	virtual void	setParameter(const char* description) = 0;
					// Set management parameters
	virtual void	showParameters(ostream& stream) const
			  = 0;
					// Display management parameters
	virtual long	writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
	virtual long	readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
	ByteSwappedStream&
			writeManagerParameterHeader(ByteSwappedStream& stream)
			  const;
					// Write header for manager parameters
	static bool	readManagerParameters(ByteSwappedStream& stream);
					// Read params for named manager
	virtual void	readInstructions(const char* fileName) const;
					// Read manager-specific instructions
	virtual void	clearInstructions(File& file) const;
					// Forget manager-specific instructions
	static ManagerBase*
			findNamedManager(const char* name);
					// Locate a named manager
	static void	runAllManagers(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Execute all managers
	static void	runDueManagers(const InternalTimer& now,
			  InternalTimer& nextManagementTime);
					// Execute managers due to run
	static void	cleanAllGlobalData();
					// Clean out manager file references
	virtual void	cleanGlobalData(); // Clean out global file references
	static void	setManagerParameters(const char* parameters);
					// Set parameters for named manager
	static void	showManagerParameters(const char* managerName,
			  ostream& stream);
					// Show parameters for named manager(s)
	virtual ManagerDataBase&
			managerData(File& file)
			  const = 0;
					// Provide data access
	virtual const ManagerDataBase&
			managerData(const File& file)
			  const = 0;
					// Provide data access
	virtual void	stateChanged(ManagerDataBase* data,
			  FileState::State oldState,
			  FileState::State newState);
					// Note that state has changed
	/*
	 * Maximum parameter-string length.  All managers must promise
	 * that their parameter strings (created by showParameters)
	 * are no longer than this value.  Parameter strings longer
	 * than this will be truncated.
	 *
	 * NEEDSWORK: there should be a way to interrogate the
	 * parameter-string length.
	 *
	 * This should be defined using 'const int', but ANSI C++
	 * forbids it, so I'll use an enum to kludge it.
	 */
	enum
	    {
	    MAX_PARAMETER_LENGTH = 2048
	    };
    protected:
	const char*	defaults;	// Default parameters
	const char*	name;		// Name of the manager
	time_t		managementInterval; // How often to run management,
					// ..if time-based manager
	InternalTimer	lastManagementTime; // Last time management ran,
					// ..if time-based manager
    private:
	bool		managerHasData;	// True if manager keeps file data
	static ManagerBase* managerTable[]; // Pointers to all managers
    };

/*
 * Class used to iterate across all known managers
 */
class ManagerIterator
    {
    public:
	inline		ManagerIterator();
	inline		~ManagerIterator();
	inline		operator bool() const; // For completion testing
	inline void	reset();	// Reset to start of manager table
	inline ManagerIterator&
			operator++();	// Step to next entry
	inline ManagerBase&
			operator*();	// Return referenced manager
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			ManagerIterator(const ManagerIterator&);
	ManagerIterator& operator=(const ManagerIterator&);
	ManagerBase**	current;	// Current entry in the manager table
    };

/*
 * Class used as a base class for manager-specific data
 */
class ManagerDataBase
    {
	friend ostream&			// Write printable version
			operator<<(ostream& stream,
			  const ManagerDataBase& data);
    public:
	inline		ManagerDataBase();
	virtual		~ManagerDataBase();
	virtual ManagerBase&
			manager() const = 0;
					// Return pointer to our manager
	virtual ostream& show(ostream& stream, ShowOptions options,
			  const char* managerName) const = 0;
					// Show manager data in printable form,
					// ..with options
	virtual ByteSwappedStream&	// Read binary version from stream
			read(ByteSwappedStream& stream, File& file) = 0;
	virtual ByteSwappedStream&	// Write binary version to stream
			write(ByteSwappedStream& stream,
			  const File& file) const = 0;
	virtual void	clear(const File& file) = 0;
					// Clear management data
	virtual void	recordReference(File& file, ReferenceType type,
			  bool isMeaningless) = 0;
					// Keep track of a file reference
	virtual void	inherit(const File& file, const File& parent) = 0;
					// Inherit manager data from parent
	virtual void	cleanReferences() = 0;
					// Clean out any dangling references
    protected:
	virtual ManagerDataBase&
			managerData(File& file)
			  const = 0;
					// Provide recursive data access
	virtual const ManagerDataBase&
			managerData(const File& file)
			  const = 0;
					// Provide recursive data access
    };

/*
 * Base class for managers that store no data in the file structure.
 */
class EmptyManagerBase : public ManagerBase
    {
    public:
	inline		EmptyManagerBase(const char* managerName,
			  const char* managerDefaults);
	virtual		~EmptyManagerBase();
    protected:
	virtual ManagerDataBase&
			managerData(File& file)
			  const;
					// Panic on data-access attempts
	virtual const ManagerDataBase&
			managerData(const File& file)
			  const;
					// Panic on data-access attempts
    };

/*
 * Base class for data for hoarding managers.
 */
class HoardingManagerDataBase : public ManagerDataBase
    {
    public:
	inline		HoardingManagerDataBase();
	virtual		~HoardingManagerDataBase();
	/*
	 * This should return HoardingManagerBase&, but that
	 * requires a cross-reference to a derived class, which at
	 * least g++ can't handle.
	 */
	virtual ManagerBase&
			manager() const = 0;
					// Return pointer to our manager
	inline unsigned long
			store(File& file);
					// Arrange to have file stored
	inline void	unstore(File& file);
					// Arrange for file to be unstored
	inline bool	isStored() const; // Tell if file is currently stored
	virtual void	interpretStatus(File& file);
					// Apply manager-specific
					// ..interpretation to file status
	virtual void	recordMissedReference(File& file,
			  ReferenceType referenceType, bool isMeaningless);
					// Record that a file miss happened
	virtual void	recordReference(File& file, ReferenceType type,
			  bool isMeaningless);
					// Keep track of a file reference
	virtual ostream& show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
	virtual ByteSwappedStream&	// Read binary version from stream
			read(ByteSwappedStream& stream, File& file);
	virtual ByteSwappedStream&	// Write binary version to stream
			write(ByteSwappedStream& stream,
			  const File& file) const;
	virtual void	clear(const File& file);
					// Clear management data
	virtual void	inherit(const File& file, const File& parent);
					// Inherit manager data from parent
	virtual void	cleanReferences();
					// Clean out any dangling references
	inline ReferenceCount
			lastReferenceTime() const;
					// Return time of last reference
	virtual void	setReferenceTime(long change, int whence);
					// Set (fake) file reference time
	virtual void	relationsChanged();
					// Note that relations have changed
    protected:
	virtual HoardingManagerDataBase&
			managerData(File& file)
			  const = 0;
					// Provide recursive data access
	virtual const HoardingManagerDataBase&
			managerData(const File& file)
			  const = 0;
					// Provide recursive data access
	/*
	 * Private data
	 */
	enum HoardingManagerStorageState
	    {
	    STORED,			// We would hoard this file
	    NOT_STORED,			// We wouldn't hoard this file
	    NOT_STORED_BUT_MISSED	// We wouldn't hoard, but miss happened
	    };
	ReferenceCount	lastReference;	// Last time we think file was ref'ed
	HoardingManagerStorageState
			ourStorageState; // How we would store this file
    };

/*
 * Base class for managers that store hoarding data in the file structure.
 */
class HoardingManagerBase : public ManagerBase
    {
    public:
			HoardingManagerBase(const char* managerName,
			  const char* managerDefaults);
	virtual		~HoardingManagerBase();
	virtual HoardingManagerDataBase&
			managerData(File& file)
			  const = 0;
					// Provide data access
	virtual const HoardingManagerDataBase&
			managerData(const File& file)
			  const = 0;
					// Provide data access
	virtual long	writeParameters(ByteSwappedStream& stream,
			    bool reallyWrite)
			  const;
					// Write management params to stream
	virtual long	readParameters(ByteSwappedStream& stream,
			  long parameterBytes);
					// Read management params from stream
	inline ReferenceCount
			currentReferenceTime() const;
					// Return current reference number
	inline void	increaseCurrentReferenceTime(ReferenceCount newTime);
					// Set current reference no. to max
	inline ReferenceCount
			nextReferenceTime(); // Produce next reference number
	inline ReferenceCount
			getBaseReferenceTime() const;
					// Return baseReference
	inline void	resetBaseReferenceTime();
					// Reset baseReference to equal
					// ..the current file reference value
	virtual void	setReferenceTime(File& file, long change, int whence);
					// Set (fake) file reference time
	static ostream&	showAllStatistics(ostream& stream);
	virtual ostream& showStatistics(ostream& stream) const;
					// Display collected quality statistics
	static void	resetAllStatistics();
					// Reset statistics for all managers
	virtual void	resetStatistics();
					// Reset quality statistics
	virtual ostream& requiredSize(ostream& stream,
			    ostream* duplicateStream)
			  const = 0;
					// Display hoard size needed for
					// ..no misses
	virtual void	interpretStatus(File& file) const;
					// Apply manager-specific
					// ..interpretation to file status
	virtual void	recordMissStatistics(File& file,
			  ReferenceType referenceType, bool isMeaningless);
					// Record statistics about a file miss
    private:
	/*
	 * Private data
	 */
	ReferenceCount	currentReference; // Current reference number
	ReferenceCount	firstMissReference; // Reference no. of first miss
	unsigned long	missCount;	// Count of total hoard misses
	ReferenceCount	referenceBase;	// Base that reference no.'s
					// ..are referred to
    };

/*
 * Class used to iterate across all known hoarding managers
 */
class HoardingManagerIterator : public ManagerIterator
    {
    public:
	inline		HoardingManagerIterator();
	inline		~HoardingManagerIterator();
	inline void	reset();	// Reset to start of manager table
	inline HoardingManagerIterator&
			operator++();	// Step to next entry
	inline HoardingManagerBase&
			operator*();	// Return referenced manager
    private:
			// I'm disallowing copying/assignment only because
			// I don't want to write them yet.  There's no
			// fundamental reason not to provide them.
			HoardingManagerIterator(
			  const HoardingManagerIterator&);
	HoardingManagerIterator& operator=(const HoardingManagerIterator&);
    };

#define _MANAGERBASE_ICC_NEEDED

#endif // _MANAGERBASE_HH
