#ifndef _MODLRUMANAGER_HH
#define _MODLRUMANAGER_HH
/*
 * $Id: modlrumanager.hh,v 1.8 1998/10/23 23:15:48 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: modlrumanager.hh,v $
 * Revision 1.8  1998/10/23 23:15:48  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.7  1998/05/14 01:15:38  geoff
 * Add MIME-encoding support.
 *
 * Revision 1.6  1998/03/05 03:18:17  geoff
 * Get rid of DERIVED_CLASS_BUG.
 *
 * Revision 1.5  1996/10/04 07:12:17  geoff
 * Add the new linear LRU manager.
 *
 * Revision 1.4  1996/05/03 01:12:44  geoff
 * Update the copyright.  Add new routines and modify declarations now
 * that instructions can be reread.
 *
 * Revision 1.3  1996/04/10  01:28:41  geoff
 * Add the new ManagerData::manager() routine.  Add support for
 * manager-specific storage statuses.
 *
 * Revision 1.2  1996/03/19  18:43:09  geoff
 * Get rid of the show/read/write/clear-manager-data and record-reference
 * routines now that they can be inherited.  Inherit from
 * HoardingManagerBase.  Add support for the DERIVED_CLASS_BUG ifdef.
 *
 * Revision 1.1  1996/02/02  17:16:42  geoff
 * Initial revision
 *
 */

#include "basictypes.hh"
#include "lrumanager.hh"
#include "managerbase.hh"

class BoundedLRU_Manager;
class BoundedLRU_ManagerData;
class ByteSwappedStream;
class File;
class LinearLRU_Manager;
class LinearLRU_ManagerData;
class WeightedLRU_Manager;
class WeightedLRU_ManagerData;
class ostream;

/*
 * Class to hold bounded-LRU management data for files
 */
class BoundedLRU_ManagerData : public LRU_ManagerData
    {
	friend		BoundedLRU_Manager;
					// So compareLRU can access us
    public:
	inline		BoundedLRU_ManagerData();
	virtual		~BoundedLRU_ManagerData();
	/*
	 * This should return BoundedLRU_Manager&, but that requires a
	 * cross-reference to a derived class, which g++, at least,
	 * can't handle.
	 */
	virtual HoardingManagerBase&
			manager() const; // Return pointer to our manager
	virtual ByteSwappedStream&	// Read binary version from stream
			read(ByteSwappedStream& stream, File& file);
	virtual ByteSwappedStream&	// Write binary version to stream
			write(ByteSwappedStream& stream,
			    const File& file)
			  const;
	virtual void	inherit(const File& file, const File& parent);
					// Inherit information from parent
	inline void	setBound(ReferenceCount newBound);
					// Set a new bound
	friend ostream&			// Write printable version
			operator<<(ostream& stream,
			  const BoundedLRU_ManagerData& data);
	ostream&	show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
    private:
			BoundedLRU_ManagerData(
			  const BoundedLRU_ManagerData& that);
					// Disable copy
	BoundedLRU_ManagerData&	operator=(const BoundedLRU_ManagerData& that);
					// Disable assignment
	virtual BoundedLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide recursive data access
	virtual const BoundedLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide recursive data access
	/*
	 * Private data
	 */
	ReferenceCount	bound;		// Bound applied to LRU age
    };

class BoundedLRU_Manager : public LRU_Manager
    {
    friend BoundedLRU_ManagerData;
    public:
			BoundedLRU_Manager(const char* managerName,
			  const char* managerDefaults);
	virtual		~BoundedLRU_Manager();
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
					// Change a management parameter
	virtual void	showParameters(class ostream& stream) const;
					// Display management parameters
	virtual void	readInstructions(const char* fileName) const;
					// Read manager-specific instructions
	virtual void	clearInstructions(File& file) const;
					// Clear manager-specific instructions
    private:
			BoundedLRU_Manager(const BoundedLRU_Manager& that);
					// Disable assignment
	BoundedLRU_Manager&
			operator=(const BoundedLRU_Manager& that);
					// Disable copy
	virtual BoundedLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const BoundedLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide data access
	static int	compareLRU(class File** file1, class File** file2);
    };

/*
 * Class to hold linear-LRU management data for files
 */
class LinearLRU_ManagerData : public LRU_ManagerData
    {
	friend		LinearLRU_Manager;
					// So compareLRU can access us
    public:
	inline		LinearLRU_ManagerData();
	virtual		~LinearLRU_ManagerData();
	/*
	 * This should return LinearLRU_Manager&, but that requires
	 * a cross-reference to a derived class, which g++, at least,
	 * can't handle.
	 */
	virtual HoardingManagerBase&
			manager() const; // Return pointer to our manager
	ByteSwappedStream&		// Read binary version from stream
			read(ByteSwappedStream& stream, File& file);
	ByteSwappedStream&		// Write binary version to stream
			write(ByteSwappedStream& stream,
			  const File& file) const;
	virtual void	inherit(const File& file, const File& parent);
					// Inherit information from parent
	inline void	setLineParameters(double newSlope,
			  double newIntercept, double newBound);
					// Set new linear parameters
	friend ostream&			// Write printable version
			operator<<(ostream& stream,
			  const LinearLRU_ManagerData& data);
	ostream&	show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
    private:
			LinearLRU_ManagerData(
			  const LinearLRU_ManagerData& that);
					// Disable copy
	LinearLRU_ManagerData&
			operator=(const LinearLRU_ManagerData& that);
					// Disable assignment
	virtual LinearLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide recursive data access
	virtual const LinearLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide recursive data access
	/*
	 * Private data
	 */
	double		intercept;	// Intercept of linear transformation
	double		slope;		// Slope of linear transformation
	double		bound;		// Bound on transformed value
    };

class LinearLRU_Manager : public LRU_Manager
    {
    friend LinearLRU_ManagerData;
    public:
			LinearLRU_Manager(const char* managerName,
			  const char* managerDefaults);
	virtual		~LinearLRU_Manager();
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
	virtual void	showParameters(class ostream& stream) const;
					// Display management parameters
	virtual void	readInstructions(const char* fileName) const;
					// Read manager-specific instructions
	virtual void	clearInstructions(File& file) const;
					// Clear manager-specific instructions
    private:
			LinearLRU_Manager(const LinearLRU_Manager& that);
					// Disable assignment
	LinearLRU_Manager&
			operator=(const LinearLRU_Manager& that);
					// Disable copy
	virtual LinearLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const LinearLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide data access
	static int	compareLRU(class File** file1, class File** file2);
    };

/*
 * Class to hold weighted-LRU management data for files
 */
class WeightedLRU_ManagerData : public LRU_ManagerData
    {
	friend		WeightedLRU_Manager;
					// So compareLRU can access us
    public:
	inline		WeightedLRU_ManagerData();
	virtual		~WeightedLRU_ManagerData();
	/*
	 * This should return WeightedLRU_Manager&, but that requires
	 * a cross-reference to a derived class, which g++, at least,
	 * can't handle.
	 */
	virtual HoardingManagerBase&
			manager() const; // Return pointer to our manager
	ByteSwappedStream&		// Read binary version from stream
			read(ByteSwappedStream& stream, File& file);
	ByteSwappedStream&		// Write binary version to stream
			write(ByteSwappedStream& stream,
			  const File& file) const;
	virtual void	inherit(const File& file, const File& parent);
					// Inherit information from parent
	inline void	setWeight(double newWeight);
					// Set a new weight
	friend ostream&			// Write printable version
			operator<<(ostream& stream,
			  const WeightedLRU_ManagerData& data);
	ostream&	show(ostream& stream, ShowOptions options,
			  const char* managerName) const;
					// Show manager data in printable form,
					// ..with options
    private:
			WeightedLRU_ManagerData(
			  const WeightedLRU_ManagerData& that);
					// Disable copy
	WeightedLRU_ManagerData&
			operator=(const WeightedLRU_ManagerData& that);
					// Disable assignment
	virtual WeightedLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide recursive data access
	virtual const WeightedLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide recursive data access
	/*
	 * Private data
	 */
	double		weight;		// Weight applied to LRU age
    };

class WeightedLRU_Manager : public LRU_Manager
    {
    friend WeightedLRU_ManagerData;
    public:
			WeightedLRU_Manager(const char* managerName,
			  const char* managerDefaults);
	virtual		~WeightedLRU_Manager();
	virtual void	manage(const InternalTimer& now,
			  ByteSwappedStream* dumpStream, bool encodeFilenames);
					// Run hoard management
	virtual void	setParameter(const char* description);
	virtual void	showParameters(class ostream& stream) const;
					// Display management parameters
	virtual void	readInstructions(const char* fileName) const;
					// Read manager-specific instructions
	virtual void	clearInstructions(File& file) const;
					// Clear manager-specific instructions
    private:
			WeightedLRU_Manager(const WeightedLRU_Manager& that);
					// Disable assignment
	WeightedLRU_Manager&
			operator=(const WeightedLRU_Manager& that);
					// Disable copy
	virtual WeightedLRU_ManagerData&
			managerData(File& file)
			  const;
					// Provide data access
	virtual const WeightedLRU_ManagerData&
			managerData(const File& file)
			  const;
					// Provide data access
	static int	compareLRU(class File** file1, class File** file2);
    };

#define _MODLRUMANAGER_ICC_NEEDED

#endif // _MODLRUMANAGER_HH
