#ifndef _PACKET_HH
#define _PACKET_HH
/*
 * $Id: packet.hh,v 1.12 1999/02/04 21:05:13 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: packet.hh,v $
 * Revision 1.12  1999/02/04 21:05:13  geoff
 * Add Inputlist::hasFd.
 *
 * Revision 1.11  1998/10/23 23:15:49  geoff
 * Add support for the new automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.10  1997/12/18 01:25:26  geoff
 * Get rid of some obsolete development ifdefs.
 *
 * Revision 1.9  1997/12/04 23:16:29  geoff
 * Add InputList::removeAll.
 *
 * Revision 1.8  1996/10/04 07:12:22  geoff
 * Don't declare fd_set as an anonymous class; it screws up some
 * compilation situations.
 *
 * Revision 1.7  1996/09/11 23:00:15  geoff
 * Add REPORT_SELECTED_FAILURES support.  Convert from dlists to slists.
 *
 * Revision 1.6  1996/05/03  01:12:54  geoff
 * Update the copyright.
 *
 * Revision 1.5  1996/02/02  18:29:16  geoff
 * Declare inline functions as such in the class definition.
 *
 * Revision 1.4  1995/12/19  01:02:41  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.3  1995/12/19  00:34:49  geoff
 * Add process-tracking support.
 *
 * Revision 1.2  1995/11/22  00:22:14  geoff
 * Get rid of the TIME_SORTED_INPUT and HAND_BYTE_SWAPPING ifdefs.
 *
 * Revision 1.1  1995/11/10  02:09:40  geoff
 * Initial revision
 *
 */

/*
 * Classes to handle packets incoming to the correlator, including
 * time-sorting them.
 */

#include "communication.hh"
#include "hashtable.hh"
#include "list.hh"
#include "process.hh"
#include <sys/time.h>
#include <sys/types.h>

class ByteSwappedStream;
class InputList;

/*
 * Class used to handle input packets from the observer.  Note
 * that "path1" and "path2" are both pointers into "buffer".
 */
class InputPacket
    {
    friend		InputList;
    public:
			InputPacket(ByteSwappedStream* stream,
			   bool blockForInput);
			InputPacket(int fd, bool blockForInput);
			~InputPacket();
	bool		get();		// Collect a packet from the stream
	inline void	consume();	// Consume a packet (mark it invalid)
	bool		write(ByteSwappedStream& stream);
					// Write a packet to a stream
	inline bool	succeeded() const; // Return packet success status
	inline ReferenceType
			packetType() const; // Return type of the packet
	inline const struct timeval&
			packetTime() const; // Return packet's timestamp
	inline unsigned long
			pid() const;	// Return packet's process ID
	inline unsigned long
			parentPid() const; // Return packet's parent PID
	inline Process*	process() const; // Return current process
	inline void	processExit();	// Handle exit of a process
	inline void	loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
	inline const char*
			path1() const;	// Return first path
	inline int	path1Length() const; // Return length of first path
	inline const char*
			path2() const;	// Return second path
	inline int	path2Length() const; // Return length of second path
	inline bool	isBlocking() const; // Tell whether stream blocks
	inline void	dontBlock();	// Mark stream as no-longer-blocking
    private:
	/*
	 * Private data
	 */
	ByteSwappedStream* sourceStream; // Stream this packet came from
	bool		blockingStream;	// True if we block for input on stream
	bool		deleteStreamOnDestruction; // True if we "own" stream
	bool		valid;		// True if this packet is valid
	bool		callSucceeded;	// True if reference was successful
	ReferenceType	type;		// Type of reference packet represents
	struct timeval	timestamp;	// Timestamp assigned to packet
	Process*	currentProcess;	// Structure for referencing process
	int		length1;	// Length of first path in packet
	const char*	pathName1;	// First path in packet
	int		length2;	// Length of second path in packet
	const char*	pathName2;	// Second path in packet
	char		buffer[(MAXPATHLEN + 1) * 2];
					// Buffer for pathnames in packet
	ProcessTable	processes;	// Table of known processes
    };

/*
 * Class used to manage multiple input packet streams.  The primary
 * feature of this class is that it attempts to deliver packets in
 * timestamp order.
 */
class InputList
    {
    public:
	inline		InputList();
			~InputList();
	void		addFd(int fd, bool blockForInput);
					// Add a new packet source
	bool		hasFd(int fd) const;
					// See if fd is in list's fd_set
	void		removeSource(InputPacket* oldSource);
					// Remove a packet source
	void		removeAll();	// Remove all packet sources
	void		dontBlock(InputPacket* subjectPacket);
					// Mark a stream as no-longer-blocking
	InputPacket*	get(fd_set& control_FDs, fd_set& badFDs,
			  struct timeval* waitTime);
					// Return next input packet,
					// ..or NULL if none to be
					// ..processed
	void		parseFds(const char* descriptor);
					// Parse ASCII into an input list
	char*		saveFds() const; // Turn input list into ASCII string
	void		loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
    private:
	SList		packets;	// Input packets and streams
    };

#define _PACKET_ICC_NEEDED

#endif // _PACKET_HH
