#ifndef _PROCESS_HH
#define _PROCESS_HH
/*
 * $Id: process.hh,v 1.8 1998/10/23 23:15:51 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: process.hh,v $
 * Revision 1.8  1998/10/23 23:15:51  geoff
 * Fix things to compile with -pedantic.  Add support for the new
 * automatically-created zzinlinefunctions.hh.
 *
 * Revision 1.7  1996/09/11 23:00:17  geoff
 * Convert from dlists to slists.
 *
 * Revision 1.6  1996/06/11  07:05:04  geoff
 * Get rid of USE_PROBABLY_MEANINGLESS.
 *
 * Revision 1.5  1996/06/11  06:13:26  geoff
 * Add hasDirectoryOpen, getcwd support, and always/currently meaningless
 * routines.
 *
 * Revision 1.4  1996/05/03  01:12:57  geoff
 * Update the copyright.  Add the open-directory count,
 * probablyMeaningless, and PF_INFERRED_MEANINGLESS.
 *
 * Revision 1.3  1996/02/02  18:29:18  geoff
 * Declare inline functions as such in the class definition.  Add
 * isMeaningless and the meaningless flag.
 *
 * Revision 1.2  1995/12/19  01:02:42  geoff
 * Get rid of the TRACK_PROCESSES ifdef.
 *
 * Revision 1.1  1995/12/19  00:05:36  geoff
 * Initial revision
 *
 */

/*
 * Classes used to track active processes on the observed machines
 */
#include "basictypes.hh"
#include "filehistory.hh"
#include "list.hh"

class File;
class FinalProcessTableDestructor;
class Process;
class ProcessTable;
class ostream;

/*
 * Class used to track a single process
 */
class Process
    {
	friend		FinalProcessTableDestructor;
	friend		ProcessTable;
	friend ostream&	operator<<(ostream& stream, const Process& process);
    public:
	inline		Process(unsigned long id);
	inline		~Process();
	inline unsigned long
			pid() const;	// Return process ID
	inline Process*	parent() const;	// Return parent of this process
	inline File*	program() const; // Return program that is running
	inline FileHistory&
			fileHistory();	// Return file-history table
	inline void	exec(File* file); // Set executing file
	void		loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
	inline void	openFile(const File& file);
					// Record that a file is open
	inline bool	closeFile(const File& file);
					// Record that a file was closed
	inline bool	hasDirectoryOpen() const;
					// Test whether any directory is open
	inline bool	isOpenByProcess(const File& file) const;
					// Test whether process has file open
	inline bool	alwaysMeaningless() const;
					// Test whether refs are always
					// ..meaningless
	inline bool	currentlyMeaningless() const;
					// Test whether refs are currently
					// ..meaningless
	inline void	setInGetcwd(bool value);
					// Set getcwd flag
	inline bool	inGetcwd() const; // Test getcwd flag
    private:
			Process(Process& that); // Disable assignment
	Process&	operator=(const Process& that); // Disable copy
	/*
	 * Private data
	 */
	unsigned long	processId;	// PID for this process on its machine
	int		flags;		// Flag bits; see below
	Process*	parentProcess;	// Process that spawned us
	File*		runningProgram;	// Program that is executing
	FileHistory	history;	// History of process' references
	SList		openFiles;	// Files that process has open
	unsigned long	openDirectoryCount; // Count of directories that the
					// ..process has open
#ifdef CHECK_PROCESS_EXITS
	bool		exitCalled;	// True after processExit runs on us
#endif // CHECK_PROCESS_EXITS
	/*
	 * Process flags.  These should be defined using 'const int',
	 * but ANSI C++ forbids it, so I'll kludge it with an enum.
	 */
	enum
	    {
	    PF_MEANINGLESS = 0x01,	// References from this process and its
					// ..children are meaningless
	    PF_IN_GETCWD = 0x02,	// We think that this process
					// ..is in the midst of doing
					// ..a getcwd.
	    PF_INHERIT_LIST = PF_MEANINGLESS
					// Flags inherited from parent procs
	    };
    };

/*
 * Dummy class used solely to provide an on-exit destructor for
 * ProcessTable.
 */
class FinalProcessTableDestructor
    {
	friend		ProcessTable;
    public:
			~FinalProcessTableDestructor();
    private:
			FinalProcessTableDestructor();
    };

/*
 * Class used to keep track of all processes on a given observed
 * machine.  Note that there is one instance of a ProcessTable per
 * input stream.
 */
class ProcessTable : private HashTable
    {
	friend		FinalProcessTableDestructor;
    public:
			ProcessTable();
			~ProcessTable();
	Process*	findProcess(unsigned long processId,
			  unsigned long parentProcessId, bool create);
					// Find or create a process entry
	void		processExit(Process* process);
					// Record the exit of a given process
	void		loseForgettingPointers();
					// Eliminate pointers to files being
					// forgotten
    private:
	static unsigned int processHash(Process* process);
					// Hash process-table entries
	static int	processCompare(Process* process1, Process* process2);
					// Compare process-table entries, for
					// ..hash-table routines
	/*
	 * Static data
	 */
	static FinalProcessTableDestructor
			finalProcessTableDestructor;
	static Process*	newEntry;	// Scratch for table lookups/inserts
    };

#define _PROCESS_ICC_NEEDED

#endif // _PROCESS_HH
