#ifndef lint
static const char Rcs_Id[] =
    "$Id: sanitize.cc,v 1.5 2005/04/13 21:03:02 geoff Exp $";
#endif

/*
 * $Id: sanitize.cc,v 1.5 2005/04/13 21:03:02 geoff Exp $
 *
 * UCLA Ficus Project
 *
 * Copyright (c) 1997 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Sanitize a Seer observer trace file or an ASCII trace file.
 *
 * Usage:
 *
 * sanitize [-c] [-d] [-f regex-file] ... [-m seed] [-p c] [-s c] [-S] [-u]
 *   [-v] [input-files]
 *
 * Options:
 *
 *	-c	Automatically break components at special characters.
 *		This reduces the number of regular expressions required.
 *	-d	Sanitize into decimal instead of base 64.
 *	-f	Specify a file containing controlling regular
 *		expressions (see below).
 *	-m	Use seeded MD5 digests instead of a hash table.  The seed is
 *		an arbitrary string of characters; at least 64 bits of
 *		entropy are recommended.  A good source of the
 *		characters is to encode /dev/random into printable
 *		ASCII using a base-64 encoding.
 *	-p c	Specify a different prefix character for the sanitized
 *		strings.  If the null string is given, no prefix will
 *		be used.
 *	-s c	Specify a different suffix character for the sanitized
 *		strings.  If the null string is given, no suffix will
 *		be used.
 *	-S	The input files are Seer observer dump files rather
 *		than standardized ASCII input files.  In this case,
 *		the output will also be in Seer format.
 *	-u	Sanitize UIDs in Seer dump files.
 *	-v	Report statistics to cerr at the end of the run.
 *
 * By default, input and output files are in a standardized ASCII
 * format that can be easily produced by any program.  Each line in a
 * standardized file consists of an odd number of ASCII integers
 * separated by whitespace, and followed by an arbitrary-length
 * string.  The first ASCII integer gives the number of pairs of
 * integers following.  Each subsequent pair identifies one substring
 * that is to be sanitized by giving a zero-indexed offset from the
 * beginning of the arbitrary string, and the length of the substring.
 * The final pair must be followed by a single whitespace character
 * and then the arbitrary string.
 *
 * Examples:
 *
 *	0 abcdef
 *		The string "abcdef" is passed through without
 *		sanitization.
 *	1 2 3 abcdef
 *		The string "cde" (offset 2, length 3) is sanitized.
 *		The remainder of the input ("ab" and "f") is passed
 *		through unchanged.
 *	3 0 3 3 1 7 2 abcdef ghi
 *		The strings "abc", "d", and "gh" are sanitized
 *		independently.  The strings "ef" and "i" are passed
 *		through unchanged.
 *
 * The standardized output file has the same format as the
 * standardized input file, with the length fields for the sanitized
 * strings having been adjusted appropriately.  For example, if all
 * sanitized output strings have length 3, the last example line above
 * might be changed into:
 *
 *	3 0 3 3 3 9 3 |0||1|ef |2|i
 *
 * The regex-files contain lists of regular expressions that will
 * match describe acceptable pathnames, one per line, as described
 * further below.  The regular-expression files are combined and
 * loaded into the internal r.e. table (in the order given).  The
 * input files are then read and sanitized, with the results written
 * to standard output.
 *
 * The regular-expression file can hold two types of expressions: an
 * "accept" pattern or a "reject" pattern.  "Accept" patterns are
 * marked by the word "pass" in the first field of the line; "reject"
 * patterns are marked by "clean."  Each pathname is compared against
 * all expressions, in the order given in the file.  If a "reject"
 * pattern matches, the substrings matched by the subexpressions of
 * the pattern (or the string matched by the entire pattern, if there
 * are no subexpressions) is marked for sanitization.  If an "accept"
 * pattern matches, the corresponding subexpressions are removed from
 * the sanitization list.
 *
 * The beginning sanitization list contains the entire path, so that
 * the entire path will be cleaned unless an "accept" pattern
 * specifies otherwise.
 *
 * All output entries are written in the current seer buffer format.
 *
 * When a pathname is sanitized, all or part of it is replaced by
 * generated strings, as follows: within the section of the path that
 * will be sanitized, each sequence of alphanumeric characters is
 * looked up in a hash table.  If the component is found, it is
 * replaced by the string from the table; otherwise a new entry is
 * created.  The replacement strings are generated by a sequence
 * generator that produces X0, X1, ..., X10, X11, ...
 *
 * The result of this is that the same pathname component is always
 * replaced by the same sanitized string regardless of where it
 * appears, and the number of components does not change.  For
 * example, if "bin" translates to "X0", then "/usr/bin" might be
 * sanitized as "/X3/X0", while "/foo/bar/bin" became "/X9/X70/X0".
 * This can be considered as a feature or a bug, depending on your
 * point of view.  Note that, regardless of sanitization, an
 * intelligent investigator can probably figure out which paths are
 * /usr/bin, /bin, and similar popular directories, and possible even
 * identify some commonly used programs such as shells.  With quite a
 * bit more work, many programs can be identified from their
 * system-call signature (a similar method has been used to detect
 * security breaches).
 *
 * Examples:
 *
 * To sanitize everything except /usr/bin, but hide the identities of
 * programs inside /usr/bin:
 *
 *	pass ^/usr/bin/\|^/usr/bin$
 *
 * To accept everything, but clean the string "geoff" everywhere (even
 * if it is a substring of a longer string):
 *
 *	pass ^.*$
 *	clean geoff
 *
 * To accept everything in /usr, including their full pathnames, but
 * hide game files (note: the [/] is to prevent ending the C comment
 * prematurely; it's not necessary in a real file):
 *
 *	pass ^/usr/.*$\|^/usr$
 *	clean ^/usr/.*[/]games$\|/usr.*[/]games/.*$
 *
 * BUGS AND RESTRICTIONS:
 *
 * For historical reasons having to do with the way the Seer
 * file-access functions work, this program makes use of the g++
 * feature that allows a C++ program to intermix the use of C-style
 * malloc and stdio with C++-style new and streams.
 */

/*
 * $Log: sanitize.cc,v $
 * Revision 1.5  2005/04/13 21:03:02  geoff
 * Major improvements to make it Seer-independent and implement features
 * described in the UCSC tech report.
 *
 * Revision 1.4  1998/10/23 23:15:53  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.3  1998/04/09 00:33:03  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.2  1998/01/31 05:38:54  geoff
 * Correct the capitalization of a class name.  Don't get stuck in an
 * infinite loop when a regular expression matches the null string (this
 * can happen if a pathname contains a newline).  Generate names of the
 * form Y%dY, to avoid accidental collisions with X11 and possible
 * all-alphabetic names that might have been generated by X%x.
 *
 * Revision 1.1  1998/01/29 20:42:05  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "hashtable.hh"
#include "list.hh"
extern "C"
    {
#include "md5.h"    
    };
#include "simplestring.hh"
#include <ctype.h>
#include <errno.h>
#include <fstream.h>
#include <iostream.h>
/*
 * This program uses the GNU regular-expression library.  This is not
 * by choice; I'd personally prefer the more portable POSIX interface.
 * But the GNU one is what's available on my machines, and I'm in a
 * hurry.
 */
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <seer/seer.h>
#include <sys/param.h>

/*
 * Entries in the component-name hash tables.
 *
 * Each pathname component that gets translated is entered into a hash
 * table.  This structure holds the original component and its
 * translation.
 */
class ComponentTranslation
    {
    public:
			ComponentTranslation();
	SimpleString	original;	// Original component
	SimpleString	rewrite;	// Rewritten component
    };

/*
 * Entries in the UID hash tables.
 *
 * Each UID that gets translated is entered into a hash
 * table.  This structure holds the original UID and its
 * translation.
 */
class UID_Translation
    {
    public:
			UID_Translation(unsigned short original_);
	unsigned short	original;	// Original UID
	unsigned short	rewrite;	// Rewritten UID
	static unsigned short
			nextUID;
    };

/*
 * Next UID to assign (more accurately, last UID assigned)
 */
unsigned short		UID_Translation::nextUID = 0;

int			main(int argc, char* argv[]);
					// Sanitize an observer dump
static void		usage();	// Issue a usage message
void			nomem(const char* where, int line, unsigned int size);
					// Report memory exhaustion
static bool		loadRegex(const char* fileName);
					// Load expressions from a file
static bool		sanitizeSeerFile(FILE* file, bool sanitizeUID);
					// Sanitize one Seer file
static bool		sanitizeASCII_File(istream& file,
			  const char* fileName);
					// Sanitize one ASCII file
static SimpleString	maybeSanitizePath(const SimpleString& path);
					// Check regexes and maybe sanitize
static SimpleString	sanitizePath(const SimpleString& dirtyPath,
			  bool* mask);
					// Sanitize one pathname
static SimpleString	sanitizePortion(const SimpleString& dirtyPath,
			  unsigned int start, unsigned int end);
					// Sanitize a substring
static SimpleString	sanitizeComponent(const SimpleString& component);
					// Sanitize a pathname component
static unsigned int	componentHash(ComponentTranslation* component);
					// Hash a translation entry
static int		componentCompare(ComponentTranslation* component1,
			  ComponentTranslation* component2);
					// Compare two translation entries
static unsigned short	cleanUID(unsigned short uid);
					// Sanitize a UID
static unsigned int	UID_Hash(UID_Translation* uid);
					// Hash a UID entry
static int		UID_Compare(UID_Translation* uid1,
			  UID_Translation* uid2);
					// Compare two UID entries
static SimpleString	encode(unsigned int value, unsigned int numberBase,
			  int minDigits = 0);
					// Encode a number into a base
static seerbuf_t*	recreateEntry(seerbuf_t* entry,
			  const SimpleString& newPath1,
			  const SimpleString& newPath2);
					// Recreate an observer trace entry

/*
 * Types of expression specifications
 */
enum ExpressionType
    {
    ACCEPT,				// Accept matching expressions
    REJECT				// Reject matching expressions
    };

/*
 * Structure used to hold GNU regexp information.  This is needed
 * partly because of undocumented requirements in the !@#$%^&
 * interface.
 */
typedef struct
    {
    ExpressionType	type;		// Whether to accept or reject
    re_pattern_buffer	expression;	// Compiled regular expression
    struct re_registers	matches;	// Positions of matches
    }
			expressionControl;


/*
 * Characters used to delimit sanitized integers.
 */
char			sanitizePrefix = '|';
char			sanitizeSuffix = '|';

/*
 * Width of a byte
 */
const unsigned int	BYTE_WIDTH = 8;

/*
 * Numeric base of encoding system.  To allow us to use base 64, we add
 * underscores and dashes to the alphanumeric space.  That's safe
 * since all encoded strings are surrounded by sanitizePrefix/Suffix.
 */
const unsigned int	ENCODING_BASE_BITS = 6;
const unsigned int	ENCODING_BASE = 1 << ENCODING_BASE_BITS;
const unsigned int	ENCODING_MASK = (1 << ENCODING_BASE_BITS) - 1;

/*
 * Base-64 encoding table
 */
static char encoding[ENCODING_BASE] =
    {
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
    'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
    'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
    '_', '-'
    };

/*
 * Size of a maximal encoding.  There are 2 single boundary
 * characters, and a 32-bit number is encoded into 32/6 + 1 = 6
 * characters.  For MD5, the maximal encoding is 16 characters plus a
 * boundary character.
 */
unsigned int		maximalEncoding = 2
			  + sizeof (unsigned long)
			    * BYTE_WIDTH / ENCODING_BASE_BITS
			  + 1;
const unsigned int	MD5_MAXIMAL_ENCODING = 16 + 1;

/*
 * Global variables.
 */
unsigned int		base(ENCODING_BASE);
					// Base to use for encoding output
static SimpleString	md5Seed;	// Seed for MD5 sanitization
static bool		passSpecials(false);
					// True to auto-pass special chars
char*			progname;	// Name we were run under
static DList		regexList(false); // List of regular expressions
static HashTable	translations(10000,
			  (unsigned int (*)(const void*))componentHash,
			  (int (*)(const void*, const void*))componentCompare);
static HashTable	UID_translations(10000,
			  (unsigned int (*)(const void*))UID_Hash,
			  (int (*)(const void*, const void*))UID_Compare);
static bool		useMD5(false);	// True to do MD5 sanitization

/*
 * External functions not declared in header files.
 */
extern seerbuf_t*	readent(FILE* replayfile);
extern int		seerclose(FILE* replayFile);
extern FILE*		seeropen(const char* fileName);
extern int		seerversion();
extern "C" char*	strerr(int errnum);

int main(				// Dump an observer trace file
    int			argc,		// Argument count
    char*		argv[])		// Argument vector
    {
    bool		sanitizeUID = false;
					// True to sanitize UID in Seer format
    bool		seerFormat = false;
					// True if input is in Seer format
    bool		verbose = false;
					// True to report statistics at end

    progname = argv[0];
    bool trouble = false;
    while (argc > 1  &&  argv[1][0] == '-')
	{
	argc--;
	argv++;
	switch (argv[0][1])
	    {
	    case 'c':
		{
		passSpecials = true;
		break;
		}
	    case 'd':
		{
		base = 10;
		break;
		}
	    case 'f':
		{
		if (--argc < 1)
		    usage();
		argv++;
		trouble |= loadRegex(argv[0]);
		break;
		}
	    case 'm':
		{
		if (--argc < 1)
		    usage();
		argv++;
		useMD5 = true;
		maximalEncoding = MD5_MAXIMAL_ENCODING;
		md5Seed = argv[0];
		break;
		}
	    case 'p':
		{
		if (--argc < 1  ||  strlen(argv[1]) > 1)
		    usage();
		argv++;
		sanitizePrefix = argv[0][0];
		break;
		}
	    case 's':
		{
		if (--argc < 1  ||  strlen(argv[1]) > 1)
		    usage();
		argv++;
		sanitizeSuffix = argv[0][0];
		break;
		}
	    case 'S':
		{
		seerFormat = true;
		break;
		}
	    case 'u':
		{
		sanitizeUID = true;
		break;
		}
	    case 'v':
		{
		verbose = true;
		break;
		}
	    default:
		usage();
		break;
	    }
	}

    if (trouble)
	return 1;

    if (seerFormat)
	{
	if (argc == 1)
	    {
	    FILE* infile = seeropen((char*)NULL);
	    trouble |= sanitizeSeerFile(infile, sanitizeUID);
	    }
	else
	    {
	    for (int argno = 1;  argno < argc;  argno++)
		{
		FILE* infile = seeropen(argv[argno]);
		if (infile == NULL)
		    {
		    int saveErrno = errno;
		    cerr << progname << ":  couldn't open '" << argv[argno]
		      << "': " << strerr(saveErrno) << '\n';
		    trouble = true;
		    }
		else
		    {
		    trouble |= sanitizeSeerFile(infile, sanitizeUID);
		    (void)seerclose(infile);
		    }
		}
	    }
	}
    else
	{
	if (sanitizeUID)
	    usage();
	if (argc == 1)
	    trouble |= sanitizeASCII_File(cin, "<stdin>");
	else
	    {
	    for (int argno = 1;  argno < argc;  argno++)
		{
		ifstream infile(argv[argno]);
		if (infile)
		    trouble |= sanitizeASCII_File(infile, argv[argno]);
		else
		    {
		    int saveErrno = errno;
		    cerr << progname << ":  couldn't open '" << argv[argno]
		      << "': " << strerr(saveErrno) << '\n';
		    trouble = true;
		    }
		}
	    }
	}

    if (verbose)
	{
	translations.showStatistics(cerr);
	cerr << '\n';
	if (sanitizeUID)
	    {
	    UID_translations.showStatistics(cerr);
	    cerr << '\n';
	    }
	}

    return trouble;
    }

static void usage()			// Issue a usage message and exit
    {
    cerr << "Usage: sanitize [-c] [-d] [-f regex-file] ... [-m] [-p c] [-s c] "
      << "[-S] [-u] [-v] [input-files]\n";
    exit(2);
    }

void nomem(			// Issue an out-of-memory message
    const char*		where,	// Routine that failed to allocate
    int			line,	// Line number of failure
    unsigned int	size)	// Size that was wanted
    {
    cout.flush();
    cerr << progname << ":  out of memory at " << line << " in " << where;
    if (size)
	cerr << ", couldn't get " << size << " bytes";
    cerr << "!\n";
    cerr.flush();
    abort();
    }

static bool loadRegex(			// Load regular expressions into table
    const char*		fileName)	// File containing expressions
    {
    ifstream file(fileName);
    if (!file)
	{
	int saveErrno = errno;
	cerr << progname << ":  couldn't open '" << fileName
	  << "': " << strerr(saveErrno) << '\n';
	return true;
	}
    bool trouble = false;
    SimpleString nextLine;
    while (nextLine.gets(file))
	{
	ExpressionType type;
	SimpleString nextRegex;
	if (nextLine.strlen() == 0  ||  nextLine.subString(0, 1) == "#")
	    continue;			// Ignore blank lines and comments
	else if (nextLine.subString(0, 5) == "pass ")
	    {
	    type = ACCEPT;
	    nextRegex = nextLine.subString(5, nextLine.strlen());
	    }
	else if (nextLine.subString(0, 6) == "clean ")
	    {
	    type = REJECT;
	    nextRegex = nextLine.subString(6, nextLine.strlen());
	    }
	else
	    {
	    cerr << progname << ": bad control line '" << nextLine
	      << "' in file '" << fileName << "'.\n";
	    continue;
	    }
#ifdef __sparc__
	/*
	 * GNU regex support is broken on SPARCs (or maybe I was
	 * getting bitten by all those stupid initialization
	 * assumptions).  To make it easier for me to test on those
	 * machine, I'll fake it in a very simple fashion.
	 */
	SimpleString* compiledRegex = new SimpleString(nextRegex);
	regexList.pushTail(compiledRegex);
#else // __sparc__
	/*
	 * Initializing for GNU regex is a pain...
	 */
	expressionControl* compiledRegex =
	  (expressionControl*)calloc(1, sizeof (expressionControl));
	if (compiledRegex == NULL)
	    nomem("loadRegex", __LINE__, sizeof *compiledRegex);
	compiledRegex->type = type;
	compiledRegex->expression.allocated = nextRegex.strlen();
	compiledRegex->expression.buffer =
	  (unsigned char*)calloc(compiledRegex->expression.allocated, 1);
	if (compiledRegex->expression.buffer == NULL)
	    nomem("loadRegex", __LINE__, compiledRegex->expression.allocated);
	compiledRegex->expression.translate = NULL;
	compiledRegex->expression.fastmap = (char*)malloc(256);

	compiledRegex->expression.fastmap_accurate = 0;
	compiledRegex->expression.regs_allocated = REGS_UNALLOCATED;
	const char* errorMessage =
	  re_compile_pattern((const char*)nextRegex, nextRegex.strlen(),
	    &compiledRegex->expression);
	if (errorMessage != NULL)
	    {
	    cerr << progname << ": " << errorMessage;
	    trouble = true;
	    free(compiledRegex->expression.buffer);
	    free(compiledRegex->expression.fastmap);
	    free(compiledRegex);
	    }
	regexList.pushTail(compiledRegex);
#endif // __sparc__
	}
    return trouble;
    }

static bool sanitizeSeerFile(		// Sanitize one Seer file
    FILE*		file,		// File to be sanitized
    bool		sanitizeUID)	// True to sanitize UIDs
    {
    seerbuf_t* entry;
    while ((entry = readent(file)) != NULL)
	{
	seerbuf_t* cleanEntry = entry;
	if (entry->flags & (SEER_RESTART | SEER_TIMESTAMP))
	    ;
	else
	    {
	    /*
	     * For coding convenience, we put pointers to the paths to be
	     * sanitized into "path1" and "path2".
	     */
	    SimpleString path1;
	    SimpleString path2;
	    switch (entry->callcode)
		{
		case SEER_SYS_open:
		case SEER_SYS_creat:
		case SEER_SYS_unlink:
		case SEER_SYS_execv:
		case SEER_SYS_execve:
		case SEER_SYS_chdir:
		case SEER_SYS_mknod:
		case SEER_SYS_mkfifo:
		case SEER_SYS_oldstat:
		case SEER_SYS_access:
		case SEER_SYS_stat:
		case SEER_SYS_lstat:
		case SEER_SYS_oldlstat:
		case SEER_SYS_readlink:
		case SEER_SYS_uselib:
		case SEER_SYS_symlink:
		case SEER_SYS_chroot:
		case SEER_SYS_truncate:
		case SEER_SYS_mkdir:
		case SEER_SYS_rmdir:
		case SEER_SYS_utime:
		case SEER_SYS_utimes:
		    {
		    path1 = (const char*)&entry->args[0];
		    break;
		    }
		case SEER_SYS_link:
		case SEER_SYS_rename:
		    {
		    path1 = (const char*)&entry->args[0];
		    /*
		     * Code stolen (and simplified) from findarg in observer.cc
		     */
		    path2 = (const char*)&entry->args[0]
		      + ((strlen(path1) + sizeof entry->args[0] - 1 + 1)
			/ sizeof entry->args[0]) * sizeof entry->args[0];
		    break;
		    }
		}
	    if (path1 != NULL)
		{
		SimpleString cleanPath1 = maybeSanitizePath(path1);
		SimpleString cleanPath2 = maybeSanitizePath(path2);
		cleanEntry = recreateEntry(entry, cleanPath1, cleanPath2);
		}
	    if (sanitizeUID)
		entry->uid = cleanUID(entry->uid);
	    }
	int entrySize = sizeof(seerbuf_t)
	  + (cleanEntry->argsize - 1) * sizeof cleanEntry->args[0];
	(void)fwrite((char*) cleanEntry, entrySize, 1, stdout);
	}
    return false;
    }

static bool sanitizeASCII_File(		// Sanitize one ASCII file
    istream&		file,		// File to be sanitized
    const char*		fileName)	// Name of file being sanitized
    {
    /*
     * For each line, first read the number of pairs using normal C++
     * input parsinig.  Then use that value to control the reading of
     * the remainder of the line.
     */
    int numPairs;
    while (file >> numPairs)
	{
	struct pairs
	    {
	    int			offset;
	    int			length;
	    SimpleString	sanitizedPath;
	    };
	pairs* pairList = new pairs[numPairs];
	for (int i = 0;  i < numPairs;  i++)
	    {
	    if (!(file >> pairList[i].offset)
	      ||  !(file >> pairList[i].length))
		{
		cerr << progname << ": short line while reading " << fileName
		  << '\n';
		return true;
		}
	    }

	/*
	 * Read the string that we want to sanitize.  The first
	 * character that we read is the separator.
	 */
	SimpleString sanitizeTarget;
	if (!sanitizeTarget.gets(file))
	    {
	    cerr << progname << ": short line while reading " << fileName
	      << '\n';
	    return true;
	    }
	else if (!isspace(sanitizeTarget[0]))
	    {
	    cerr << progname << ": missing whitespace while reading "
	      << fileName << '\n';
	    return true;
	    }

	/*
	 * Sanitize the chosen substrings.
	 */
	for (int i = 0;  i < numPairs;  i++)
	    pairList[i].sanitizedPath =
	      maybeSanitizePath(
	        sanitizeTarget.subString(pairList[i].offset + 1,
		pairList[i].length));


	/*
	 * Adjust the pair offsets and create the output string.
	 * Recall that the first character of the output string is the
	 * whitespace character.
	 */
	SimpleString outputString =
	  sanitizeTarget.subString(1, pairList[0].offset);
	for (int i = 0;  i < numPairs;  i++)
	    {
	    int newOffset = outputString.strlen();
	    outputString += pairList[i].sanitizedPath;
	    if (i < numPairs - 1)
		outputString +=
		  sanitizeTarget.subString(
		    1 + pairList[i].offset + pairList[i].length,
		      pairList[i + 1].offset
			- pairList[i].offset - pairList[i].length);
	    else
		outputString +=
		  sanitizeTarget.subString(
		    1 + pairList[i].offset + pairList[i].length,
		      sanitizeTarget.strlen());
	    pairList[i].offset = newOffset;
	    pairList[i].length = pairList[i].sanitizedPath.strlen();
	    }

	/*
	 * Sanitization is done, produce output.
	 */
	cout << numPairs << ' ';
	for (int i = 0;  i < numPairs;  i++)
	    cout << pairList[i].offset << ' ' << pairList[i].length << ' ';
	cout << outputString;
	cout << '\n';

	delete[] pairList;
	}
    return false;
    }

/*
 * Search the regular-expression list and sanitize a path as described
 * by the list.  Returns NULL if the path doesn't need to be cleaned
 * in any fashion; otherwise returns a malloc'ed area containing the
 * clean path.
 */
static SimpleString maybeSanitizePath(
    const SimpleString&	path)		// Path to be checked
    {
    if (path == NULL)
	return (const char*)NULL;
    int length = path.strlen();
    if (length == 0)
	return (const char*)NULL;
    static bool* sanitizeMask = NULL;
    static int maskSize = 0;
    if (length > maskSize)
	{
	maskSize = length;
	delete[] sanitizeMask;
	sanitizeMask = new bool[length];
	}
    /*
     * Elements of sanitizeMask are true if the associated
     * character needs to be cleaned up.  Start by assuming the entire
     * path must be sanitized.
     */
    for (int maskItem = 0;  maskItem < length;  maskItem++)
	sanitizeMask[maskItem] = true;
    for (DList::Iterator regex(regexList);  regex;  ++regex)
	{
#ifdef __sparc__
	SimpleString* expression = (SimpleString*)(regexList[regex]);
	if (strncmp(path, *expression, expression->strlen()) == 0)
	    return NULL;		// No sanitization needed
#else // __sparc__
	expressionControl* nextExpression =
	  (expressionControl*)regexList[regex];
	for (int startPosition = 0;
	  startPosition < length
	  &&  re_search(&nextExpression->expression,
		(const char*)path, length, startPosition,
		length - startPosition, &nextExpression->matches)
	    >= 0;
	  startPosition =
	    startPosition == nextExpression->matches.end[0]
	      ? startPosition + 1
	      : nextExpression->matches.end[0])
	    {
	    /*
	     * The regular expression matches.  Modify the
	     * sanitization mask to reflect whether it is accepting or
	     * rejecting stuff.  If there are no subexpressions, we
	     * modify the whole matched range; otherwise we modify
	     * just what's matched by the subexpressions.
	     */
	    bool newMaskValue = nextExpression->type == REJECT;
	    for (unsigned int matchNo =
		nextExpression->expression.re_nsub == 0 ? 0 : 1;
	      matchNo <= nextExpression->expression.re_nsub;
	      matchNo++)
		{
		for (int maskItem = nextExpression->matches.start[matchNo];
		  maskItem < nextExpression->matches.end[matchNo];
		  maskItem++)
		    sanitizeMask[maskItem] = newMaskValue;
		}
	    }
#endif // __sparc__
	}
    SimpleString cleanPath;
    cleanPath = sanitizePath(path, sanitizeMask);
    return cleanPath;
    }

/*
 * Sanitize a single path, based on a mask buffer.
 */
static SimpleString sanitizePath(
    const SimpleString&	dirtyPath,	// Path to be sanitized
    bool*		mask)		// Mask of positions to sanitize
    {
    if (dirtyPath == NULL)
	return (const char*)NULL;
    unsigned int dirtyLength = dirtyPath.strlen();
    SimpleString cleanPath(dirtyLength);
    unsigned int lastEnd = 0;
    for (unsigned int startPosition = 0;
      startPosition < dirtyLength;
      startPosition++)
	{
	if (mask[startPosition])
	    {
	    /*
	     * We've found the beginning of a string to sanitize.
	     * Find the end.
	     */
	    if (lastEnd != startPosition)
		cleanPath += dirtyPath.subString(lastEnd,
		  startPosition - lastEnd);
	    unsigned int endPosition = startPosition + 1;
	    for (  ;
	      endPosition < dirtyLength  &&  mask[endPosition];
	      endPosition++)
		;
	    cleanPath +=
	      sanitizePortion(dirtyPath, startPosition, endPosition);
	    lastEnd = startPosition = endPosition;
	    }
	}
    if (lastEnd == 0)
	return (const char*)NULL;
    else if (lastEnd < dirtyLength)
	cleanPath += dirtyPath.subString(lastEnd, dirtyLength - lastEnd);
    return cleanPath;
    }

static SimpleString sanitizePortion(	// Sanitize a substring of a path
    const SimpleString&	dirtyPath,	// Path to be sanitized
    unsigned int	start,		// Offset of first ch to sanitize
    unsigned int	end)		// Offset+1 of end of substring
    {
    SimpleString dirtyCopy(dirtyPath);
    WritableSimpleString cleanPath(dirtyCopy.strlen());
    unsigned int componentStart = start;
    unsigned int componentEnd;
    for (componentEnd = start;  componentEnd < end;  componentEnd++)
	{
	char ch = dirtyCopy[componentEnd];
	/*
	 * If we are auto-passing special characters and this is a
	 * special, break the component here and sanitize it.
	 */
	if (passSpecials  &&  !isalnum(ch))
	    {
	    if (componentStart < componentEnd)
		cleanPath += sanitizeComponent(
		  dirtyCopy.subString(componentStart,
		    componentEnd - componentStart));
	    cleanPath += ch;
	    componentStart = componentEnd + 1;
	    }
	}
    if (componentStart < componentEnd)
	cleanPath += sanitizeComponent(
	  dirtyCopy.subString(componentStart, componentEnd - componentStart));
    return cleanPath;
    }

static SimpleString sanitizeComponent(	// Sanitize one path component
    const SimpleString&	component)	// Component to sanitize
    {
    if (useMD5)
	{
	static MD5_CTX	context;	// Context to use for sanitizing

	MD5Init(&context);
	MD5Update(&context, (unsigned char*)(const char*)md5Seed,
	  md5Seed.strlen());
	MD5Update(&context, (unsigned char*)(const char*)component,
	  component.strlen());
	MD5Final(&context);
	unsigned int finalLength;	// Output length in chars
	switch (component.strlen())
	    {
	    case 0:
	    case 1:
	    case 2:
		{
		finalLength = 4;
		break;
		}
	    case 3:
	    case 4:
		{
		finalLength = 6;
		break;
		}
	    case 5:
	    case 6:
		{
		finalLength = 8;
		break;
		}
	    default:
		{
		finalLength = 10;
		break;
		}
	    }
	/*
	 * Turn each character of the digest into a 6-bit value.
	 */
	for (unsigned int i = 0;  i < sizeof context.digest;  i++)
	    context.digest[i] =
	      (context.digest[i] & ENCODING_MASK)
		^ (context.digest[i] >> ENCODING_BASE_BITS);
	/*
	 * Fold stuff down to the final length
	 */
	for (unsigned int i = finalLength, j = 0;
	  i < sizeof context.digest;
	  ++i, ++j)
	    {
	    if (j >= finalLength)
		j = 0;
	    context.digest[j] ^= context.digest[i];
	    }

	/*
	 * Because of how SimpleString works, it's faster to construct
	 * the result in a scratch variable than in a
	 * compiler-allocated temporary.
	 */
	SimpleString result(maximalEncoding);
	if (sanitizePrefix != '\0')
	    result += sanitizePrefix;
	for (unsigned int i = 0;  i < finalLength;  ++i)
	    result += encoding[context.digest[i]];
	if (sanitizeSuffix != '\0')
	    result += sanitizeSuffix;
	return result;
	}
    else
	{
	static ComponentTranslation* newTranslation = NULL;

	if (newTranslation == NULL)
	    newTranslation = new ComponentTranslation;
	newTranslation->original = component;
	ComponentTranslation* translation =
	  (ComponentTranslation*)translations.find(newTranslation,
	    HashTable::INSERT);
	if (translation == newTranslation)
	    newTranslation = NULL;
	return translation->rewrite;
	}
    }

static unsigned int componentHash(	// Hash a component translation
    ComponentTranslation* component)	// Component to be hashed
    {
    return hashString(component->original);
    }

static int componentCompare(		// Compare two translation entries
    ComponentTranslation* component1,	// Entries to compare
    ComponentTranslation* component2)	// ...
    {
    return component1->original != component2->original;
    }

/*
 * Construct a new component translation.  The "original" field is
 * left NULL, but the "rewrite" field is assigned a generated string.
 */
ComponentTranslation::ComponentTranslation()
    : original(),			// Original handled by user
      rewrite(maximalEncoding)		// Make sure there's room for encoding
    {
    /*
     * We generate names by the simple expedient of appending a
     * numeric sequence to a single alphabetic character.  The numeric
     * sequence comes from a 32-bit counter; we use base 62 so that the
     * names will be shorter.  (I thought about generating random
     * names, but since it's trivially easy for a reader to determine
     * which names were produced first, that seems silly.)
     *
     * It's faster to use += than +.
     */    
    if (sanitizePrefix != '\0')
	rewrite += sanitizePrefix;
    static unsigned int nameCounter = 0;
    rewrite += encode(nameCounter++, ENCODING_BASE);
    if (sanitizeSuffix != '\0')
	rewrite += sanitizeSuffix;
    }

/*
 * Sanitize a UID.  As a special case, a UID of zero is returned
 * unsanitized, so that root activity can be detected.
 *
 * Warning: in MD5 mode, there is a significant risk of collisions in
 * the sanitized output.
 */
static unsigned short cleanUID(
    unsigned short	uid)		// UID to sanitize
    {
    static UID_Translation* newTranslation = NULL;

    if (uid == 0)
	return 0;

    if (useMD5)
	{
	static MD5_CTX	context;	// Context to use for sanitizing

	MD5Init(&context);
	MD5Update(&context, (unsigned char*)(const char*)md5Seed,
	  md5Seed.strlen());
	MD5Update(&context, (unsigned char*)&uid, sizeof uid);
	MD5Final(&context);

	/*
	 * Fold stuff down to the final length
	 */
	for (unsigned int i = sizeof uid, j = 0;
	  i < sizeof context.digest;
	  ++i, ++j)
	    {
	    if (j >= sizeof uid)
		j = 0;
	    context.digest[j] ^= context.digest[i];
	    }

	return (context.digest[0] << 8) | context.digest[1];
	}
    else
	{
	if (newTranslation == NULL)
	    newTranslation = new UID_Translation(uid);
	newTranslation->original = uid;
	UID_Translation* translation =
	  (UID_Translation*)UID_translations.find(newTranslation,
	    HashTable::INSERT);
	if (translation == newTranslation)
	    newTranslation = NULL;
	return translation->rewrite;
	}
    }

static unsigned int UID_Hash(		// Hash a UID translation
    UID_Translation*	uid)		// UID to be hashed
    {
    return hashInteger((int)uid->original);
    }

static int UID_Compare(			// Compare two UID entries
    UID_Translation*	uid1,		// Entries to compare
    UID_Translation*	uid2)		// ...
    {
    return uid1->original != uid2->original;
    }

/*
 * Construct a new UID translation.  The "original" field is
 * set to the user-supplied value, but the "rewrite" field is assigned
 * a generated value.
 */
UID_Translation::UID_Translation(
    unsigned short	original_)	// Original UID to translate
    : original(original_),		// Original provided by user
      rewrite(++nextUID)		// Create new UID for translation
    {
    }

/*
 * Encode a number in a given base.  This function is relatively
 * inefficient because it does a lot of string copying.  A
 * nonrecursive version would be faster but isn't worth the toruble.
 */
static SimpleString encode(
    unsigned int	value,		// Value to encode
    unsigned int	numberBase,	// Base to use
    int			minDigits)	// Minimum number of digits to build
    {
    if (value < numberBase  &&  minDigits <= 0)
	return SimpleString(encoding[value], maximalEncoding);
    else
	{
	SimpleString
	  result(encode(value / numberBase, numberBase, minDigits - 1));
	result += encoding[value % numberBase];
	return result;
	}
    }

/*
 * If necessary, recreate an entry with new pathnames.  Any argument
 * bytes that aren't a member of either path are duplicated in the
 * returned entry.  Note that this routine can return either a pointer
 * to the original, or one to a static area.  Existing pathnames are
 * replaced only if they are non-null.
 */
static seerbuf_t* recreateEntry(
    seerbuf_t*		entry,		// Entry to recreate
    const SimpleString&	newPath1,	// First new path to insert
    const SimpleString&	newPath2)	// Second new path to insert
    {
    if (newPath1 == NULL  &&  newPath2 == NULL)
	return entry;
    /*
     * Buffer used to build sanitized entries.  The union is just
     * to make sure it is padded to an adequate size.
     */
    static union			// Space for sanitized entry
	{
	seerbuf_t	buf;		// Buffer to hold entry
	char		padding[MAXPATHLEN * 2
			  + sizeof(seerbuf_t) - sizeof entry->args[0]
			  + 2 * sizeof entry->args[0]];
	}
			convertedEntry;
    convertedEntry.buf.callcode = entry->callcode;
    convertedEntry.buf.errcode = entry->errcode;
    convertedEntry.buf.flags = entry->flags;
    convertedEntry.buf.uid = entry->uid;
    convertedEntry.buf.pid = entry->pid;
    convertedEntry.buf.calltime = entry->calltime;
    /*
     * WARNING: don't try to optimize out the division; it's needed to
     * cause proper rounding.
     */
    int argumentSize =
      (strlen((const char*)&entry->args[0]) + 1 + sizeof entry->args[0] - 1)
	/ sizeof entry->args[0];
    if (newPath1 == NULL)
	{
	memcpy((char*)&convertedEntry.buf.args[0],
	  (const char*)&entry->args[0],
	  argumentSize * sizeof entry->args[0]);
	convertedEntry.buf.argsize = argumentSize;
	}
    else
	{
	(void)strcpy((char*)&convertedEntry.buf.args[0], newPath1);
	convertedEntry.buf.argsize =
	  (newPath1.strlen() + 1 + sizeof entry->args[0] - 1)
	    / sizeof entry->args[0];
	}
    if (newPath2 != NULL)
	{
	(void)strcpy(
	  (char*)&convertedEntry.buf.args[convertedEntry.buf.argsize],
	  newPath2);
	int newArgument2Size =
	  (newPath2.strlen() + 1 + sizeof entry->args[0] - 1)
	    / sizeof entry->args[0];
	convertedEntry.buf.argsize += newArgument2Size;
	argumentSize += newArgument2Size;
	}
    int difference = entry->argsize - argumentSize;
    if (difference > 0)
	{
	memcpy((char*)&convertedEntry.buf.args[convertedEntry.buf.argsize],
	  (const char*)&entry->args[argumentSize],
	  difference * sizeof entry->args[0]);
	convertedEntry.buf.argsize += difference;
	}
    return &convertedEntry.buf;
    }
