#ifndef lint
static const char Rcs_Id[] =
    "$Id: streamset.cc,v 1.15 1998/10/23 23:15:55 geoff Exp $";
#endif

/*
 * UCLA Ficus Project
 *
 * Copyright (c) 1995, 1996 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Log: streamset.cc,v $
 * Revision 1.15  1998/10/23 23:15:55  geoff
 * Fix things to compile with -pedantic.
 *
 * Revision 1.14  1998/04/09 00:33:03  geoff
 * Use the new notation for list iterators.
 *
 * Revision 1.13  1997/12/04 23:16:37  geoff
 * Use correlatorPanic to do aborts.
 *
 * Revision 1.12  1996/09/11 23:00:23  geoff
 * Convert from dlists to slists.
 *
 * Revision 1.11  1996/05/03  01:13:08  geoff
 * Update the copyright.
 *
 * Revision 1.10  1996/02/02  20:09:58  geoff
 * Use the new panic() routine to do aborts.
 *
 * Revision 1.9  1996/02/02  18:29:26  geoff
 * Fix some minor problems detected by the new C++ compiler.
 *
 * Revision 1.8  1995/11/10  03:00:28  geoff
 * Add zzinlinefunctions.hh.
 *
 * Revision 1.7  1995/09/15  00:06:41  geoff
 * Change the RCS ID to be a const.  Clean out the list when destroying.
 *
 * Revision 1.6  1995/06/30  19:05:05  geoff
 * Change the interpretation of maxfd to be max+1 so that the
 * conventional less-than tests will work.  When removing an fd, search
 * allfds to get a new maxfd.
 *
 * Revision 1.5  1995/06/17  00:12:24  geoff
 * Flush cerr/cout before all aborts.
 *
 * Revision 1.4  1995/05/23  00:40:32  geoff
 * Add UNBUFFERED_STREAMS support.  Abort on errors instead of just
 * exiting.
 *
 * Revision 1.3  1995/05/05  22:42:23  geoff
 * Explicitly include sys/socket.h.
 *
 * Revision 1.2  1995/03/29  21:05:11  geoff
 * Switch to byte-swapped streams, and fix some bugs.
 *
 * Revision 1.1  1995/03/07  22:30:56  geoff
 * Initial revision
 *
 */

#include "compileoptions.hh"
#include "byteswappedstream.hh"
#include "cdecls.hh"
#include "panic.hh"
#include "streamset.hh"
#include "zzinlinefunctions.hh"
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

/*
 * Table of contents.
 *
 * This is #if'ed out because with -pedantic, the compiler warns me
 * that it's not a definition.  Idiots.
 */
#if 0
			StreamSet::~StreamSet();
void			StreamSet::parseFds(const char* descriptor);
					// Parse ASCII into a stream set
void			StreamSet::addFd(int fd);
					// Add a newly-opened FD to set
int			StreamSet::removeFd(int fd);
					// Remove an FD from the set
char*			StreamSet::saveFds() // Turn set into an ASCII string
			  const;
ByteSwappedStream&	StreamSet::streamFor(int fd)
			  const;
					// Return stream for given FD
#endif // 0

extern fd_set		allfds;		// List of all open FD's in the system
extern int		maxfd;		// Maximum known FD + 1

extern void		correlatorPanic(const char* message);
extern void		nomem(const char* where, int line, unsigned int size);

StreamSet::~StreamSet()
    {
    for (SList::Iterator iter(streams);  iter;  ++iter)
	{
	ByteSwappedStream* stream = (ByteSwappedStream*)streams[iter];
	delete stream;
	iter.removeCurrent();
	}
    }

/*
 * Parse an ASCII list of fd's into a StreamSet.  The routine assumes
 * that the fd's are already open.
 */
void		StreamSet::parseFds(
    const char*		descriptor)	// String describing the FD's
    {
    /*
     * An fd_set is represented as a comma-separated list of numbers.
     * We do almost no error checking.
     */
    FD_ZERO(&fds);			// Start with nothing
    for (const char* cp = descriptor;  *cp != '\0';  cp++)
	{
	int fd = atoi(cp);
	FD_SET(fd, &fds);
	FD_SET(fd, &allfds);
	ByteSwappedStream* stream =
	  new ByteSwappedStream(fd, ByteSwappedStream::bigEnded);
	streams.pushHead(stream);
	if (fd >= maxfd)
	    maxfd = fd + 1;
	cp = strchr(cp, ',');
	if (cp == NULL)
	    break;
	}
    }

/*
 * Add a newly-opened FD to a stream set.  The fd is inserted into the
 * stream's fd_set, as well as the "allfds" fd_set.  Also, if we were
 * compiled with UNBUFFERED_STREAMS, the stream is set to unbuffered
 * mode.
 */
void StreamSet::addFd(			// Add a newly-opened FD to set
    int			fd)		// Open fd to be added
    {
    ByteSwappedStream* stream =
      new ByteSwappedStream(fd, ByteSwappedStream::bigEnded);
    streams.pushHead(stream);
    FD_SET(fd, &fds);
    FD_SET(fd, &allfds);
    if (fd >= maxfd)
	maxfd = fd + 1;
#ifdef UNBUFFERED_STREAMS
    stream->rdbuf()->unbuffered(true);
#endif // UNBUFFERED_STREAMS
    }

int StreamSet::removeFd(		// Remove an FD from the set
    int			fd)		// Open fd to be removed
    {
    for (SList::Iterator iter(streams);  iter;  ++iter)
	{
	ByteSwappedStream* stream = (ByteSwappedStream*)streams[iter];
	if (stream->rdbuf()->fd() == fd)
	    {
	    iter.removeCurrent();
	    stream->flush();
	    stream->close();
	    int result = stream->good();
	    delete stream;
	    FD_CLR(fd, &fds);
	    FD_CLR(fd, &allfds);
	    while (--maxfd >= 0)
		{
		if (FD_ISSET(maxfd, &allfds))
		    break;
		}
	    maxfd++;
	    return result;
	    }
	}
    cerr << "correlator: removeFd didn't find file " << fd;
    correlatorPanic("");
    // NOTREACHED
    return 0;
    }

/*
 * Convert a streamSet into a ASCII list of FD's.  Returns a pointer
 * to a new[]'ed storage area containing the descriptive string.
 *
 * This routine assumes that every fd fits into three digits or less.
 */
char* StreamSet::saveFds()
    const
    {
    char* returnString = new char[streams.count() * 4];
    if (returnString == NULL)
	nomem("saveFds", __LINE__, streams.count() * 4);
    char* nextSpot = returnString;
    *nextSpot = '\0';
    for (SList::Iterator iter((SList&)streams);  iter;  ++iter)
	{
	ByteSwappedStream* stream = (ByteSwappedStream*)streams[iter];
	stream->flush();
	if (nextSpot != returnString)
	    *nextSpot++ = ',';
	(void) sprintf(nextSpot, "%d", stream->rdbuf()->fd());
	nextSpot += strlen(nextSpot);
	}
    return returnString;
    }

ByteSwappedStream& StreamSet::streamFor( // Return stream for given FD
    int			fd)		// Fd to locate
      const
    {
    for (SList::Iterator iter((SList&)streams);  iter;  ++iter)
	{
	ByteSwappedStream* stream = (ByteSwappedStream*)streams[iter];
	if (stream->rdbuf()->fd() == fd)
	    return *stream;
	}
    // NEEDSWORK: should throw exception
    cerr << "correlator:  couldn't find stream for fd " << fd;
    correlatorPanic("");
    // NOTREACHED
    return *(new ByteSwappedStream(0));
    }
